# -*- coding: utf-8 -*-

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import logging
import netCDF4
import datetime
import syntool_converter.aquarius as aquarius

logger = logging.getLogger(__name__)


def build_mask(dset):
    """"""
    land_frac = dset.variables['land_frac'][:]
    ice_frac = dset.variables['ice_frac'][:]
    anc_surface_temp = dset.variables['anc_surface_temp'][:]
    cap_flag = dset.variables['cap_flag'][:]

    """
    cap_flags = {'wind_speed_below_15ms': 0,
                 'wind_speed_below_30ms': 1,
                 'wind_speed_above_30ms': 2,
                 'invalid_cap_wind_speed_or_sss': 3,
                 'no_retrieval': 4,
                 'brightness_temperature_error_above_0.4K': 5,
                 'wind_speed_diff_below_15ms_while_raining': 10,
                 'wind_speed_diff_below_30ms_while_raining': 11,
                 'wind_speed_diff_above_30ms_while_raining': 12,
                 'invalid_cap_wind_speed_or_ss_while_raining': 13,
                 'no_retrieval_while_raining': 14,
                 'brightness_temperature_error_above_0.4k': 15}
    """

    """
    aquarius/jpl/v4/l2 valid if
        land_frac < 0.01
        ice_frac < 0.0005,
        anc_surface_temp > 273
        cap_flag < 3 or 10<=cap_flag < 13
        <=> cap_flag in (0, 1, 2, 10 ,11, 12)
    """

    # +100 if abs(rad_TaV - rad_TfV) >= 1 or if abs(rad_TaH - rad_TfH) >= 1
    # rfi: radio frequency interference
    # Ta: Brightness temperature (raw)
    # Tf: Brightness temperature (rfi-filtered)
    # rad_TaH: Radiometer Ta H polarization
    # rad_TaV: Radiometer Ta V polarization
    # rad_TfH: Radiometer Tf H polarization
    # rad_TfV: Radiometer Tf V polarization
    mask = ((land_frac >= 0.1) |
            (ice_frac >= 0.0005) |
            (anc_surface_temp <= 273) |
            ((cap_flag % 10) > 2))

    del land_frac
    del ice_frac
    del anc_surface_temp
    del cap_flag

    return mask


def get_chunk_lon(dset, chunk_rows, beam):
    """"""
    return dset.variables['beam_clon'][chunk_rows, beam]


def get_chunk_lat(dset, chunk_rows, beam):
    """"""
    return dset.variables['beam_clat'][chunk_rows, beam]


def get_chunk_sss(dset, chunk_rows, beam):
    """"""
    values = dset.variables['SSS'][chunk_rows, beam]
    values = numpy.ma.masked_where(values == -999., values)
    return values


def aquarius_l2_sss_cap(infile, outdir, vmin=31.825, vmax=38.175,
                        vmin_pal=32, vmax_pal=38, write_netcdf=False):
    """"""
    # Read/Process data
    datagroup = os.path.splitext(os.path.basename(infile))[0]
    header, _ = datagroup.split('.', 1)
    _dt = datetime.datetime.strptime(header[1:], '%Y%j%H%M%S')
    _date = datetime.datetime(_dt.year, _dt.month, _dt.day)
    dset = netCDF4.Dataset(infile, 'r')
    reltime = dset['sec'][:]
    time_start = dset['sec'][0]
    reltime[reltime < time_start] += 86400
    _lat = dset['beam_clat'][:]
    _lon = dset['beam_clon'][:]

    mask = build_mask(dset)

    version = None
    if datagroup.endswith('V4.0'):
        version = 'v4'
    elif datagroup.endswith('V5.0'):
        version = 'v5'
    else:
        raise Exception('This reader only supports v4.0 and v5.0 data')

    config = {'file_handler': dset,
              'datagroup': datagroup,
              'input_path': os.path.abspath(infile),
              'output_path': os.path.abspath(outdir),
              'mask': mask,
              'reference_time': _date,
              'lon': _lon,
              'lat': _lat,
              'relative_time': reltime,
              'vmin': vmin,
              'vmax': vmax,
              'vmin_pal': vmin_pal,
              'vmax_pal': vmax_pal,
              'lon_getter': get_chunk_lon,
              'lat_getter': get_chunk_lat,
              'sss_getter': get_chunk_sss,
              'product_name': 'AQUARIUS_L2_SSS_CAP_{}'.format(version),
              'min_chunk_size': 10}

    aquarius.process_l2(config)

    dset.close()
