# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-02-19

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import netCDF4
import logging
import datetime
import syntool_converter.aquarius as aquarius

logger = logging.getLogger(__name__)


def extract_var(hdf_group, var_name):
    """"""
    var_obj = hdf_group.variables[var_name]
    fill_value = var_obj._FillValue
    nan_ind = numpy.where(numpy.isnan(var_obj))
    var_array = numpy.ma.array(var_obj[:])
    var_array.mask = False
    if 0 < len(nan_ind[0]):
        var_array[nan_ind] = fill_value
    return var_array, fill_value


def build_mask(f_handler):
    """Masked inputs must be replaced by valid values (odd choice...)."""

    data_group = f_handler.groups['Aquarius Data']

    (rad_land_frac,
     rad_land_frac_fill) = extract_var(data_group, 'rad_land_frac')

    (rad_ice_frac,
     rad_ice_frac_fill) = extract_var(data_group, 'rad_ice_frac')

    (rad_hh_wind_speed,
     rad_hh_wind_speed_fill) = extract_var(data_group, 'rad_hh_wind_speed')

    (rad_tb_consistency,
     rad_tb_consistency_fill) = extract_var(data_group, 'rad_Tb_consistency')

    (rad_moon_ta_ref_v,
     rad_moon_ta_ref_v_fill) = extract_var(data_group, 'rad_moon_Ta_ref_V')

    (rad_moon_ta_ref_h,
     rad_moon_ta_ref_h_fill) = extract_var(data_group, 'rad_moon_Ta_ref_H')

    (rad_galact_ta_ref_v,
     rad_galact_ta_ref_v_fill) = extract_var(data_group, 'rad_galact_Ta_ref_V')

    (rad_galact_ta_ref_h,
     rad_galact_ta_ref_h_fill) = extract_var(data_group, 'rad_galact_Ta_ref_H')

    (anc_sst,
     anc_sst_fill) = extract_var(data_group, 'anc_surface_temp')

    nav_group = f_handler.groups['Navigation']
    (att_ang,
     att_ang_fill) = extract_var(nav_group, 'att_ang')

    land_mask = ((rad_land_frac != rad_land_frac_fill) &
                 (rad_land_frac >= 0.01))

    ice_mask = ((rad_ice_frac != rad_ice_frac_fill) &
                (rad_ice_frac >= 0.01))

    wind_mask = ((rad_hh_wind_speed != rad_hh_wind_speed_fill) &
                 (rad_hh_wind_speed >= 20.0))

    brightness_mask = ((rad_tb_consistency != rad_tb_consistency_fill) &
                       (rad_tb_consistency >= 0.40))

    rad_moon_ta_ref_h[rad_moon_ta_ref_h == rad_moon_ta_ref_h_fill] = 0.0
    rad_moon_ta_ref_v[rad_moon_ta_ref_v == rad_moon_ta_ref_v_fill] = 0.0
    moon_reflect_mask = ((rad_moon_ta_ref_h + rad_moon_ta_ref_v) >= 0.25)

    rad_galact_ta_ref_h[rad_galact_ta_ref_h == rad_galact_ta_ref_h_fill] = 0.0
    rad_galact_ta_ref_v[rad_galact_ta_ref_v == rad_galact_ta_ref_v_fill] = 0.0
    galaxy_reflect_mask = ((rad_galact_ta_ref_h + rad_galact_ta_ref_v) >= 5.60)

    sst_mask = ((anc_sst != anc_sst_fill) &
                (anc_sst <= 0.0))

    att_ang[att_ang == att_ang_fill] = 0.0
    _orientation_mask = ((numpy.abs(att_ang[:, 0]) >= 1.0) |
                         (numpy.abs(att_ang[:, 1]) >= 1.0) |
                         (numpy.abs(att_ang[:, 2]) >= 5.0))

    # Replicate the spacecraft orientation mask for each beam
    orientation_mask = numpy.repeat(_orientation_mask[:, numpy.newaxis], 3,
                                    axis=1)

    # Debug mask components
    logger.debug('Masks')
    logger.debug('-----')
    logger.debug('Land: {} {}'.format(numpy.any(land_mask),
                                      numpy.all(land_mask)))
    logger.debug('Ice: {} {}'.format(numpy.any(ice_mask),
                                     numpy.all(ice_mask)))
    logger.debug('Wind: {} {}'.format(numpy.any(wind_mask),
                                      numpy.all(wind_mask)))
    logger.debug('Brightness temp: {} {}'.format(numpy.any(brightness_mask),
                                                 numpy.all(brightness_mask)))
    logger.debug('Moon: {} {}'.format(numpy.any(moon_reflect_mask),
                                      numpy.all(moon_reflect_mask)))
    logger.debug('Galaxy: {} {}'.format(numpy.any(galaxy_reflect_mask),
                                        numpy.all(galaxy_reflect_mask)))
    logger.debug('SST: {} {}'.format(numpy.any(sst_mask), numpy.all(sst_mask)))
    logger.debug('Orientation: {} {}'.format(numpy.any(orientation_mask),
                                             numpy.any(orientation_mask)))

    _mask = (land_mask |
             ice_mask |
             wind_mask |
             brightness_mask |
             moon_reflect_mask |
             galaxy_reflect_mask |
             sst_mask |
             orientation_mask)

    logger.debug('Result: {} {}'.format(numpy.any(_mask), numpy.all(_mask)))

    return _mask


def get_chunk_lon(dset, chunk_rows, beam):
    """"""
    return dset.groups['Navigation'].variables['beam_clon'][chunk_rows, beam]


def get_chunk_lat(dset, chunk_rows, beam):
    """"""
    return dset.groups['Navigation'].variables['beam_clat'][chunk_rows, beam]


def get_chunk_sss(dset, chunk_rows, beam):
    """"""
    _values, fill = extract_var(dset.groups['Aquarius Data'], 'SSS')
    values = _values[chunk_rows, beam]
    values = numpy.ma.masked_where(values == fill, values)
    return values


def aquarius_l2_sss_or_v5(in_path, out_path, vmin=31.825, vmax=38.175,
                          vmin_pal=32, vmax_pal=38, write_netcdf=False):
    """"""
    datagroup = os.path.basename(in_path)
    f_handler = netCDF4.Dataset(in_path, 'r')
    t_start = '{}000Z'.format(f_handler.time_coverage_start[:-1])
    _dt = datetime.datetime.strptime(t_start, '%Y-%m-%dT%H:%M:%S.%fZ')
    _date = datetime.datetime(_dt.year, _dt.month, _dt.day)
    reltime, fill = extract_var(f_handler.groups['Block Attributes'], 'sec')
    time_start = reltime[0]
    reltime[reltime < time_start] += 86400
    _lat = f_handler.groups['Navigation'].variables['beam_clat'][:]
    _lon = f_handler.groups['Navigation'].variables['beam_clon'][:]

    mask = build_mask(f_handler)

    config = {'file_handler': f_handler,
              'datagroup': datagroup,
              'input_path': os.path.abspath(in_path),
              'output_path': os.path.abspath(out_path),
              'mask': mask,
              'reference_time': _date,
              'lon': _lon,
              'lat': _lat,
              'relative_time': reltime,
              'vmin': vmin,
              'vmax': vmax,
              'vmin_pal': vmin_pal,
              'vmax_pal': vmax_pal,
              'lon_getter': get_chunk_lon,
              'lat_getter': get_chunk_lat,
              'sss_getter': get_chunk_sss,
              'product_name': 'AQUARIUS_L2_SSS_OR_v5',
              'min_chunk_size': 10}

    aquarius.process_l2(config)

    f_handler.close()
