# -*- coding: utf-8 -*-

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import logging
import netCDF4
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def extract_var(hdf_group, var_name):
    """"""
    var_obj = hdf_group.variables[var_name]
    fill_value = var_obj._FillValue
    nan_ind = numpy.where(numpy.isnan(var_obj))
    var_array = numpy.ma.array(var_obj[:])
    var_array.mask = False
    if 0 < len(nan_ind[0]):
        var_array[nan_ind] = fill_value
    return var_array, fill_value


def aquarius_l3_sss_or_v5(infile, outdir, vmin=31.825, vmax=38.175,
                          vmin_pal=32, vmax_pal=38):
    """"""
    # Read/Process data
    logger.info('Read/Process data')
    f_handler = netCDF4.Dataset(infile, 'r')
    lat_min = f_handler.southernmost_latitude
    lat_max = f_handler.northernmost_latitude
    lon_min = f_handler.westernmost_longitude
    lon_max = f_handler.easternmost_longitude
    _sss, fill = extract_var(f_handler, 'l3m_data')
    t_start = '{}000Z'.format(f_handler.time_coverage_start[:-1])
    t_end = '{}000Z'.format(f_handler.time_coverage_end[:-1])
    f_handler.close()

    sss = _sss[::-1, :]
    invalid_mask = (sss == fill)

    dt_start = datetime.datetime.strptime(t_start, '%Y-%m-%dT%H:%M:%S.%fZ')
    dt_end = datetime.datetime.strptime(t_end, '%Y-%m-%dT%H:%M:%S.%fZ')

    dt_middle = dt_start + (dt_end - dt_start) / 2
    time_start = datetime.datetime(dt_middle.year, dt_middle.month,
                                   dt_middle.day)
    time_stop = time_start + datetime.timedelta(days=1)

    lat0 = lat_min
    dlat = (lat_max - lat_min) / numpy.shape(sss)[0]
    lon0 = lon_min
    dlon = (lon_max - lon_min) / numpy.shape(sss)[1]

    # Construct metadata/geolocation/band(s)
    logger.info('Construct metadata/geolocation/band(s)')
    dtime, time_range = stfmt.format_time_and_range(time_start, time_stop,
                                                    units='h')
    now = datetime.datetime.utcnow()
    metadata = {}
    metadata['product_name'] = 'AQUARIUS_L3_SSS_OR_v5'
    metadata['name'] = os.path.basename(infile)
    metadata['datetime'] = dtime
    metadata['time_range'] = time_range
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'Jet Propulsion Laboratory'
    metadata['processing_center'] = ''
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'sea surface salinity'

    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['geotransform'] = [lon0, dlon, 0,
                                   lat0, 0, dlat]
    band = []
    array, offset, scale = pack.ubytes_0_254(sss.data, vmin, vmax)
    if numpy.any(invalid_mask):
        array[numpy.where(invalid_mask)] = 255
    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin_pal, vmax_pal=vmax_pal)
    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea surface salinity',
                 'unittype': 'PSS',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    # Write geotiff
    logger.info('Write geotiff')
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)
