# -*- coding: utf-8 -*-

"""
@author <sylvain.herledan@oceandatalab.com>
@author <lucile.gaultier@oceandatalab.com>
@date 2019-11-13

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# Note: this reader has been written for mono-mission L3 products available at:
# ftp://my.cmems-du.eu/Core/SEALEVEL_GLO_PHY_L3_REP_OBSERVATIONS_008_045/
#
# Supported missions: alg, c2, j2n, j3, sral

import os
import numpy
import scipy
import logging
import netCDF4
import datetime
import syntool_converter
import syntool_converter.utils.tools_for_gcp as tools_for_gcp
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)

TIMEFMT = '%Y-%m-%dT%H:%M:%S.%fZ'

MIN_WIND = 0
MAX_WIND = 25.4
MIN_SWH = 0
MAX_SWH = 8.
MIN_SIGM0 = 10
MAX_SIGM0 = 50
VARIABLES = {'wind': {'varname': 'nadir_wind_nsec',
                      'minmax': (MIN_WIND, MAX_WIND, MIN_WIND, MAX_WIND),
                      'colormap':'noaa_wind',
                      'flag': 'flag_valid_sigma0_nsec',
                      'unit': 'knots',
                      },

             'wave': {'varname': 'nadir_swh_nsec',
                      'minmax': (MIN_SWH, MAX_SWH, MIN_SWH, MAX_SWH),
                      'colormap': 'matplotlib_jet',
                      'flag': 'flag_valid_swh_nsec',
                      'unit': 'm',
                      },
             'sigma0': {'varname': 'nadir_sigma0_nsec',
                      'minmax': (MIN_SIGM0, MAX_SIGM0, MIN_SIGM0, MAX_SIGM0),
                      'colormap': 'matplotlib_jet',
                      'flag': 'flag_valid_sigma0_nsec',
                      'unit': '',
                      },
            }


class UnknownMission(Exception):
    """"""
    def __init__(self, mission, *args, **kwargs):
        """"""
        self.mission = mission
        super(UnknownMission, self).__init__(*args, **kwargs)


def fill_gaps(pass_time, pass_values):
    """"""
    # Detect modal time step and detect data slices that require some
    # interpolation (longer than 1.5 time step)
    pass_dtime = pass_time[1:] - pass_time[:-1]
    time_step = scipy.stats.mode(pass_dtime)[0][0]
    steps_count = numpy.round(pass_dtime / time_step).astype('int')
    needs_interpolation_index = numpy.where(steps_count > 1)[0]

    if 0 >= needs_interpolation_index.size:
        # No need to interpolate
        return pass_time, pass_values

    varnames = pass_values.keys()
    first_index = needs_interpolation_index[0]
    filled_time = pass_time[:first_index + 1]
    filled_values = {}
    for varname in varnames:
        filled_values[varname] = pass_values[varname][:first_index + 1]

    for i in range(0, len(needs_interpolation_index)):
        start_index = needs_interpolation_index[i]
        stop_index = start_index + 1
        interpolated_time = numpy.linspace(pass_time[start_index],
                                           pass_time[stop_index],
                                           num=steps_count[start_index],
                                           endpoint=False)

        # Create fillers with interpolated values for time and NaNs for the
        # observations
        # First value in the interpolated result is already included in the
        # accumulator so we skip it
        time_fill = interpolated_time[1:]
        fill_shape = numpy.shape(time_fill)
        value_fill = numpy.zeros(fill_shape) * numpy.nan

        # Build slice for the next complete (i.e. which does not require
        # interpolation) data chunk
        if i == len(needs_interpolation_index) - 1:
            next_full_slice = slice(stop_index, None)
        else:
            next_index = needs_interpolation_index[i + 1]
            next_full_slice = slice(stop_index, next_index + 1)

        # Concatenate previous results with the interpolated values and the
        # next complete data chunk.
        filled_time = numpy.hstack([filled_time, time_fill])
        filled_time = numpy.hstack([filled_time, pass_time[next_full_slice]])
        for varname in varnames:
            _values = filled_values[varname]
            _values = numpy.hstack([_values, value_fill])
            _values = numpy.hstack([_values,
                                    pass_values[varname][next_full_slice]])
            filled_values[varname] = _values

    return filled_time, filled_values


def get_index_for_changes(tracks, time):
    """"""

    track = tracks.split('/')
    track_change_index = [0]
    name_track = []
    for itrack in track:
        ntrack, time0, time1 = itrack.split('_')
        date0 = datetime.datetime.strptime(time0, '%Y-%m-%dT%H:%M:%S%fZ')
        date1 = datetime.datetime.strptime(time1, '%Y-%m-%dT%H:%M:%S%fZ')
        time_ind = numpy.where((time > date0) & (time < date1))
        if time_ind[0].shape[0]!=0 :
            name_track.append(ntrack)
            track_change_index.append(time_ind[0][-1])
    #track_diff = track[1:] - track[:-1]
    #track_change_index = numpy.where(track_diff != 0)[0]

    #track_change_index += 1
    # = numpy.hstack([0, track_change_index, len(track)])
    return track_change_index, name_track


def track_monotonic_longitudes(time_sorted_lon):
    """"""
    # Ensure longitude continuity (longitudinal movement is monotonic)
    lon0 = numpy.mod(time_sorted_lon[0] + 360.0, 360.0) + 0.0  # [0, 360[
    lon1 = numpy.mod(time_sorted_lon[1] + 360.0, 360.0) + 0.0  # [0, 360[
    lon2 = numpy.mod(time_sorted_lon[2] + 360.0, 360.0) + 0.0  # [0, 360[
    dlon = lon1 - lon0
    if 180 < numpy.abs(dlon):
        # Probability that the first two longitudes are around the 0° meridian
        # is very high: in this case the sign of dlon has an absolute value
        # close to 360 and its sign cannot be trusted to find out if longitudes
        # are increasing or decreasing.
        # If 0° meridian is crossed between the first and second points,
        # crossing it again between the second and third would mean that the
        # track made a U-turn (not possible), so try again with second and
        # third point.
        dlon = lon2 - lon1

    lon_diff = time_sorted_lon - time_sorted_lon[0]
    if 0 < dlon:
        pass_lon = time_sorted_lon[0] + numpy.mod(lon_diff, 360.0)
    else:
        # compute dlon opposite in order to feed numpy.mod with a positive
        # value
        pass_lon = time_sorted_lon[0] - numpy.mod(-1 * lon_diff, 360.0)

    return pass_lon


def read_from_file(f_handler, data_ranges):
    """"""
    variables = data_ranges.keys()

    if 'time_nadir_native' not in f_handler.variables.keys():
        logger.warning('Nadir data not available in this file')
        raise syntool_converter.EarlyExit()

    # Keep everything in chronological order
    time_us = numpy.ma.array(f_handler.variables['time_nadir_native'][:, 1])
    _time = numpy.ma.array(f_handler.variables['time_nadir_native'][:, 0]) + time_us * 10**-6
    _mask = numpy.ma.getmaskarray(_time)
    time_valid = numpy.where(numpy.isfinite(_time) & ~_mask)
    time = _time[time_valid]
    time_ind = numpy.argsort(time)

    latitude = f_handler.variables['lat_l2anad_0'][time_valid]
    latitude = latitude[time_ind]
    longitude = f_handler.variables['lon_l2anad_0'][time_valid]
    longitude = longitude[time_ind]
    #track = f_handler.variables['track'][time_ind]
    #cycle = f_handler.variables['cycle'][time_ind]
    flags_surf = f_handler.variables['wf_surf_flag'][time_valid]
    flags_surf = flags_surf[time_ind]
    time_units = 'seconds since 2009-01-01 00:00:00 0:00'
    _dtime = netCDF4.num2date(time, time_units)
    tracks = f_handler.trace
    values = {}

    for ivar in variables:
        dic_var = data_ranges[ivar]
        values[ivar] = f_handler.variables[dic_var['varname']][time_valid]
        values[ivar] = values[ivar][time_ind]
        flags = f_handler.variables[dic_var['flag']][time_valid]
        flags = flags[time_ind]
        fill_value = f_handler.variables[dic_var['varname']]._FillValue
        masked_index = numpy.where((values[ivar] == fill_value)
                                   | (flags == 1) | (flags_surf == 1))
        values[ivar][masked_index] = numpy.nan
    track_change_index, name_pass = get_index_for_changes(tracks, _dtime)
    for pass_num in range(0, len(track_change_index) - 1):
        pass_first = track_change_index[pass_num]
        pass_last = track_change_index[pass_num + 1] - 1
        pass_slice = slice(pass_first, pass_last + 1)
        pass_length = 1 + pass_last - pass_first
        npass = name_pass[pass_num]
        if 3 > pass_length:
            logger.warning('Pass is too short (less than 3 points). Skipped.')
            continue

        pass_lon = longitude[pass_slice]
        pass_lat = latitude[pass_slice]
        pass_time = time[pass_slice]
        pass_values = {}
        for varname in variables:
            pass_values[varname] = values[varname][pass_slice]

        # Make sure longitudinal variation is monotonic
        pass_lon = track_monotonic_longitudes(pass_lon)

        filled_time, filled_values = fill_gaps(pass_time, pass_values)

        if numpy.any(numpy.ma.getmaskarray(filled_time)):
            logger.error('Time array should not have any masked value')

        filled_time = numpy.ma.getdata(filled_time)
        _pass_time = numpy.ma.getdata(pass_time)
        _pass_lon = numpy.ma.getdata(pass_lon)
        _pass_lat = numpy.ma.getdata(pass_lat)
        # From scipy documentation:
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.PchipInterpolator.html
        #
        # The interpolator preserves monotonicity in the interpolation data and
        # does not overshoot if the data is not smooth.
        lon_interp = scipy.interpolate.PchipInterpolator(_pass_time, _pass_lon)
        lat_interp = scipy.interpolate.PchipInterpolator(_pass_time, _pass_lat)
        lat = lat_interp(filled_time)
        _lon = lon_interp(filled_time)

        lon = _lon

        dist_gcp = None
        gcps = tools_for_gcp.make_gcps_v2(lon, lat, 'linear',
                                          dist_gcp=dist_gcp)
        geolocation = {}
        geolocation['projection'] = stfmt.format_gdalprojection()
        geolocation['gcps'] = stfmt.format_gdalgcps(*gcps)

        timestamps = numpy.float64(filled_time)
        datetimes = netCDF4.num2date(timestamps, units=time_units)
        start_time = datetimes[0]
        end_time = datetimes[-1]
        (dtime, time_range) = stfmt.format_time_and_range(start_time,
                                                          end_time, units='s')
        now = datetime.datetime.utcnow()

        meta = {'datetime': dtime,
                'time_range': time_range,
                'begin_datetime': start_time.strftime(TIMEFMT),
                'end_datetime': end_time.strftime(TIMEFMT),
                'source_provider': '',
                'processing_center': '',
                'conversion_software': 'Syntool',
                'conversion_version': '0.0.0',
                'conversion_datetime': stfmt.format_time(now),
                'type': 'along_track',
                'cycle': 0,
                'pass': npass,
                'spatial_resolution': 7000.0}

        for ivar in variables:
            dic_var = data_ranges[ivar]
            gap_mask = numpy.isnan(filled_values[ivar])
            mask = numpy.ma.getmaskarray(filled_values[ivar])

            if (gap_mask | mask).all():
                _msg = 'Pass {} contains only masked values for {}. Skipped.'
                logger.warning(_msg.format(meta['pass'], ivar))
                continue

            vmin, vmax, vmin_pal, vmax_pal = dic_var['minmax']
            colormap = dic_var['colormap']
            colortable = stfmt.format_colortable(colormap,
                                                 vmax=vmax, vmax_pal=vmax_pal,
                                                 vmin=vmin, vmin_pal=vmin_pal)

            _array = filled_values[ivar]
            array, offset, scale = pack.ubytes_0_254(_array, vmin, vmax)
            if numpy.any(mask):
                array[numpy.where(mask)] = 255
            if numpy.any(gap_mask):
                array[numpy.where(gap_mask)] = 255

            # Add a second axis for Geotiff compatibility
            array = array[:, numpy.newaxis]

            data = [{'array': array,
                     'scale': scale,
                     'offset': offset,
                     'description': ivar,
                     'name': ivar,
                     'unittype': dic_var['unit'],
                     'nodatavalue': 255,
                     'parameter_range': [vmin, vmax],
                     'colortable': colortable}]

            extra = {'varname': ivar}

            yield (meta, geolocation, data, extra)


def convert_from_file(input_path, output_path):
    """"""
    granule_filename = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_filename)

    data_ranges = VARIABLES

    f_handler = netCDF4.Dataset(input_path, 'r')
    for (meta, geolocation, data, extra) in read_from_file(f_handler,
                                                           data_ranges):
        product_name = '{}_{}'.format('CFOSAT_L2anad', extra['varname'])
        pass_name = '{}_p{}'.format(granule_name,
                                    str(meta['pass']).zfill(3))

        meta['name'] = pass_name
        meta['datagroup'] = granule_name
        meta['source_URI'] = input_path
        meta['product_name'] = product_name

        tifffile = stfmt.format_tifffilename(output_path, meta,
                                             create_dir=True)
        stfmt.write_geotiff(tifffile, meta, geolocation, data)
    f_handler.close()


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARN)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)
