#!/usr/bin/env python
# coding=utf-8
"""
"""

# IMPORTANT : use agg backend for fig2rgb() working
# the next two lines must be called before any matplotlib import
import matplotlib
matplotlib.use('agg')

from datetime import timedelta
import matplotlib.pyplot as plt
import numpy as np
from netCDF4 import Dataset, num2date
import os
import pyproj

import syntool_converter.utils.syntoolformat as stfmt


def ax_txt_corner(ax, txt, pos, **kwargs):
    """
    """
    # pos2set = {'tl': {'x':0.01, 'y':0.99, 'ha':'left', 'va':'top'},
    #            'tr': {'x':0.99, 'y':0.99, 'ha':'right', 'va':'top'},
    #            'bl': {'x':0.01, 'y':0.01, 'ha':'left', 'va':'bottom'},
    #            'br': {'x':0.99, 'y':0.01, 'ha':'right', 'va':'bottom'},
    #            'tcr': {'x':0.51, 'y':0.99, 'ha':'left', 'va':'top'}}
    pos2set = {'tl1': {'x':0.015, 'y':0.985, 'ha':'left', 'va':'top'},
               'tl2': {'x':0.015, 'y':0.955, 'ha':'left', 'va':'top'},
               'tl3': {'x':0.015, 'y':0.925, 'ha':'left', 'va':'top'},
               'tr1': {'x':0.985, 'y':0.985, 'ha':'right', 'va':'top'},
               'tr2': {'x':0.985, 'y':0.955, 'ha':'right', 'va':'top'},
               'tr3': {'x':0.985, 'y':0.925, 'ha':'right', 'va':'top'},
               'bl3': {'x':0.015, 'y':0.065, 'ha':'left', 'va':'bottom'},
               'bl2': {'x':0.015, 'y':0.04, 'ha':'left', 'va':'bottom'},
               'bl1': {'x':0.015, 'y':0.015, 'ha':'left', 'va':'bottom'},
               'br3': {'x':0.985, 'y':0.065, 'ha':'right', 'va':'bottom'},
               'br2': {'x':0.985, 'y':0.04, 'ha':'right', 'va':'bottom'},
               'br1': {'x':0.985, 'y':0.015, 'ha':'right', 'va':'bottom'},
               'tc1': {'x':0.49, 'y':0.975, 'ha':'right', 'va':'top'},
               'tc2': {'x':0.51, 'y':0.975, 'ha':'left', 'va':'top'},
               'tc3': {'x':0.50, 'y':0.95, 'ha':'center', 'va':'top'}}
    xlim = ax.get_xlim()
    x = pos2set[pos]['x']
    xpos = xlim[0] + x*(xlim[1]-xlim[0])
    ylim = ax.get_ylim()
    y = pos2set[pos]['y']
    ypos = ylim[0] + y*(ylim[1]-ylim[0])
    ha = pos2set[pos]['ha']
    va = pos2set[pos]['va']
    ax.text(xpos, ypos, txt, ha=ha, va=va, **kwargs)


def make_spec_fig(spec, kctr, phictr, heading,
                  labels=None, params=None, partorder=None,
                  vmax=None, vmin=None,
                  spec_size=(512, 512), kmax=2*np.pi/50,
                  kcircle=[2*np.pi/400, 2*np.pi/200, 2*np.pi/100, 2*np.pi/50],
                  pcolor=['r', 'b', 'g', 'c'], ppos=['tl', 'tr', 'br', 'bl'],
                  title1=None, title2=None, title3=None,
                  cmap=None, fontsize=10.):
    """
    """
    # Constants
    nphi, nk = spec.shape
    klim = np.concatenate([[kctr[0] - (kctr[1] - kctr[0]) / 2.],
                           (kctr[0:-1] + kctr[1:]) / 2.,
                           [kctr[-1] + (kctr[-1] - kctr[-2]) / 2.]])
    dphi = 360. / phictr.size
    philim = np.append(phictr - dphi / 2., phictr[-1] + dphi / 2.)
    klimgrid, philimgrid = np.meshgrid(klim, philim)
    theta = 2. * np.pi * np.linspace(0, 1, num=361)
    if vmax is None:
        vmax = spec.max()
    if vmin is None:
        vmin = 0.

    # Create figure
    dpi = 100.
    fig_size = np.asarray(spec_size, dtype='float')
    fig_size_inch = fig_size / dpi + 0.0001
    fig = plt.figure(figsize=fig_size_inch, dpi=dpi, facecolor='w')

    # Make figure
    # Set xlim, ylim for pcolormesh
    philimgrid2 = 90. - philimgrid + heading
    if np.sin((90 - heading) * np.pi / 180) < 0:
        # descending pass
        philimgrid2 += 180
    xlim = klimgrid * np.cos(philimgrid2 * np.pi / 180)
    ylim = klimgrid * np.sin(philimgrid2 * np.pi / 180)

    # Create axes and plot
    ax = fig.add_axes([0., 0., 1., 1.])
    axkmin = -kmax
    axkmax = kmax
    ax.set_ylim([axkmin, axkmax])
    ax.set_xlim([axkmin, axkmax])

    ax.pcolormesh(xlim, ylim, spec, cmap=cmap, vmin=vmin, vmax=vmax)

    # Partition contours
    if labels is not None:
        nlab = labels.max()
        _partorder = partorder
        if _partorder is None:
            _partorder = np.arange(nlab)
        for i in range(nlab):
            ilab = _partorder[i] + 1 # labels in [1,N], 0 being no partition
            pcol = pcolor[i]
            indpart = np.where(labels == ilab)
            for x, y in zip(indpart[0], indpart[1]):
                if y == 0:
                    plt.plot(xlim[x:x+2, y], ylim[x:x+2, y], color=pcol)
                else:
                    ilabneigh = labels[x, y-1]
                    if ilabneigh == 0:
                        plt.plot(xlim[x:x+2, y], ylim[x:x+2, y], color=pcol)
                    elif ilabneigh != ilab:
                        plt.plot(xlim[x:x+2, y], ylim[x:x+2, y], color=pcol, alpha=0.5)
                if y == nk-1:
                    plt.plot(xlim[x:x+2, y+1], ylim[x:x+2, y+1], color=pcol)
                else:
                    ilabneigh = labels[x, y+1]
                    if ilabneigh == 0:
                        plt.plot(xlim[x:x+2, y+1], ylim[x:x+2, y+1], color=pcol)
                    elif ilabneigh != ilab:
                        plt.plot(xlim[x:x+2, y+1], ylim[x:x+2, y+1], color=pcol, alpha=0.5)
                ilabneigh = labels[x-1, y]
                if ilabneigh == 0:
                    plt.plot(xlim[x, y:y+2], ylim[x, y:y+2], color=pcol)
                elif ilabneigh != ilab:
                    plt.plot(xlim[x, y:y+2], ylim[x, y:y+2], color=pcol, alpha=0.5)
                ilabneigh = labels[(x+1) % nphi, y]
                if ilabneigh == 0:
                    plt.plot(xlim[x+1, y:y+2], ylim[x+1, y:y+2], color=pcol)
                elif ilabneigh != ilab:
                    plt.plot(xlim[x+1, y:y+2], ylim[x+1, y:y+2], color=pcol, alpha=0.5)

    # Partition parameters
    if params is not None:
        nlab = params.shape[1]
        _partorder = partorder
        if _partorder is None:
            _partorder = np.arange(nlab)
        for i in range(nlab):
            ilab = _partorder[i]
            hs, wl, dire = params[:, ilab]
            if np.ma.is_masked(dire):
                txt1 = 'dir=--deg (+-180)'
            else:
                txt1 = 'dir={:3d}deg (+-180)'.format(int(np.round(dire)))
            ax_txt_corner(ax, txt1, ppos[i] + '1', color=pcolor[i], fontsize=fontsize)
            if np.ma.is_masked(wl):
                txt2 = 'lambda=--m'
            else:
                txt2 = 'lambda={:3d}m'.format(int(np.round(wl)))
            ax_txt_corner(ax, txt2, ppos[i] + '2', color=pcolor[i], fontsize=fontsize)
            if np.ma.is_masked(hs):
                txt3 = 'Hs=--m'
            else:
                txt3 = 'Hs={0:.2f}m'.format(hs)
            ax_txt_corner(ax, txt3, ppos[i] + '3', color=pcolor[i], fontsize=fontsize)
            if not np.ma.is_masked(dire) and not np.ma.is_masked(wl) and wl != 0:
                _phi = np.deg2rad(90. - dire + heading)
                _k = 2. * np.pi / wl
                plt.plot(_k * np.cos(_phi), _k * np.sin(_phi), '+' + pcolor[i],
                         markersize=10., markeredgewidth=2.)
                plt.plot(_k * np.cos(_phi + np.pi), _k * np.sin(_phi + np.pi),
                         '+' + pcolor[i], markersize=10., markeredgewidth=2.)

    # Add north for debug
    # north = 90. - 0. + heading
    # if np.sin((90 - heading) * np.pi / 180) < 0:
    #     # descending pass
    #     north += 180
    # ax.plot([0., 2. * np.pi / 200. * np.cos(np.deg2rad(north))],
    #         [0., 2. * np.pi / 200. * np.sin(np.deg2rad(north))], '-k')

    if title1 is not None:
        ax_txt_corner(ax, title1, 'tc1', color='k', fontsize=fontsize)
    if title2 is not None:
        ax_txt_corner(ax, title2, 'tc2', color='k', fontsize=fontsize)
    if title3 is not None:
        ax_txt_corner(ax, title3, 'tc3', color='k', fontsize=fontsize)

    ax.set_axis_off()
    ax.plot([axkmin, axkmax], [0, 0], ':k')
    ax.plot([0, 0], [axkmin, axkmax], ':k')
    if kcircle != None:
        for kcir in kcircle:
            ax.plot(kcir*np.cos(theta), kcir*np.sin(theta), ':k')
            kcirstr = '%im' % (np.round(2*np.pi/kcir))
            if kcir != kmax:
                ax.text(0, -kcir, kcirstr, ha='center', va='top',
                        fontsize=fontsize)
            else:
                ax.text(0, -kcir, kcirstr, ha='center', va='bottom',
                        fontsize=fontsize)

    return fig


def fig2rgb(fig):
    """
    """
    fig.canvas.draw()
    #fig.canvas.print_png('/tmp/test_sm_xspec.png')
    wid, hei = fig.canvas.get_width_height()
    rgb = np.fromstring(fig.canvas.tostring_rgb(), dtype=np.uint8)
    rgb.shape = (hei, wid, 3)
    return rgb


def swim_l2_spec(infile, outdir, latmax=80., nocontour=True):
    """
    """
    # Read
    dset = Dataset(infile, 'r')
    npos = len(dset.dimensions['n_posneg'])
    nbox = len(dset.dimensions['n_box'])
    nbeam = len(dset.dimensions['n_beam_l2'])
    npart = len(dset.dimensions['npartitions'])
    #lat = dset.variables['lat_l2'][:]
    #lon = dset.variables['lon_l2'][:]
    lat_spec = dset.variables['lat_spec_l2'][:]
    lon_spec = dset.variables['lon_spec_l2'][:]
    #min_lat = dset.variables['min_lat_l2'][:]
    #max_lat = dset.variables['max_lat_l2'][:]
    #min_lon = dset.variables['min_lon_l2'][:]
    #max_lon = dset.variables['max_lon_l2'][:]
    #phi_orbit_box = dset.variables['phi_orbit_box'][:]
    #lat_nadir = dset.variables['lat_nadir_l2'][:]
    #lon_nadir = dset.variables['lon_nadir_l2'][:]
    #time_nadir = dset.variables['time_nadir_l2'][:]
    start_time = dset.variables['start_time_l2'][:]
    _time_units = dset.variables['start_time_l2'].units
    stop_time = dset.variables['stop_time_l2'][:]
    k = dset.variables['k_spectra'][:]
    phi = dset.variables['phi_vector'][:]
    #pp = dset.variables['pp_mean_raw'][:]
    pp = dset.variables['pp_mean'][:]
    flag_part = np.ma.filled(dset.variables['flag_partition'][:], 0)
    param_part = dset.variables['wave_param_part'][:]
    mask_part = np.ma.filled(dset.variables['mask'][:], 0)
    pp_combi = dset.variables['p_combined'][:]
    flag_part_combi = np.ma.filled(dset.variables['flag_partition_combined'][:], 0)
    param_part_combi = dset.variables['wave_param_part_combined'][:]
    mask_part_combi = np.ma.filled(dset.variables['mask_combined'][:], 0)
    macrocycle_angle = dset.macrocycle_angle
    seaice_cov = dset.variables['sea_ice_coverage'][:]
    land_cov = dset.variables['land_coverage'][:]
    seaice_cov_combi = dset.variables['sea_ice_coverage_combined'][:]
    land_cov_combi = dset.variables['land_coverage_combined'][:]
    #mask_along_track = dset.variables['mask_along_track_box'][:]
    #along_track_masking = dset.along_track_masking
    dset.close()

    # Set time_units
    if not _time_units.startswith('s+us '):
        raise Exception('time units has changed (not s+us)')
    time_units = 'seconds {}'.format(_time_units.replace('s+us ', ''))

    # Identify L2 beams
    macrocycle = [int(a) for a in macrocycle_angle.split(';')]
    beam_name = []
    for inc in [6, 8, 10]:
        if inc in macrocycle:
            beam_name.append('{:02d}'.format(inc))
    if len(beam_name) != nbeam:
        raise Exception('Found beams {} but n_beam_l2={}'.format(beam_name, nbeam))

    # Concatenate L2 beams and combined spectra
    pp = np.ma.concatenate((pp, pp_combi[..., np.newaxis]), axis=-1)
    flag_part = np.ma.concatenate((flag_part, flag_part_combi[..., np.newaxis]), axis=-1)
    param_part = np.ma.concatenate((param_part, param_part_combi[..., np.newaxis]), axis=-1)
    mask_part = np.concatenate((mask_part, mask_part_combi[..., np.newaxis]), axis=-1)
    seaice_cov = np.ma.concatenate((seaice_cov, seaice_cov_combi[..., np.newaxis]), axis=-1)
    land_cov = np.ma.concatenate((land_cov, land_cov_combi[..., np.newaxis]), axis=-1)
    beam_name.append('combined')
    nbeam += 1
    beam_name_tit = list(beam_name)
    beam_name_tit[-1] = 'cb'

    # Set heading and boxes polygons
    geod = pyproj.Geod(ellps='WGS84')
    ll_mask = np.ma.getmaskarray(lon_spec) | np.ma.getmaskarray(lat_spec)
    _lon_spec = np.ma.filled(lon_spec, fill_value=0.)
    _lat_spec = np.ma.filled(lat_spec, fill_value=0.)
    _forw, _back, _dist = geod.inv(_lon_spec[:, :-1], _lat_spec[:, :-1],
                                   _lon_spec[:, 1:], _lat_spec[:, 1:])
    _mask = ll_mask[:, :-1] | ll_mask[:, 1:]
    forw = np.concatenate((_back[:, [0]], _forw, _forw[:, [-1]]), axis=1)
    dist = np.concatenate((_dist[:, [0]], _dist, _dist[:, [-1]]), axis=1)
    mask = np.concatenate((_mask[:, [0]], _mask, _mask[:, [-1]]), axis=1)
    lonmid, latmid, _ = geod.fwd(np.concatenate((_lon_spec[:, [0]], _lon_spec), axis=1),
                                 np.concatenate((_lat_spec[:, [0]], _lat_spec), axis=1),
                                 forw, dist / 2.)
    forw[:, 0] = _forw[:, 0]
    lonmidl, latmidl, _ = geod.fwd(lonmid, latmid, forw - 90., dist / 2.)
    lonmidr, latmidr, _ = geod.fwd(lonmid, latmid, forw + 90., dist / 2.)

    #box_lon = np.stack((lonmidl[:, 1:], lonmidr[:, 1:], lonmidr[:, :-1], lonmidl[:, :-1]))
    #box_lat = np.stack((latmidl[:, 1:], latmidr[:, 1:], latmidr[:, :-1], latmidl[:, :-1]))
    box_lon = np.stack((lonmidl[:, 1:], lonmid[:, 1:], lonmidr[:, 1:],
                        lonmidr[:, :-1], lonmid[:, :-1], lonmidl[:, :-1]))
    box_lat = np.stack((latmidl[:, 1:], latmid[:, 1:], latmidr[:, 1:],
                        latmidr[:, :-1], latmid[:, :-1], latmidl[:, :-1]))
    mask[dist > 100000.] = True # in case of holes in product
    box_okgeo = ~(mask[:, 1:] | mask[:, :-1])
    heading, _, _ = geod.inv(lonmid[:, :-1], latmid[:, :-1], lonmid[:, 1:], latmid[:, 1:])

    # Plot boxes for debug
    # for ipos in range(npos):
    #     for ibox in range(nbox):
    #         p = plt.plot(box_lon[[0, 1, 2, 3, 0], ipos, ibox],
    #                      box_lat[[0, 1, 2, 3, 0], ipos, ibox], '-')
    #         p = plt.plot(box_lon[[0, 1, 2, 3, 4, 5, 0], ipos, ibox],
    #                      box_lat[[0, 1, 2, 3, 4, 5, 0], ipos, ibox], '-')
    #         plt.plot(box_lon[0, ipos, ibox], box_lat[0, ipos, ibox], 'o', color=p[-1].get_color())
    #         plt.plot(box_lon[1, ipos, ibox], box_lat[1, ipos, ibox], 'x', color=p[-1].get_color())
    # plt.plot(lon_spec.flatten(), lat_spec.flatten(), '+k')
    # #plt.plot(lon.flatten(), lat.flatten(), 'xr')
    # #plt.plot(lon_nadir.flatten(), lat_nadir.flatten(), 'o-k')
    # plt.show()
    # import pdb ; pdb.set_trace()

    # Constants
    phictr = np.concatenate((phi, phi + 180))
    indk = np.where(k > 2. * np.pi / 1000)[0]
    ksli = slice(indk[0] - 1, None)
    kctr = k[ksli]
    cmap, _ = stfmt.load_colormap('cerbere_wind')
    cmap.set_bad('lightgrey')
    spec_size = (512, 512)
    dataname = os.path.splitext(os.path.basename(infile))[0]

    # Loop on boxes and beams
    for ibox in range(nbox):
        for ipos in range(npos):

            if not box_okgeo[ipos, ibox]:
                continue
            if np.all(np.ma.getmaskarray(pp[ksli, :, ipos, ibox, :])):
                continue

            # Set time
            if np.ma.is_masked(start_time[ipos, ibox, :]) or \
               np.ma.is_masked(stop_time[ipos, ibox, :]):
                continue
            _start_time = num2date(start_time[ipos, ibox, 0], time_units) + \
                timedelta(microseconds=float(start_time[ipos, ibox, 1]))
            _stop_time = num2date(stop_time[ipos, ibox, 0], time_units) + \
                timedelta(microseconds=float(stop_time[ipos, ibox, 1]))
            (dtime, time_range) = stfmt.format_time_and_range(_start_time, _stop_time,
                                                              units='ms')

            # Set geoloc
            _heading = heading[ipos, ibox]
            gcplon = box_lon[:, ipos, ibox]
            gcplat = box_lat[:, ipos, ibox]
            if np.abs(gcplat).max() > latmax:
                continue
            # gcppix = np.array([0, spec_size[0], spec_size[0], 0])
            # gcplin = np.array([1., 1., spec_size[1] - 1., spec_size[1] - 1.])
            gcppix = np.array([0., spec_size[0] / 2., spec_size[0],
                               spec_size[0], spec_size[0] / 2., 0])
            gcplin = np.array([1., 1., 1., spec_size[1] - 1., spec_size[1] - 1.,
                               spec_size[1] - 1.])
            if np.sin((90 - _heading) * np.pi / 180) < 0:
                # descending pass
                gcppix = spec_size[0] - gcppix
                gcplin = spec_size[1] - gcplin
            gcphei = np.zeros(gcplin.size)
            if gcplon.min() < -135 and gcplon.max() > 135:
                gcplon[np.where(gcplon < 0)] += 360.

            geolocation = {}
            geolocation['projection'] = stfmt.format_gdalprojection()
            geolocation['gcps'] = stfmt.format_gdalgcps(gcplon, gcplat, gcphei,
                                                        gcppix, gcplin)

            for ibeam in range(nbeam):

                # Get spectra, labels, params
                _pp = pp[ksli, :, ipos, ibox, ibeam]
                if np.all(np.ma.getmaskarray(_pp)):
                    continue
                #if np.any(np.ma.getmaskarray(_pp)):
                #    continue

                spec = np.rollaxis(np.ma.concatenate((_pp, _pp), axis=1), 1)

                labels = None
                if not nocontour and mask_part[ksli, :, :, ipos, ibox, ibeam].max() != 0:
                    _lab = np.zeros((kctr.size, phi.size), dtype='int8')
                    for ipart in range(npart):
                        _lab += mask_part[ksli, :, ipart, ipos, ibox, ibeam] * (ipart + 1)
                    labels = np.rollaxis(np.concatenate((_lab, _lab), axis=1), 1)
                params = None
                fpart = flag_part[ipos, ibox, ibeam]
                if fpart != 0:
                    params = param_part[:, :fpart, ipos, ibox, ibeam] # hs,wl,dir first dim
                    # if np.all(np.ma.getmaskarray(params)):
                    #     params = None
                partorder = None
                if params is not None:
                    # Order partitions from directions
                    partorder = params[2, :].argsort()

                # Make spectrum figure
                tit1 = 'beam={}'.format(beam_name_tit[ibeam])
                tit2 = 'box={:03d}/{:01d}'.format(ibox, ipos)
                tit3 = []
                if np.ma.filled(land_cov[ipos, ibox, ibeam], 0) != 0:
                    tit3.append('LAND={:d}%'.format(land_cov[ipos, ibox, ibeam]))
                if np.ma.filled(seaice_cov[ipos, ibox, ibeam], 0) != 0:
                    tit3.append('SEAICE={:d}%'.format(seaice_cov[ipos, ibox, ibeam]))
                if len(tit3) == 0:
                    tit3 = None
                elif len(tit3) == 1:
                    tit3 = tit3[0]
                else:
                    tit3 = '/'.join(tit3)
                ## contrast linear
                #vmin = 0.
                #vmax = 0.5
                ## contrast log
                vmin = -4. #-3.5 #-6.
                vmax = 1. #0.
                spec = np.ma.log(np.ma.maximum(spec, np.exp(vmin)))

                fig = make_spec_fig(spec, kctr, phictr, _heading,
                                    labels=labels, params=params,
                                    partorder=partorder,
                                    vmax=vmax, vmin=vmin, cmap=cmap,
                                    spec_size=spec_size,
                                    title1=tit1, title2=tit2, title3=tit3)
                rgb = fig2rgb(fig)
                plt.close(fig)

                # Write geotiff
                metadata = {}
                metadata['product_name'] = 'SWIM_{}_L2_wave_spectrum'.format(beam_name[ibeam])
                metadata['datagroup'] = dataname
                metadata['name'] = '{}-{:03d}-{:1d}'.format(dataname, ibox, ipos)
                metadata['datetime'] = dtime
                metadata['time_range'] = time_range

                band = []
                band.append({'array': rgb[:, :, 0]})
                band.append({'array': rgb[:, :, 1]})
                band.append({'array': rgb[:, :, 2]})

                tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
                stfmt.write_geotiff(tifffile, metadata, geolocation, band)


### For debug merged shape
# def get_shape(metadata_path):
#     """
#     """
#     import json
#     import ogr
#     with open(metadata_path, 'r') as f:
#         shape_str = json.load(f)['shape_str']
#     return ogr.CreateGeometryFromWkt(shape_str)

# def get_shape_xy(geom):
#     """
#     """
#     xy = []
#     for i in range(geom.GetGeometryCount()):
#         _geom = geom.GetGeometryRef(i)
#         __geom = _geom.GetGeometryRef(0)
#         xy.append(np.array(__geom.GetPoints()))
#     return xy

