# -*- encoding: utf-8 -*-

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

@author <sylvain.herledan@oceandatalab.com>
@date 2018-10-22
"""

import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def build_scalar_data(f_handler, ind_lat, ind_lon, t, vname, _min, _max, units,
                      cmap):
    """"""
    v_data = f_handler.variables[vname][t, ind_lat, ind_lon]
    v_data = numpy.ma.masked_invalid(v_data)
    mask = numpy.ma.getmaskarray(v_data)

    if 'CHL' == vname:
        vname = 'Chlorophyll concentration'
        units = 'log({})'.format(units)
        v_data = numpy.ma.log(v_data)
        _min = numpy.percentile(v_data.compressed(), 1)
        _max = numpy.percentile(v_data.compressed(), 99.99)
        error_data = f_handler.variables['CHL_error'][t, ind_lat, ind_lon]
        error_mask = (error_data >= 100.0)
        mask = (mask | error_mask)

    masked_ind = numpy.where(mask)
    valid_ind = numpy.where(~mask)
    logger.debug('{} min: {}'.format(vname, numpy.nanmin(v_data[valid_ind])))
    logger.debug('{} max: {}'.format(vname, numpy.nanmax(v_data[valid_ind])))

    colortable = stfmt.format_colortable(cmap,
                                         vmin=_min, vmax=_max,
                                         vmin_pal=_min, vmax_pal=_max)

    array, offset, scale = pack.ubytes_0_254(v_data, _min, _max)
    if numpy.any(mask):
        array[masked_ind] = 255

    data = []
    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': vname,
                 'unittype': units,
                 'nodatavalue': 255,
                 'parameter_range': [_min, _max],
                 'colortable': colortable})
    return data


def read_from_file(input_path, process_var):
    """"""
    f_handler = netCDF4.Dataset(input_path, 'r')
    lon = f_handler.variables['lon'][:]
    lat = f_handler.variables['lat'][:]
    _t = f_handler.variables['time'][:]
    _t_units = f_handler.variables['time'].units

    # Longitudes are provided in [0, 360] must the landmask is defined with
    # [-180, 180] so shift longitudes to match the landmask.
    lon[lon > 180.0] = lon[lon > 180.0] - 360.0
    ind_lon = numpy.argsort(lon)
    lon = lon[ind_lon]

    ind_lat = numpy.where(numpy.abs(lat) < 86)[0]
    lat = lat[ind_lat]

    lon0, dlon = lon[0], numpy.mean(lon[1:] - lon[:-1])
    lat0, dlat = lat[0], numpy.mean(lat[1:] - lat[:-1])
    geo = {'projection': stfmt.format_gdalprojection(),
           'geotransform': [lon0 - 0.5 * dlon, dlon, 0,
                            lat0 - 0.5 * dlat, 0, dlat]}
    now = datetime.datetime.utcnow()

    meta = {'product_name': None,  # Set later
            'name': None,  # Set later
            'datetime': None,  # Set later
            'time_range': None,  # Set later
            'source_URI': input_path,
            'source_provider': 'WEkEO - Copernicus',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_version': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'parameter': None}  # Set later

    dtimes = netCDF4.num2date(_t, _t_units)

    _vars = {'chlorophyll': (('CHL',), None, None, 'mg.m-3', 'chla_jet'),
             'chlerror': (('CHL_error',), 0.0, 100.0, '%', 'matplotlib_jet')}

    if 'all' == process_var:
        process_var = _vars.keys()

    product_pattern = 'CMEMS_NRT_OI_daily_{}'
    for variable_name in process_var:
        fields, _min, _max, units, cmap = _vars[variable_name]

        meta['product_name'] = product_pattern.format(variable_name)
        meta['parameter'] = variable_name

        for t, dtime in enumerate(dtimes):
            granule_name = '{}_{}'.format(dtime.strftime('%Y%m%d%H%M%S'),
                                          variable_name)
            meta['name'] = granule_name
            meta['datetime'] = stfmt.format_time(dtime)
            meta['time_range'] = ['-12h', '+12h']  # Daily product

            data = build_scalar_data(f_handler, ind_lat, ind_lon, t, fields[0],
                                     _min, _max, units, cmap)

            yield (meta, geo, data, None)
    f_handler.close()


def convert_from_file(input_path, output_path, process_var='all'):
    """"""
    for (meta, geo, data, extra) in read_from_file(input_path, process_var):
        tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                               create_dir=True)
        stfmt.write_geotiff(tiff_fname, meta, geo, data)


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.DEBUG)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)
