#!/usr/bin/env python
# coding=utf-8
"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import re
import numpy
import netCDF4
from datetime import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack


class ProcessingDateNotFound(ValueError):
    """Raised when the processing date cannot be extracted from the input file
    or its path"""
    pass


class InvalidDepth(ValueError):
    """Raised when the depth value passed to the reader is not supported"""
    pass


class FormatNotSupported(TypeError):
    """Raised when the reader cannot extract the expected information from the
    input file"""
    pass


def data_geos(ncfile):
    ucur = ncfile.variables['ugos'][:, ::-1, :]
    vcur = ncfile.variables['vgos'][:, ::-1, :]
    sla = ncfile.variables['sla'][:, ::-1, :]
    adt = ncfile.variables['adt'][:, ::-1, :]
    ulong = ncfile.variables['ugos'].long_name
    vlong = ncfile.variables['vgos'].long_name
    slalong = ncfile.variables['sla'].long_name
    adtlong = ncfile.variables['adt'].long_name

    return (ucur, vcur, sla, adt, ulong, vlong, slalong, adtlong)


def data_total(ncfile, depth):
    ucur = ncfile.variables['uo'][:, depth, ::-1, :]
    vcur = ncfile.variables['vo'][:, depth, ::-1, :]
    ulong = ncfile.variables['uo'].long_name
    vlong = ncfile.variables['vo'].long_name

    return (ucur, vcur, ulong, vlong)


def make_metadata(infile, time_range, product_name, dtime):
    metadata = {}

    bn, _ = os.path.splitext(os.path.basename(infile))
    _dataset_name, _ = bn.rsplit('_', 1)
    dataset_name = '{}_{}'.format(_dataset_name,
                                  dtime.strftime('%Y%m%dT%H%M%SZ'))
    metadata['product_name'] = product_name
    metadata['name'] = dataset_name
    metadata['datetime'] = stfmt.format_time(dtime)
    metadata['time_range'] = time_range
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'CMEMS'
    metadata['processing_center'] = 'CLS'
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(datetime.utcnow())
    metadata['parameter'] = ['current velocity', 'current direction']

    return metadata


def cmems_l4(infile, outdir, time=[0, ], depth=0,
             vmin=0., vmax=5.08, vmin_pal=0., vmax_pal=2.,
             write_netcdf=False):
    """
    """
    # Read/Process data
    ncfile = netCDF4.Dataset(infile, 'r')
    lon = ncfile.variables['longitude'][:]
    lat = ncfile.variables['latitude'][-1:-3:-1]
    vtime = ncfile.variables['time']
    dtime_units = vtime.units
    dtime = netCDF4.num2date(vtime[:], dtime_units)

    # Wrap data in [-180.0, 180[ instead of [0, 360[
    lon = numpy.mod(lon + 180.0, 360.0) - 180.0
    shift = -numpy.where(lon < 0)[0][0]
    lon = lon[shift: shift + 2]

    lon0 = lon[0]
    dlon = lon[1] - lon[0]
    lat0 = lat[0]
    dlat = lat[1] - lat[0]
    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['geotransform'] = [lon0 - dlon/2., dlon, 0,
                                   lat0 - dlat/2., 0, dlat]

    processing_date = None
    if 'date_modified' in ncfile.ncattrs():
        processing_date = ncfile.date_modified
    else:
        bn, _ = os.path.splitext(os.path.basename(infile))
        _, pdate = bn.rsplit('_', 1)
        if pdate.startswith('P'):
            processing_date = '{}-{}-{}T{}:{}:00Z'.format(pdate[1:5],
                                                          pdate[5:7],
                                                          pdate[7:9],
                                                          pdate[10:12],
                                                          pdate[12:14])
        else:
            raise ProcessingDateNotFound

    adt = None
    sla = None
    ucur = None
    vcur = None
    if 'ugos' in ncfile.variables.keys():
        product_name = 'cmems_geos_current'
        time_range = ['-12h', '+12h']

        uname = 'ugos'
        vname = 'vgos'
        slaname = 'sla'
        adtname = 'adt'
        ucur, vcur, sla, adt, ulong, vlong, slalong, adtlong = data_geos(ncfile)

        # Apply longitude shift
        ucur = numpy.roll(ucur, shift, axis=2)
        vcur = numpy.roll(vcur, shift, axis=2)
        sla = numpy.roll(sla, shift, axis=2)
        adt = numpy.roll(adt, shift, axis=2)
    elif 'uo' in ncfile.variables.keys():
        if 0 == depth:
            d_ind = 0
        elif 15 == depth:
            d_ind = 1
        else:
            raise InvalidDepth

        product_name = 'cmems_total_current_{}m'.format(depth)
        time_range = ['-3h', '+3h']

        uname = 'uo'
        vname = 'vo'
        ucur, vcur, ulong, vlong, = data_total(ncfile, d_ind)

        # Apply longitude shift
        ucur = numpy.roll(ucur, shift, axis=2)
        vcur = numpy.roll(vcur, shift, axis=2)
    else:
        logger.error('Unknown format for CMEMS L4 file')
        raise FormatNotSupported
    ncfile.close()

    for itime in range(numpy.shape(ucur)[0]):
        # Construct metadata/geolocation/band(s)
        metadata = make_metadata(infile, time_range, product_name,
                                 dtime[itime])
        metadata['EXPORTED_created'] = processing_date

        ucur_t = + ucur[itime, :, :]
        vcur_t = + vcur[itime, :, :]
        if sla is not None:
            sla_t = + sla[itime, :, :]
        if adt is not None:
            adt_t = + adt[itime, :, :]
        band = []
        mask = ucur_t.mask | vcur_t.mask
        if write_netcdf is False:
            curvel = numpy.sqrt(ucur_t.data**2 + vcur_t.data**2)
            curdir = numpy.mod(numpy.rad2deg(numpy.arctan2(vcur_t.data,
                               ucur_t.data)) + 360., 360.)
            array, offset, scale = pack.ubytes_0_254(curvel, vmin, vmax)
            array[mask] = 255
            colortable = stfmt.format_colortable('matplotlib_jet', vmin=vmin,
                                                 vmax=vmax, vmin_pal=vmin_pal,
                                                 vmax_pal=vmax_pal)
            band.append({'array': array, 'scale': scale, 'offset': offset,
                         'description': 'current velocity', 'unittype': 'm/s',
                         'nodatavalue': 255, 'parameter_range': [vmin, vmax],
                         'colortable': colortable})
            array, offset, scale = pack.ubytes_0_254(curdir, 0, 360)
            array[mask] = 255
            band.append({'array': array, 'scale': scale, 'offset': offset,
                         'description': 'current direction', 'unittype': 'deg',
                         'nodatavalue': 255, 'parameter_range': [0, 360.]})
        # Write geotiff
            tifffile = stfmt.format_tifffilename(outdir, metadata,
                                                 create_dir=True)
            stfmt.write_geotiff(tifffile, metadata, geolocation, band)
        else:
            # u/v -> bands
            vmin = -vmax
            array, offset, scale = pack.ubytes_0_254(ucur_t, vmin, vmax)
            array[mask] = 255
            band.append({'array': array, 'scale': scale, 'offset': offset,
                         'description': 'current u', 'unittype': 'm s-1',
                         'nodatavalue': 255, 'parameter_range': [vmin, vmax]})
            band[0]['name'] = uname
            band[0]['long_name'] = ulong
            array, offset, scale = pack.ubytes_0_254(vcur_t, vmin, vmax)
            array[mask] = 255
            band.append({'array': array, 'scale': scale, 'offset': offset,
                         'description': 'current v', 'unittype': 'm s-1',
                         'nodatavalue': 255, 'parameter_range': [vmin, vmax]})
            band[1]['name'] = vname
            band[1]['long_name'] = vlong
            vmin = numpy.min(sla_t)
            vmax = numpy.max(sla_t)
            array, offset, scale = pack.ubytes_0_254(sla_t, vmin, vmax)
            array[mask] = 255
            band.append({'array': array, 'scale': scale, 'offset': offset,
                         'description': 'current v', 'unittype': 'm s-1',
                         'nodatavalue': 255, 'parameter_range': [vmin, vmax]})
            band[2]['name'] = slaname
            band[2]['long_name'] = slalong
            vmin = numpy.min(adt_t)
            vmax = numpy.max(adt_t)
            array, offset, scale = pack.ubytes_0_254(adt_t, vmin, vmax)
            array[mask] = 255
            band.append({'array': array, 'scale': scale, 'offset': offset,
                         'description': 'current v', 'unittype': 'm s-1',
                         'nodatavalue': 255, 'parameter_range': [vmin, vmax]})
            band[3]['name'] = adtname
            band[3]['long_name'] = adtlong
            # Write
            ncfile = stfmt.format_ncfilename(outdir, metadata, create_dir=True)
            metadata['spatial_resolution'] = min([abs(dlat),
                                                  abs(dlon)]) * 111000.
            dgcps = numpy.round(1. / numpy.abs(numpy.array([dlat,
                                dlon]))).astype('int')
            stfmt.write_netcdf(ncfile, metadata, geolocation, band,
                               'grid_lonlat', dgcps=dgcps)
