# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2019-07-05

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys
import json
import numpy
import errno
import pyproj
import netCDF4
import logging
import argparse
import datetime
import calendar
import dateutil.relativedelta

import syntool_converter.utils.geojson as geojson

logger = logging.getLogger(__name__)


def time_lerp(geod, p0, p1, t):
    """"""
    (t0, lon0, lat0, data0) = p0
    (t1, lon1, lat1, data1) = p1

    time_dist = t1 - t0
    rel_factor = float((t - t0)) / time_dist

    try:
        (fwd, _, geo_dist) = geod.inv(lon0, lat0, lon1, lat1, radians=False)
        (int_lon, int_lat, _) = geod.fwd(lon0, lat0, fwd, rel_factor * geo_dist,
                                         radians=False)
    except:
        print(lon0, lat0, lon1, lat1)
        import sys
        sys.exit(1)

    int_data = {'t': t, 'lon': int_lon, 'lat': int_lat}
    for field_name in data0:
        field_dist = data1[field_name] - data0[field_name]
        int_value = data0[field_name] + rel_factor * field_dist
        int_data[field_name] = int_value
    return int_data


def read_data(input_path, ends_after=None):
    """"""
    f_handler = netCDF4.Dataset(input_path, 'r')

    time_units = f_handler.variables['time'].units

    # List unique identifiers
    _tracks = f_handler.variables['track'][:]
    tracks = list(set(_tracks))

    in_proj = pyproj.Proj(init='epsg:4326')
    out_proj = pyproj.Proj(init='epsg:3857')

    dummy_id = 0
    for eddy_id in tracks:
        eddy_index = numpy.where(_tracks == eddy_id)
        obs_count = len(eddy_index[0])
        if 2 > obs_count:
            logger.warn('Less than two positions for eddy {}.'.format(eddy_id))
            logger.warn('Skipped.')
            continue

        granule_name = 'eddy_{}'.format(eddy_id)
        logger.info('Processing {}'.format(granule_name))
        logger.debug('\tNumber of observations: {}'.format(obs_count))

        _time = f_handler.variables['time'][eddy_index]
        lon = f_handler.variables['longitude'][eddy_index]
        lat = f_handler.variables['latitude'][eddy_index]
        sla_diff = f_handler.variables['amplitude'][eddy_index]
        rotation = f_handler.variables['cyclonic_type'][eddy_index]
        radius = f_handler.variables['speed_radius'][eddy_index]
        speed = f_handler.variables['speed_average'][eddy_index]

        # Shift longitudes from [0.0, 360.0] to [-180.0, 180.0]
        lon = numpy.mod(lon + 180.0, 360.0) - 180.0

        dtimes = netCDF4.num2date(_time, time_units)
        t = [calendar.timegm(x.timetuple()) for x in dtimes]
        granule_data = {'sla_diff': sla_diff,
                        'rotation': rotation,
                        'radius': radius,
                        'speed': speed}

        if (ends_after is not None) and (max(t) <= ends_after):
            continue

        for result in process(eddy_id, t, lat, lon, granule_data):
            yield result


def get_time_slices(_time, mode, _delta):
    """"""
    try:
        start_dt = datetime.datetime.utcfromtimestamp(_time[0])
    except:
        logger.error('No time... {}'.format(_time[:]))
        import sys
        sys.exit(1)
    if 'day' == mode:
        min_dt = datetime.date(start_dt.year, start_dt.month, start_dt.day)
        delta = datetime.timedelta(days=_delta)
        stop_dt = datetime.datetime.utcfromtimestamp(_time[-1])
        end_dt = stop_dt + delta
        end_ts = calendar.timegm(end_dt.timetuple())
        time_fmt = '%Y%m%d'
    elif 'month' == mode:
        min_dt = datetime.date(start_dt.year, start_dt.month, 1)
        delta = dateutil.relativedelta.relativedelta(months=_delta)
        stop_dt = datetime.datetime.utcfromtimestamp(_time[-1])
        end_dt = stop_dt + delta
        end_ts = calendar.timegm(end_dt.timetuple())
        time_fmt = '%Y%m01'
    else:
        raise Exception('{} mode is not supported'.format(mode))

    max_dt = min_dt + delta
    lower_limit = calendar.timegm(min_dt.timetuple())
    upper_limit = calendar.timegm(max_dt.timetuple())

    slices = {}
    slice_start = 0
    while upper_limit <= end_ts:
        last_valid_index = numpy.max(numpy.where(_time < upper_limit))
        slice_stop = last_valid_index + 1
        if slice_start < slice_stop:
            _slice = slice(slice_start, slice_stop)
            slice_label = '{}_{}'.format(min_dt.strftime(time_fmt),
                                         max_dt.strftime(time_fmt))
            slice_range = (lower_limit, upper_limit - 1)
            slices[slice_label] = (slice_range, _slice)
        slice_start = slice_stop
        min_dt = max_dt
        max_dt = max_dt + delta
        lower_limit = upper_limit
        upper_limit = calendar.timegm(max_dt.timetuple())

    return slices

def process(_eddy, eddy_time, eddy_lat, eddy_lon, eddy_data):
    """"""
    # Make sure everything is a NumPy array
    valid_t = numpy.array(eddy_time)
    valid_lat = numpy.array(eddy_lat)
    valid_lon = numpy.array(eddy_lon)
    valid_lon = numpy.mod(valid_lon + 180.0, 360.0) - 180.0
    valid_data = {}
    _last = {'t': None, 'lon': None, 'lat': None}
    for field_name in eddy_data:
        valid_data[field_name] = numpy.array(eddy_data[field_name])
        _last[field_name] = None

    valid_count = len(valid_t)
    geod = pyproj.Geod(ellps='WGS84')
    time_slices = get_time_slices(valid_t, 'month', 24)
    for chunk_label in sorted(time_slices):
        chunk_range, chunk_slice = time_slices[chunk_label]
        granule_name = '{}_{}'.format(_eddy, chunk_label)
        granule_time = [int(x) for x in valid_t[chunk_slice]]
        granule_lon = [float(x) for x in valid_lon[chunk_slice]]
        granule_lat = [float(x) for x in valid_lat[chunk_slice]]
        granule_data = {}
        for field_name in valid_data:
            granule_data[field_name] = [float(x) for x in valid_data[field_name][chunk_slice]]

        if None not in _last.values():
            granule_time.insert(0, _last['t'])
            granule_lon.insert(0, _last['lon'])
            granule_lat.insert(0, _last['lat'])
            for field_name in valid_data:
                granule_data[field_name].insert(0, _last[field_name])

        if chunk_slice.stop < valid_count:
            # Compute midpoint between the last location of this slice and the
            # first of the next slice
            t0 = granule_time[-1]
            lon0 = granule_lon[-1]
            lat0 = granule_lat[-1]
            data0 = {}
            for field_name in valid_data:
                data0[field_name] = granule_data[field_name][-1]

            t1 = valid_t[chunk_slice.stop]  # slice is [start, stop[
            lon1 = valid_lon[chunk_slice.stop]
            lat1 = valid_lat[chunk_slice.stop]
            data1 = {}
            for field_name in valid_data:
                data1[field_name] = valid_data[field_name][chunk_slice.stop]

            p0 = (t0, lon0, lat0, data0)
            p1 = (t1, lon1, lat1, data1)
            _last = time_lerp(geod, p0, p1, chunk_range[1])

            # Add the computed values at the end of current slice
            granule_time.append(_last['t'])
            granule_lon.append(_last['lon'])
            granule_lat.append(_last['lat'])
            for field_name in valid_data:
                granule_data[field_name].append(_last[field_name])

        locs = zip(granule_lon, granule_lat)
        (_geojson, bbox, east, west, idl) = geojson.process_data(granule_time,
                                                                 locs,
                                                                 granule_data)
        _geojson['properties']['granule_name'] = granule_name

        shape_wkt, bbox_wkt, w_bbox, e_bbox = geojson._bbox_shape(idl, west,
                                                                  east)
        eddy_meta = _create_metadata(_geojson, shape_wkt, bbox_wkt,
                                     w_bbox, e_bbox)
        eddy_meta['dataset'] = granule_name
        result = {'geojson': _geojson, 'metadata': eddy_meta}
        yield result


def _create_metadata(_geojson, shape, bbox, w_bbox, e_bbox):
    """ """
    metadata = {'begin_datetime': _geojson['properties']['timeRange'][0],
                'end_datetime': _geojson['properties']['timeRange'][1],
                'output_type': 'TRAJECTORY',
                'min_zoom_level': 0,
                'max_zoom_level': 0,
                'resolutions': [],
                'bbox_str': bbox,
                'shape_str': shape}
    if None not in (w_bbox, e_bbox):
        metadata['w_bbox'] = w_bbox
        metadata['e_bbox'] = e_bbox
    return metadata


def process_eddies_atlas(input_path, output_path, ends_after):
    """"""

    for granule_data in read_data(input_path, ends_after):
        if granule_data is None:
            continue
        geo = granule_data['geojson']
        meta = granule_data['metadata']
        meta['product'] = 'Eddies_Atlas'
        meta['syntool_id'] = '3857_Eddies_Atlas'
        granule_name = geo['properties']['granule_name']
        granule_path = os.path.join(output_path, '3857_Eddies_Atlas',
                                    granule_name)
        try:
            os.makedirs(granule_path)
        except OSError:
            _, e, _ = sys.exc_info()
            if e.errno != errno.EEXIST:
                raise

        geojson_path = os.path.join(granule_path, 'geojson.json')
        with open(geojson_path, 'wt') as f:
            json.dump(geo, f)

        metadata_path = os.path.join(granule_path, 'metadata.json')
        with open(metadata_path, 'wt') as f:
            json.dump(meta, f)


if '__main__' == __name__:
    input_path = sys.argv[1]
    output_path = sys.argv[2]


    # Only 2015 or after
    process_eddies_atlas(input_path, output_path, 1420070399)
