"""
@author <sylvain.herledan@oceandatalab.com>
@date 2018-11-02
"""
import os
import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def read_from_file(input_path, latest_path):
    """"""
    granule_filename = os.path.basename(input_path)
    granule_prefix, _ = os.path.splitext(granule_filename)

    latest_dt = None
    if latest_path is not None:
        with open(latest_path, 'rt') as f:
            latest_str = f.read().decode('utf-8').strip()
        latest_dt = datetime.datetime.strptime(latest_str,
                                               '%Y-%m-%dT%H:%M:%SZ')

    f_handler = netCDF4.Dataset(input_path, 'r')
    _t_units = f_handler.variables['time'].units
    _t = f_handler.variables['time'][:]
    lon = f_handler.variables['longitude'][:]
    lat = f_handler.variables['latitude'][:]
    if 'ewct' in f_handler.variables:
        u_cur = f_handler.variables['ewct'][:]
    else:
        u_cur = f_handler.variables['u'][:]

    if 'nsct' in f_handler.variables:
        v_cur = f_handler.variables['nsct'][:]
    else:
        v_cur = f_handler.variables['v'][:]
    quality = f_handler.variables['flag'][:]
    f_handler.close()

    lon0, dlon = lon[0], numpy.mean(lon[1:] - lon[:-1])
    lat0, dlat = lat[0], numpy.mean(lat[1:] - lat[:-1])
    geo = {'projection': stfmt.format_gdalprojection(),
           'geotransform': [lon0, dlon, 0, lat0, 0, dlat]}

    now = datetime.datetime.utcnow()
    dtimes = netCDF4.num2date(_t, _t_units)

    for i, dtime in enumerate(dtimes):
        if (latest_dt is not None) and (dtime <= latest_dt):
            # Already processed
            logger.info('Skipped: {}'.format(dtime.strftime('%H%M%S')))
            continue

        start_dt = dtime
        stop_dt = start_dt + datetime.timedelta(seconds=1)
        _dtime, time_range = stfmt.format_time_and_range(start_dt, stop_dt,
                                                        units='s')

        granule_name = '{}_{}'.format(granule_prefix, dtime.strftime('%H%M%S'))
        meta = {'product_name': 'SHOM_HF_radar_current',
                'name': granule_name,
                'datetime': _dtime,
                'time_range': time_range,
                'source_URI': input_path,
                'source_provider': 'SHOM',
                'processing_center': '',
                'conversion_software': 'Syntool',
                'conversion_software': '0.0.0',
                'conversion_datetime': stfmt.format_time(now),
                'parameter': 'current'}

        u_data = u_cur[i]
        v_data = v_cur[i]
        mask = (u_data.mask | v_data.mask | (quality[i] == 0))
        masked_ind = numpy.where(mask)
        u_data[masked_ind] = 1.0  # avoid runtime warnings, masked later anyway
        v_data[masked_ind] = 1.0  # avoid runtime warning, masked later anyway

        data = []

        vmin = 0.0
        vmax = 2.5
        cur = numpy.hypot(v_data, u_data)

        colortable = stfmt.format_colortable('matplotlib_jet',
                                             vmin=vmin, vmax=vmax,
                                             vmin_pal=vmin, vmax_pal=vmax)

        array, offset, scale = pack.ubytes_0_254(cur, vmin, vmax)
        if numpy.any(mask):
            array[numpy.where(mask)] = 255
        data.append({'array': array,
                     'scale': scale,
                     'offset': offset,
                     'description': 'current speed',
                     'unit': 'm.s-1',
                     'nodatavalue': 255,
                     'parameter_range': [vmin, vmax],
                     'colortable': colortable})

        vmin = 0
        vmax = 360
        cur_dir = numpy.rad2deg(numpy.arctan2(v_data, u_data))
        cur_dir = numpy.mod(cur_dir + 360.0, 360.0)

        array, offset, scale = pack.ubytes_0_254(cur_dir, vmin, vmax)
        if numpy.any(mask):
            array[numpy.where(mask)] = 255
        data.append({'array': array,
                     'scale': scale,
                     'offset': offset,
                     'description': 'current direction',
                     'unit': 'degrees',
                     'nodatavalue': 255,
                     'parameter_range': [vmin, vmax]})
        yield (meta, geo, data, None)

    max_dt = max(dtimes)
    if (latest_dt is not None) and (max_dt > latest_dt):
        with open(latest_path, 'wt') as f:
            f.write(max_dt.strftime('%Y-%m-%dT%H:%M:%SZ'))


def convert_from_file(input_path, output_path, latest_path=None):
    """"""
    for (meta, geo, data, extra) in read_from_file(input_path, latest_path):
        tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                               create_dir=True)
        stfmt.write_geotiff(tiff_fname, meta, geo, data)


if '__main__' == __name__:
    import os
    import sys

    # Main logger
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.INFO)

    # TODO: use argparse
    input_path = sys.argv[1]
    output_path = sys.argv[2]
    latest_path = sys.argv[3]

    if not os.path.exists(latest_path):
        logger.error('{} not found'.format(latest_path))
        sys.exit(1)
    convert_from_file(input_path, output_path, latest_path)
