# -*- encoding: utf-8 -*-
# vim: set ts=4:sts=4:sw=4:noet

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""
@author <sylvain.herledan@oceandatalab.com>
@date 2019-01-24
"""

import os
import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt

logger = logging.getLogger(__name__)


VARIABLES = {'temperature': (('thetao',), -2.1, 36.0, '°C',
                             'cerbere_medspiration',),
             'salinity': (('so',), 32.0, 38.0, '',
                          'matplotlib_jet',),
             'sea_water_velocity': (('uo', 'vo'), 0.0, 2.0, 'm.s-1',
                                    'matplotlib_jet',),
             'sea_ice_velocity': (('usi', 'vsi'), 0.0, 1.0, 'm.s-1',
                                  'matplotlib_jet',)}


class UnsupportedVariables(ValueError):
    """Error triggered when a conversion is requested for one or several
    variables that are not supported by this reader."""

    def __init__(self, variables):
        self.variables = variables


class FieldsCountError(ValueError):
    """Error triggered when the variable is not defined by 1 or 2 fields."""
    def __init__(self, fields):
        """"""
        self.fields = fields


def build_scalar_data(f_handler, vnames, _min, _max, units, cmap, t, d):
    """"""
    dim_count = len(f_handler.variables[vnames[0]].dimensions)
    if 3 == dim_count:
        v_data = f_handler.variables[vnames[0]][t, ::-1, :]
    elif 4 == dim_count:
        v_data = f_handler.variables[vnames[0]][t, d, ::-1, :]
    v_data = numpy.ma.masked_invalid(v_data)
    mask = numpy.ma.getmaskarray(v_data)
    masked_ind = numpy.where(mask)

    # Remove mask and set dummy value to prevent issues with numpy
    # It has no side-effect as mask is saved and reapplied later
    v_data._sharedmask = False
    v_data.mask = False
    v_data[masked_ind] = 0

    valid_ind = numpy.where(~numpy.ma.getmaskarray(v_data))
    logger.debug('{} min: {}'.format(vnames, numpy.nanmin(v_data[valid_ind])))
    logger.debug('{} max: {}'.format(vnames, numpy.nanmax(v_data[valid_ind])))

    colortable = stfmt.format_colortable(cmap,
                                         vmin=_min, vmax=_max,
                                         vmin_pal=_min, vmax_pal=_max)

    array, offset, scale = pack.ubytes_0_254(v_data, _min, _max)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data = []
    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': vnames[0],
                 'unit': units,
                 'nodatavalue': 255,
                 'parameter_range': [_min, _max],
                 'colortable': colortable})
    return data


def build_directional_data(f_handler, vnames, _min, _max, units, cmap, t, d):
    """"""
    u_dim_count = len(f_handler.variables[vnames[0]].dimensions)
    if 3 == u_dim_count:
        u_data = f_handler.variables[vnames[0]][t, ::-1, :]
    elif 4 == u_dim_count:
        u_data = f_handler.variables[vnames[0]][t, d, ::-1, :]
    u_data = numpy.ma.masked_invalid(u_data)

    v_dim_count = len(f_handler.variables[vnames[1]].dimensions)
    if 3 == v_dim_count:
        v_data = f_handler.variables[vnames[1]][t, ::-1, :]
    elif 4 == v_dim_count:
        v_data = f_handler.variables[vnames[1]][t, d, ::-1, :]
    v_data = numpy.ma.masked_invalid(v_data)

    mask = (numpy.ma.getmaskarray(u_data) | numpy.ma.getmaskarray(v_data))
    masked_ind = numpy.where(mask)

    u_data[masked_ind] = 0
    v_data[masked_ind] = 0

    norm = numpy.hypot(u_data, v_data)
    _direction = numpy.rad2deg(numpy.arctan2(v_data, u_data))
    direction = numpy.mod(_direction, 360.0)

    valid_ind = numpy.where(~mask)
    logger.debug('{} min: {}'.format(vnames, numpy.nanmin(norm[valid_ind])))
    logger.debug('{} max: {}'.format(vnames, numpy.nanmax(norm[valid_ind])))
    colortable = stfmt.format_colortable(cmap,
                                         vmin=_min, vmax=_max,
                                         vmin_pal=_min, vmax_pal=_max)

    data = []

    array, offset, scale = pack.ubytes_0_254(norm, _min, _max)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'amplitude',
                 'unit': units,
                 'nodatavalue': 255,
                 'parameter_range': [_min, _max],
                 'colortable': colortable})

    array, offset, scale = pack.ubytes_0_254(direction, 0, 360)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'direction',
                 'unit': 'degrees',
                 'nodatavalue': 255,
                 'parameter_range': [0, 360]})
    return data


def model_001_024(input_path, f_handler, surface_only, process_var):
    """"""
    granule_filename = os.path.basename(input_path)
    granule_prefix, _ = os.path.splitext(granule_filename)

    # Read/Process data
    logger.info('Read/Process data')
    if 'all' == process_var:
        process_var = VARIABLES.keys()

    unsupported_vars = [x for x in process_var if x not in VARIABLES.keys()]
    if 0 < len(unsupported_vars):
        raise UnsupportedVariables(unsupported_vars)

    time = f_handler.variables['time']
    source_provider = f_handler.source
    lon = f_handler.variables['longitude'][:]
    lat = f_handler.variables['latitude'][::-1]

    # Construct metadata/geolocation/band(s)
    now = datetime.datetime.utcnow()
    metadata = {}
    metadata['time_range'] = ['-12h', '+12h']  # daily product
    metadata['source_provider'] = source_provider
    metadata['processing_center'] = source_provider
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = None

    # The coordinates found in the input file already correspond to the
    # top-left corner of the pixel => the usual shift by (dx/2, dy/2) is not
    # required.
    x0, dx, y0, dy = lon[0], lon[1] - lon[0], lat[0], lat[1] - lat[0]
    geolocation = {}
    geolocation['projection'] = 'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]'  # noqa
    geolocation['geotransform'] = [x0, dx, 0, y0, 0, dy]

    _depths = [0.0]
    if ((surface_only is False) and ('depth' in f_handler.variables)):
        _depths = f_handler.variables['depth'][:]

    for var_name in process_var:
        metadata['parameter'] = var_name
        fields, _min, _max, units, cmap = VARIABLES[var_name]

        if 1 == len(fields):
            extract_method = build_scalar_data
        elif 2 == len(fields):
            extract_method = build_directional_data
        else:
            raise FieldsCountError(fields)

        if 'depth' in f_handler.variables[fields[0]].dimensions:
            depths = _depths
        else:
            depths = [0.0]

        for d_ind in range(0, len(depths)):
            depth = int(numpy.round(depths[d_ind]))

            # Update metadata
            product_name = '001_024_{}_{:04d}'.format(var_name, depth)
            product_name = product_name[:51]  # truncate if too long (51 max)
            metadata['product_name'] = product_name

            for t_ind in range(0, len(time)):
                # Update metadata
                dtime = netCDF4.num2date(time[t_ind], units=time.units)
                dtime_str = dtime.strftime('%Y%m%d%H%M%S')
                granule_name = '{}_{}_{:04d}_{}'.format(granule_prefix,
                                                        var_name, depth,
                                                        dtime_str)
                metadata['datetime'] = stfmt.format_time(dtime)
                metadata['name'] = granule_name

                data = extract_method(f_handler, fields, _min, _max, units,
                                      cmap, t_ind, d_ind)

                yield (metadata, geolocation, data, None,)


def convert_from_file(input_path, output_path, surface_only=True,
                      process_var='all'):
    """Extract data from input  and save the results as GeoTIFF files."""
    f_handler = netCDF4.Dataset(input_path, 'r')
    for meta, geo, data, extra in model_001_024(input_path, f_handler,
                                                surface_only, process_var):
        meta['source_URI'] = input_path
        tifffile = stfmt.format_tifffilename(output_path, meta,
                                             create_dir=True)
        stfmt.write_geotiff(tifffile, meta, geo, data)
    f_handler.close()


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler(stream=sys.stderr)
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.DEBUG)

    _input_path = sys.argv[1]
    _output_path = sys.argv[2]
    input_path = os.path.abspath(_input_path)
    output_path = os.path.abspath(_output_path)
    convert_from_file(input_path, output_path)
