# -*- encoding: utf-8 -*-
# vim: set ts=4:sts=4:sw=4:noet
#
# Copyright (C) 2014-2021 OceanDataLab
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
@author <sylvain.herledan@oceandatalab.com>
@date 2019-06-03
"""

import os
import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt

logger = logging.getLogger(__name__)


def _process_3d(f_handler, t, lat_slice, lon_slice, hs_name, dir_name,
                period_name, min_swh_threshold=0.0):
    """"""
    hs_data = f_handler.variables[hs_name][(t, lat_slice, lon_slice)]
    hs_data = hs_data[::-1, :]
    hs_data = numpy.ma.masked_invalid(hs_data)

    dir_data = f_handler.variables[dir_name][(t, lat_slice, lon_slice)]
    dir_data = dir_data[::-1, :]
    dir_data = numpy.ma.masked_invalid(dir_data)
    # Convert from meteo (direction from) to oceano (direction to) convention
    dir_data = numpy.mod(180.0 + dir_data, 360.0)
    # Convert from clockwise from north to counter-clockwise from east
    dir_data = numpy.mod(90.0 - dir_data, 360.0)

    period_data = f_handler.variables[period_name][(t, lat_slice, lon_slice)]
    period_data = period_data[::-1, :]
    period_data = numpy.ma.masked_invalid(period_data)

    mask = (numpy.ma.getmaskarray(hs_data) |
            numpy.ma.getmaskarray(dir_data) |
            numpy.ma.getmaskarray(period_data) |
            (hs_data < min_swh_threshold))
    if numpy.all(mask):
        logger.info('No valid data for ({}, {}, {})'.format(hs_name, dir_name,
                                                            period_name))
        return None

    masked_ind = numpy.where(mask)

    hs_data[masked_ind] = 0
    dir_data[masked_ind] = 0
    period_data[masked_ind] = 0

    data = []

    # Hs
    array, offset, scale = pack.ubytes_0_254(hs_data, 0.0, 8.0)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'amplitude',
                 'unit': 'm',
                 'nodatavalue': 255,
                 'parameter_range': [0.0, 8.0]})

    # Period
    array, offset, scale = pack.ubytes_0_254(period_data, 0.0, 25.4)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'period',
                 'unit': 's',
                 'nodatavalue': 255,
                 'parameter_range': [0, 25.4]})

    # Direction
    array, offset, scale = pack.ubytes_0_254(dir_data, 0, 360)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'direction',
                 'unit': 'degrees',
                 'nodatavalue': 255,
                 'parameter_range': [0, 360]})

    return data


def process_swh(f_handler, t, lat_slice, lon_slice):
    """"""
    v_data = f_handler.variables['VHM0'][(t, lat_slice, lon_slice)]
    v_data = v_data[::-1, :]
    v_data = numpy.ma.masked_invalid(v_data)
    mask = numpy.ma.getmaskarray(v_data)
    masked_ind = numpy.where(mask)

    # Remove mask and set dummy value to prevent issues with numpy
    # It has no side-effect as mask is saved and reapplied later
    v_data._sharedmask = False
    v_data.mask = False
    v_data[masked_ind] = 0

    _min = 0.0
    _max = 8.0
    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=_min, vmax=_max,
                                         vmin_pal=_min, vmax_pal=_max)

    array, offset, scale = pack.ubytes_0_254(v_data, _min, _max)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data = []
    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'significant wave height',
                 'unit': 'm',
                 'nodatavalue': 255,
                 'parameter_range': [_min, _max],
                 'colortable': colortable})
    return data


def process_stokes_drift(f_handler, t, lat_slice, lon_slice):
    """"""
    u_data = f_handler.variables['VSDX'][(t, lat_slice, lon_slice)]
    u_data = u_data[::-1, :]
    u_data = numpy.ma.masked_invalid(u_data)

    v_data = f_handler.variables['VSDY'][(t, lat_slice, lon_slice)]
    v_data = v_data[::-1, :]
    v_data = numpy.ma.masked_invalid(v_data)

    mask = (numpy.ma.getmaskarray(u_data) | numpy.ma.getmaskarray(v_data))
    masked_ind = numpy.where(mask)

    u_data[masked_ind] = 0
    v_data[masked_ind] = 0

    norm = numpy.hypot(u_data, v_data)
    _direction = numpy.rad2deg(numpy.arctan2(v_data, u_data))
    direction = numpy.mod(_direction, 360.0)

    _min = 0.0
    _max = 2.5
    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=_min, vmax=_max,
                                         vmin_pal=_min, vmax_pal=_max)

    data = []

    array, offset, scale = pack.ubytes_0_254(norm, _min, _max)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'amplitude',
                 'unit': 'm.s-1',
                 'nodatavalue': 255,
                 'parameter_range': [_min, _max],
                 'colortable': colortable})

    array, offset, scale = pack.ubytes_0_254(direction, 0, 360)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'direction',
                 'unit': 'degrees',
                 'nodatavalue': 255,
                 'parameter_range': [0, 360]})
    return data


def process_swell1(f_handler, t, lat_slice, lon_slice):
    """"""
    return _process_3d(f_handler, t, lat_slice, lon_slice, 'VHM0_SW1',
                       'VMDR_SW1', 'VTM01_SW1', 0.0)


def process_swell2(f_handler, t, lat_slice, lon_slice):
    """"""
    return _process_3d(f_handler, t, lat_slice, lon_slice, 'VHM0_SW2',
                       'VMDR_SW2', 'VTM01_SW2', 0.5)


def process_wind_waves(f_handler, t, lat_slice, lon_slice):
    """"""
    return _process_3d(f_handler, t, lat_slice, lon_slice, 'VHM0_WW',
                       'VMDR_WW', 'VTM01_WW', 0.1)


def cmems_001_027(input_path, f_handler, lat_crop=85.4):
    """"""
    granule_filename = os.path.basename(input_path)
    granule_prefix, _ = os.path.splitext(granule_filename)

    # Read/Process data
    logger.info('Read/Process data')
    time = f_handler.variables['time']
    source_provider = f_handler.producer
    lat = f_handler.variables['latitude'][::-1]

    # Construct metadata/geolocation/band(s)
    now = datetime.datetime.utcnow()
    metadata = {}
    metadata['time_range'] = ['-5400s', '+5399s']  # 3-hourly product
    metadata['source_provider'] = source_provider
    metadata['processing_center'] = source_provider
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = None

    # Crop latitudes
    valid_lat_ind = numpy.where(numpy.abs(lat) <= lat_crop)
    lat_slice = slice(numpy.min(valid_lat_ind), numpy.max(valid_lat_ind) + 1)

    lon_slice = slice(None, None)  # take all longitudes

    # Coordinates appear to be the top-left corner of each pixel => offset by
    # half-pixels does not seem to be required
    geolocation = {}
    pref = 'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"]],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",0.0174532925199433,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG","4326"]]'  # noqa
    geolocation['projection'] = pref
    dx = f_handler.geospatial_lon_step
    dy = -1 * f_handler.geospatial_lat_step  # latitudes are read backwards
    geolocation['geotransform'] = [-180.0, dx, 0,
                                   lat[0], 0, dy]

    # Update metadata
    for t_ind in range(0, len(time)):
        # Update metadata
        dtime = netCDF4.num2date(time[t_ind], units=time.units)
        dtime_str = dtime.strftime('%Y%m%d%H%M%S')
        metadata['datetime'] = stfmt.format_time(dtime)

        # SWH
        var_name = 'swh'
        granule_name = '{}_mfwamglocep_{}'.format(dtime_str, var_name)
        metadata['name'] = granule_name
        product_name = '001_027_{}'.format(var_name)
        product_name = product_name[:51]  # truncate if too long (51 max)
        metadata['product_name'] = product_name
        data = process_swh(f_handler, t_ind, lat_slice, lon_slice)
        yield (metadata, geolocation, data, None,)

        # Swell 1
        var_name = 'swell1'
        granule_name = '{}_mfwamglocep_{}'.format(dtime_str, var_name)
        metadata['name'] = granule_name
        product_name = '001_027_{}'.format(var_name)
        product_name = product_name[:51]  # truncate if too long (51 max)
        metadata['product_name'] = product_name
        data = process_swell1(f_handler, t_ind, lat_slice, lon_slice)
        if data is not None:
            yield (metadata, geolocation, data, None,)
        else:
            logger.info('No valid swell1')

        # Swell 2
        var_name = 'swell2'
        granule_name = '{}_mfwamglocep_{}'.format(dtime_str, var_name)
        metadata['name'] = granule_name
        product_name = '001_027_{}'.format(var_name)
        product_name = product_name[:51]  # truncate if too long (51 max)
        metadata['product_name'] = product_name
        data = process_swell2(f_handler, t_ind, lat_slice, lon_slice)
        if data is not None:
            yield (metadata, geolocation, data, None,)
        else:
            logger.info('No valid swell 2')

        # Wind waves
        var_name = 'windwaves'
        granule_name = '{}_mfwamglocep_{}'.format(dtime_str, var_name)
        metadata['name'] = granule_name
        product_name = '001_027_{}'.format(var_name)
        product_name = product_name[:51]  # truncate if too long (51 max)
        metadata['product_name'] = product_name
        data = process_wind_waves(f_handler, t_ind, lat_slice, lon_slice)
        if data is not None:
            yield (metadata, geolocation, data, None,)
        else:
            logger.info('No valid wind waves')

        # Stokes drift
        var_name = 'stokes'
        granule_name = '{}_mfwamglocep_{}'.format(dtime_str, var_name)
        metadata['name'] = granule_name
        product_name = '001_027_{}'.format(var_name)
        product_name = product_name[:51]  # truncate if too long (51 max)
        metadata['product_name'] = product_name
        data = process_stokes_drift(f_handler, t_ind, lat_slice, lon_slice)
        yield (metadata, geolocation, data, None,)


def convert_from_file(input_path, output_path):
    """Extract data from input  and save the results as GeoTIFF files."""
    f_handler = netCDF4.Dataset(input_path, 'r')
    for meta, geo, data, extra in cmems_001_027(input_path, f_handler):
        meta['source_URI'] = input_path

        # Add extra metadata
        meta['EXPORTED_type'] = f_handler.product_type
        meta['EXPORTED_creation time'] = f_handler.date_created
        meta['EXPORTED_run time'] = f_handler.product_ref_date

        tifffile = stfmt.format_tifffilename(output_path, meta,
                                             create_dir=True)
        # Write geotiff
        if 2 < len(data):
            # 3D data have no colortable
            stfmt.write_geotiff(tifffile, meta, geo, data,
                                drv_opts=['PHOTOMETRIC=MINISBLACK'])
        else:
            stfmt.write_geotiff(tifffile, meta, geo, data)
    f_handler.close()


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler(stream=sys.stderr)
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.DEBUG)

    _input_path = sys.argv[1]
    _output_path = sys.argv[2]
    input_path = os.path.abspath(_input_path)
    output_path = os.path.abspath(_output_path)
    convert_from_file(input_path, output_path)
