# -*- encoding: utf-8 -*-

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-10-22
"""

import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack
try:
    from osgeo import gdal
    from osgeo import ogr
except ImportError:
    import gdal
    import ogr

logger = logging.getLogger(__name__)


def compute_landmask(xsize, ysize, geotf, landmaskpath):
    """
    """
    # Set grid georeferencing
    srs = stfmt.epsg2srs(4326)
    # Get landmask layer (polygons) and set spatial filter
    datasource = ogr.Open(landmaskpath)
    layer = datasource.GetLayer(0)
    ring = ogr.Geometry(ogr.wkbLinearRing)
    ring.AddPoint(geotf[0] + geotf[1] * 0, geotf[3] + geotf[5] * 0)
    ring.AddPoint(geotf[0] + geotf[1] * xsize, geotf[3] + geotf[5] * 0)
    ring.AddPoint(geotf[0] + geotf[1] * xsize, geotf[3] + geotf[5] * ysize)
    ring.AddPoint(geotf[0] + geotf[1] * 0, geotf[3] + geotf[5] * ysize)
    ring.AddPoint(geotf[0] + geotf[1] * 0, geotf[3] + geotf[5] * 0)
    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(ring)
    poly.AssignSpatialReference(srs)
    poly.TransformTo(layer.GetSpatialRef())
    env = poly.GetEnvelope()
    layer.SetSpatialFilterRect(env[0], env[2], env[1], env[3])
    # Rasterize
    driver = gdal.GetDriverByName('MEM')
    dataset = driver.Create('', xsize, ysize, 1, gdal.GDT_Byte)
    dataset.SetProjection(srs.ExportToWkt())
    dataset.SetGeoTransform(geotf)
    gdal.RasterizeLayer(dataset, [1], layer, burn_values=[1],
                        options=['ALL_TOUCHED=TRUE'])
    landmask = dataset.GetRasterBand(1).ReadAsArray().astype('bool')
    return landmask


def build_scalar_data(f_handler, ind_lat, ind_lon, landmask, t, vname, _min,
                      _max, units, cmap):
    """"""
    v_data = f_handler.variables[vname][t, ind_lat, ind_lon]
    v_data = numpy.ma.masked_invalid(v_data)
    mask = (landmask | numpy.ma.getmaskarray(v_data))

    if 'tcc' == vname:
        masked_ind = numpy.where(mask)
        # increase contrast where cloud coverage is important 
        v_data[masked_ind] = 0
        v_data = numpy.log(101 - 100 * v_data)
        _min = 0.0  # log(1)
        _max = numpy.log(101)
    elif 'tp' == vname:
        mask = (mask | (1e-4 >= v_data))  # mask below 0.1 mm.h-1
        masked_ind = numpy.where(mask)
        v_data[masked_ind] = 0
        v_data = 1000 * v_data  # convert from m.h-1 to mm.h-1
        _min = 1  # mm.h-1
        _max = 5  # mm.h-1
    else:
        masked_ind = numpy.where(mask)

    # Remove mask and set dummy value to prevent issues with numpy
    # It has no side-effect as mask is saved and reapplied later
    v_data._sharedmask = False
    v_data.mask = False
    v_data[masked_ind] = 0

    valid_ind = numpy.where(~numpy.ma.getmaskarray(v_data))
    logger.debug('{} min: {}'.format(vname, numpy.nanmin(v_data[valid_ind])))
    logger.debug('{} max: {}'.format(vname, numpy.nanmax(v_data[valid_ind])))

    colortable = stfmt.format_colortable(cmap,
                                         vmin=_min, vmax=_max,
                                         vmin_pal=_min, vmax_pal=_max)

    array, offset, scale = pack.ubytes_0_254(v_data, _min, _max)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data = []
    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': vname,
                 'unit': units,
                 'nodatavalue': 255,
                 'parameter_range': [_min, _max],
                 'colortable': colortable})
    return data


def build_directional_data(f_handler, ind_lat, ind_lon, landmask, t, vnames,
                           _min, _max, units, cmap):
    """"""
    u_data = f_handler.variables[vnames[0]][t, ind_lat, ind_lon]
    u_data = numpy.ma.masked_invalid(u_data)

    v_data = f_handler.variables[vnames[1]][t, ind_lat, ind_lon]
    v_data = numpy.ma.masked_invalid(v_data)

    mask = (landmask |
            numpy.ma.getmaskarray(u_data) | numpy.ma.getmaskarray(v_data))
    masked_ind = numpy.where(mask)

    u_data[masked_ind] = 0
    v_data[masked_ind] = 0

    norm = numpy.hypot(u_data, v_data)
    _direction = numpy.rad2deg(numpy.arctan2(v_data, u_data))
    direction = numpy.mod(_direction, 360.0)

    valid_ind = numpy.where(~mask)
    logger.debug('{} min: {}'.format(vnames, numpy.nanmin(norm[valid_ind])))
    logger.debug('{} max: {}'.format(vnames, numpy.nanmax(norm[valid_ind])))
    colortable = stfmt.format_colortable(cmap,
                                         vmin=_min, vmax=_max,
                                         vmin_pal=_min, vmax_pal=_max)

    data = []

    array, offset, scale = pack.ubytes_0_254(norm, _min, _max)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'amplitude',
                 'unit': units,
                 'nodatavalue': 255,
                 'parameter_range': [_min, _max],
                 'colortable': colortable})

    array, offset, scale = pack.ubytes_0_254(direction, 0, 360)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'direction',
                 'unit': 'degrees',
                 'nodatavalue': 255,
                 'parameter_range': [0, 360]})
    return data


def read_from_file(input_path, landmask_path, pressure_level=1000):
    """"""
    f_handler = netCDF4.Dataset(input_path, 'r')
    lon = f_handler.variables['longitude'][:]
    lat = f_handler.variables['latitude'][:]
    _t = f_handler.variables['time'][:]
    _t_units = f_handler.variables['time'].units

    # Longitudes are provided in [0, 360] must the landmask is defined with
    # [-180, 180] so shift longitudes to match the landmask.
    lon[lon > 180.0] = lon[lon > 180.0] - 360.0
    ind_lon = numpy.argsort(lon)
    lon = lon[ind_lon]

    ind_lat = numpy.where(numpy.abs(lat) < 86)[0]
    lat = lat[ind_lat]

    lon0, dlon = lon[0], numpy.mean(lon[1:] - lon[:-1])
    lat0, dlat = lat[0], numpy.mean(lat[1:] - lat[:-1])
    geo = {'projection': stfmt.format_gdalprojection(),
           'geotransform': [lon0 - 0.5 * dlon, dlon, 0,
                            lat0 - 0.5 * dlat, 0, dlat]}
    landmask = compute_landmask(numpy.shape(lon)[0],
                                numpy.shape(lat)[0],
                                geo['geotransform'],
                                landmask_path)
    now = datetime.datetime.utcnow()

    meta = {'product_name': None,  # Set later
            'name': None,  # Set later
            'datetime': None, # Set later
            'time_range': None, # Set later
            'source_URI': input_path,
            'source_provider': 'WEkEO - Copernicus',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_software': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'parameter': None} # Set later

    dtimes = netCDF4.num2date(_t, _t_units)

    # 'vertical_velocity': (('w',), -0.5, 0.5, 'm.s-1'),
    variables = {'clouds': (('tcc',), None, None, '%', 'matplotlib_bone_r'),
                 'wind': (('u10', 'v10'), 0, 25.4, 'm.s-1', 'noaa_wind'),
                 'rain': (('tp',), None, None, 'm', 'matplotlib_viridis')}


    product_pattern = 'ERA5_hourly_{}'
    for variable_name in variables:
        fields, _min, _max, units, cmap = variables[variable_name]

        meta['product_name'] = product_pattern.format(variable_name)
        meta['parameter'] = variable_name

        for t, dtime in enumerate(dtimes):
            granule_name = '{}_{}'.format(dtime.strftime('%Y%m%d%H%M%S'),
                                          variable_name)
            meta['name'] = granule_name
            meta['datetime'] = stfmt.format_time(dtime)
            meta['time_range'] = ['-30m', '+30m']  # Hourly product

            if 1 == len(fields):
                data = build_scalar_data(f_handler, ind_lat, ind_lon, landmask,
                                         t, fields[0], _min, _max, units, cmap)
            elif 2 == len(fields):
                data = build_directional_data(f_handler, ind_lat, ind_lon,
                                              landmask, t, fields, _min, _max,
                                              units, cmap)

            yield (meta, geo, data, None)
    f_handler.close()


def convert_from_file(input_path, output_path, landmaskpath):
    """"""
    for (meta, geo, data, extra) in read_from_file(input_path, landmaskpath):
        tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                               create_dir=True)
        stfmt.write_geotiff(tiff_fname, meta, geo, data)


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARN)

    input_path = sys.argv[1]
    output_path = sys.argv[2]
    landmask_path = sys.argv[3]

    convert_from_file(input_path, output_path, landmask_path)
