# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-10-22

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import struct
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def read_from_file(input_path):
    """"""
    granule_fname = os.path.basename(input_path)
    _granule_name = granule_fname  # decompressed binary data have no extension
    _, granule_date = _granule_name.rsplit('_', 1)
    _, version, process, _ = _granule_name.split('_', 3)

    with open(input_path, 'rb') as f:
        file_content = f.read()

    lon0 = -179.875
    dlon = 0.25
    lat0 = -59.875
    dlat = 0.25

    now = datetime.datetime.utcnow()

    product_name = 'CMORPH_{}_{}'.format(version, process)
    meta = {'product_name': product_name,
            'name': None,  # Set later
            'datetime': None, # Set later
            'time_range': None, # Set later
            'source_URI': input_path,
            'source_provider': 'NOAA',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_software': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'parameter': 'precipitation'}

    dtime_base = datetime.datetime.strptime(granule_date, '%Y%m%d')
    geo = {'projection': stfmt.format_gdalprojection(),
           'geotransform': [lon0, dlon, 0, lat0, 0, dlat]}

    fmt = '<{}f'.format(1440*480*8)
    _all_data = struct.unpack(fmt, file_content)
    all_data = numpy.reshape(_all_data, (8, 480, 1440,))
    all_data = numpy.roll(all_data, 720, axis=2)

    vmin = 0.0
    vmax = 75
    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin, vmax_pal=vmax)
    for i in range(0, 8):
        hour = 3 * i
        granule_name = '{}_{:02d}'.format(_granule_name, hour)

        start_dt = dtime_base + datetime.timedelta(hours=hour)
        stop_dt = start_dt + datetime.timedelta(hours=3)
        dtime, time_range = stfmt.format_time_and_range(start_dt, stop_dt,
                                                        units='m')
        meta['name'] = granule_name
        meta['datetime'] = dtime
        meta['time_range'] = time_range

        _data = all_data[i, :, :]
        mask_ind = numpy.where(_data <= 0.0)  # nodata is -999
        v_data = _data * 3  # input contains 3-hourly rate, we want the amount

        array, offset, scale = pack.ubytes_0_254(v_data, vmin, vmax)
        if numpy.any(mask_ind):
            array[mask_ind] = 255

        data = []
        data.append({'array': array,
                     'scale': scale,
                     'offset': offset,
                     'description': 'precipitation',
                     'unit': 'mm',
                     'nodatavalue': 255,
                     'parameter_range': [vmin, vmax],
                     'colortable': colortable})

        yield (meta, geo, data, {})


def convert_from_file(input_path, output_path):
    """"""
    for (meta, geo, data, extra) in read_from_file(input_path):
        tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                               create_dir=True)
        stfmt.write_geotiff(tiff_fname, meta, geo, data)


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARN)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)

