# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-10-22

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy
import netCDF4
import logging
import datetime
import dateutil.relativedelta
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)

MONTHS= {'January': 1,
         'February': 2,
         'March': 3,
         'April': 4,
         'May': 5,
         'June': 6,
         'July': 7,
         'August': 8,
         'September': 9,
         'October': 10,
         'November': 11,
         'December': 12}

def read_from_file(input_path, variables, max_depth=0.0):
    """"""
    f_handler = netCDF4.Dataset(input_path, 'r')
    lon = f_handler.variables['LONGITUDE'][:]
    lat = f_handler.variables['LATITUDE'][:]
    depths = f_handler.variables['LEVEL'][:]

    lon = numpy.mod(lon + 180.0, 360.0) - 180.0
    lon_ind = numpy.argsort(lon)
    lon = lon[lon_ind]

    lon0, dlon = lon[0], numpy.mean(lon[1:] - lon[:-1])
    lat0, dlat = lat[0], numpy.mean(lat[1:] - lat[:-1])
    geo = {'projection': stfmt.format_gdalprojection(),
           'geotransform': [lon0 - 0.5 * dlon, dlon, 0, lat0 - 0.5 * dlat, 0, dlat]}

    now = datetime.datetime.utcnow()
    split_title = [_ for _ in f_handler.title.split(' ') if 0 < len(_)]
    year = int(split_title[5])
    month_str = split_title[6]
    month = MONTHS.get(month_str)
    start_dt = datetime.datetime(year, month, 1, 0, 0, 0)
    stop_dt = start_dt + dateutil.relativedelta.relativedelta(months=1)

    dtime, time_range = stfmt.format_time_and_range(start_dt, stop_dt,
                                                    units='h')

    meta = {'product_name': None,  # Set later
            'name': None,  # Set later
            'datetime': dtime,
            'time_range': time_range,
            'source_URI': input_path,
            'source_provider': 'IPRC',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_software': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'parameter': None} # Set later

    # Pi-MEP context
    for d, depth in enumerate(depths):
        if abs(depth) > max_depth:
            continue
        for variable in variables:
            vname, vmin, vmax, vunit = variable
            if vname not in f_handler.variables.keys():
                continue

            depth_str = '{}'.format(numpy.floor(depth).astype(int))
            meta['product_name'] = 'ARGO_IPRC_{}_{}'.format(depth_str, vname)
            meta['parameter'] = vname
            colortable = stfmt.format_colortable('matplotlib_jet',
                                                 vmin=vmin, vmax=vmax,
                                                 vmin_pal=vmin, vmax_pal=vmax)

            granule_name = '{}_{}_{}'.format(start_dt.strftime('%Y%m'),
                                             depth_str, vname)
            meta['name'] = granule_name

            data = []

            v_data = f_handler.variables[vname][d, :, :]
            v_data = v_data[:, lon_ind]
            array, offset, scale = pack.ubytes_0_254(v_data, vmin, vmax)
            if numpy.any(v_data.mask):
                array[numpy.where(v_data.mask)] = 255
            data.append({'array': array,
                         'scale': scale,
                         'offset': offset,
                         'description': vname,
                         'unit': vunit,
                         'nodatavalue': 255,
                         'parameter_range': [vmin, vmax],
                         'colortable': colortable})

            yield (meta, geo, data, None)
    f_handler.close()


def convert_from_file(input_path, output_path, variables=None):
    """"""
    if variables is None:
        variables = [('SALT', 32, 38, 'PSS'), ('TEMP', 0, 35, '°C'),]

    for (meta, geo, data, extra) in read_from_file(input_path, variables):
        tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                               create_dir=True)
        stfmt.write_geotiff(tiff_fname, meta, geo, data)


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARN)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)
