# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2019-08-13

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import pyproj
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack
try:
    from osgeo import osr
except ImportError:
    import osr

logger = logging.getLogger(__name__)


class VersionNotSupported(Exception):
    """"""
    def __init__(self, *args, **kwargs):
        """"""
        _msg = ('This reader only supports v1.5 and v1.7 of the CCI Salinity '
                'product. Other versions might be supported but have not been '
                'tested yet.')
        super(VersionNotSupported, self).__init__(_msg, *args, **kwargs)


def read_from_file(input_path):
    """"""
    granule_filename = os.path.basename(input_path)
    _granule_name, _ = os.path.splitext(granule_filename)

    version = None
    spatial_grid = None

    if input_path.endswith('-fv1.5.nc'):
        import h5py

        logger.warning('Data will be read as if input was an HDF5 file because'
                       'v1.5 ESA CCI Salinity files are not valid NetCDF')

        f_handler = h5py.File(input_path, 'r')
        lon = numpy.array(f_handler['lon'][:])
        lat = numpy.array(f_handler['lat'][:])
        _t = numpy.array(f_handler['time'][0])
        _t_units = f_handler['time'].attrs['units']
        sss = numpy.ma.array(f_handler['sss'][:])
        sss_qc = numpy.ma.array(f_handler['sss_qc'][:])
        f_handler.close()

        qc_mask = (sss_qc == 0)  # 1=Good 0=Bad
        dtime = netCDF4.num2date(_t, _t_units)
        version = '1.5'
    elif input_path.endswith('.nc'):
        f_handler = netCDF4.Dataset(input_path, 'r')
        lon = f_handler.variables['lon'][:]
        lat = f_handler.variables['lat'][:]
        _t = f_handler.variables['time'][:]
        _t_units = f_handler.variables['time'].units
        sss = f_handler.variables['sss'][0, :, :]
        sss_qc = f_handler.variables['sss_qc'][0, :, :]
        product_version = f_handler.product_version
        spatial_grid = getattr(f_handler, 'spatial_grid', None)
        f_handler.close()

        qc_mask = (sss_qc == 1)  # 0=Good 1=Bad
        dtime = netCDF4.num2date(_t[0], _t_units)
        version = '{}'.format(float(product_version))
    else:
        raise VersionNotSupported

    if '25km regular grid' == spatial_grid:

        lat_crop = 85.4
        valid_lat_ind = numpy.where(numpy.abs(lat) <= lat_crop)
        lat_slice = slice(numpy.min(valid_lat_ind),
                          numpy.max(valid_lat_ind) + 1)
        lon_slice = slice(None, None)  # take all longitudes

        lat = lat[lat_slice]
        sss = sss[lat_slice, lon_slice]
        qc_mask = qc_mask[lat_slice, lon_slice]

        dx = lon[1] - lon[0]
        dy = lat[1] - lat[0]
        x0 = lon[0]
        y0 = lat[0]
        projection = stfmt.format_gdalprojection()
    else:
        # Definition for EPSG:6933 (for EASE grid)
        proj_txt = '+proj=cea +lon_0=0 +lat_ts=30 +x_0=0 +y_0=0 +ellps=WGS84'
        proj_txt += ' +towgs84=0,0,0,0,0,0,0 +units=m +no_defs'

        # The product was created using an EPSG:6933 grid, so the geotransform
        # should be expressed using the parameters of this grid instead of
        # lat/lon.
        #srs = stfmt.proj2srs(str(proj_txt))  # for Syntool converter with GDAL3
        srs = osr.SpatialReference()
        srs.ImportFromProj4(str(proj_txt))
        iproj = pyproj.Proj(init='epsg:4326')
        oproj = pyproj.Proj(srs.ExportToProj4())

        xx, yy = numpy.meshgrid(lon, lat)

        xp, yp = pyproj.transform(iproj, oproj, xx, yy)
        mxp = numpy.mean(xp, axis=0)
        myp = numpy.mean(yp, axis=1)
        dx = (mxp[-1] - mxp[0]) / (numpy.shape(lon)[0] - 1)
        dy = (myp[-1] - myp[0]) / (numpy.shape(lat)[0] - 1)
        x0 = mxp[0]
        y0 = myp[0]
        projection = srs.ExportToWkt()

    geo = {}
    geo['projection'] = projection
    geo['geotransform'] = [x0 - 0.5 * dx, dx, 0, y0 - 0.5 * dy, 0, dy]

    now = datetime.datetime.utcnow()

    if version in ('1.5', '1.7',):
        # actual center of the rolling window
        dtime = dtime + datetime.timedelta(hours=12)

    granule_name = '{}_{}'.format(_granule_name, 'sss')
    meta = {'product_name': 'ESA_CCI_Salinity_sss',
            'name': granule_name,
            'datetime': stfmt.format_time(dtime),
            'time_range': ['-43200s', '+43199s'],  # Daily
            'source_URI': input_path,
            'source_provider': 'ESA',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_software': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'parameter': 'sea_surface_salinity'}

    vmin = 32
    vmax = 38
    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin, vmax_pal=vmax)

    mask = (sss.mask | qc_mask)
    masked_ind = numpy.where(mask)

    data = []

    # Remove mask and set dummy value to prevent issues with numpy
    # It has no side-effect as mask is saved and reapplied later
    sss._sharedmask = False
    sss.mask = False
    sss[masked_ind] = 0

    array, offset, scale = pack.ubytes_0_254(sss, vmin, vmax)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sss',
                 'unit': '',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    return (meta, geo, data, None)



def convert_from_file(input_path, output_path):
    """"""
    (meta, geo, data, extra) = read_from_file(input_path)
    tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                           create_dir=True)
    stfmt.write_geotiff(tiff_fname, meta, geo, data)


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARN)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)
