# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-10-22

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def read_from_file(input_path, variables):
    """"""
    granule_filename = os.path.basename(input_path)
    _granule_name, _ = os.path.splitext(granule_filename)

    f_handler = netCDF4.Dataset(input_path, 'r')
    lon = f_handler.variables['lon'][:]
    lat = f_handler.variables['lat'][:]
    _t = f_handler.variables['time'][:]
    _t_units = f_handler.variables['time'].units

    lon = numpy.mod(lon + 180.0, 360.0) - 180.0
    lon_ind = numpy.argsort(lon)
    lon = lon[lon_ind]

    lon0, dlon = lon[0], numpy.mean(lon[1:] - lon[:-1])
    lat0, dlat = lat[0], numpy.mean(lat[1:] - lat[:-1])
    geo = {'projection': stfmt.format_gdalprojection(),
           'geotransform': [lon0 - 0.5 * dlon, dlon, 0,
                            lat0 - 0.5 * dlat, 0, dlat]}

    now = datetime.datetime.utcnow()

    meta = {'product_name': None,  # Set later
            'name': None,  # Set later
            'datetime': None, # Set later
            'time_range': None, # Set later
            'source_URI': input_path,
            'source_provider': 'CMEMS',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_software': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'parameter': None} # Set later

    dtimes = netCDF4.num2date(_t, _t_units)

    for variable in variables:
        vname, vmin, vmax, vunit = variable
        if vname not in f_handler.variables.keys():
            continue

        meta['product_name'] = 'CMEMS_015_013_{}'.format(vname)
        meta['parameter'] = vname
        colortable = stfmt.format_colortable('matplotlib_jet',
                                             vmin=vmin, vmax=vmax,
                                             vmin_pal=vmin, vmax_pal=vmax)

        for t, dtime in enumerate(dtimes):
            granule_name = '{}_{}'.format(_granule_name, vname)
            meta['name'] = granule_name
            meta['datetime'] = stfmt.format_time(dtime)
            # This is a weekly product: -3.5 days, +3.5 days
            meta['time_range'] = ['-84h', '+84h']

            data = []

            v_data = f_handler.variables[vname][t, 0, :, :]  # assume 1 depth
            v_data = v_data[:, lon_ind]
            masked_ind = numpy.where(numpy.ma.getmaskarray(v_data) |
                                     numpy.isnan(v_data))

            # Remove mask and set dummy value to prevent issues with numpy
            # It has no side-effect as mask is saved and reapplied later
            v_data._sharedmask = False
            v_data.mask = False
            v_data[masked_ind] = 0

            array, offset, scale = pack.ubytes_0_254(v_data, vmin, vmax)
            if numpy.any(masked_ind):
                array[masked_ind] = 255
            data.append({'array': array,
                         'scale': scale,
                         'offset': offset,
                         'description': vname,
                         'unit': vunit,
                         'nodatavalue': 255,
                         'parameter_range': [vmin, vmax],
                         'colortable': colortable})

            yield (meta, geo, data, None)
    f_handler.close()



def convert_from_file(input_path, output_path, variables=None):
    """"""
    if variables is None:
        variables = [('sos', 32, 38, 'PSS'),]

    for (meta, geo, data, extra) in read_from_file(input_path, variables):
        tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                               create_dir=True)
        stfmt.write_geotiff(tiff_fname, meta, geo, data)


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARN)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)
