# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2024-03-21

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import pyproj
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def read_from_file(input_path):
    """"""
    granule_filename = os.path.basename(input_path)
    _granule_name, _ = os.path.splitext(granule_filename)

    f_handler = netCDF4.Dataset(input_path, 'r')
    lon = f_handler.variables['longitude'][:]
    lat = f_handler.variables['latitude'][:]
    _t = f_handler.variables['time'][:]
    _t_units = f_handler.variables['time'].units
    sss = f_handler.variables['sss'][:, :, 0]
    f_handler.close()

    dtime = netCDF4.num2date(_t[0], _t_units)

    lat_crop = 85.4
    valid_lat_ind = numpy.where(numpy.abs(lat) <= lat_crop)
    lat_slice = slice(numpy.min(valid_lat_ind),
                      numpy.max(valid_lat_ind) + 1)
    lon_slice = slice(None, None)  # take all longitudes

    lat = lat[lat_slice]
    sss = sss[lat_slice, lon_slice]

    dx = lon[1] - lon[0]
    dy = lat[1] - lat[0]
    x0 = lon[0]
    y0 = lat[0]

    geo = {}
    geo['projection'] = stfmt.format_gdalprojection()
    geo['geotransform'] = [x0 - 0.5 * dx, dx, 0, y0 - 0.5 * dy, 0, dy]

    now = datetime.datetime.utcnow()

    granule_name = '{}_{}'.format(_granule_name, 'sss')
    meta = {'product_name': 'ESR_OISSS_v2',
            'name': granule_name,
            'datetime': stfmt.format_time(dtime),
            'time_range': ['-172800s', '+172799s'],  # 4 days
            'source_URI': input_path,
            'source_provider': 'ESR',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_software': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'parameter': 'sea_surface_salinity'}

    vmin = 32
    vmax = 38
    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin, vmax_pal=vmax)

    masked_ind = numpy.where(sss.mask)

    data = []

    # Remove mask and set dummy value to prevent issues with numpy
    # It has no side-effect as mask is saved and reapplied later
    sss._sharedmask = False
    sss.mask = False
    sss[masked_ind] = 0

    array, offset, scale = pack.ubytes_0_254(sss, vmin, vmax)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sss',
                 'unit': '',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    return (meta, geo, data, None)



def convert_from_file(input_path, output_path):
    """"""
    (meta, geo, data, extra) = read_from_file(input_path)
    tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                           create_dir=True)
    stfmt.write_geotiff(tiff_fname, meta, geo, data)
