# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2019-08-13

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import pyproj
import netCDF4
import logging
import datetime
import dateutil.relativedelta
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack
try:
    from osgeo import osr
except ImportError:
    import osr

logger = logging.getLogger(__name__)


class DepthNotFound(Exception):
    """"""
    def __init__(self, depth, *args, **kwargs):
        self.depth = depth
        _msg = ('"{}" not found in the depth values available in the input '
                'file'.format(depth))
        super(DepthNotFound, self).__init__(_msg, *args, **kwargs)


def process_var(varname, values, vmin, vmax, cmap, units=''):
    """"""
    colortable = stfmt.format_colortable(cmap,
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin, vmax_pal=vmax)

    mask = numpy.ma.getmaskarray(values)
    masked_ind = numpy.where(mask)

    data = []

    # Remove mask and set dummy value to prevent issues with numpy
    # It has no side-effect as mask is saved and reapplied later
    values._sharedmask = False
    values.mask = False
    values[masked_ind] = 0

    array, offset, scale = pack.ubytes_0_254(values, vmin, vmax)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': varname,
                 'unit': units,
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})
    return data


def read_from_file(f_handler, depths):
    """"""
    _t = f_handler.variables['time'][:]
    _t_units = f_handler.variables['time'].units
    lon = f_handler.variables['longitude'][:]
    _lat = f_handler.variables['latitude'][:]
    ind_lat = numpy.where(_lat < 86)[0]
    lat = _lat[ind_lat]
    depth = f_handler.variables['depth'][:]
    _t_units = _t_units.replace('19500101T000000Z', '1950-01-01T00:00:00Z')


    _dtime = netCDF4.num2date(_t[0], _t_units)
    start_dt = datetime.datetime(_dtime.year, _dtime.month, 1, 0, 0, 0, 0)
    stop_dt = start_dt + dateutil.relativedelta.relativedelta(months=1)
    stop_dt = stop_dt - datetime.timedelta(seconds=1)
    dtime, time_range = stfmt.format_time_and_range(start_dt, stop_dt,
                                                    units='h')

    now = datetime.datetime.utcnow()

    meta = {'product_name': None,  # Set later
            'name': None,  # Set later
            'datetime': dtime,
            'time_range': time_range,
            'source_URI': None,  # Set later
            'source_provider': '',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_software': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'parameter': None}  # Set later

    nrow = len(lat)
    ncell = len(lon)
    gcp_nrows = 20
    gcp_ncells = 20

    geo_row = numpy.round(numpy.linspace(0, nrow - 1, num=gcp_nrows)).astype('int')
    geo_cell = numpy.round(numpy.linspace(0, ncell - 1, num=gcp_ncells)).astype('int')

    geo_lon = numpy.ma.getdata(lon)[geo_cell]
    geo_lat = numpy.ma.getdata(lat)[geo_row]
    gcplon, gcplat = numpy.meshgrid(geo_lon, geo_lat)
    gcphei = numpy.zeros(gcplon.shape)

    geo_cell = geo_cell + 0.5
    geo_row = geo_row + 0.5
    gcppix, gcplin = numpy.meshgrid(geo_cell, geo_row)

    geo = {'projection': stfmt.format_gdalprojection(),
           'gcps': stfmt.format_gdalgcps(gcplon, gcplat, gcphei,gcppix,gcplin)}

    for d in depths:
        depth_indices = numpy.where(depth == d)[0]
        if 0 >= len(depth_indices):
            raise DepthNotFound(d)
        d_ind = depth_indices[0]

        if 'PSAL' in f_handler.variables.keys():
            psal = f_handler.variables['PSAL'][0, d_ind, ind_lat, :]
            vmin = 32
            vmax = 38
            data = process_var('psal', psal, vmin, vmax, 'matplotlib_jet', '')
            meta['parameter'] = 'salinity'
            yield (meta, geo, data, {'depth': d, 'param': 'psal'})

        if 'TEMP' in f_handler.variables.keys():
            temp = f_handler.variables['TEMP'][0, d_ind, ind_lat, :]
            vmin = 0
            vmax = 32
            data = process_var('temperature', temp, vmin, vmax,
                               'cerbere_medspiration', '')
            meta['parameter'] = 'temperature'
            yield (meta, geo, data, {'depth': d, 'param': 'temp'})


def convert_from_file(input_path, output_path, depths=None):
    """"""
    if depths is None:
        depths = (5,)

    granule_filename = os.path.basename(input_path)
    _granule_name, _ = os.path.splitext(granule_filename)

    f_handler = netCDF4.Dataset(input_path, 'r')

    for (meta, geo, data, extra) in read_from_file(f_handler, depths):
        meta['source_URI'] = input_path
        meta['name'] = '{}_{}m_{}'.format(_granule_name, extra['depth'],
                                          extra['param'])
        meta['product_name'] = 'ISAS_monthly_{}m_{}'.format(extra['depth'],
                                                            extra['param'])
        tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                               create_dir=True)
        stfmt.write_geotiff(tiff_fname, meta, geo, data)

    f_handler.close()

if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARN)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)
