# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-10-22

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy
import netCDF4
import logging
import datetime
import dateutil.relativedelta
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def read_from_file(input_path, vmin=32, vmax=38, max_depth=10.0):
    """"""
    f_handler = netCDF4.Dataset(input_path, 'r')
    lon = f_handler.variables['LONGITUDE'][:]
    lat = f_handler.variables['LATITUDE'][:]
    depths = f_handler.variables['PRESSURE'][:]
    _t = f_handler.variables['TIME'][:]

    lon = numpy.mod(lon + 180.0, 360.0) - 180.0
    lon_ind = numpy.argsort(lon)
    lon = lon[lon_ind]

    lon0, dlon = lon[0], numpy.mean(lon[1:] - lon[:-1])
    lat0, dlat = lat[0], numpy.mean(lat[1:] - lat[:-1])
    geo = {'projection': stfmt.format_gdalprojection(),
           'geotransform': [lon0, dlon, 0, lat0, 0, dlat]}

    now = datetime.datetime.utcnow()

    meta = {'product_name': None,  # Set later
            'name': None,  # Set later
            'datetime': None, # Set later
            'time_range': None, # Set later
            'source_URI': input_path,
            'source_provider': 'Scripps',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_software': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'parameter': 'salinity'}

    product_base = 'roemmich_gilson_salinity'
    base_dt = datetime.datetime(2004, 1, 1)
    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin, vmax_pal=vmax)

    for d, depth in enumerate(depths):
        if abs(depth) >= max_depth:
            continue

        s_mean = f_handler.variables['ARGO_SALINITY_MEAN'][d, :, :]
        s_mean = s_mean[:, lon_ind]
        depth_str = '{}'.format(numpy.floor(depth).astype(int))
        meta['product_name'] = '{}_{}'.format(product_base, depth_str)

        for t, _dmonths in enumerate(_t):
            dmonths = numpy.floor(_dmonths).astype(int)
            start_offset = dateutil.relativedelta.relativedelta(months=dmonths)
            stop_offset = dateutil.relativedelta.relativedelta(months=1)
            start_dt = base_dt + start_offset
            stop_dt = start_dt + stop_offset
            dtime, time_range = stfmt.format_time_and_range(start_dt, stop_dt,
                                                            units='h')
            meta['datetime'] = dtime
            meta['time_range'] = time_range

            granule_name = '{}_{}_salinity'.format(dtime, depth_str)
            meta['name'] = granule_name

            data = []

            v_data = f_handler.variables['ARGO_SALINITY_ANOMALY'][t, d, :, :]
            v_data = v_data[:, lon_ind]
            n_data = v_data + s_mean  # total = mean + anomaly
            masked_ind = numpy.where(n_data.mask)

            # Remove mask and set dummy value to prevent issues with numpy
            # It has no side-effect as mask is saved and reapplied later
            n_data._sharedmask = False
            n_data.mask = False
            n_data[masked_ind] = 0

            array, offset, scale = pack.ubytes_0_254(n_data, vmin, vmax)
            if numpy.any(masked_ind):
                array[masked_ind] = 255
            data.append({'array': array,
                         'scale': scale,
                         'offset': offset,
                         'description': 'salinity',
                         'unit': 'PSS',
                         'nodatavalue': 255,
                         'parameter_range': [vmin, vmax],
                         'colortable': colortable})

            yield (meta, geo, data, None)
    f_handler.close()


def convert_from_file(input_path, output_path):
    """"""
    for (meta, geo, data, extra) in read_from_file(input_path):
        tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                               create_dir=True)
        stfmt.write_geotiff(tiff_fname, meta, geo, data)


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARN)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)
