# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-10-22

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def read_from_file(input_path, variables, max_depth=0):
    """"""
    f_handler = netCDF4.Dataset(input_path, 'r')
    lon = f_handler.variables['lon'][:]
    lat = f_handler.variables['lat'][:]
    depths = f_handler.variables['depth'][:]
    _t = f_handler.variables['time'][:]
    _t_units = 'days since 2000-01-01'

    lon = numpy.mod(lon + 180.0, 360.0) - 180.0
    lon_ind = numpy.argsort(lon)
    lon = lon[lon_ind]

    lon0, dlon = lon[0], numpy.mean(lon[1:] - lon[:-1])
    lat0, dlat = lat[0], numpy.mean(lat[1:] - lat[:-1])
    geo = {'projection': stfmt.format_gdalprojection(),
           'geotransform': [lon0 - 0.5 * dlon, dlon, 0,
                            lat0 - 0.5 * dlat, 0, dlat]}

    now = datetime.datetime.utcnow()

    meta = {'product_name': None,  # Set later
            'name': None,  # Set later
            'datetime': None, # Set later
            'time_range': None, # Set later
            'source_URI': input_path,
            'source_provider': 'NODC',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_software': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'parameter': None} # Set later

    dtimes = netCDF4.num2date(_t, _t_units)

    product_base = 'WOA2009'
    if '2013' in f_handler.title:
        product_base = 'WOA2013'
        _, month_str, _ = f_handler.time_coverage_start.split('-')
        dtimes = [datetime.datetime(2000, int(month_str), 15, 0, 0, 0)]

    for d, depth in enumerate(depths):
        if abs(depth) > max_depth:
            continue

        for variable in variables:
            vname, vmin, vmax, vunit = variable
            if vname not in f_handler.variables.keys():
                continue

            depth_str = '{}'.format(numpy.floor(depth).astype(int))
            meta['product_name'] = '{}_{}_{}'.format(product_base, depth_str,
                                                     vname)
            meta['parameter'] = vname
            colortable = stfmt.format_colortable('matplotlib_jet',
                                                 vmin=vmin, vmax=vmax,
                                                 vmin_pal=vmin, vmax_pal=vmax)

            for t, dtime in enumerate(dtimes):
                granule_name = '{}_{}_{}'.format(dtime.strftime('%m'),
                                                 depth_str, vname)
                meta['name'] = granule_name
                meta['datetime'] = stfmt.format_time(dtime)
                # It is not possible to handle the time range properly for
                # climatologies using Syntool paradigms, so set it arbitrarily
                # and use a script to generate symlinks and SQL statements
                # later...
                meta['time_range'] = ['-1d', '+1d']

                data = []

                v_data = f_handler.variables[vname][t, d, :, :]
                v_data = v_data[:, lon_ind]
                masked_ind = numpy.where(v_data.mask)

                # Remove mask and set dummy value to prevent issues with numpy
                # It has no side-effect as mask is saved and reapplied later
                v_data._sharedmask = False
                v_data.mask = False
                v_data[masked_ind] = 0

                array, offset, scale = pack.ubytes_0_254(v_data, vmin, vmax)
                if numpy.any(masked_ind):
                    array[masked_ind] = 255
                data.append({'array': array,
                             'scale': scale,
                             'offset': offset,
                             'description': vname,
                             'unit': vunit,
                             'nodatavalue': 255,
                             'parameter_range': [vmin, vmax],
                             'colortable': colortable})

                yield (meta, geo, data, None)
    f_handler.close()



def convert_from_file(input_path, output_path, variables=None):
    """"""
    if variables is None:
        variables = [('s_an', 32, 38, 'PSS'),]

    for (meta, geo, data, extra) in read_from_file(input_path, variables):
        tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                               create_dir=True)
        stfmt.write_geotiff(tiff_fname, meta, geo, data)
    logger.warn('Please note that syntool-converter cannot handle cyclic data'
                ' as it requires each granule to have a single time range.'
                'You will have to write a script to generate symbolic links '
                'and SQL statements for each year you want this product to '
                'appear in.')


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARN)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)
