#!/usr/bin/env python
# coding=utf-8

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from cerbere.mapper.safeocnncfile import SAFEOCNNCFile
import os
import numpy
import logging
import datetime
import syntool_converter
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt

logger = logging.getLogger(__name__)


def sar_wind(infile, outdir, pngkml=False, valid_percent_min=1.,
             vmin=0., vmax=25.4, vmin_pal=0., vmax_pal=50*0.514):
    """
    """
    granule_filename = os.path.basename(infile)
    granule_name, _ = os.path.splitext(granule_filename)
    measurement_dir = os.path.dirname(infile)
    safe_dir = os.path.dirname(measurement_dir)
    safe_name = os.path.basename(safe_dir)
    sensor_mode = safe_name.split('_')[1]
    sensor_swath = granule_name.split('-')[1].upper()
    datagroup = safe_name.replace('.SAFE', '')
    pid = datagroup.split('_')[-1]
    dataname = '{}-{}'.format(granule_name, pid)

    # Read/Process data
    logger.info('Read/Process data')
    sarwind = SAFEOCNNCFile(infile, product='WIND')
    mission = sarwind.read_global_attribute('missionName')
    if mission == 'S1A':
        sensor_name = 'Sentinel-1A'
        sensor_platform = 'Sentinel-1A'
        source_provider = 'ESA'
    elif mission == 'S1B':
        sensor_name = 'Sentinel-1B'
        sensor_platform = 'Sentinel-1B'
        source_provider = 'ESA'
    else:
        raise Exception('S1A/S1B missions expected.')
    start_time = sarwind.get_start_time()
    stop_time = sarwind.get_end_time()
    heading = sarwind.read_values('owiHeading')
    if numpy.sin((90 - heading[0, 0]) * numpy.pi / 180) > 0:
        sensor_pass = 'Ascending'
    else:
        sensor_pass = 'Descending'
    if sensor_mode not in ['S1', 'S2', 'S3', 'S4', 'S5', 'S6', 'IW', 'EW']:
        raise Exception('S[1-6]/IW/EW modes expected.')
    sensor_polarisation = sarwind.read_global_attribute('polarisation')
    windspeed = sarwind.read_values('owiWindSpeed')
    if windspeed.shape == (1, 1):
        logger.warning('owiRaSize and owiAzSize equals 1 !')
        raise syntool_converter.EarlyExit()
    winddirection = sarwind.read_values('owiWindDirection')
    if 'owiLandFlag' in sarwind.get_fieldnames():
        # IPF version < 2.9
        landflag = sarwind.read_values('owiLandFlag')
        inversionquality = sarwind.read_values('owiInversionQuality')
        windquality = sarwind.read_values('owiWindQuality')
        windmask = (landflag != 0) | \
                   ((windspeed == 0) & (winddirection == 180)) | \
                   ((windspeed == 0) & (windquality == 3)) | \
                   ((windspeed == 0) & (inversionquality == 2))
        windmask = numpy.ma.getdata(windmask)  # we don't want a masked mask
    else:
        # IPF version >= 2.9
        owimask = sarwind.read_values('owiMask')
        windmask = numpy.ma.getmaskarray(windspeed) | \
                   numpy.ma.getmaskarray(winddirection) | \
                   (owimask != 0)

    # Check that there are enough valid values before doing more computations
    valid_percent = numpy.sum(~windmask) / float(windmask.size) * 100
    if valid_percent <= valid_percent_min:
        logger.warn('Not enough valid data: {:0.3f}%'.format(valid_percent))
        # Not an error, it just means there is no need to continue
        raise syntool_converter.EarlyExit()

    lon = sarwind.read_values('lon')
    lat = sarwind.read_values('lat')
    if numpy.ma.is_masked(lon) or numpy.ma.is_masked(lat):
        raise Exception('Some lon and/or lat is masked.')
    if numpy.all(lon == 0) or numpy.all(lat == 0):
        raise Exception('All lon and/or lat set to 0.')
    ngcps = numpy.ceil(numpy.array(lon.shape) / 10.).astype('int') + 1
    pix = numpy.linspace(0, lon.shape[1] - 1,
                         num=ngcps[1]).round().astype('int32')
    lin = numpy.linspace(0, lon.shape[0] - 1,
                         num=ngcps[0]).round().astype('int32')
    pix2d, lin2d = numpy.meshgrid(pix, lin)
    gcplon = lon[lin2d, pix2d]
    gcplat = lat[lin2d, pix2d]
    gcppix = pix2d + 0.5
    gcplin = lin2d + 0.5
    gcphei = numpy.zeros(ngcps)
    """
    # Make sure lon are continuous (no jump because of IDL crossing)
    # (if IDL crossing, by convention we make lon to be around 180deg)
    if gcplon.min() < -135 and gcplon.max() > 135:
         gcplon[numpy.where(gcplon < 0)] += 360.
    """
    gcplonmid = gcplon[ngcps[0] / 2, ngcps[1] / 2]
    gcplon = numpy.mod(gcplon - (gcplonmid - 180.), 360.) + gcplonmid - 180.
    gcplonmin = gcplon.min()
    gcplon = gcplon - numpy.floor((gcplonmin + 180.) / 360.) * 360.

    # Construct metadata/geolocation/band(s)
    logger.info('Construct metadata/geolocation/band(s)')
    now = datetime.datetime.utcnow()
    (dtime, time_range) = stfmt.format_time_and_range(start_time, stop_time,
                                                      units='ms')
    metadata = {}
    metadata['product_name'] = 'SAR_wind'
    metadata['name'] = dataname
    metadata['datetime'] = dtime
    metadata['time_range'] = time_range
    metadata['source_URI'] = infile
    metadata['source_provider'] = source_provider
    metadata['processing_center'] = ''
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = ['wind speed', 'wind direction']
    metadata['type'] = 'remote sensing'
    metadata['sensor_type'] = 'SAR'
    metadata['sensor_name'] = sensor_name
    metadata['sensor_platform'] = sensor_platform
    metadata['sensor_mode'] = sensor_mode
    metadata['sensor_swath'] = sensor_swath
    metadata['sensor_polarisation'] = sensor_polarisation
    metadata['sensor_pass'] = sensor_pass
    metadata['datagroup'] = datagroup
    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['gcps'] = stfmt.format_gdalgcps(gcplon, gcplat, gcphei,
                                                gcppix, gcplin)

    colortable = stfmt.format_colortable('noaa_wind',
                                         vmax=vmax, vmax_pal=vmax_pal,
                                         vmin=vmin, vmin_pal=vmin_pal)

    band = []

    array, offset, scale = pack.ubytes_0_254(windspeed, vmin, vmax)
    if numpy.any(windmask):
        array[numpy.where(windmask)] = 255
    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'wind speed',
                 'unittype': 'm/s',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    winddirection = numpy.mod(90. - winddirection + 180., 360.)
    array, offset, scale = pack.ubytes_0_254(winddirection, 0.0, 360.0)
    if numpy.any(windmask):
        array[numpy.where(windmask)] = 255
    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'wind direction',
                 'unittype': 'deg',
                 'nodatavalue': 255,
                 'parameter_range': [0, 360.]})

    # Write geotiff
    logger.info('Write geotiff')
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)
    # Write projected png/kml
    if pngkml is True:
        logger.info('Write projected png/kml')
        stfmt.write_pngkml_proj(tifffile)
