#!/usr/bin/env python
# coding=utf-8

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt

logger = logging.getLogger(__name__)

# File patterns :
# asi-AMSR2-s6250-20140922-v5.4.nc
# asi-AMSR2-n6250-20140922-v5.4.nc


def amsr_sea_ice_conc(infile, outdir,
                      vmin=0., vmax=100., vmin_pal=0., vmax_pal=100.,
                      write_netcdf=False):
    """
    """
    f_handler = netCDF4.Dataset(infile, 'r')
    conc = f_handler.variables['z'][::-1, :]
    _proj_var = f_handler.variables['polar_stereographic']
    geot_str = _proj_var.getncattr('GeoTransform')
    f_handler.close()

    geotransform = [int(_.strip()) for _ in geot_str.split(' ')
                    if 0 < len(_.strip())]

    granule_file = os.path.basename(infile)
    granule_name, _ = os.path.splitext(granule_file)

    _, sensor, loc, date_str, version = granule_name.split('-')

    pole = loc[0]
    if 'n' == pole:
        srs = stfmt.epsg2srs(3411)  # use 3413 if problems with ellipsoid
    elif 's' == pole:
        srs = stfmt.epsg2srs(3412)  # use 3976 if problems with ellipsoid
    else:
        raise Exception('Unknown pole: {}'.format(pole))
    geolocation = {'projection': srs.ExportToWkt(),
                   'geotransform': geotransform}

    spatial_res = int(loc[1:])
    spatial_res_km = float(spatial_res) / 1000.0

    # if the resolution is not the default (6.25km), then the product name
    # must be changed
    product_name = 'AMSR_sea_ice_concentration'
    if 6250 != spatial_res:
        res_m = numpy.floor(spatial_res).astype('int')
        product_name = '{}_{}'.format(product_name, res_m)

    # Read/Process data
    now = datetime.datetime.utcnow()
    start_dtime = datetime.datetime.strptime(date_str, '%Y%m%d')
    dtime = start_dtime + datetime.timedelta(0.5)

    metadata = {}
    metadata['product_name'] = product_name
    metadata['name'] = granule_name
    metadata['datetime'] = stfmt.format_time(dtime)
    metadata['time_range'] = ['-12h', '+12h']
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'Universität Bremen'
    metadata['processing_center'] = 'Universität Bremen'
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'sea ice concentration'
    metadata['type'] = 'remote sensing'
    metadata['sensor_type'] = 'radiometer'

    colortable = stfmt.format_colortable('sea_ice_conc', vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin_pal, vmax_pal=vmax_pal)

    band = []
    conc[numpy.where(~numpy.isfinite(conc))] = -1  # land
    conc[numpy.where(conc == 0)] = -1  # no ice at all
    data_mask = (conc < vmin)
    array, offset, scale = pack.ubytes_0_254(conc, vmin, vmax)
    if numpy.any(data_mask):
        array[numpy.where(data_mask)] = 255
    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea ice concentration',
                 'unittype': '%',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    # Write
    if write_netcdf is False:
        logger.info('Write geotiff')
        tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
        stfmt.write_geotiff(tifffile, metadata, geolocation, band)
    else:
        logger.info('Write netcdf')
        ncfile = stfmt.format_ncfilename(outdir, metadata, create_dir=True)
        band[0]['name'] = 'sea_ice_concentration'
        band[0]['long_name'] = 'ASI Ice Concentration'
        band[0]['unittype'] = 'percent'
        metadata['spatial_resolution'] = spatial_res
        resolutions_km = numpy.array([spatial_res_km, spatial_res_km])
        dgcps = numpy.round(100. / resolutions_km).astype('int')
        stfmt.write_netcdf(ncfile, metadata, geolocation, band, 'grid_proj',
                           dgcps=dgcps)


if '__main__' == __name__:
    import sys

    main_logger = logging.getLogger()
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.DEBUG)

    amsr_sea_ice_conc(sys.argv[1], sys.argv[2])
