#!/usr/bin/env python
# coding=utf-8

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import pyproj
import netCDF4
import logging
import datetime
import syntool_converter
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt

logger = logging.getLogger(__name__)


def convert_raster(divergence, shear, vorticity, data_mask, outdir, metadata,
                   geolocation):
    """"""
    divergence = numpy.ma.masked_invalid(divergence)
    shear = numpy.ma.masked_invalid(shear)
    vorticity = numpy.ma.masked_invalid(vorticity)

    mask = (data_mask | numpy.ma.getmaskarray(divergence))
    if not numpy.all(mask):
        vmin = -3.0
        vmax = 3.0
        logger.debug('divergence: {} {}'.format(divergence.min(),
                                                divergence.max()))
        array, offset, scale = pack.ubytes_0_254(divergence, vmin, vmax)
        if (mask is not None) and (numpy.any(mask)):
            array[numpy.where(mask)] = 255

        colortable = stfmt.format_colortable('matplotlib_bwr',
                                             vmin=vmin, vmax=vmax,
                                             vmin_pal=vmin, vmax_pal=vmax)
        band = []
        band.append({'array': array,
                     'scale': scale,
                     'offset': offset,
                     'description': 'sea ice drift divergence',
                     'unittype': '% day-1',
                     'nodatavalue': 255,
                     'parameter_range': [vmin, vmax],
                     'colortable': colortable})

        metadata['product_name'] = 'infranor_sea_ice_drift_divergence'
        tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
        stfmt.write_geotiff(tifffile, metadata, geolocation, band)
    else:
        logger.info('All divergence values are masked, skipping.')

    mask = (data_mask | numpy.ma.getmaskarray(shear))
    if not numpy.all(mask):
        vmin = 0.0
        vmax = 3.0
        logger.debug('shear: {} {}'.format(shear.min(), shear.max()))
        array, offset, scale = pack.ubytes_0_254(shear, vmin, vmax)
        if (mask is not None) and (numpy.any(mask)):
            array[numpy.where(mask)] = 255

        colortable = stfmt.format_colortable('matplotlib_jet',
                                             vmin=vmin, vmax=vmax,
                                             vmin_pal=vmin, vmax_pal=vmax)
        band = []
        band.append({'array': array,
                     'scale': scale,
                     'offset': offset,
                     'description': 'sea ice drift shear',
                     'unittype': '% day-1',
                     'nodatavalue': 255,
                     'parameter_range': [vmin, vmax],
                     'colortable': colortable})

        metadata['product_name'] = 'infranor_sea_ice_drift_shear'
        tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
        stfmt.write_geotiff(tifffile, metadata, geolocation, band)
    else:
        logger.info('All shear values are masked, skipping.')

    mask = (data_mask | numpy.ma.getmaskarray(vorticity))
    if not numpy.all(mask):
        vmin = -3.0
        vmax = 3.0
        logger.debug('vorticity: {} {}'.format(vorticity.min(),
                                               vorticity.max()))
        array, offset, scale = pack.ubytes_0_254(vorticity, vmin, vmax)
        if (mask is not None) and (numpy.any(mask)):
            array[numpy.where(mask)] = 255

        colortable = stfmt.format_colortable('matplotlib_bwr',
                                             vmin=vmin, vmax=vmax,
                                             vmin_pal=vmin, vmax_pal=vmax)
        band = []
        band.append({'array': array,
                     'scale': scale,
                     'offset': offset,
                     'description': 'sea ice drift vorticity',
                     'unittype': '% day-1',
                     'nodatavalue': 255,
                     'parameter_range': [vmin, vmax],
                     'colortable': colortable})

        metadata['product_name'] = 'infranor_sea_ice_drift_vorticity'
        tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
        stfmt.write_geotiff(tifffile, metadata, geolocation, band)
    else:
        logger.info('All vorticity values are masked, skipping.')


def convert_vfield(x, y, u, v, data_mask, outdir, metadata, geolocation):
    """"""
    u_velocity = numpy.ma.masked_invalid(u)
    v_velocity = numpy.ma.masked_invalid(v)

    mask = (data_mask |
            numpy.ma.getmaskarray(u_velocity) |
            numpy.ma.getmaskarray(v_velocity))
    if numpy.all(mask):
        logger.warning('No valid value for velocity. Skipping')
        return

    x0, y0 = numpy.meshgrid(x, y)
    x1 = x0 + u_velocity
    y1 = y0 + v_velocity

    p_in = pyproj.Proj('EPSG:3411')
    p_out = pyproj.Proj('EPSG:4326')
    coords_t = pyproj.Transformer.from_proj(p_in, p_out, always_xy=True)
    lon0, lat0 = coords_t.transform(x0, y0)
    lon1, lat1 = coords_t.transform(x1, y1)

    geod = pyproj.Geod(ellps='WGS84')
    _displacement_dir, _, _ = geod.inv(lon0, lat0, lon1, lat1)
    displacement_dir = 90.0 - _displacement_dir

    displacement_velocity = numpy.hypot(u_velocity, v_velocity)
    logger.debug('velocity: {} {}'.format(displacement_velocity.min(),
                                          displacement_velocity.max()))

    band = []

    vmin = 0.0
    vmax = 0.58  # ~ 50km/day
    array, offset, scale = pack.ubytes_0_254(displacement_velocity, vmin,
                                             vmax)
    if (mask is not None) and (numpy.any(mask)):
        array[numpy.where(mask)] = 255
    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea ice drift module',
                 'unittype': 'm',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax]})

    vmin = 0.0
    vmax = 360.0
    array, offset, scale = pack.ubytes_0_254(displacement_dir, vmin, vmax)
    if (mask is not None) and (numpy.any(mask)):
        array[numpy.where(mask)] = 255
    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea ice drift direction',
                 'unittype': 'degrees',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax]})

    metadata['product_name'] = 'infranor_sea_ice_drift_velocity'
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)


def infranor_sea_ice_drift(infile, outdir,
                           vmin=0., vmax=100., vmin_pal=0., vmax_pal=100.):
    """
    """
    start_dt = None
    stop_dt = None
    # spatial_res_km = 2
    data_slices = (0, slice(None, None, -1), slice(None, None, None))

    f_handler = netCDF4.Dataset(infile, 'r')
    geotransform_str = f_handler.variables['polar_stereographic'].GeoTransform
    institution = f_handler.institution
    product_name = 'infranor_sea_ice_drift'

    time_var = f_handler.variables['time']
    t = netCDF4.num2date(time_var[0], time_var.units)

    divergence = f_handler.variables['e1_med'][data_slices]
    shear = f_handler.variables['e2_med'][data_slices]
    vorticity = f_handler.variables['e3_med'][data_slices]
    x = f_handler.variables['x'][:]
    y = f_handler.variables['y'][:]
    u = f_handler.variables['u_med'][data_slices]
    v = f_handler.variables['v_med'][data_slices]

    # data_mask = numpy.full(u.shape, False, dtype='bool')
    data_mask = (0 >= f_handler.variables['mask'][data_slices].astype('int'))
    f_handler.close()

    if numpy.all(data_mask):
        raise syntool_converter.EarlyExit()

    start_dt = t
    stop_dt = t + datetime.timedelta(days=1)

    granule_file = os.path.basename(infile)
    granule_name, _ = os.path.splitext(granule_file)
    del _

    # Spatial coverage
    srs = stfmt.epsg2srs(3411)
    geotransform = [float(_) for _ in geotransform_str.split(' ')
                    if 0 < len(_)]

    geolocation = {'projection': srs.ExportToWkt(),
                   'geotransform': geotransform}

    # Temporal coverage
    dtime_str, time_range = stfmt.format_time_and_range(start_dt, stop_dt, 's')

    now = datetime.datetime.utcnow()
    metadata = {}
    metadata['product_name'] = product_name
    metadata['name'] = granule_name
    metadata['datetime'] = dtime_str
    metadata['time_range'] = time_range
    metadata['source_URI'] = infile
    metadata['source_provider'] = institution
    metadata['processing_center'] = institution
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'sea ice drift'
    metadata['type'] = 'remote sensing'

    convert_vfield(x, y, u, v, data_mask, outdir, metadata, geolocation)

    convert_raster(divergence, shear, vorticity, data_mask, outdir, metadata,
                   geolocation)


if '__main__' == __name__:
    import sys

    main_logger = logging.getLogger()
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.DEBUG)

    infranor_sea_ice_drift(sys.argv[1], sys.argv[2])
