#!/usr/bin/env python
# coding=utf-8

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import pyproj
import netCDF4
import logging
import datetime
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt
try:
    from osgeo import gdal
except ImportError:
    import gdal

logger = logging.getLogger(__name__)


def projection_workaround(tiff_path):
    """"""
    filename, _ = os.path.splitext(tiff_path)
    fixed_path = '{}_fixed.tiff'.format(filename)
    gdal.Warp(fixed_path, tiff_path, dstSRS='epsg:3413')
    os.remove(tiff_path)
    os.rename(fixed_path, tiff_path)


def convert_raster(sic, sit, snt, outdir, metadata, geolocation):
    """"""
    # Sea ice concentration
    sic = numpy.ma.masked_invalid(sic)
    mask = numpy.ma.getmaskarray(sic)
    mask = (mask | (sic <= 0.0))
    if not numpy.all(mask):
        vmin = 0.0
        vmax = 1.0
        logger.debug('sea ice conc.: {} {}'.format(sic.min(), sic.max()))
        array, offset, scale = pack.ubytes_0_254(sic, vmin, vmax)
        if (mask is not None) and (numpy.any(mask)):
            array[numpy.where(mask)] = 255

        colortable = stfmt.format_colortable('cmocean_ice',
                                             vmin=vmin, vmax=vmax,
                                             vmin_pal=vmin, vmax_pal=vmax)
        band = []
        band.append({'array': array,
                     'scale': scale,
                     'offset': offset,
                     'description': 'sea ice concentration',
                     'unittype': '%',
                     'nodatavalue': 255,
                     'parameter_range': [vmin, vmax],
                     'colortable': colortable})

        metadata['product_name'] = 'nextsim_sea_ice_concentration'
        tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
        stfmt.write_geotiff(tifffile, metadata, geolocation, band)
        projection_workaround(tifffile)
    else:
        logger.info('All sea ice concentration values are masked, skipping.')

    # Sea ice thickness
    sit = numpy.ma.masked_invalid(sit)
    mask = numpy.ma.getmaskarray(sit)
    mask = (mask | (sit <= 0.0))
    if not numpy.all(mask):
        vmin = 0.0
        vmax = 4.0
        logger.debug('sea ice thickness: {} {}'.format(sit.min(), sit.max()))
        array, offset, scale = pack.ubytes_0_254(sit, vmin, vmax)
        if (mask is not None) and (numpy.any(mask)):
            array[numpy.where(mask)] = 255

        colortable = stfmt.format_colortable('matplotlib_viridis',
                                             vmin=vmin, vmax=vmax,
                                             vmin_pal=vmin, vmax_pal=vmax)
        band = []
        band.append({'array': array,
                     'scale': scale,
                     'offset': offset,
                     'description': 'sea ice thickness',
                     'unittype': 'm',
                     'nodatavalue': 255,
                     'parameter_range': [vmin, vmax],
                     'colortable': colortable})

        metadata['product_name'] = 'nextsim_sea_ice_thickness'
        tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
        stfmt.write_geotiff(tifffile, metadata, geolocation, band)
        projection_workaround(tifffile)
    else:
        logger.info('All sea ice thickness values are masked, skipping.')

    # snow thickness
    snt = numpy.ma.masked_invalid(snt)
    mask = numpy.ma.getmaskarray(snt)
    mask = (mask | (snt <= 0))
    if not numpy.all(mask):
        vmin = 0.0
        vmax = 0.5
        logger.debug('snow thickness: {} {}'.format(snt.min(), snt.max()))
        array, offset, scale = pack.ubytes_0_254(snt, vmin, vmax)
        if (mask is not None) and (numpy.any(mask)):
            array[numpy.where(mask)] = 255

        colortable = stfmt.format_colortable('matplotlib_viridis',
                                             vmin=vmin, vmax=vmax,
                                             vmin_pal=vmin, vmax_pal=vmax)
        band = []
        band.append({'array': array,
                     'scale': scale,
                     'offset': offset,
                     'description': 'snow thickness',
                     'unittype': 'm',
                     'nodatavalue': 255,
                     'parameter_range': [vmin, vmax],
                     'colortable': colortable})

        metadata['product_name'] = 'nextsim_snow_thickness'
        tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
        stfmt.write_geotiff(tifffile, metadata, geolocation, band)
        projection_workaround(tifffile)
    else:
        logger.info('All snow thickness values are masked, skipping.')


def convert_vfield(x, y, u, v, proj4_str, outdir, metadata, geolocation):
    """"""
    u_velocity = numpy.ma.masked_invalid(u)
    v_velocity = numpy.ma.masked_invalid(v)

    mask = (numpy.ma.getmaskarray(u_velocity) |
            numpy.ma.getmaskarray(v_velocity))
    if numpy.all(mask):
        logger.warning('No valid value for velocity. Skipping')
        return

    x0, y0 = numpy.meshgrid(x, y)
    x1 = x0 + u_velocity
    y1 = y0 + v_velocity

    p_in = pyproj.Proj(proj4_str)
    p_out = pyproj.Proj('EPSG:4326')
    coords_t = pyproj.Transformer.from_proj(p_in, p_out, always_xy=True)
    lon0, lat0 = coords_t.transform(x0, y0)
    lon1, lat1 = coords_t.transform(x1, y1)

    geod = pyproj.Geod(ellps='WGS84')
    _displacement_dir, _, _ = geod.inv(lon0, lat0, lon1, lat1)
    displacement_dir = 90.0 - _displacement_dir

    displacement_velocity = numpy.hypot(u_velocity, v_velocity)
    mask = (mask | (displacement_velocity <= 0.0))
    logger.debug('velocity: {} {}'.format(displacement_velocity.min(),
                                          displacement_velocity.max()))

    band = []

    vmin = 0.0
    vmax = 0.58  # ~ 50km/day
    array, offset, scale = pack.ubytes_0_254(displacement_velocity, vmin,
                                             vmax)
    if (mask is not None) and (numpy.any(mask)):
        array[numpy.where(mask)] = 255
    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea ice drift module',
                 'unittype': 'm',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax]})

    vmin = 0.0
    vmax = 360.0
    array, offset, scale = pack.ubytes_0_254(displacement_dir, vmin, vmax)
    if (mask is not None) and (numpy.any(mask)):
        array[numpy.where(mask)] = 255
    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea ice drift direction',
                 'unittype': 'degrees',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax]})

    metadata['product_name'] = 'nextsim_sea_ice_drift_velocity'
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)
    projection_workaround(tifffile)


def nextsim(infile, outdir, vmin=0., vmax=100., vmin_pal=0., vmax_pal=100.):
    """
    """
    start_dt = None
    stop_dt = None

    f_handler = netCDF4.Dataset(infile, 'r')
    proj4_str = f_handler.variables['stereographic'].proj4
    institution = f_handler.institution
    product_name = 'nextsim_sea_ice_drift'

    timebnds_var = f_handler.variables['time_bnds']
    timebnds = netCDF4.num2date(timebnds_var[:], timebnds_var.units)
    x = f_handler.variables['x'][:]
    y = f_handler.variables['y'][:]

    sic = f_handler.variables['siconc'][:, :, :]
    sit = f_handler.variables['sithick'][:, :, :]
    snt = f_handler.variables['sisnthick'][:, :, :]
    u = f_handler.variables['vxsi'][:, :, :]
    v = f_handler.variables['vysi'][:, :, :]

    f_handler.close()

    granule_file = os.path.basename(infile)
    granule_name, _ = os.path.splitext(granule_file)
    del _

    # Spatial coverage
    srs = stfmt.proj2srs(proj4_str)
    x0 = x[0]
    y0 = y[0]

    geotransform = [x0, 3000, 0, y0, 0, 3000]
    geolocation = {'projection': srs.ExportToWkt(),
                   'geotransform': geotransform}

    for i in range(0, sic.shape[0]):
        start_dt = timebnds[i][0]
        stop_dt = timebnds[i][1]

        # Temporal coverage
        dtime_str, time_range = stfmt.format_time_and_range(start_dt, stop_dt,
                                                            's')

        now = datetime.datetime.utcnow()
        metadata = {}
        metadata['product_name'] = product_name
        metadata['name'] = '{}_{}'.format(granule_name, dtime_str)
        metadata['datetime'] = dtime_str
        metadata['time_range'] = time_range
        metadata['source_URI'] = infile
        metadata['source_provider'] = institution
        metadata['processing_center'] = institution
        metadata['conversion_software'] = 'Syntool'
        metadata['conversion_version'] = '0.0.0'
        metadata['conversion_datetime'] = stfmt.format_time(now)
        metadata['parameter'] = 'sea ice drift'
        metadata['type'] = 'remote sensing'

        convert_vfield(x, y, u[i], v[i], proj4_str, outdir, metadata,
                       geolocation)
        convert_raster(sic[i], sit[i], snt[i], outdir, metadata, geolocation)


if '__main__' == __name__:
    import sys

    main_logger = logging.getLogger()
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.DEBUG)

    nextsim(sys.argv[1], sys.argv[2])
