#!/usr/bin/env python
# coding=utf-8

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import pyproj
import netCDF4
import logging
import datetime
import syntool_converter
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt

logger = logging.getLogger(__name__)


class FilenameNotSupported(Exception):
    pass


def osisaf_sea_ice_drift(infile, outdir,
                         vmin=0., vmax=100., vmin_pal=0., vmax_pal=100.):
    """
    """
    start_dt = None
    stop_dt = None
    # displacement_duration = None  # in seconds

    bname = os.path.basename(infile)
    if not (bname.startswith('ice_drift_nh_polstere-') or
            bname.startswith('ice_drift_sh_polstere-')):
        raise FilenameNotSupported()

    tokens = bname.split('_', 7)
    if 7 < len(tokens):
        raise FilenameNotSupported()

    subproduct = tokens[4]

    f_handler = netCDF4.Dataset(infile, 'r')
    proj4_str = f_handler.variables['Polar_Stereographic_Grid'].proj4_string
    institution = f_handler.institution
    product_name = '{}_{}'.format(f_handler.product_name, subproduct)

    # Adapt CRS for GCPs depending on the hemisphere
    output_epsg = 3411
    if 'lat_ts=-' in proj4_str:
        output_epsg = 3412

    if product_name.startswith('osi_saf_mr_ice_drift'):
        spatial_res_km = 20
        displacement_max_km = 25
        data_slices = (slice(None, None, None), slice(None, None, None))

        data_mask = (0 < f_handler.variables['data_status'][data_slices])

        start_dt = datetime.datetime.strptime(f_handler.start_date,
                                              '%Y-%m-%d %H:%M:%S UTC')
        stop_dt = datetime.datetime.strptime(f_handler.stop_date,
                                             '%Y-%m-%d %H:%M:%S UTC')

        x = f_handler.variables['xc'][:]
        y = f_handler.variables['yc'][:]

        displacement_x = 1000 * f_handler.variables['dX'][data_slices]
        displacement_y = 1000 * f_handler.variables['dY'][data_slices]

        if numpy.any(data_mask):
            displacement_x[numpy.where(data_mask)] = 1
            displacement_y[numpy.where(data_mask)] = 1

        displacement_length = numpy.hypot(displacement_x, displacement_y)
        _displacement_dir = numpy.arctan2(displacement_y, displacement_x)
        displacement_dir = numpy.rad2deg(_displacement_dir)
    else:
        spatial_res_km = 62.5
        displacement_max_km = 100
        data_slices = (0, slice(None, None, None), slice(None, None, None))

        _status_flag = f_handler.variables['status_flag'][data_slices]
        data_mask = (20 > _status_flag)

        time_bnds = f_handler.variables['time_bnds'][0, :]
        time_bnds_units = f_handler.variables['time_bnds'].units
        start_dt, stop_dt = netCDF4.num2date(time_bnds, time_bnds_units)

        _xc = f_handler.variables['xc'][:]
        _yc = f_handler.variables['yc'][:]
        _x, _y = numpy.meshgrid(_xc, _yc)

        p_lon0 = pyproj.Proj('+units=km {}'.format(proj4_str))
        p_lon45 = pyproj.Proj('EPSG:{}'.format(output_epsg))
        coords_t = pyproj.Transformer.from_proj(p_lon0, p_lon45,
                                                always_xy=True)
        x_out, y_out = coords_t.transform(_x, _y)
        x = x_out[0, :]
        y = y_out[:, 0]

        lon0 = f_handler.variables['lon'][:, :]
        lat0 = f_handler.variables['lat'][:, :]
        lon1 = f_handler.variables['lon1'][data_slices]
        lat1 = f_handler.variables['lat1'][data_slices]

        geod = pyproj.Geod(ellps='WGS84')
        _displacement_dir, _, displacement_length = geod.inv(lon0, lat0,
                                                             lon1, lat1)
        displacement_dir = 90.0 - _displacement_dir

    f_handler.close()

    if numpy.all(data_mask):
        raise syntool_converter.EarlyExit()

    granule_file = os.path.basename(infile)
    granule_name, _ = os.path.splitext(granule_file)

    # Spatial coverage
    srs = stfmt.epsg2srs(3411)
    x0 = x[0]
    dxx = numpy.mean(x[1:] - x[:-1])
    dxy = 0
    y0 = y[0]
    dyx = 0
    dyy = numpy.mean(y[1:] - y[:-1])
    geotransform = [x0 - 0.5 * dxx, dxx, dxy, y0 - 0.5 * dyy, dyx, dyy]
    geolocation = {'projection': srs.ExportToWkt(),
                   'geotransform': geotransform}

    # Temporal coverage
    dtime_str, time_range = stfmt.format_time_and_range(start_dt, stop_dt, 's')

    now = datetime.datetime.utcnow()
    metadata = {}
    metadata['product_name'] = product_name
    metadata['name'] = granule_name
    metadata['datetime'] = dtime_str
    metadata['time_range'] = time_range
    metadata['source_URI'] = infile
    metadata['source_provider'] = institution
    metadata['processing_center'] = institution
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'sea ice drift'
    metadata['type'] = 'remote sensing'

    band = []

    displacement_length_km = displacement_length / 1000.0
    vmin = 0.0
    vmax = displacement_max_km
    print(displacement_length_km.min(), displacement_length_km.max(),
          spatial_res_km)
    array, offset, scale = pack.ubytes_0_254(displacement_length_km, vmin,
                                             vmax)
    if (data_mask is not None) and (numpy.any(data_mask)):
        array[numpy.where(data_mask)] = 255
    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea ice drift module',
                 'unittype': 'm',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax]})

    vmin = 0.0
    vmax = 360.0
    array, offset, scale = pack.ubytes_0_254(displacement_dir, vmin, vmax)
    if (data_mask is not None) and (numpy.any(data_mask)):
        array[numpy.where(data_mask)] = 255
    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea ice drift direction',
                 'unittype': 'degrees',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax]})

    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)


if '__main__' == __name__:
    import sys

    main_logger = logging.getLogger()
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.DEBUG)

    osisaf_sea_ice_drift(sys.argv[1], sys.argv[2])
