# -*- encoding: utf-8 -*-

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys
import errno
import numpy
import netCDF4
import logging
import syntool_converter
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack
import scipy.interpolate
from datetime import datetime
import matplotlib
matplotlib.use('Agg')  # allow matplotlib to render figures in headless env
from matplotlib import pyplot

logger = logging.getLogger(__name__)

# https://sentinel.esa.int/web/sentinel/user-guides/sentinel-3-olci/overview/heritage
# https://sentinel.esa.int/web/sentinel/technical-guides/sentinel-3-olci/level-1/fr-or-rr-toa-radiances
# Oa1   400 	15
# Oa2 	412.5 	10
# Oa3 	442.5 	10
# Oa4 	490 	10   BLUE
# Oa5 	510 	10
# Oa6 	560 	10   GREEN
# Oa7 	620 	10
# Oa8 	665 	10   RED
# Oa9 	673.75 	7.5
# Oa10 	681.25 	7.5
# Oa11 	708.75 	10
# Oa12 	753.75 	7.5
# Oa13 	761.25 	2.5
# Oa14 	764.375 3.75
# Oa15 	767.5 	2.5
# Oa16 	778.75 	15
# Oa17 	865 	20 SUN GLITTER
# Oa18 	885 	10
# Oa19 	900 	10
# Oa20 	940 	20
# Oa21 	1 020 	40


def interp_tie_ac(tie_values_list, tie_ac_factor, ac_index):
    """"""
    # Linear interpolation over across-track dimension
    tie_ac_index = numpy.arange(tie_values_list[0].shape[1],
                                dtype='float32') * tie_ac_factor
    indint = numpy.clip(numpy.searchsorted(tie_ac_index, ac_index) - 1,
                        0, tie_ac_index.size - 2)
    indratio = (ac_index - tie_ac_index[indint]) / \
               (tie_ac_index[indint + 1] - tie_ac_index[indint])
    indratio = indratio[numpy.newaxis, :]
    int_values = []
    for val in tie_values_list:
        int_val = val[:, indint] + (val[:, indint + 1] - val[:, indint]) * indratio
        int_values.append(int_val)
    return int_values


def get_atmospheric_rho(lut_path, band_index, sza, saa, oza, oaa):
    """Compute atmospheric correction from LUT"""
    # Read LUT
    sza_lut = []
    oza_lut = []
    delta_lut = []
    rho_lut = []
    oza_resol = 5
    # sza_resol = 5
    delta_resol = 10
    ntot_delta = 360 / delta_resol + 1
    ntot_oza = 90 / oza_resol + 1
    with open(lut_path, 'r') as f:
        for line in f.readlines():
            if line.startswith('#'):
                continue
            split_line = line.split(' ')
            sza_lut.append(split_line[0])
            oza_lut.append(split_line[1])
            delta_lut.append(split_line[2])
            rho_lut.append(float(split_line[band_index + 3]))

    # Check the size of the LUT
    lut_size = len(rho_lut)
    lut_size_factor = ntot_oza * ntot_delta
    if 0 != (lut_size % lut_size_factor):
        raise Exception('The size of the LUT does not match the hardcoded '
                        'parameters: the number of lines per sza value should '
                        'be {}'.format(lut_size_factor))

    # Interpolate atmospheric correction according to angles
    delta = numpy.mod(saa - oaa, 360.)
    _points = numpy.array((sza.flatten(), oza.flatten(),
                           delta.flatten())).transpose()
    sza_lut = numpy.unique(numpy.array(sza_lut, dtype='float32'))
    oza_lut = numpy.unique(numpy.array(oza_lut, dtype='float32'))
    delta_lut = numpy.unique(numpy.array(delta_lut, dtype='float32'))
    lut_points = (sza_lut, oza_lut, delta_lut)
    lut_shape = (sza_lut.size, oza_lut.size, delta_lut.size)
    lut_values = numpy.array(rho_lut, dtype='float32').reshape(lut_shape)
    interpolator = scipy.interpolate.RegularGridInterpolator
    lut_func = interpolator(lut_points, lut_values, method='linear',
                            bounds_error=False, fill_value=None)
    rho_atm = lut_func(_points).reshape(sza.shape)
    return rho_atm


def studpdf(z, wind, n):
    """
    """
    mss = 0.003 + 5.12e-3 * wind
    p = n / (n - 1.) / (numpy.pi * mss) / \
        (1. + (2. * z ** 2. / mss) / (n - 1.)) ** (n / 2. + 1)
    return p


def tilt_angle(sza, saa, oza, oaa):
    zx = -((numpy.sin(numpy.deg2rad(sza)) * numpy.cos(numpy.deg2rad(saa))
            + numpy.sin(numpy.deg2rad(oza)) * numpy.cos(numpy.deg2rad(oaa)))
            / (numpy.cos(numpy.deg2rad(sza)) + numpy.cos(numpy.deg2rad(oza))))
    zy = -((numpy.sin(numpy.deg2rad(sza)) * numpy.sin(numpy.deg2rad(saa))
            + numpy.sin(numpy.deg2rad(oza)) * numpy.sin(numpy.deg2rad(oaa)))
            / (numpy.cos(numpy.deg2rad(sza)) + numpy.cos(numpy.deg2rad(oza))))
    z = numpy.sqrt(zx ** 2. + zy ** 2.)
    return zx, zy, z


def tam_figure(zang, satellite_id):
    """
    """

    if '' == satellite_id:  # S3A
        color = 'm'
    elif 'B' == satellite_id.upper():  # S3B
        color = 'k'
    nrow, ncell = zang.shape
    dpi = 100.
    figsize = numpy.array([ncell, nrow]) / dpi
    fig = pyplot.figure(figsize=figsize, dpi=dpi, facecolor='w')
    ax = fig.add_axes([0., 0., 1., 1.])

    level0 = numpy.ceil(numpy.nanmin(zang) / 5.) * 5
    level1 = numpy.floor(numpy.nanmax(zang) / 5.) * 5
    levels = numpy.arange(level0, level1 + 1, 5).astype('int')
    cs = ax.contour(zang, levels, colors=color, linewidths=5)
    ax.clabel(cs, fmt='%i', fontsize=12 * 5)

    ax.set_axis_off()
    return fig


def fig2rgb(fig):
    """
    """
    fig.canvas.draw()
    wid, hei = fig.canvas.get_width_height()
    rgb = numpy.fromstring(fig.canvas.tostring_rgb(), dtype=numpy.uint8)
    rgb.shape = (hei, wid, 3)
    rgb = rgb[:, :]
    return rgb


def sentinel3_olci_bc(infile, outdir, vmin=None, vmax=None,
                      channels='nir', write_netcdf=False, lut_path=None,
                      log_path=None, lat_crop=85.0):
    """"""
    t0 = datetime.utcnow()

    if type(channels) is list or type(channels) is tuple:
        bandnames = channels
        _product_name = 'Sentinel-3{}_OLCI_BC'
        if len(bandnames) != 1:
            raise Exception('only 1 band in channels.')
    elif 'nir' == channels:
        bandnames = ('Oa17',)
        _product_name = 'Sentinel-3{}_OLCI_NIR_BC'
    else:
        raise Exception('channels must be "nir"')

    syntool_stats = {}
    for bandname in bandnames:
        syntool_stats[bandname] = {}
    if log_path is not None and not os.path.exists(log_path):
        try:
            os.makedirs(log_path)
        except OSError:
            _, e, _ = sys.exc_info()
            if e.errno != errno.EEXIST:
                raise

    full_path = os.path.normpath(infile)
    file_path = os.path.basename(full_path)
    file_name, _ = os.path.splitext(file_path)

    geo_path = os.path.join(infile, 'geo_coordinates.nc')
    time_path = os.path.join(infile, 'time_coordinates.nc')
    quality_path = os.path.join(infile, 'qualityFlags.nc')

    # Extract geo coordinates information
    geo_handler = netCDF4.Dataset(geo_path, 'r')
    nrow = geo_handler.dimensions['rows'].size
    nrow_all = nrow
    ncell = geo_handler.dimensions['columns'].size
    ncell_all = ncell
    lon = geo_handler.variables['longitude'][:]
    tie_lon = numpy.ma.array(lon)
    lat = geo_handler.variables['latitude'][:]
    tie_lat = numpy.ma.array(lat)
    satellite_id = geo_handler.product_name[2]
    geo_handler.close()

    if 'A' == satellite_id:
        satellite_id = ''  # for historical reason
    product_name = _product_name.format(satellite_id)

    coord_tie_path = os.path.join(infile, 'tie_geo_coordinates.nc')
    coord_tie = netCDF4.Dataset(coord_tie_path, 'r')
    ss_lat = coord_tie.variables['latitude'][:]
    ss_lon = coord_tie.variables['longitude'][:]
    coord_tie.close()

    # Handle longitude continuity
    dlon = lon[1:, :] - lon[:-1, :]
    if 180.0 <= numpy.max(numpy.abs(dlon)):
        lon[lon < 0.0] = lon[lon < 0.0] + 360.0
        ss_lon[ss_lon < 0.0] = ss_lon[ss_lon < 0.0] + 360.0

    # Extract time coordinates information
    time_handler = netCDF4.Dataset(time_path, 'r')
    start_timestamp = time_handler.variables['time_stamp'][0]
    end_timestamp = time_handler.variables['time_stamp'][-1]
    timestamp_units = time_handler.variables['time_stamp'].units
    time_handler.close()

    # Format time information
    start_time = netCDF4.num2date(start_timestamp, timestamp_units)
    end_time = netCDF4.num2date(end_timestamp, timestamp_units)
    (dtime, time_range) = stfmt.format_time_and_range(start_time, end_time,
                                                      units='ms')

    parameters = ['{} brightness contrast'.format(bnd) for bnd in bandnames]
    metadata = {}
    metadata['product_name'] = product_name
    metadata['name'] = file_name
    metadata['datetime'] = dtime
    metadata['time_range'] = time_range
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'ESA'
    metadata['processing_center'] = 'OceanDataLab'
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(datetime.utcnow())
    metadata['parameter'] = parameters
    metadata['type'] = 'remote sensing'
    metadata['sensor_type'] = 'medium-resolution imaging spectrometer'
    metadata['sensor_name'] = 'OLCI'
    metadata['sensor_platform'] = 'Sentinel-3'

    # Crop high latitude to avoid projection issues
    LAT_MAX = 89.0
    ind_valid_cols = numpy.where(numpy.abs(tie_lat).max(axis=0) <= LAT_MAX)[0]
    slice_lat0 = slice(None)
    slice_lat1 = slice(numpy.min(ind_valid_cols),
                       numpy.max(ind_valid_cols) + 1)
    tie_lat = tie_lat[slice_lat0, slice_lat1]
    tie_lon = tie_lon[slice_lat0, slice_lat1]
    nrow, ncell = tie_lat.shape

    # Handle longitude continuity
    dlon = tie_lon[1:, :] - tie_lon[:-1, :]
    if 180.0 <= numpy.max(numpy.abs(dlon)):
        lon0 = tie_lon[0, 0] + 180.0
        tie_lon[:, :] = numpy.mod(tie_lon[:, :] - lon0, 360.0) + lon0
        ss_lon[:, :] = numpy.mod(ss_lon[:, :] - lon0, 360.0) + lon0
    tie_lon_i = + tie_lon
    # Try to fit longitudes in [-540, 540] (ingestor constraint)
    while 540.0 < numpy.max(tie_lon):
        tie_lon = tie_lon - 540.0
    while -540.0 > numpy.min(tie_lon):
        tie_lon = tie_lon + 540.0

    # Compute GCPs
    tie_row = numpy.linspace(0, nrow - 1, num=tie_lon.shape[0])
    tie_cell = numpy.linspace(0, ncell - 1, num=tie_lon.shape[1])
    tie_facrow = (nrow - 1.) / (tie_lon.shape[0] - 1.)
    tie_faccell = (ncell - 1.) / (tie_lon.shape[1] - 1.)
    gcp_fac = 128
    gcp_fac = numpy.maximum(gcp_fac, numpy.maximum(tie_faccell, tie_facrow))
    gcp_nrow = numpy.ceil((nrow - 1.) / gcp_fac).astype('int') + 1
    gcp_ncell = numpy.ceil((ncell - 1.) / gcp_fac).astype('int') + 1
    tie_indrow = numpy.round(numpy.linspace(0, tie_lon.shape[0] - 1,
                                            num=gcp_nrow)).astype('int')
    tie_indcell = numpy.round(numpy.linspace(0, tie_lon.shape[1] - 1,
                                             num=gcp_ncell)).astype('int')
    gcp_lon = tie_lon[tie_indrow.reshape((-1, 1)),
                      tie_indcell.reshape((1, -1))]
    gcp_lat = tie_lat[tie_indrow.reshape((-1, 1)),
                      tie_indcell.reshape((1, -1))]
    gcp_row = numpy.tile(tie_row[tie_indrow].reshape((-1, 1)) + 0.5,
                         (1, gcp_ncell))
    gcp_cell = numpy.tile(tie_cell[tie_indcell].reshape((1, -1)) + 0.5,
                          (gcp_nrow, 1))
    gcp_hei = numpy.zeros(gcp_lon.shape)

    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['gcps'] = stfmt.format_gdalgcps(gcp_lon, gcp_lat, gcp_hei,
                                                gcp_cell, gcp_row)

    syntool_stats['lon_min'] = float(numpy.min(tie_lon))
    syntool_stats['lon_max'] = float(numpy.max(tie_lon))
    syntool_stats['lat_min'] = float(numpy.min(tie_lat))
    syntool_stats['lat_max'] = float(numpy.max(tie_lat))

    logger.info('Construct bands')
    t_start = datetime.utcnow()
    bands = []
    dset = netCDF4.Dataset(quality_path, 'r')
    quality_flags = dset.variables['quality_flags'][slice_lat0, slice_lat1]
    dset.close()

    bits = {'saturated@Oa21': 0,
            'saturated@Oa20': 1,
            'saturated@Oa19': 2,
            'saturated@Oa18': 3,
            'saturated@Oa17': 4,
            'saturated@Oa16': 5,
            'saturated@Oa15': 6,
            'saturated@Oa14': 7,
            'saturated@Oa13': 8,
            'saturated@Oa12': 9,
            'saturated@Oa11': 10,
            'saturated@Oa10': 11,
            'saturated@Oa09': 12,
            'saturated@Oa08': 13,
            'saturated@Oa07': 14,
            'saturated@Oa06': 15,
            'saturated@Oa05': 16,
            'saturated@Oa04': 17,
            'saturated@Oa03': 18,
            'saturated@Oa02': 19,
            'saturated@Oa01': 20,
            'dubious': 21,
            'sun-glint_risk': 22,
            'duplicated': 23,
            'cosmetic': 24,
            'invalid': 25,
            'straylight_risk': 26,
            'bright': 27,
            'tidal_region': 28,
            'fresh_inland_water': 29,
            'coastline': 30,
            'land': 31}

    lat_mask = (numpy.abs(tie_lat) > lat_crop)
    data_mask = numpy.zeros(quality_flags.shape, dtype='bool')
    data_mask = (data_mask | lat_mask)

    if numpy.all(data_mask):
        logger.warn('No data to extract.')
        raise syntool_converter.EarlyExit()

    off_invalid = numpy.uint32(1 << bits['invalid'])
    off_fresh = numpy.uint32(1 << bits['fresh_inland_water'])
    off_land = numpy.uint32(1 << bits['land'])

    contrast_mask = (numpy.bitwise_and(quality_flags, off_invalid) > 0) | \
                    ((numpy.bitwise_and(quality_flags, off_land) > 0) & \
                     (numpy.bitwise_and(quality_flags, off_fresh) == 0))
    contrast_mask = (contrast_mask | lat_mask)

    valid_ratio_lower_threshold = 0.001 # 0.1%
    valid_ratio = numpy.sum(~contrast_mask) / float(contrast_mask.size)
    if valid_ratio <= valid_ratio_lower_threshold:
        logger.warn('Not enough data.')
        raise syntool_converter.EarlyExit()

    t_stop = datetime.utcnow()
    syntool_stats['mask_computation'] = (t_stop - t_start).total_seconds()

    t_start = datetime.utcnow()

    bandname = bandnames[0]
    bandindex = int(bandname[2:]) - 1

    logger.info('\tConstruct {} band'.format(bandname))
    fieldname = '{}_radiance'.format(bandname)
    file_path = os.path.join(infile, '{}.nc'.format(fieldname))
    f_handler = netCDF4.Dataset(file_path, 'r')
    band = f_handler.variables[fieldname][:]
    band = numpy.ma.array(band)
    f_handler.close()
    band = band[slice_lat0, slice_lat1]

    logger.info('\tCompute brightness contrast')

    data_angle_path = os.path.join(infile, 'tie_geometries.nc')
    data_angle = netCDF4.Dataset(data_angle_path, 'r')
    tie_sza = data_angle.variables['SZA'][:]
    tie_saa = data_angle.variables['SAA'][:]
    tie_oza = data_angle.variables['OZA'][:]
    tie_oaa = data_angle.variables['OAA'][:]
    tie_ac_factor = data_angle.ac_subsampling_factor
    data_angle.close()

    if tie_sza.shape[0] != nrow_all:
        raise Exception('Wrong along track dimension in tie_geometries.nc')
    tie_sza = tie_sza[slice_lat0, :].astype('float32')
    tie_saa = tie_saa[slice_lat0, :].astype('float32')
    tie_oza = tie_oza[slice_lat0, :].astype('float32')
    tie_oaa = tie_oaa[slice_lat0, :].astype('float32')
    tie_zx, tie_zy, tie_z = tilt_angle(tie_sza, tie_saa, tie_oza, tie_oaa)
    ac_index = numpy.arange(slice_lat1.start, slice_lat1.stop, dtype='float32')
    _tie_oza = tie_oza * numpy.sign(numpy.sin(numpy.deg2rad(tie_oaa)))
    tie_values_list = [tie_zx, tie_zy, tie_sza, _tie_oza]
    zx, zy, sza, oza = interp_tie_ac(tie_values_list, tie_ac_factor, ac_index)
    oza = numpy.abs(oza)

    data_es_path = os.path.join(infile, 'instrument_data.nc')
    data_es = netCDF4.Dataset(data_es_path, 'r')
    solar_flux = data_es.variables['solar_flux'][:]
    data_es.close()
    solar_flux = numpy.mean(solar_flux[bandindex, :])

    rho_band = band * numpy.pi / numpy.cos(numpy.deg2rad(sza)) / solar_flux
    z = numpy.sqrt(zx ** 2. + zy ** 2.)
    del zx, zy
    offset = 0.008263
    k = 0.859117
    w = 6.766508
    n = 4.878770
    rho_back = k * studpdf(z, w, n) * numpy.pi * 0.022 / 4. / \
               numpy.cos(numpy.deg2rad(sza)) / numpy.cos(numpy.deg2rad(oza)) * \
               (1. + z ** 2.) ** 2. + offset
    del sza, oza, z
    if lut_path is not None:
        tie_rho_atm = get_atmospheric_rho(lut_path, bandindex, tie_sza,
                                          tie_saa, tie_oza, tie_oaa)
        rho_atm = interp_tie_ac([tie_rho_atm], tie_ac_factor, ac_index)[0]
        rho_back += rho_atm
        del rho_atm
    contrast = numpy.ma.log(rho_band / rho_back)
    del rho_band, rho_back
    contrast[contrast_mask] = numpy.ma.masked

    t_stop = datetime.utcnow()
    syntool_stats['brightness_contrast_computation'] = \
                (t_stop - t_start).total_seconds()

    logger.info('\tSet vmin/vmax')
    if vmin is None:
        vmin = -1.2 #-1.
    if vmax is None:
        vmax = 1.5 #1.

    logger.info('\tBytescaling')
    byte, offset, scale = pack.ubytes_0_254(contrast, vmin, vmax)
    description = '{} brightness contrast (log)'.format(bandname)
    if contrast.mask is not numpy.ma.nomask:
        byte[contrast.mask] = 255

    # mask data for extreme latitudes
    # byte[data_mask] = 255

    band_range = [numpy.exp(vmin), numpy.exp(vmax)]
    bands.append({'array': byte,
                  'scale': scale,
                  'offset': offset,
                  'description': description,
                  'unittype': '',
                  'nodatavalue': 255,
                  'parameter_range': band_range})

    if write_netcdf:
        bands[-1]['name'] = bandname
        bands[-1]['long_name'] = bandname
        bands[-1]['unittype'] = '1'
        metadata['spatial_resolution'] = 300
        ncfile = stfmt.format_ncfilename(outdir, metadata, create_dir=True)
        stfmt.write_netcdf(ncfile, metadata, geolocation, bands, 'swath',
                           ngcps=gcp_lon.shape)
    else:
        logger.info('Write geotiff')
        tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
        stfmt.write_geotiff(tifffile, metadata, geolocation, bands)
    bands = []
    metadata['product_name'] = 'Sentinel-3{}_OLCI_tilt_azimuth_angle'.format(satellite_id)
    description = 'Sentinel3{} OLCI Tilt azimuth angle'.format(satellite_id)
    interpolator = scipy.interpolate.griddata
    z_i = interpolator((ss_lon.flatten(), ss_lat.flatten()), tie_z.flatten(),
                       (tie_lon_i.flatten(), tie_lat.flatten()),
                       method='cubic')
    z_i = z_i.reshape(tie_lat.shape)
    z_i = z_i[::-1, :]
    fig = tam_figure(numpy.rad2deg(numpy.arctan(z_i)), satellite_id)
    rgb = fig2rgb(fig)
    pyplot.close(fig)
    bands.append({'array': rgb[:, :, 0],
                  'description': description,
                  'nodatavalue':255})
    bands.append({'array': rgb[:, :, 1],
                  'description': description,
                  'nodatavalue':255})
    bands.append({'array': rgb[:, :, 2],
                  'description': description,
                  'nodatavalue':255})
    if not write_netcdf:
        logger.info('Write geotiff')
        tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
        stfmt.write_geotiff(tifffile, metadata, geolocation, bands)

    logger.info(datetime.utcnow() - t0)
    syntool_stats['total_time'] = (datetime.utcnow() - t0).total_seconds()
    if log_path is not None:
        import json
        stats_path = os.path.join(log_path, '{}.json'.format(file_name))
        with open(stats_path, 'w') as f:
            json.dump(syntool_stats, f)

