# -*- encoding: utf-8 -*-

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys
import errno
import numpy
import scipy
import netCDF4
import logging
import syntool_converter
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack
from datetime import datetime
import scipy.interpolate
import scipy.signal

logger = logging.getLogger(__name__)

# S-2 : 665, 560, 490
# red : 650 nm
# green : 510 nm
# blue : 475 nm
# https://sentinel.esa.int/web/sentinel/user-guides/sentinel-3-olci/overview/heritage
# https://sentinel.esa.int/web/sentinel/technical-guides/sentinel-3-olci/level-1/fr-or-rr-toa-radiances
# Oa1   400 	15
# Oa2 	412.5 	10
# Oa3 	442.5 	10
# Oa4 	490 	10   BLUE
# Oa5 	510 	10
# Oa6 	560 	10   GREEN
# Oa7 	620 	10
# Oa8 	665 	10   RED
# Oa9 	673.75 	7.5
# Oa10 	681.25 	7.5
# Oa11 	708.75 	10
# Oa12 	753.75 	7.5
# Oa13 	761.25 	2.5
# Oa14 	764.375 3.75
# Oa15 	767.5 	2.5
# Oa16 	778.75 	15
# Oa17 	865 	20 SUN GLITTER
# Oa18 	885 	10
# Oa19 	900 	10
# Oa20 	940 	20
# Oa21 	1 020 	40

default_minmax = {'Oa01': (None, None),
                  'Oa02': (None, None),
                  'Oa03': (None, None),
                  'Oa04': (13.966, 95.000),
                  'Oa05': (None, None),
                  'Oa06': (8.213, 95.000),
                  'Oa07': (None, None),
                  'Oa08': (None, None),
                  'Oa09': (3.346, 95.000),
                  'Oa10': (None, None),
                  'Oa11': (None, None),
                  'Oa12': (None, None),
                  'Oa13': (None, None),
                  'Oa14': (None, None),
                  'Oa15': (None, None),
                  'Oa16': (None, None),
                  'Oa17': (3., 150.0),
                  'Oa18': (None, None),
                  'Oa19': (None, None),
                  'Oa20': (None, None),
                  'Oa21': (None, None),
                  'CHL_NN': (-2.1, 1.35),
                  'CHL_OC4ME': (-1.7, 1.55),
                  'TSM_NN': (-1.5, 1),
                  'KD490': (-1.5, 0.5),
                  'CDM_NN': (-2.5, 0.5)}

SPEC_ANGLE_MASKED_VARS = ['CHL_NN', 'CHL_OC4ME', 'KD490', 'KD490_M07']


class TieGeometriesNotFound(IOError):
    """Error raised when the path for the file for tie coordinates does not
    exist."""
    def __init__(self, path):
        """"""
        self.path = path


class TieGeometriesDimensionMismatch(ValueError):
    """Error raised when the alongtrack dimension has different sizes in the
    tie_geometries.nc and geo_coordinates.nc files."""
    def __init__(self, tie_size, geo_size):
        """"""
        self.tie_size = tie_size
        self.geo_size = geo_size


def atmospheric_correction(lut_name, infile, band_columns, nrow, ncell):
    '''Compute atmospheric correction from LUT '''
    # Read LUT
    sza_lut = []
    oza_lut = []
    delta_lut = []
    rho_atm = []
    oza_resol = 5
    # sza_resol = 5
    delta_resol = 10
    ntot_delta = 360 / delta_resol + 1
    ntot_oza = 90 / oza_resol + 1
    with open(lut_name, 'r') as f:
        for line in f.readlines():
            if line.startswith('#'):
                continue
            split_line = line.split(' ')
            sza_lut.append(split_line[0])
            oza_lut.append(split_line[1])
            delta_lut.append(split_line[2])
            rho_line = []
            for icolband in band_columns:
                rho_line.append(float(split_line[icolband]))
            rho_atm.append(rho_line)

    # Check the size of the LUT
    lut_size = len(rho_atm)
    lut_size_factor = ntot_oza * ntot_delta
    if 0 != (lut_size % lut_size_factor):
        raise Exception('The size of the LUT does not match the hardcoded '
                        'parameters: the number of lines per sza value should '
                        'be {}'.format(lut_size_factor))

    # Read angles from tie_geometrie from OLCI
    data_angle_path = os.path.join(infile, 'tie_geometries.nc')
    data_angle = netCDF4.Dataset(data_angle_path, 'r')
    # Sun Zenithal Angle
    sza = data_angle.variables['SZA'][:]
    # Delta = Sun Azimuthal Angle - Observation Azimuthal Angle
    delta = data_angle.variables['SAA'][:] - data_angle.variables['OAA'][:]
    delta = (delta + 360) % 360
    # Observation zenithal angle
    oza = data_angle.variables['OZA'][:]
    # Subsampling factor across track
    ac_subsampling = data_angle.ac_subsampling_factor
    data_angle.close()
    # Interpolate atmospheric correction
    nrow_geom = numpy.shape(sza)[0]
    ncell_geom = numpy.shape(sza)[1]
    if nrow_geom != nrow:
        raise Exception('Wrong along track dimension in tie_geometries.nc')
    rho = numpy.zeros((len(band_columns), nrow, ncell))
    _points = numpy.array((sza.flatten(), oza.flatten(), delta.flatten())
                          ).transpose()
    sza_lut = numpy.unique(numpy.array(sza_lut, dtype='float32'))
    oza_lut = numpy.unique(numpy.array(oza_lut, dtype='float32'))
    delta_lut = numpy.unique(numpy.array(delta_lut, dtype='float32'))
    lut_points = (sza_lut, oza_lut, delta_lut)
    rho_atm = numpy.array(rho_atm[:][:])
    lut_shape = (sza_lut.size, oza_lut.size, delta_lut.size, 1)

    dst_range = numpy.arange(ncell)
    interpolator = scipy.interpolate.RegularGridInterpolator
    if nrow != nrow_geom:
        logger.warn('Discrepency between channel and geometry dimensions')
    for iband in range(len(band_columns)):
        lut_values = numpy.array((rho_atm[:, iband])).reshape(lut_shape)
        lut_func = interpolator(lut_points, lut_values, method='linear',
                                bounds_error=False, fill_value=None)

        _values = lut_func(_points)
        _values = _values.reshape(nrow_geom, ncell_geom)
        _values = _values * numpy.cos(sza * numpy.pi / 180.)

        src_range = numpy.arange(_values.shape[1]) * ac_subsampling
        for irow in range(nrow_geom):
            src_values = _values[irow]
            if irow >= nrow:
                break
            rho[iband, irow] = numpy.interp(dst_range, src_range, src_values)

    # Convert reflectance into radiance TOA:
    # Read Extra-terrestrial solar irradiance
    data_es_path = os.path.join(infile, 'instrument_data.nc')
    data_es = netCDF4.Dataset(data_es_path, 'r')
    solar_flux = data_es.variables['solar_flux'][:]
    data_es.close()
    # 10**(-3) variation for solar flux is neglectable so we consider the mean
    L_toa = []
    for iband in range(len(band_columns)):
        # band_columns correspond to the column of the band in the LUT, which
        # is number of band + 3
        es = numpy.mean(solar_flux[band_columns[iband]-3, :])
        L_toa.append(rho[iband, :, :] * es / numpy.pi)
    return L_toa


def interp_tie_ac(tie_values_list, tie_ac_factor, ac_index):
    """"""
    # Linear interpolation over across-track dimension
    tie_ac_index = numpy.arange(tie_values_list[0].shape[1],
                                dtype='float32') * tie_ac_factor
    indint = numpy.clip(numpy.searchsorted(tie_ac_index, ac_index) - 1,
                        0, tie_ac_index.size - 2)
    indratio = (ac_index - tie_ac_index[indint]) / \
               (tie_ac_index[indint + 1] - tie_ac_index[indint])
    indratio = indratio[numpy.newaxis, :]
    int_values = []
    for val in tie_values_list:
        ac_diff = val[:, indint + 1] - val[:, indint]
        int_val = val[:, indint] + ac_diff * indratio
        int_values.append(int_val)
    return int_values


def compute_tilt(sza, saa, oza, oaa):
    """"""
    cos_sza = numpy.cos(numpy.deg2rad(sza))
    sin_sza = numpy.sin(numpy.deg2rad(sza))
    cos_saa = numpy.cos(numpy.deg2rad(saa))
    sin_saa = numpy.sin(numpy.deg2rad(saa))
    cos_oza = numpy.cos(numpy.deg2rad(oza))
    sin_oza = numpy.sin(numpy.deg2rad(oza))
    cos_oaa = numpy.cos(numpy.deg2rad(oaa))
    sin_oaa = numpy.sin(numpy.deg2rad(oaa))

    # Promoting to float64 is required to avoid underflow errors
    tie_zx_num = (sin_sza * cos_saa + sin_oza * cos_oaa).astype(numpy.float64)
    tie_zx_denom = (cos_sza + cos_oza).astype(numpy.float64)
    tie_zx = -1 * tie_zx_num / tie_zx_denom

    # Promoting to float64 is required to avoid underflow errors
    tie_zy_num = (sin_sza * sin_saa + sin_oza * sin_oaa).astype(numpy.float64)
    tie_zy_denom = (cos_sza + cos_oza).astype(numpy.float64)
    tie_zy = -1 * tie_zy_num / tie_zy_denom

    return (tie_zx, tie_zy)


def compute_spec_angle_mask(input_dir, nrow, ncell, min_angle):
    """"""
    # Compute specular angle
    data_angle_path = os.path.join(input_dir, 'tie_geometries.nc')
    if not os.path.exists(data_angle_path):
        raise TieGeometriesNotFound(data_angle_path)

    data_angle = netCDF4.Dataset(data_angle_path, 'r')
    tie_sza = data_angle.variables['SZA'][:]
    tie_saa = data_angle.variables['SAA'][:]
    tie_oza = data_angle.variables['OZA'][:]
    tie_oaa = data_angle.variables['OAA'][:]
    tie_ac_factor = data_angle.ac_subsampling_factor
    data_angle.close()

    if tie_sza.shape[0] != nrow:
        logger.error('Wrong along track dimension in tie_geometries.nc')
        raise TieGeometriesDimensionMismatch(tie_sza.shape[0], nrow)

    tie_sza = tie_sza[:, :].astype('float32')
    tie_saa = tie_saa[:, :].astype('float32')
    tie_oza = tie_oza[:, :].astype('float32')
    tie_oaa = tie_oaa[:, :].astype('float32')
    tie_zx, tie_zy = compute_tilt(tie_sza, tie_saa, tie_oza, tie_oaa)
    tie_z = numpy.sqrt(tie_zx ** 2. + tie_zy ** 2.)

    ac_index = numpy.arange(0, ncell, dtype='float32')
    z = interp_tie_ac([tie_z, ], tie_ac_factor, ac_index)[0]
    z_angle = numpy.rad2deg(numpy.arctan(z))

    # Mask data when specular angle is too low
    low_specular_angle_mask = (z_angle < min_angle)

    return low_specular_angle_mask


def sentinel3_olci(infile, outdir, vmin=None, vmax=None,
                   slope_threshold=-0.00001, channels='nir',
                   write_netcdf=False, lut_path=None, log_path=None,
                   lat_crop=85.0):
    """"""
    t0 = datetime.utcnow()

    if type(channels) is list or type(channels) is tuple:
        bandnames = channels
        _product_name = 'Sentinel-3{}_OLCI'
    elif 'nir' == channels:
        bandnames = ('Oa17',)
        _product_name = 'Sentinel-3{}_OLCI_NIR'
    elif 'true_rgb' == channels:
        bandnames = ('Oa09', 'Oa06', 'Oa04')
        _product_name = 'Sentinel-3{}_OLCI_true_RGB'
    elif 'false_rgb' == channels:
        bandnames = ('Oa17', 'Oa06', 'Oa04')
        _product_name = 'Sentinel-3{}_OLCI_false_RGB'
    elif 'CHL_NN' == channels:
        bandnames = ('CHL_NN', )
        _product_name = 'Sentinel-3{}_OLCI_Chlorophyll_a_nn'
    elif 'CHL_OC4ME' == channels:
        bandnames = ('CHL_OC4ME', )
        _product_name = 'Sentinel-3{}_OLCI_Chlorophyll_a_oc4me'
    elif 'KD490' == channels:
        bandnames = ('KD490', )
        _product_name = 'Sentinel-3{}_OLCI_KD490_nn'
    elif 'TSM_NN' == channels:
        bandnames = ('TSM_NN', )
        _product_name = 'Sentinel-3{}_OLCI_tsm_nn'
    else:
        raise Exception('channels must be either "nir", "true_rgb" '
                        'or "false_rgb" or CHL_NN or KD490 or TSM_NN'
                        ' or CHL_OC4ME')

    syntool_stats = {}
    for bandname in bandnames:
        syntool_stats[bandname] = {}
    if log_path is not None and not os.path.exists(log_path):
        try:
            os.makedirs(log_path)
        except OSError:
            _, e, _ = sys.exc_info()
            if e.errno != errno.EEXIST:
                raise

    # convert band into column number in the LUT
    band_columns = [int(x[2:])+3-1 for x in bandnames if ('Oa' in x)]

    if vmin is None:
        vmin = [None] * len(bandnames)
    if vmax is None:
        vmax = [None] * len(bandnames)

    full_path = os.path.normpath(infile)
    file_path = os.path.basename(full_path)
    file_name, _ = os.path.splitext(file_path)

    geo_path = os.path.join(infile, 'geo_coordinates.nc')
    time_path = os.path.join(infile, 'time_coordinates.nc')
    quality_path = os.path.join(infile, 'wqsf.nc')

    # Extract geo coordinates information
    geo_handler = netCDF4.Dataset(geo_path, 'r')
    nrow = geo_handler.dimensions['rows'].size
    nrow_all = nrow
    ncell = geo_handler.dimensions['columns'].size
    ncell_all = ncell
    lon = geo_handler.variables['longitude'][:]
    tie_lon = numpy.ma.array(lon)
    lat = geo_handler.variables['latitude'][:]
    tie_lat = numpy.ma.array(lat)
    satellite_id = geo_handler.product_name[2]
    geo_handler.close()

    #if 'A' == satellite_id:
    #    satellite_id = ''  # for historical reason
    product_name = _product_name.format(satellite_id)

    # Handle longitude continuity
    dlon = lon[1:, :] - lon[:-1, :]
    if 180.0 <= numpy.max(numpy.abs(dlon)):
        lon[lon < 0.0] = lon[lon < 0.0] + 360.0

    # Extract time coordinates information
    time_handler = netCDF4.Dataset(time_path, 'r')
    start_timestamp = time_handler.variables['time_stamp'][0]
    end_timestamp = time_handler.variables['time_stamp'][-1]
    timestamp_units = time_handler.variables['time_stamp'].units
    time_handler.close()

    # Format time information
    start_time = netCDF4.num2date(start_timestamp, timestamp_units)
    end_time = netCDF4.num2date(end_timestamp, timestamp_units)
    (dtime, time_range) = stfmt.format_time_and_range(start_time, end_time,
                                                      units='ms')

    parameters = ['{} TOA radiance'.format(bnd) for bnd in bandnames if ('Oa' in bnd)]
    if (len(bandnames) == 1) and ('Oa' not in bandnames[0]):
        parameters = bandnames[0]
    metadata = {}
    metadata['product_name'] = product_name
    metadata['name'] = file_name
    metadata['datetime'] = dtime
    metadata['time_range'] = time_range
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'ESA'
    metadata['processing_center'] = 'OceanDataLab'
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(datetime.utcnow())
    metadata['parameter'] = parameters
    metadata['type'] = 'remote sensing'
    metadata['sensor_type'] = 'medium-resolution imaging spectrometer'
    metadata['sensor_name'] = 'OLCI'
    metadata['sensor_platform'] = 'Sentinel-3'

    # Crop high latitude to avoid projection issues
    LAT_MAX = 89.0
    ind_valid_cols = numpy.where(numpy.abs(tie_lat).max(axis=0) <= LAT_MAX)[0]
    slice_lat0 = slice(None)
    slice_lat1 = slice(numpy.min(ind_valid_cols),
                       numpy.max(ind_valid_cols) + 1)
    tie_lat = tie_lat[slice_lat0, slice_lat1]
    tie_lon = tie_lon[slice_lat0, slice_lat1]
    nrow, ncell = tie_lat.shape

    # Handle longitude continuity
    dlon = tie_lon[1:, :] - tie_lon[:-1, :]
    if 180.0 <= numpy.max(numpy.abs(dlon)):
        lon0 = tie_lon[0, 0] + 180.0
        tie_lon[:, :] = numpy.mod(tie_lon[:, :] - lon0, 360.0) + lon0

    # Try to fit longitudes in [-540, 540] (ingestor constraint)
    while 540.0 < numpy.max(tie_lon):
        tie_lon = tie_lon - 540.0
    while -540.0 > numpy.min(tie_lon):
        tie_lon = tie_lon + 540.0

    # Compute GCPs
    tie_row = numpy.linspace(0, nrow - 1, num=tie_lon.shape[0])
    tie_cell = numpy.linspace(0, ncell - 1, num=tie_lon.shape[1])
    tie_facrow = (nrow - 1.) / (tie_lon.shape[0] - 1.)
    tie_faccell = (ncell - 1.) / (tie_lon.shape[1] - 1.)
    gcp_fac = 128
    gcp_fac = numpy.maximum(gcp_fac, numpy.maximum(tie_faccell, tie_facrow))
    gcp_nrow = numpy.ceil((nrow - 1.) / gcp_fac).astype('int') + 1
    gcp_ncell = numpy.ceil((ncell - 1.) / gcp_fac).astype('int') + 1
    tie_indrow = numpy.round(numpy.linspace(0, tie_lon.shape[0] - 1,
                                            num=gcp_nrow)).astype('int')
    tie_indcell = numpy.round(numpy.linspace(0, tie_lon.shape[1] - 1,
                                             num=gcp_ncell)).astype('int')
    gcp_lon = tie_lon[tie_indrow.reshape((-1, 1)),
                      tie_indcell.reshape((1, -1))]
    gcp_lat = tie_lat[tie_indrow.reshape((-1, 1)),
                      tie_indcell.reshape((1, -1))]
    gcp_row = numpy.tile(tie_row[tie_indrow].reshape((-1, 1)) + 0.5,
                         (1, gcp_ncell))
    gcp_cell = numpy.tile(tie_cell[tie_indcell].reshape((1, -1)) + 0.5,
                          (gcp_nrow, 1))
    gcp_hei = numpy.zeros(gcp_lon.shape)

    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['gcps'] = stfmt.format_gdalgcps(gcp_lon, gcp_lat, gcp_hei,
                                                gcp_cell, gcp_row)

    syntool_stats['lon_min'] = float(numpy.min(tie_lon))
    syntool_stats['lon_max'] = float(numpy.max(tie_lon))
    syntool_stats['lat_min'] = float(numpy.min(tie_lat))
    syntool_stats['lat_max'] = float(numpy.max(tie_lat))

    # Compute atmospheric radiance TOA correction
    L_toa = None
    if 'Oa' in bandnames[0]:
        if lut_path is not None:
            t_start = datetime.utcnow()
            # Compute atmospheric correction for the whole granule
            L_toa = atmospheric_correction(lut_path, infile, band_columns,
                                       nrow_all, ncell_all)
            # Extract the slices of atmospheric correction that match the ones of
            # the bands after the removal of the high latitude columns.
            for iband in range(len(L_toa)):
                L_toa[iband] = L_toa[iband][slice_lat0, slice_lat1]
                # An erroneous LUT could produce nan values for the atmospheric
                # correction. This is not acceptable.
                if numpy.any(~numpy.isfinite(L_toa[iband])):
                    raise Exception('Infinite or nan value found in the '
                                    'atmospheric correction, please check that the'
                                    'LUT has valid values for the angles contained'
                                    'in the instrument_data.nc file')
            t_stop = datetime.utcnow()
            syntool_stats['lut_computation'] = (t_stop - t_start).total_seconds()
    L_toa = None

    logger.info('Construct bands')
    t_start = datetime.utcnow()
    bands = []
    dset = netCDF4.Dataset(quality_path, 'r')
    quality_flags = dset.variables['WQSF'][slice_lat0, slice_lat1]
    _bits = dset.variables['WQSF'].flag_masks
    _bits_meaning_str = dset.variables['WQSF'].flag_meanings
    dset.close()

    # Build quality flags dictionary
    _bits_meaning = _bits_meaning_str.split(' ')
    bits = {}
    for iflag, ibit in zip(_bits_meaning, _bits):
        bits[iflag] = int(numpy.log(ibit)/numpy.log(2))

    # We keep inland waters, whitecaps
    _list_flag = ['INVALID','LAND', 'CLOUD', 'SNOW_ICE', 'COSMETIC', 'SUSPECT',
                  #'HISOLZEN',
                  # Remove Hisolzen flag as it removes high latitude data
                 'SATURATED', #'WHITECAPS,]
                 'ADJAC', 'WV_FAIL', 'PAR_FAIL',# 'AC_FAIL',
                 #'KDM_FAIL',
                 'CLOUD_MARGIN', 'CLOUD_AMBIGUOUS'] # 
    if 'CHL_NN' in bandnames:
        list_flag = _list_flag
        list_flag.append('OCNN_FAIL')
        #list_flag.append('MEGLINT')
        #list_flag.append('HIGHGLINT')
    if 'CHL_OC4ME' in bandnames:
        list_flag = _list_flag
        list_flag.append('OC4ME_FAIL')
        #list_flag.append('MEGLINT')
        #list_flag.append('HIGHGLINT')
    off_flags = numpy.uint32(0)

    for nflag in list_flag:
        off_flags = off_flags + numpy.uint32(1 << bits[nflag])

    lat_mask = (numpy.abs(tie_lat) > lat_crop)


    contrast_mask = numpy.zeros(quality_flags.shape, dtype='bool')
    contrast_mask = (contrast_mask | lat_mask)
    contrast_mask = (contrast_mask |
                     (numpy.bitwise_and(quality_flags, off_flags) > 0))
    if numpy.all(contrast_mask):
        logger.warn('No data to extract.')
        raise syntool_converter.EarlyExit()
    t_stop = datetime.utcnow()
    syntool_stats['mask_computation'] = (t_stop - t_start).total_seconds()

    t_start = datetime.utcnow()
    _vmin = list(vmin)
    _vmax = list(vmax)
    if len(bandnames) > 1:
        for band_index in range(len(bandnames)):
            bandname = bandnames[band_index]
            logger.info('\tConstruct {} band'.format(bandname))
            if 'Oa' in bandname:
                fieldname = '{}_reflectance'.format(bandname)
                varname = fieldname
            elif ('NN' in bandname) or ('OC4ME' in bandname):
                fieldname = bandname.lower()
                varname = bandname
            else:
                fieldname = 'trsp'
                varname = 'KD490_M07'
            file_path = os.path.join(infile, '{}.nc'.format(fieldname))
            f_handler = netCDF4.Dataset(file_path, 'r')
            band = f_handler.variables[varname][:]
            band = numpy.ma.array(band)
            f_handler.close()

            # Apply atmospheric correction
            band = band[slice_lat0, slice_lat1]
            if L_toa is not None:
                band -= L_toa[band_index][:, :]

            # Mask null and negative values: they are inferior or equal to
            # atmospheric correction and should probably have be flagged as clouds.
            if 'Oa' in bandname:
                mask_negative = (band <= 0.0)

            logger.info('\tSet contrast')
            valid_ratio_lower_threshold = 0.001  # 0.1%
            valid_data_mask = (band.mask | contrast_mask)
            lat_mask = (band.mask | lat_mask)
            if 'Oa' in bandname:
                valid_data_mask = (valid_data_mask | mask_negative)
            valid_data = band.data[~valid_data_mask]
            valid_ratio = float(valid_data.size) / float(band.data.size)
            if valid_ratio <0.01:
                sys.exit()
            syntool_stats[bandname]['valid_ratio'] = valid_ratio
            if valid_ratio_lower_threshold >= valid_ratio:
                #logger.warn('No valid values for {}'.format(bandname))
                logger.warn('Using default min/max.')

                # Use arbitrary extrema on land
                if _vmin[band_index] is None:
                    _min = default_minmax[bandname][0]
                    _vmin[band_index] = _min
                    syntool_stats[bandname]['default_min'] = float(_min)
                if _vmax[band_index] is None:
                    _max = default_minmax[bandname][1]
                    _vmax[band_index] = _max
                    syntool_stats[bandname]['default_max'] = float(_max)
            else:
                # TODO: add clipping for NIR
                # _min = numpy.clip(_min, 1.5, 10.0)
                # _max = numpy.clip(_max, 30.0, 60.0)
                if _vmin[band_index] is None:
                    _min = numpy.percentile(valid_data, 1)
                    _vmin[band_index] = _min
                    syntool_stats[bandname]['p0050'] = float(_min)
                    syntool_stats[bandname]['min'] = float(numpy.min(valid_data))
                if _vmax[band_index] is None:
                    _max = numpy.percentile(valid_data, 90)
                    _vmax[band_index] = _max
                    p99 = numpy.percentile(valid_data, 99.0)
                    syntool_stats[bandname]['p9900'] = float(p99)
                    syntool_stats[bandname]['p9999'] = float(_max)
                    syntool_stats[bandname]['max'] = float(numpy.max(valid_data))
	    logger.info('\tContrast : vmin={} / vmax={}'.format(_vmin[band_index],
								_vmax[band_index]))
    else:
        # Single band
        band_index = 0

	    # Use arbitrary extrema on land
        if _vmin[band_index] is None:
            _min = default_minmax[bandname][0]
            _vmin[band_index] = _min
            syntool_stats[bandname]['default_min'] = float(_min)
        if _vmax[band_index] is None:
            _max = default_minmax[bandname][1]
            _vmax[band_index] = _max
            syntool_stats[bandname]['default_max'] = float(_max)

    logger.warn('\tContrast : vmin={} / vmax={}'.format(_min, _max))

    min_values = [_vmin[band_index] for band_index in range(len(bandnames))]
    max_values = [_vmax[band_index] for band_index in range(len(bandnames))]

    t_stop = datetime.utcnow()
    syntool_stats['minmax_computation'] = (t_stop - t_start).total_seconds()
    syntool_stats['final_min'] = float(numpy.min(min_values))
    syntool_stats['final_max'] = float(numpy.max(max_values))

    if 'Oa' in bandname:
        _min = numpy.log(numpy.min(min_values))
        _max = numpy.log(numpy.max(max_values))
    else:
        _min = numpy.min(min_values)
        _max = numpy.max(max_values)

    spec_angle_filtered = [_ for _ in bandnames if _ in SPEC_ANGLE_MASKED_VARS]
    spec_angle_mask = None
    if 0 < len(spec_angle_filtered):
        spec_angle_mask = compute_spec_angle_mask(infile, nrow_all, ncell_all,
                                                  min_angle=9)

    for band_index in range(len(bandnames)):
        bandname = bandnames[band_index]
        logger.info('\tConstruct {} band'.format(bandname))
        if 'Oa' in bandname:
            fieldname = '{}_reflectance'.format(bandname)
            varname = fieldname
        elif ('NN' in bandname) or ('OC4ME' in bandname):
            fieldname = bandname.lower()
            varname = bandname
        else:
            fieldname = 'trsp'
            varname = 'KD490_M07'
        file_path = os.path.join(infile, '{}.nc'.format(fieldname))
        f_handler = netCDF4.Dataset(file_path, 'r')
        band = f_handler.variables[varname][:]
        band = numpy.ma.array(band)
        f_handler.close()

        # Mask values where the specular angle is too low (mostly noise)
        if varname in SPEC_ANGLE_MASKED_VARS:
            band = numpy.ma.masked_where(spec_angle_mask, band)

        # Apply atmospheric correction
        band = band[slice_lat0, slice_lat1]
        if L_toa is not None:
            band -= L_toa[band_index][:, :]

        # Mask with flags and high latitudes
        valid_data_mask = (band.mask | contrast_mask)

        # Mask null and negative values: they are inferior or equal to
        # atmospheric correction and should probably have be flagged as clouds.
        if 'Oa' in bandname:
            mask_negative = (band <= 0.0)
            valid_data_mask = (valid_data_mask | mask_negative)

        # Exit if there is not enough data (less than 1\%)
        valid_data = band.data[~valid_data_mask]
        valid_ratio = float(valid_data.size) / float(band.data.size)
        if valid_ratio <0.01:
            logger.warn('less than 1% value found for {}'.format(bandname))
            raise syntool_converter.EarlyExit()

        # Compute the logarithm only for radiance values that are higher than
        # the atmospheric correction.
        if 'Oa' in bandname:
            bnd = numpy.log(band.data, where=(~mask_negative))
        else:
            bnd = + scipy.signal.medfilt(band.data, kernel_size=(3, 3))

        logger.info('\tBytescaling')
        byte, offset, scale = pack.ubytes_0_254(bnd, _min, _max)
        description = '{} reflectance (log)'.format(bandname)
        if band.mask is not numpy.ma.nomask:
            byte[band.mask] = 255

        # mask data for extreme latitudes
        if 'Oa' in bandname:
            byte[lat_mask] = 255
        else:
            byte[valid_data_mask] = 255

        # Pixels with a radiance equal or inferior to atmospheric correction
        # are clipped to the minimal value.
        if 'Oa' in bandname:
            if 0 < mask_negative.size:
                byte[numpy.where(mask_negative == True)] = 0  # noqa

        band_range = [_vmin[band_index], _vmax[band_index]]
        colortable = stfmt.format_colortable('chla_jet',
                                              vmax=_max, vmax_pal=_max,
                                              vmin=_min, vmin_pal=_min)

        if 'Oa' in bandname:
            bands.append({'array': byte,
                          'scale': scale,
                          'offset': offset,
                          'description': description,
                          'unittype': '',
                          'nodatavalue': 255,
                          'parameter_range': band_range,
                          })
        else:
            bands.append({'array': byte,
                          'scale': scale,
                          'offset': offset,
                          'description': description,
                          'unittype': '',
                          'nodatavalue': 255,
                          'parameter_range': band_range,
                          'colortable': colortable
                          })
        if write_netcdf:
            bands[-1]['name'] = bandname
            bands[-1]['long_name'] = bandname
            bands[-1]['unittype'] = '1'

    logger.info('Make sure nodata are at the same locations in all bands')
    mask = numpy.any([_band['array'] == 255 for _band in bands], axis=0)
    for band in bands:
        band['array'][mask] = 255

    t_stop = datetime.utcnow()
    syntool_stats['bytescaling'] = (t_stop - t_start).total_seconds()

    if write_netcdf:
        metadata['spatial_resolution'] = 300
        ncfile = stfmt.format_ncfilename(outdir, metadata, create_dir=True)
        stfmt.write_netcdf(ncfile, metadata, geolocation, bands, 'swath',
                           ngcps=gcp_lon.shape)
    else:
        logger.info('Write geotiff')
        tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
        stfmt.write_geotiff(tifffile, metadata, geolocation, bands)

    logger.info(datetime.utcnow() - t0)
    syntool_stats['total_time'] = (datetime.utcnow() - t0).total_seconds()
    if log_path is not None:
        import json
        stats_path = os.path.join(log_path, '{}.json'.format(file_name))
        with open(stats_path, 'w') as f:
            json.dump(syntool_stats, f)
