# -*- encoding: utf-8 -*-

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys
import numpy
import logging
from datetime import datetime
import syntool_converter
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.sentinel3.slstr as slstr
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)

# https://earth.esa.int/web/sentinel/technical-guides/sentinel-3-slstr/level-1/vis-and-swir-radiances
# https://sentinel.esa.int/web/sentinel/user-guides/sentinel-3-slstr/data-formats/level-1
# https://earth.esa.int/web/sentinel/user-guides/sentinel-3-slstr/product-types/level-1b
# https://sentinel.esa.int/documents/247904/685236/Sentinel-3_User_Handbook#%5B%7B%22num%22%3A235%2C%22gen%22%3A0%7D%2C%7B%22name%22%3A%22XYZ%22%7D%2C54%2C628%2C0%5D
# S1 555   nm            B
# S2 659   nm            G
# S3 865   nm            R
# S4 1.375 microm
# S5 1.610 microm
# S6 2.25  microm
# S7 3.74  microm        IR
# S8 10.85 microm        IR
# S9 12    microm        IR


default_minmax = {'S7': (220, 310),
                  'S8': (197.4, 306),
                  'S9': (190.15, 306.15)}


def build_mask_ir(channels, quality_flags, raw_cloud_flags, tie_lat, lat_crop):
    """Build masks to compute contrast and to mask invalid data"""

    bits = {'coastline': 0,
            'ocean': 1,
            'tidal': 2,
            'land': 3,
            'inland_water': 4,
            'unfilled': 5,
            'spare': 6,
            'spare_': 7,
            'cosmetic': 8,
            'duplicate': 9,
            'day': 10,
            'twilight': 11,
            'sun_glint': 12,
            'snow': 13,
            'summary_cloud': 14,
            'summary_pointing': 15}

    # Invalidate if these flags are all off
    on_flags = numpy.ushort(0)
    on_flags = on_flags + numpy.ushort(1 << bits['day'])

    # Invalidate if any of these flags is on
    off_flags = numpy.ushort(0)
    off_flags = off_flags + numpy.ushort(1 << bits['coastline'])
    off_flags = off_flags + numpy.ushort(1 << bits['land'])
    off_flags = off_flags + numpy.ushort(1 << bits['snow'])

    # Intialize mask to compute histograms on selected valid values
    contrast_mask = numpy.zeros(quality_flags.shape, dtype='bool')

    # Initialize mask to remove invalid data
    data_mask = numpy.zeros(quality_flags.shape, dtype='bool')

    lat_mask = (numpy.abs(tie_lat) > lat_crop)
    contrast_mask = (contrast_mask | lat_mask)
    data_mask = (data_mask | lat_mask)

    return contrast_mask, data_mask


def sentinel3_slstr_bt(infile, outdir, vmin=None, vmax=None,
                       min_percentile=2.0, channels='S9', file_range=None,
                       write_netcdf=False, log_path=None, lat_crop=80.0):

    t0 = datetime.utcnow()
    # Process nadir data
    view = 'n'
    fname = 'BT'
    sltype = 'i'
    if type(channels) is list or type(channels) is tuple:
        bandnames = channels
        _product_name = 'Sentinel-3{}_SLSTR'
    elif 'S8' == channels:
        bandnames = ('S8',)
        _product_name = 'Sentinel-3{}_SLSTR_S8'
    elif 'S7' == channels:
        bandnames = ('S7',)
        _product_name = 'Sentinel-3{}_SLSTR_S7'
    elif 'S9' == channels:
        bandnames = ('S9',)
        _product_name = 'Sentinel-3{}_SLSTR_S9'
    else:
        raise Exception('channels must be either "S7", "S8", "S9" or a tuple '
                        'containing several of these bands')

    satellite_id = slstr.get_satellite(infile, sltype, view)
    if 'A' == satellite_id:
        satellite_id = ''  # for historical reason
    product_name = _product_name.format(satellite_id)

    # Read coordinates and compute gcps
    (syntool_stats, metadata, geolocation, tie_lon, tie_lat, slice_lat0,
        slice_lat1, __, __, ngcps, month) = slstr.read_geometry(infile,
                                             bandnames, fname, sltype, view,
                                             product_name, vmin, vmax,
                                             log_path)

    # Compute masks
    logger.info('Build masks')
    t_start = datetime.utcnow()
    quality_flags = slstr.read_mask(infile, sltype, view, slice_lat0,
                                    slice_lat1)
    raw_cloud_flags = slstr.read_cloud_mask(infile, sltype, view, slice_lat0,
                                            slice_lat1)
    contrast_mask, data_mask = build_mask_ir(channels, quality_flags,
                                             raw_cloud_flags, tie_lat,
                                             lat_crop)
    t_stop = datetime.utcnow()
    syntool_stats['mask_computation'] = (t_stop - t_start).total_seconds()

    # Read band to compute histograms
    logger.info('Construct bands')
    t_start = datetime.utcnow()

    bands = []
    # Initialize min and max values
    if vmin is None:
        vmin = [None] * len(bandnames)
    if vmax is None:
        vmax = [None] * len(bandnames)
    _vmin = list(vmin)
    _vmax = list(vmax)
    for band_index in range(len(bandnames)):
        bandname = bandnames[band_index]
        fieldname = slstr.get_field_name(fname, bandname, sltype, view)
        band = slstr.read_band(infile, bandname, fieldname, slice_lat0,
                               slice_lat1)

        logger.info('\tSet contrast')

        # Select valid data to compute histograms
        valid_data_mask = (band.mask | data_mask) # | contrast_mask)
        data_mask = (data_mask | valid_data_mask)

        # No need to produce an output if all data values are masked
        if numpy.all(data_mask):
            logger.warn('No valid value found for band {}'.format(bandname))
            raise syntool_converter.EarlyExit()

        # Retrieve minimum and maximum values from default or valid_data
        # histograms
        _min, _max = slstr.apply_default_min_max(default_minmax, bandname,
                                                 _vmin[band_index],
                                                 _vmax[band_index],
                                                 syntool_stats)
        _vmin[band_index] = _min
        _vmax[band_index] = _max
        logger.info('\tContrast : vmin={} / vmax={}'.format(_vmin[band_index],
                                                            _vmax[band_index]))
    min_values = [_vmin[band_index] for band_index in range(len(bandnames))]
    max_values = [_vmax[band_index] for band_index in range(len(bandnames))]

    t_stop = datetime.utcnow()
    syntool_stats['minmax_computation'] = (t_stop - t_start).total_seconds()
    syntool_stats['final_min'] = float(numpy.min(min_values))
    syntool_stats['final_max'] = float(numpy.max(max_values))

    _min = numpy.min(min_values)
    _max = numpy.max(max_values)
    # Construct bands
    for band_index in range(len(bandnames)):
        bandname = bandnames[band_index]
        fieldname = slstr.get_field_name(fname, bandname, sltype, view)
        band = slstr.read_band(infile, bandname, fieldname, slice_lat0,
                               slice_lat1)

        bnd = band.data

        logger.info('\tBytescaling')
        byte, offset, scale = pack.ubytes_0_254(bnd, _min, _max)
        description = '{} {} (log)'.format(bandname, fname)
        if band.mask is not numpy.ma.nomask:
            byte[band.mask] = 255

        # mask night data for rgb and invalid data for ir (cloud, land,
        # range value). Also mask data for extreme latitudes
        byte[data_mask] = 255

        band_range = [_vmin[band_index], _vmax[band_index]]
        description = '{} {}'.format(bandname, fname)  # no log for IR
        colortable = stfmt.format_colortable('medspi_contrast',
                                             vmax=_max, vmax_pal=_max,
                                             vmin=_min, vmin_pal=_min)
        bands.append({'array': byte,
                      'plot': band.data,
                      'scale': scale,
                      'offset': offset,
                      'description': description,
                      'unittype': 'K',
                      'nodatavalue': 255,
                      'parameter_range': band_range,
                      'colortable': colortable})

        if write_netcdf:
            bands[-1]['name'] = bandname
            bands[-1]['long_name'] = bandname
            bands[-1]['unittype'] = 'K'

    logger.info('Make sure nodata are at the same locations in all bands')
    mask = numpy.any([_band['array'] == 255 for _band in bands], axis=0)
    for band in bands:
        band['array'][mask] = 255

    t_stop = datetime.utcnow()
    syntool_stats['bytescaling'] = (t_stop - t_start).total_seconds()

    if write_netcdf:
        metadata['spatial_resolution'] = 1000
        ncfile = stfmt.format_ncfilename(outdir, metadata, create_dir=True)
        stfmt.write_netcdf(ncfile, metadata, geolocation, bands, 'swath',
                           ngcps=ngcps)
    else:
        logger.info('Write geotiff')
        tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
        stfmt.write_geotiff(tifffile, metadata, geolocation, bands)

    logger.info(datetime.utcnow() - t0)
    syntool_stats['total_time'] = (datetime.utcnow() - t0).total_seconds()
    if log_path is not None:
        import json
        full_path = os.path.normpath(infile)
        file_path = os.path.basename(full_path)
        file_name, _ = os.path.splitext(file_path)
        stats_path = os.path.join(log_path, '{}.json'.format(file_name))
        with open(stats_path, 'w') as f:
            json.dump(syntool_stats, f)
