# -*- encoding: utf-8 -*-

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import sys
import numpy
import netCDF4
import logging
import bisect
from scipy.signal import convolve2d
import syntool_converter
import syntool_converter.utils.syntoolformat as stfmt
from scipy.misc import bytescale
from datetime import datetime
import syntool_converter.sentinel3.slstr as slstr

logger = logging.getLogger(__name__)

# https://earth.esa.int/web/sentinel/technical-guides/sentinel-3-slstr/level-1/vis-and-swir-radiances
# https://sentinel.esa.int/web/sentinel/user-guides/sentinel-3-slstr/data-formats/level-1
# https://earth.esa.int/web/sentinel/user-guides/sentinel-3-slstr/product-types/level-1b
# https://sentinel.esa.int/documents/247904/685236/Sentinel-3_User_Handbook#%5B%7B%22num%22%3A235%2C%22gen%22%3A0%7D%2C%7B%22name%22%3A%22XYZ%22%7D%2C54%2C628%2C0%5D
# S1 555   nm            B
# S2 659   nm            G
# S3 865   nm            R
# S4 1.375 microm
# S5 1.610 microm
# S6 2.25  microm
# S7 3.74  microm        IR
# S8 10.85 microm        IR
# S9 12    microm        IR


default_minmax = {'sst': (270, 305),
                  'high_latitude': (271.15, 301.15),
                  'med_latitude': (274.15, 306.15),
                  'warm_sea':(284.15, 310.15)}


def read_manifest(nfile):
    """ Read metadata in xml file. Return start and track Offset, number of
    lines and elements spatial resolution for Tie Points, 1 km nadir and 1 km
    oblique grid."""
    import xml.etree.ElementTree as ET
    tree = ET.parse(nfile)
    root = tree.getroot()
    param = {}
    s3path = '{http://www.esa.int/safe/sentinel/sentinel-3/1.0}'
    slstrpath = '{http://www.esa.int/safe/sentinel/sentinel-3/slstr/1.0}'
    for a_child in root.findall("./*/metadataObject/metadataWrap/xmlData/*"):
        for child in a_child:
            if child.tag == '{}nadirImageSize'.format(slstrpath):
                if child.attrib['grid'] == '1 km':
                    for item in child.findall("./*"):
                        if item.tag == '{}startOffset'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['startOffset_n'] = float(_item)
                        if item.tag == '{}trackOffset'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['trackOffset_n'] = float(_item)
                        if item.tag == '{}rows'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['line_n'] = int(_item)
                        if item.tag == '{}columns'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['elem_n'] = int(_item)
                if child.attrib['grid'] == 'Tie Points':
                     for item in child.findall("./*"):
                        if item.tag == '{}startOffset'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['startOffset_tx'] = float(_item)
                        if item.tag == '{}trackOffset'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['trackOffset_tx'] = float(_item)
                        if item.tag == '{}rows'.format(s3path):
                            param['line_tx'] = int(item.text.encode('utf8'))
                        if item.tag == '{}columns'.format(s3path):
                            param['elem_tx'] = int(item.text.encode('utf8'))
            if child.tag == '{}obliqueImageSize'.format(slstrpath):
                if child.attrib['grid'] == '1 km':
                    for item in child.findall("./*"):
                        if item.tag == '{}startOffset'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['startOffset_o'] = float(_item)
                        if item.tag == '{}trackOffset'.format(s3path):
                            _item = item.text.encode('utf8')
                            param['trackOffset_o'] = float(_item)
                        if item.tag == '{}rows'.format(s3path):
                            param['line_o'] = int(item.text.encode('utf8'))
                        if item.tag == '{}columns'.format(s3path):
                            param['elem_o'] = int(item.text.encode('utf8'))
            if child.tag == '{}resolution'.format(slstrpath):
                if child.attrib['grid'] == '1 km':
                    for item in child.findall("./*"):
                        if item.tag == '{}spatialResolution'.format(slstrpath):
                            param['res_n'] = float(item.text.encode('utf8'))
                if child.attrib['grid'] == 'Tie Points':
                    for item in child.findall("./*"):
                        if item.tag == '{}spatialResolution'.format(slstrpath):
                            param['res_tx'] = float(item.text.encode('utf8'))
    return param


def coordstoaffin(lon, lat):
    """ Function to compute affine parameters corresponding to the polyline
    coordinates. """
    alpha = []
    beta = []
    for i in range(len(lon)-1):
        _alpha = (lat[i + 1] - lat[i]) / (lon[i + 1] - lon[i])
        _beta = lat[i] - lon[i] * _alpha
        alpha.append(_alpha)
        beta.append(_beta)
    return alpha, beta


def polyline_north():
    """ Polyline to define high latitudes in northern atmosphere (annular ring)
    """
    lon = [-180, -169.01, -156.43, -150.60, -95.60, -86.59,
           -75.77, -65.60, -9.143, 0.14, 8.30, 32.85, 34.47,
           37.16, 56.47, 120.44, 180]
    lat = [46.78, 52.84, 57.27, 62.66, 48.36, 36.23,
           36.06, 43.47, 43.00, 43.00, 47.13, 47.66, 45.25,
           45.20, 48.07, 36.81, 46.78]
    alpha, beta = coordstoaffin(lon, lat)
    return lon, lat, alpha, beta


def polyline_south():
    """ Polyline to define high latitudes in southern atmosphere (annular ring)
    """
    lon = [-180, -71.03, -64.94, 167.45, 180]
    lat = [-45.43, -44.87, -42.98, -45.43, -45.43]
    alpha, beta = coordstoaffin(lon, lat)
    return lon, lat, alpha, beta


def polyline_baja():
    """ Polyline to define warm pool for Baja region"""
    lon = [-116.24, -109.88, -105.68, -100]
    lat = [32.1, 22.92, 20.40, 20.40]
    alpha, beta = coordstoaffin(lon, lat)
    return lon, lat, alpha, beta


def polygon_warm_sea():
    """ define warm pool box: [lon_min, lon_max, lat_min, lat_max]"""
    box = []
    # Red sea / Gulf of Aden
    box.append([27.73, 50.89, 10.27, 30.64])
    # Persian / Oman Gulf
    box.append([42.67, 59.798, 22.24, 33.06])
    # Baja
    # box.append([-115.07, -112.02, 29.5, 32.16])
    # box.append([-113.97, -109.77, 27.03, 29.56])
    # box.append([-112.03, -110.39, 24.85, 27.27])
    # box.append([-110.90,-107.29, 23.99, 27.32])
    # box.append([-109.97, -106.05, 22.88, 27.14])
    # Gulf of Mexico
    # box.append([-100.12, -81.89, 17.13, 31.03])
    # box.append([-89.67, -80.44, 15.20, 26.98])

    return box


def mask_from_polyline(list_lon, list_lat, alpha, beta, tie_lon_masked,
                       tie_lat_masked, tie_lat, above=False):
    """ Compute mask from polyline and granule """
    min_lon = numpy.min(tie_lon_masked)
    max_lon = numpy.max(tie_lon_masked)
    mask_out = numpy.full(numpy.shape(tie_lat), False, dtype=numpy.bool)
    bisect_l = bisect.bisect_left

    polyline_lon_min = numpy.min(list_lon)
    polyline_lon_max = numpy.max(list_lon)
    _tie_lon_masked = tie_lon_masked[numpy.where((tie_lon_masked <= polyline_lon_max) & (tie_lon_masked >= polyline_lon_min))]
    lon_l = numpy.mod(numpy.min(_tie_lon_masked) + 180.0, 360.0) - 180.0
    lon_r = numpy.mod(numpy.max(_tie_lon_masked) + 180.0, 360.0) - 180.0
    i_min_l = numpy.max([0, bisect_l(list_lon, lon_l) - 1])
    i_min_r = numpy.max([0, bisect_l(list_lon, lon_r) - 1])
    def mask_above(data, ref):
        return (abs(data) >= abs(ref))
    def mask_under(data, ref):
        return (abs(data) <= abs(ref))
    lat_mask = mask_under
    if above is True:
        lat_mask = mask_above
    if i_min_l == i_min_r:
        test_lat = alpha[i_min_l] * tie_lon_masked + beta[i_min_l]
        mask_out = lat_mask(tie_lat_masked, test_lat)
    else:
        if (max_lon - min_lon) > 180:
            for k in range(0, len(alpha)):
                test_lat = alpha[k] * tie_lon_masked + beta[k]
                # polylines have to be wrapped
                segment_mask = (lat_mask(tie_lat_masked, test_lat)
                                & (tie_lon_masked <= list_lon[k + 1])
                                & (tie_lon_masked > list_lon[k]))
                mask_out = (mask_out | segment_mask)
        else:
            for k in range(i_min_l, i_min_r + 1):
                test_lat = alpha[k] * tie_lon_masked + beta[k]
                segment_mask = (lat_mask(tie_lat_masked, test_lat)
                                 & (tie_lon_masked <= list_lon[k + 1])
                                 & (tie_lon_masked > list_lon[k]))
                mask_out = (mask_out | segment_mask)
    return mask_out


def build_mask_minmax(tie_lon, tie_lat, data_mask):
    """ Compute mask and corresponding min max for each area (warm pool, med
    latitude or high latitude). """
    tie_lat_masked = numpy.ma.masked_array(tie_lat, mask=data_mask)
    tie_lon_masked = numpy.ma.masked_array(tie_lon, mask=data_mask)
    tie_lon_masked = numpy.mod(tie_lon_masked + 180.0, 360.0) - 180.0
    min_lat = numpy.min(tie_lat_masked)
    max_lat = numpy.max(tie_lat_masked)
    min_lon = numpy.min(tie_lon_masked)
    max_lon = numpy.max(tie_lon_masked)
    # check if granule min max corresponds to high or low latitude, granule can
    # be splitted.
    list_mask = []
    # Check if all latitudes are higher than the northern polyline point
    if (min_lat > 62.66):
        vmin = (default_minmax['high_latitude'][0], )
        vmax = (default_minmax['high_latitude'][1], )
        list_mask.append(numpy.full(numpy.shape(tie_lat), True,
                         dtype=numpy.bool))
    # Check if all latitudes are higher than the southern polyline point
    elif (max_lat < -48.13).all() is True:
        vmin = (default_minmax['high_latitude'][0], )
        vmax = (default_minmax['high_latitude'][1], )
        list_mask.append(numpy.full(numpy.shape(tie_lat), True,
                         dtype=numpy.bool))
    # Check if all latitudes are lower than the lowest latitude in polyline
    else:
        vmin = []
        vmax = []
        #vmin = (default_minmax['med_latitude'][0], )
        #vmax = (default_minmax['med_latitude'][1], )
        # Build mask for warm pool
        mask_b = numpy.full(numpy.shape(tie_lat), False, dtype=numpy.bool)
        bbox = polygon_warm_sea()
        for box in bbox:
            box_mask = ((tie_lon_masked > box[0]) & (tie_lon_masked < box[1])
                        & (tie_lat_masked > box[2]) & (tie_lat_masked < box[3]))
            mask_b = (mask_b | box_mask)
        # If granule is near Baja box
        baja_box = [-115.1, -104.44, 20.38, 32.10]
        baja_box_mask = ((tie_lon_masked > baja_box[0])
                         & (tie_lon_masked < baja_box[1])
                         & (tie_lat_masked > baja_box[2])
                         & (tie_lat_masked < baja_box[3]))
        if numpy.any(baja_box_mask):
            list_lon, list_lat, alpha, beta = polyline_baja()
            mask_baja = mask_from_polyline(list_lon, list_lat, alpha, beta,
                                           tie_lon_masked, tie_lat_masked,
                                           tie_lat, above=True)
            mask_b = (mask_b | mask_baja)
        # if there are points in warm pool
        if numpy.any(mask_b):
            vmin.append(default_minmax['warm_sea'][0])
            vmax.append(default_minmax['warm_sea'][1])
            list_mask.append(mask_b)
        # Build mask for high latitude
        if (numpy.mean(tie_lat_masked) > 0):
            list_lon, list_lat, alpha, beta = polyline_north()
        else:
            list_lon, list_lat, alpha, beta = polyline_south()
        # Get index for affine function parameters 
        mask = mask_from_polyline(list_lon, list_lat, alpha, beta,
                                  tie_lon_masked, tie_lat_masked, tie_lat)
        if numpy.all(mask):
            mask = (mask & ~mask_b)
            vmin.append(default_minmax['med_latitude'][0])
            vmax.append(default_minmax['med_latitude'][1])
            list_mask.append(mask)
        elif not numpy.any(mask):
            mask = (~mask & ~mask_b)
            vmin.append(default_minmax['high_latitude'][0])
            vmax.append(default_minmax['high_latitude'][1])
            list_mask.append(mask)
        else:
            vmin.append(default_minmax['high_latitude'][0])
            vmin.append(default_minmax['med_latitude'][0])
            vmax.append(default_minmax['high_latitude'][1])
            vmax.append(default_minmax['med_latitude'][1])
            list_mask.append(~mask)
            # One assumes that no warm pool is in high latitude area
            mask = (mask & ~mask_b)
            list_mask.append(mask)
    return vmin, vmax, list_mask


def build_mask_ir(quality_flags, raw_cloud_flags, collection, tie_lat,
                  lat_crop):
    """Build masks to compute contrast and to mask invalid data"""

    bits = {'coastline': 0,
            'ocean': 1,
            'tidal': 2,
            'land': 3,
            'inland_water': 4,
            'unfilled': 5,
            'spare': 6,
            'spare_': 7,
            'cosmetic': 8,
            'duplicate': 9,
            'day': 10,
            'twilight': 11,
            'sun_glint': 12,
            'snow': 13,
            'summary_cloud': 14,
            'summary_pointing': 15}

    if collection == '002':
        cloud_bits_to_mask = ('visible',
                              'threshold',
                              #'small_histogram1.6',
                              #'small_histogram2.25',
                              'spatial_coherence',
                              'gross_cloud',
                              'thin_cirrus',
                              'medium_high',
                              'fog_low_stratus',
                              #'11_12_view_difference',
                              )
    elif collection in ('003', '004',):
        cloud_bits_to_mask = ('visible',
                              'threshold',
                              'small_histogram1.6',
                              'large_histogram1.6',
                              'small_histogram2.25',
                              'large_histogram2.25',
                              'spatial_coherence',
                              'gross_cloud',
                              'thin_cirrus',
                              'medium_high',
                              'fog_low_stratus',
                              '11_12_view_difference',
                              ##'3.7_11_view_difference',
                              )
    else:
        logger.error('collection {} should be 002, 003 or 004'.format(collection))
        sys.exit(1)

    # Invalidate if these flags are all off
    on_flags = numpy.ushort(0)
    on_flags = on_flags + numpy.ushort(1 << bits['day'])

    # Invalidate if any of these flags is on
    off_flags = numpy.ushort(0)
    off_flags = off_flags + numpy.ushort(1 << bits['coastline'])
    off_flags = off_flags + numpy.ushort(1 << bits['land'])
    off_flags = off_flags + numpy.ushort(1 << bits['snow'])

    # Initialize mask to remove invalid data
    data_mask = numpy.zeros(quality_flags.shape, dtype='bool')

    lat_mask = (numpy.abs(tie_lat) > lat_crop)
    data_mask = (data_mask | lat_mask)

    land_mask = slstr.build_land_mask(quality_flags, bits)

    cloud_mask = slstr.build_cloud_mask(raw_cloud_flags, cloud_bits_to_mask)

    # Do not mask data at night or twilight for ir channels.
    data_mask = (data_mask |
                 cloud_mask |
                 (numpy.bitwise_and(quality_flags, off_flags) > 0))

    return data_mask


def gen_fractional_grid(elem_n, line_n, param):
    """" Generate fractional grid from xml file parameters """
    # retrieve parameters
    elem_n = param['elem_n'] # Across-track
    line_n = param['line_n'] # Along-track
    # Track offset nadir
    toffset_n = param['trackOffset_n']
    # Track offset tie point
    toffset_tx = param['trackOffset_tx']
    # Start offset nadir
    soffset_n = param['startOffset_n']
    # Start offset tie point
    soffset_tx = param['startOffset_tx']
    # Spatial resolution nadir
    res_n = param['res_n']
    # Spatial resolution tie points
    res_tx = param['res_tx']
    # Interpolating tie point fractional grid
    dj_in = numpy.zeros((elem_n), dtype='float')
    di_in = numpy.zeros((line_n), dtype='float')

    # Build fractional grid
    x_tx = toffset_tx * res_tx
    y_tx = soffset_tx * res_n
    start_index = 0
    for j in range (elem_n):
        x_in = (toffset_n - j) * res_n
        dj_in[j] = (x_tx - x_in) / res_tx

    for j in range(start_index, line_n + start_index):
        y_in = (soffset_n + j) * res_n
        di_in[j-start_index] = (y_in - y_tx) / res_n
    return di_in, dj_in


def filter_sst(sst, bt8, kernel_size):
    filter_kernel = numpy.ones(kernel_size) / (kernel_size[0] * kernel_size[1])
    sst_diff = sst - bt8
    diff_mean = convolve2d(sst_diff, filter_kernel, mode='same',
                                  boundary='fill', fillvalue=numpy.nan)
    sst_filtered = diff_mean + bt8
    sst_filtered[0, :] = sst[0, :]
    sst_filtered[-1, :] = sst[-1, :]

    return sst_filtered


def get_alongtrack_contour(values, offset):
    """Expects along-track to be the first axis"""
    l = numpy.shape(values)[0]
    idx_i = []
    idx_j = []
    for i in range(0, l):
        valid_values = numpy.where(values.mask[i] == False)
        if 0 >= numpy.size(valid_values):
            continue
        first = numpy.min(valid_values[0])
        idx_i.append(i)
        idx_j.append(first)
        for o in range(0, offset):
            idx_i.append(i)
            idx_j.append(first + 1 + o)

        last = numpy.max(valid_values[0])
        if first != last:
            idx_i.append(i)
            idx_j.append(last)
            for o in range(0, offset):
                idx_i.append(i)
                idx_j.append(min(0, last - (1 + o)))
    return (numpy.array(idx_i, dtype=int), numpy.array(idx_j, dtype=int))


def sentinel3_slstr_sst(infile, outdir, vmin=None, vmax=None,
                        min_percentile=2.0, file_range=None, coeff_path=None,
                        write_netcdf=False, log_path=None, lat_crop=80.0):
    t0 = datetime.utcnow()
    # Process nadir data
    _product_name = 'Sentinel-3{}_SLSTR_SST'
    fname = 'BT'
    view = 'n'
    sltype = 'i'
    bandnames = ['sst', ]
    while infile.endswith('/'):
        infile = infile[:-1]
    collection = infile[-8:-5]
    if coeff_path is None:
        logger.error('Please provide coefficient file')
        sys.exit(1)

    satellite_id = slstr.get_satellite(infile, sltype, view)
    if 'A' == satellite_id:
        satellite_id = ''  # for historical reason
    product_name = _product_name.format(satellite_id)

    # Read coordinates and compute gcps
    (syntool_stats, metadata, geolocation, tie_lon, tie_lat, slice_lat0,
        slice_lat1, __, __, ngcps, month) = slstr.read_geometry(infile,
                                             bandnames, fname, sltype, view,
                                             product_name, vmin, vmax,
                                             log_path)
    # Compute masks
    logger.info('Build masks')
    t_start = datetime.utcnow()
    quality_flags = slstr.read_mask(infile, sltype, view, slice_lat0,
                                    slice_lat1)
    raw_cloud_flags = slstr.read_cloud_mask(infile, sltype, view, slice_lat0,
                                            slice_lat1)
    data_mask = build_mask_ir(quality_flags, raw_cloud_flags, collection,
                              tie_lat, lat_crop)

    # No need to produce an output if all data values are masked
    if numpy.all(data_mask):
        logger.warn('No valid value found for sst')
        raise syntool_converter.EarlyExit()

    t_stop = datetime.utcnow()
    syntool_stats['mask_computation'] = (t_stop - t_start).total_seconds()

    # Read band to compute sst
    logger.info('Construct bands')
    t_start = datetime.utcnow()

    bands = []

    bandname = 'sst'
    fieldname = slstr.get_field_name(fname, 'S8', sltype, view)
    n11 = slstr.read_band(infile, 'S8', fieldname, slice_lat0, slice_lat1)
    if numpy.all(numpy.ma.getmaskarray(n11)):
        logger.warn('All S8 values are masked')
        raise syntool_converter.EarlyExit()

    fieldname = slstr.get_field_name(fname, 'S9', sltype, view)
    n12 = slstr.read_band(infile, 'S9', fieldname, slice_lat0, slice_lat1)
    if numpy.all(numpy.ma.getmaskarray(n12)):
        logger.warn('All S9 values are masked')
        raise syntool_converter.EarlyExit()

    # -- Read view offsets and info from manifest file
    nfile = os.path.join(infile, 'xfdumanifest.xml')
    param = read_manifest(nfile)

    # - Get satellite angles
    nfile = os.path.join(infile, 'geometry_t{}.nc'.format(view))
    fid = netCDF4.Dataset(nfile, 'r')
    sat_zen_n = fid.variables['sat_zenith_t{}'.format(view)][:]
    fid.close()
    # Convert zenith angles into pathlength
    path_tn = 1/numpy.cos(numpy.deg2rad(sat_zen_n))

    # - Get TCWV from MET file
    nfile = os.path.join(infile, 'met_tx.nc')
    fid = netCDF4.Dataset(nfile, 'r')
    tcwv = fid.variables['total_column_water_vapour_tx'][0,:,:]
    fid.close()

    # - Generate fractional grid
    elem_n = param['elem_{}'.format(view)]
    line_n = param['line_{}'.format(view)]
    di_in, dj_in = gen_fractional_grid(elem_n, line_n, param)

    # - Interpolate TCWV and Pathlength to fractional grid
    n2_in = slstr.interp_to_grid(coeff_path, tcwv, path_tn, di_in, dj_in,
                                 elem_n, line_n)
    # Look for the first column which has a non-masked value
    # Transpose so that the first index corresponds to the first valid
    # longitude (and not the first valid latitude)
    contour_n11 = get_alongtrack_contour(n11, 0)
    n11.mask[contour_n11] = True


    # - Calculate N2 SST
    _alpha = n2_in[:, :, 0].transpose()
    _beta = n2_in[:, :, 1].transpose()
    _gamma = n2_in[:, :, 2].transpose()
    alpha = _alpha[slice_lat0, slice_lat1]
    beta = _beta[slice_lat0, slice_lat1]
    gamma = _gamma[slice_lat0, slice_lat1]
    band = (alpha + beta * n12[:,:] + gamma * n11[:,:])
    kernel_size = (3, 3)
    band = filter_sst(band, n11, kernel_size)
    band.mask = (numpy.ma.getmaskarray(n11) | numpy.ma.getmaskarray(n12))
    if numpy.all(numpy.ma.getmaskarray(band)):
        logger.warn('All pixels have either a masked S8 or a masked S9')
        raise syntool_converter.EarlyExit()

    logger.info('\tSet contrast')

    vmin, vmax, list_mask = build_mask_minmax(tie_lon, tie_lat, data_mask)
    if vmin is None or vmax is None:
        sys.exit(1)
    _vmin = min(vmin)
    _vmax = max(vmax)

    logger.info('\tContrast : vmin={} / vmax={}'.format(vmin, vmax))

    t_stop = datetime.utcnow()
    syntool_stats['minmax_computation'] = (t_stop - t_start).total_seconds()
    syntool_stats['final_min'] = vmin
    syntool_stats['final_max'] = vmax

    bnd = band.data
    # Initialize extremum variable to be stored in metadata
    extremum_range = []
    for ilist in range(len(vmin)):
        _min = vmin[ilist]
        _max = vmax[ilist]
        extremum_range.append((_min, _max))
        scale = (_max - _min) / 254.
        if scale == 0.0:
            scale = 1.0
        offset = _min
        # Construct bands
        logger.info('\tBytescaling')
        #byte = bytescale(bnd, cmin=_min, cmax=_max, low=0, high=254)
        scaled = numpy.round((bnd - offset) / scale)
        _byte = numpy.clip(scaled, 0, 254).astype('ubyte')
        if ilist == 0:
            byte = + _byte
        byte[list_mask[ilist]] = _byte[list_mask[ilist]]

    if band.mask is not numpy.ma.nomask:
        byte[band.mask] = 255

    # mask night data for rgb and invalid data for ir (cloud, land,
    # range value). Also mask data for extreme latitudes
    byte[data_mask] = 255
    # Remove borders + n to remove pixels affected by the kernel_size kernel
    n = int((kernel_size[0] - 1) / 2)
    contour_byte = get_alongtrack_contour(n11, 1)
    byte[contour_byte] = 255

    band_range = [_vmin, _vmax]
    description = 'Sea Surface Temperature'  # no log for IR
    colortable = stfmt.format_colortable('medspi_contrast')
                                        # vmax=_vmax, vmax_pal=_vmax,
                                        # vmin=_vmin, vmin_pal=_vmin)
    bands.append({'array': byte,
                  'plot': band.data,
                  'scale': scale,
                  'offset': offset,
                  'description': description,
                  'unittype': 'K',
                  'nodatavalue': 255,
                  'parameter_range': band_range,
                  'colortable': colortable,
                  'granule_range': extremum_range})

    if write_netcdf:
        bands[-1]['name'] = 'sst'
        bands[-1]['long_name'] = 'Sea Surface Temperature'
        bands[-1]['unittype'] = '1'


    t_stop = datetime.utcnow()
    syntool_stats['bytescaling'] = (t_stop - t_start).total_seconds()

    if write_netcdf:
        metadata['spatial_resolution'] = 1000
        ncfile = stfmt.format_ncfilename(outdir, metadata, create_dir=True)
        stfmt.write_netcdf(ncfile, metadata, geolocation, bands, 'swath',
                           ngcps=ngcps)
    else:
        logger.info('Write geotiff')
        tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
        stfmt.write_geotiff(tifffile, metadata, geolocation, bands)

    logger.info(datetime.utcnow() - t0)
    syntool_stats['total_time'] = (datetime.utcnow() - t0).total_seconds()
    if log_path is not None:
        import json
        full_path = os.path.normpath(infile)
        file_path = os.path.basename(full_path)
        file_name, _ = os.path.splitext(file_path)
        stats_path = os.path.join(log_path, '{}.json'.format(file_name))
        with open(stats_path, 'w') as f:
            json.dump(syntool_stats, f)
