# -*- encoding: utf-8 -*-

"""
@author: <lucile.gaultier@oceandatalab.com>
@author: <sylvain.herledan@oceandatalab.com>
@date: 2019-05-14

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.


This reader has been developed for the Non Time Critical version of the
Sentinel-3 SLSTR SST product distributed by Eumetsat.
"""

import os
import numpy
import bisect
import logging
import netCDF4
import datetime
import syntool_converter
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt

logger = logging.getLogger(__name__)

# https://earth.esa.int/web/sentinel/technical-guides/sentinel-3-slstr/level-1/vis-and-swir-radiances
# https://sentinel.esa.int/web/sentinel/user-guides/sentinel-3-slstr/data-formats/level-1
# https://earth.esa.int/web/sentinel/user-guides/sentinel-3-slstr/product-types/level-1b
# https://sentinel.esa.int/documents/247904/685236/Sentinel-3_User_Handbook#%5B%7B%22num%22%3A235%2C%22gen%22%3A0%7D%2C%7B%22name%22%3A%22XYZ%22%7D%2C54%2C628%2C0%5D
# S1 555   nm            B
# S2 659   nm            G
# S3 865   nm            R
# S4 1.375 microm
# S5 1.610 microm
# S6 2.25  microm
# S7 3.74  microm        IR
# S8 10.85 microm        IR
# S9 12    microm        IR

GDAL_MAX_GCPS = 8000
MAX_ROWS_PER_CHUNK = 2000

default_minmax = {'sst': (270, 305),
                  'high_latitude': (271.15, 301.15),
                  'med_latitude': (274.15, 306.15),
                  'warm_sea': (284.15, 310.15)}


class MissingMinMax(ValueError):
    """Error raised when build_mask_minmax does not return valid min and max
    values (i.e. None)."""
    pass


class LongitudinalExtentTooWide(Exception):
    """Error raised if the difference between the min and max longitudes of the
    GCPs exceeds (or equals) 120"""
    pass


def coordstoaffin(lon, lat):
    """ Function to compute affine parameters corresponding to the polyline
    coordinates. """
    alpha = []
    beta = []
    for i in range(len(lon)-1):
        _alpha = (lat[i + 1] - lat[i]) / (lon[i + 1] - lon[i])
        _beta = lat[i] - lon[i] * _alpha
        alpha.append(_alpha)
        beta.append(_beta)
    return alpha, beta


def polyline_north():
    """ Polyline to define high latitudes in northern atmosphere (annular ring)
    """
    lon = [-180, -169.01, -156.43, -150.60, -95.60, -86.59,
           -75.77, -65.60, -9.143, 0.14, 8.30, 32.85, 34.47,
           37.16, 56.47, 120.44, 180]
    lat = [46.78, 52.84, 57.27, 62.66, 48.36, 36.23,
           36.06, 43.47, 43.00, 43.00, 47.13, 47.66, 45.25,
           45.20, 48.07, 36.81, 46.78]
    alpha, beta = coordstoaffin(lon, lat)
    return lon, lat, alpha, beta


def polyline_south():
    """ Polyline to define high latitudes in southern atmosphere (annular ring)
    """
    lon = [-180, -71.03, -64.94, 167.45, 180]
    lat = [-45.43, -44.87, -42.98, -45.43, -45.43]
    alpha, beta = coordstoaffin(lon, lat)
    return lon, lat, alpha, beta


def polyline_baja():
    """ Polyline to define warm pool for Baja region"""
    lon = [-116.24, -109.88, -105.68, -100]
    lat = [32.1, 22.92, 20.40, 20.40]
    alpha, beta = coordstoaffin(lon, lat)
    return lon, lat, alpha, beta


def polygon_warm_sea():
    """ define warm pool box: [lon_min, lon_max, lat_min, lat_max]"""
    box = []
    # Red sea / Gulf of Aden
    box.append([27.73, 50.89, 10.27, 30.64])
    # Persian / Oman Gulf
    box.append([42.67, 59.798, 22.24, 33.06])
    # Baja
    # box.append([-115.07, -112.02, 29.5, 32.16])
    # box.append([-113.97, -109.77, 27.03, 29.56])
    # box.append([-112.03, -110.39, 24.85, 27.27])
    # box.append([-110.90,-107.29, 23.99, 27.32])
    # box.append([-109.97, -106.05, 22.88, 27.14])
    # Gulf of Mexico
    # box.append([-100.12, -81.89, 17.13, 31.03])
    # box.append([-89.67, -80.44, 15.20, 26.98])

    return box


def mask_from_polyline(list_lon, list_lat, alpha, beta, tie_lon_masked,
                       tie_lat_masked, tie_lat, above=False):
    """ Compute mask from polyline and granule """
    min_lon = numpy.min(tie_lon_masked)
    max_lon = numpy.max(tie_lon_masked)
    mask_out = numpy.full(numpy.shape(tie_lat), False, dtype=numpy.bool)
    bisect_l = bisect.bisect_left

    polyline_lon_min = numpy.min(list_lon)
    polyline_lon_max = numpy.max(list_lon)
    in_polyline_range = ((tie_lon_masked <= polyline_lon_max) &
                         (tie_lon_masked >= polyline_lon_min))
    _tie_lon_masked = tie_lon_masked[numpy.where(in_polyline_range)]
    lon_l = numpy.mod(numpy.min(_tie_lon_masked) + 180.0, 360.0) - 180.0
    lon_r = numpy.mod(numpy.max(_tie_lon_masked) + 180.0, 360.0) - 180.0
    i_min_l = numpy.max([0, bisect_l(list_lon, lon_l) - 1])
    i_min_r = numpy.max([0, bisect_l(list_lon, lon_r) - 1])

    def mask_above(data, ref):
        return (abs(data) >= abs(ref))

    def mask_under(data, ref):
        return (abs(data) <= abs(ref))

    lat_mask = mask_under
    if above is True:
        lat_mask = mask_above
    if i_min_l == i_min_r:
        test_lat = alpha[i_min_l] * tie_lon_masked + beta[i_min_l]
        mask_out = lat_mask(tie_lat_masked, test_lat)
    else:
        if (max_lon - min_lon) > 180:
            for k in range(0, len(alpha)):
                test_lat = alpha[k] * tie_lon_masked + beta[k]
                # polylines have to be wrapped
                segment_mask = (lat_mask(tie_lat_masked, test_lat)
                                & (tie_lon_masked <= list_lon[k + 1])
                                & (tie_lon_masked > list_lon[k]))
                mask_out = (mask_out | segment_mask)
        else:
            for k in range(i_min_l, i_min_r + 1):
                test_lat = alpha[k] * tie_lon_masked + beta[k]
                segment_mask = (lat_mask(tie_lat_masked, test_lat)
                                & (tie_lon_masked <= list_lon[k + 1])
                                & (tie_lon_masked > list_lon[k]))
                mask_out = (mask_out | segment_mask)
    return mask_out


def build_mask_minmax(tie_lon, tie_lat, data_mask):
    """ Compute mask and corresponding min max for each area (warm pool, med
    latitude or high latitude). """
    tie_lat_masked = numpy.ma.masked_array(tie_lat, mask=data_mask)
    tie_lon_masked = numpy.ma.masked_array(tie_lon, mask=data_mask)
    tie_lon_masked = numpy.mod(tie_lon_masked + 180.0, 360.0) - 180.0
    min_lat = numpy.min(tie_lat_masked)
    max_lat = numpy.max(tie_lat_masked)
    # check if granule min max corresponds to high or low latitude, granule can
    # be splitted.
    list_mask = []
    # Check if all latitudes are higher than the northern polyline point
    if (min_lat > 62.66):
        vmin = (default_minmax['high_latitude'][0], )
        vmax = (default_minmax['high_latitude'][1], )
        list_mask.append(numpy.full(numpy.shape(tie_lat), True,
                         dtype=numpy.bool))
    # Check if all latitudes are higher than the southern polyline point
    elif (max_lat < -48.13).all() is True:
        vmin = (default_minmax['high_latitude'][0], )
        vmax = (default_minmax['high_latitude'][1], )
        list_mask.append(numpy.full(numpy.shape(tie_lat), True,
                         dtype=numpy.bool))
    # Check if all latitudes are lower than the lowest latitude in polyline
    else:
        vmin = []
        vmax = []
        # vmin = (default_minmax['med_latitude'][0], )
        # vmax = (default_minmax['med_latitude'][1], )

        # Build mask for warm pool
        mask_b = numpy.full(numpy.shape(tie_lat), False, dtype=numpy.bool)
        bbox = polygon_warm_sea()
        for box in bbox:
            box_mask = ((tie_lon_masked > box[0]) & (tie_lon_masked < box[1]) &
                        (tie_lat_masked > box[2]) & (tie_lat_masked < box[3]))
            mask_b = (mask_b | box_mask)
        # If granule is near Baja box
        baja_box = [-115.1, -104.44, 20.38, 32.10]
        baja_box_mask = ((tie_lon_masked > baja_box[0])
                         & (tie_lon_masked < baja_box[1])
                         & (tie_lat_masked > baja_box[2])
                         & (tie_lat_masked < baja_box[3]))
        if numpy.any(baja_box_mask):
            list_lon, list_lat, alpha, beta = polyline_baja()
            mask_baja = mask_from_polyline(list_lon, list_lat, alpha, beta,
                                           tie_lon_masked, tie_lat_masked,
                                           tie_lat, above=True)
            mask_b = (mask_b | mask_baja)
        # if there are points in warm pool
        if numpy.any(mask_b):
            vmin.append(default_minmax['warm_sea'][0])
            vmax.append(default_minmax['warm_sea'][1])
            list_mask.append(mask_b)
        # Build mask for high latitude
        if (numpy.mean(tie_lat_masked) > 0):
            list_lon, list_lat, alpha, beta = polyline_north()
        else:
            list_lon, list_lat, alpha, beta = polyline_south()
        # Get index for affine function parameters
        mask = mask_from_polyline(list_lon, list_lat, alpha, beta,
                                  tie_lon_masked, tie_lat_masked, tie_lat)
        if numpy.all(mask):
            mask = (mask & ~mask_b)
            vmin.append(default_minmax['med_latitude'][0])
            vmax.append(default_minmax['med_latitude'][1])
            list_mask.append(mask)
        elif not numpy.any(mask):
            mask = (~mask & ~mask_b)
            vmin.append(default_minmax['high_latitude'][0])
            vmax.append(default_minmax['high_latitude'][1])
            list_mask.append(mask)
        else:
            vmin.append(default_minmax['high_latitude'][0])
            vmin.append(default_minmax['med_latitude'][0])
            vmax.append(default_minmax['high_latitude'][1])
            vmax.append(default_minmax['med_latitude'][1])
            list_mask.append(~mask)
            # One assumes that no warm pool is in high latitude area
            mask = (mask & ~mask_b)
            list_mask.append(mask)
    return vmin, vmax, list_mask


def get_alongtrack_slices(lon, lat_mask, data_mask):
    """"""
    data_slices = []
    nrows = lon.shape[0]
    extra_row = 1
    i = 0
    slice_start = None
    while i < nrows:
        # Remove rows that are partially cropped and contain no valid data
        # This is required to avoid GDAL projection issues at high latitudes.
        # Note that this method is not bullet-proof and seems to work for this
        # product because rows that are above or cross the latitude threshold
        # (85 degrees by default) are unlikely to contain any valid SST value.
        if numpy.any(lat_mask[i]) and numpy.all(data_mask[i]):
            if (slice_start is not None):
                if 2 < (i - slice_start):
                    data_slices.append(slice(slice_start, i))
                else:
                    _msg = 'Skipped [{}, {}[: less than 3 rows is too short'
                    logger.warning(_msg.format(slice_start, i))
            slice_start = None
            i = i + 1
            continue

        if slice_start is None:
            slice_start = i

        # GDAL cannot handle data with too many GCPs (threshold not very clear,
        # apparently it would be around 8000), so swaths are split into chunks
        # of MAX_ROWS_PER_CHUNK rows (except for chunks that cross the IDL but
        # these should not contain too many rows anyway).
        # The value of MAX_ROWS_PER_CHUNK may be too conservative as it creates
        # a lot of chunks (datasets) but they will be concatenated afterwards
        # by syntool-merger anyway.
        if MAX_ROWS_PER_CHUNK < (i - slice_start):
            if 2 < (i + extra_row - slice_start):
                # Include current row so that there is an overlap (merge)
                data_slices.append(slice(slice_start, i + extra_row))
            else:
                _msg = 'Skipped [{}, {}[: less than 3 rows is too short'
                logger.warning(_msg.format(slice_start, i + extra_row))

            # Set slice_start to current row in order to have a two-ways
            # overlap between slices that have been cut due to their size
            slice_start = i

        i = i + 1

    if (slice_start is not None) and (slice_start < nrows):
        data_slices.append(slice(slice_start, nrows))

    return data_slices


def _process_alongtrack_slice(alongtrack_slice, lon, lat, sst, discarded_mask):
    """"""
    slice_lon = lon[alongtrack_slice, :]
    slice_lat = lat[alongtrack_slice, :]
    slice_sst = sst[alongtrack_slice, :]
    slice_discarded_mask = discarded_mask[alongtrack_slice, :]

    # No need to produce an output if all data values are masked
    if numpy.all(slice_discarded_mask):
        logger.warn('No valid value found for sst')
        return None, None
        raise syntool_converter.EarlyExit()

    # Build GCPs
    dgcp = 128.
    min_abs_lat = numpy.min(numpy.abs(slice_lat))
    max_abs_lat = numpy.max(numpy.abs(slice_lat))
    if numpy.all(70 < min_abs_lat):
        dgcp = 32.
    if 87.0 < max_abs_lat:
        dgcp = 16.
    if 88.0 < max_abs_lat:
        dgcp = 8.
    if 89.0 < max_abs_lat:
        dgcp = 4.
    if 10 > (alongtrack_slice.stop - alongtrack_slice.start):
        dgcp = dgcp * 2

    slice_shape = numpy.array(slice_lon.shape)
    ngcps = (numpy.ceil(slice_shape / dgcp) + 1.).astype('int32')
    while GDAL_MAX_GCPS < ngcps[0] * ngcps[1]:
        dgcp = dgcp * 2.0
        ngcps = (numpy.ceil(slice_shape / dgcp) + 1.).astype('int32')

    pix = numpy.linspace(0, slice_lon.shape[1] - 1, num=ngcps[1]).round()
    lin = numpy.linspace(0, slice_lon.shape[0] - 1, num=ngcps[0]).round()
    pix2d, lin2d = numpy.meshgrid(pix.astype('int32'),
                                  lin.astype('int32'))
    _gcp_lon = slice_lon[lin2d, pix2d]
    gcp_lat = slice_lat[lin2d, pix2d]
    gcp_pix = pix2d + 0.5
    gcp_lin = lin2d + 0.5
    gcp_hei = numpy.zeros(ngcps)

    # Recreate the GCP longitudes grid using the following assumption:
    # "Longitudinal distance between two adjacent GCPs should never exceed 180
    # degrees"
    #
    # Maybe there is a cleaner way to achieve longitudinal continuity even
    # across the IDL but at least this method is simple and produces the
    # expected result.
    gcp_lon = numpy.full(numpy.shape(_gcp_lon), _gcp_lon[0, 0],
                         dtype=_gcp_lon.dtype)

    _dlon0_dy = _gcp_lon[1:, 0] - _gcp_lon[:-1, 0]
    dlon0_dy = numpy.mod(_dlon0_dy + 180.0, 360.0) - 180.0
    for y in range(dlon0_dy.shape[0]):
        gcp_lon[y + 1, 0] = gcp_lon[y, 0] + dlon0_dy[y]

    _dlon_dx = _gcp_lon[:, 1:] - _gcp_lon[:, :-1]
    dlon_dx = numpy.mod(_dlon_dx + 180.0, 360.0) - 180.0
    for x in range(dlon_dx.shape[1]):
        gcp_lon[:, x + 1] = gcp_lon[:, x] + dlon_dx[:, x]

    if 120.0 <= (numpy.max(gcp_lon) - numpy.min(gcp_lon)):
        raise LongitudinalExtentTooWide()

    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['gcps'] = stfmt.format_gdalgcps(gcp_lon, gcp_lat, gcp_hei,
                                                gcp_pix, gcp_lin)

    # Read band to compute sst
    logger.info('Construct bands')

    slice_sst.mask = slice_discarded_mask

    vmin, vmax, list_mask = build_mask_minmax(slice_lon, slice_lat,
                                              slice_discarded_mask)
    # TODO: check if this case can actually happen
    if vmin is None or vmax is None:
        raise MissingMinMax()

    _vmin = min(vmin)
    _vmax = max(vmax)

    logger.info('\tContrast : vmin={} / vmax={}'.format(vmin, vmax))

    logger.debug('Final min: {}'.format(vmin))
    logger.debug('Final max: {}'.format(vmax))
    bnd = slice_sst.data
    # Initialize extremum variable to be stored in metadata
    extremum_range = []
    for ilist in range(0, len(vmin)):
        _min = vmin[ilist]
        _max = vmax[ilist]
        extremum_range.append((_min, _max))
        _byte, offset, scale = pack.ubytes_0_254(bnd, _min, _max)
        if ilist == 0:
            byte = + _byte
        bytes_idx = numpy.where(list_mask[ilist] == True)  # noqa
        byte[bytes_idx] = _byte[bytes_idx]

    if slice_sst.mask is not numpy.ma.nomask:
        byte[numpy.where(slice_sst.mask == True)] = 255  # noqa

    band_range = [_vmin, _vmax]
    description = 'Sea Surface Temperature'  # no log for IR
    colortable = stfmt.format_colortable('medspi_contrast')
    bands = []
    bands.append({'array': byte,
                  'scale': scale,
                  'offset': offset,
                  'description': description,
                  'unittype': 'K',
                  'nodatavalue': 255,
                  'parameter_range': band_range,
                  'colortable': colortable,
                  'granule_range': extremum_range})

    return bands, geolocation


def process_alongtrack_slice(alongtrack_slice, lon, lat, sst, discarded_mask):
    """"""
    try:
        data, geo = _process_alongtrack_slice(alongtrack_slice, lon, lat, sst,
                                              discarded_mask)
        yield (data, geo)
    except LongitudinalExtentTooWide:
        # Split the chunk again, hoping that it will be enough to get a
        # longtidunal extent shorter than 120 degrees
        start_idx = alongtrack_slice.start
        stop_idx = alongtrack_slice.stop
        mid_idx = numpy.floor(0.5 * (start_idx + stop_idx)).astype('int')

        # First half
        alongtrack_slice = slice(start_idx, mid_idx)
        if 2 < (alongtrack_slice.stop - alongtrack_slice.start):
            for data, geo in process_alongtrack_slice(alongtrack_slice, lon,
                                                      lat, sst,
                                                      discarded_mask):
                if None not in (data, geo):
                    yield (data, geo)

        # Second half
        alongtrack_slice = slice(mid_idx, stop_idx)
        if 2 < (alongtrack_slice.stop - alongtrack_slice.start):
            for data, geo in process_alongtrack_slice(alongtrack_slice, lon,
                                                      lat, sst,
                                                      discarded_mask):
                if None not in (data, geo):
                    yield (data, geo)


def process_part(variables, dtime0):
    """"""
    time1, lon, lat, sst, ql_mask, lat_mask = variables
    sst_mask = numpy.ma.getmaskarray(sst)
    data_mask = (sst_mask | ql_mask)
    discarded_mask = (data_mask | lat_mask)

    start_time = dtime0 + datetime.timedelta(seconds=numpy.min(time1))
    end_time = dtime0 + datetime.timedelta(seconds=numpy.max(time1))
    (dtime, time_range) = stfmt.format_time_and_range(start_time, end_time,
                                                      units='s')
    meta = {}
    meta['product_name'] = None  # set later
    meta['name'] = None  # set later
    meta['datagroup'] = None  # set later
    meta['datetime'] = dtime
    meta['time_range'] = time_range
    meta['source_URI'] = None  # set later
    meta['source_provider'] = 'eumetsat'
    meta['processing_center'] = 'OceanDataLab'
    meta['conversion_software'] = 'Syntool'
    meta['conversion_version'] = '0.0.0'
    meta['conversion_datetime'] = None  # set later
    meta['parameter'] = 'sea surface temperature'
    meta['type'] = 'remote sensing'
    meta['sensor_type'] = 'dual scan temperature radiometer'
    meta['sensor_name'] = 'SLSTR'
    meta['sensor_platform'] = None  # set later

    alongtrack_slices = get_alongtrack_slices(lon, lat_mask, data_mask)
    logger.info('{} slices found'.format(len(alongtrack_slices)))
    for i, alongtrack_slice in enumerate(alongtrack_slices):
        for data, geo in process_alongtrack_slice(alongtrack_slice, lon, lat,
                                                  sst, discarded_mask):
            if None not in (data, geo):
                yield (meta, geo, data, None)



def sentinel3_slstr_sst_eumetsat(infile, outdir, vmin=None, vmax=None,
                                 min_percentile=2.0, file_range=None,
                                 coeff_path=None, write_netcdf=False,
                                 log_path=None, lat_crop=85.0):
    """"""
    t0 = datetime.datetime.utcnow()
    # Process nadir data
    _product_name = '{}_SLSTR_SST_eumetsat'
    basename = os.path.basename(infile)
    granule_name, _ = os.path.splitext(basename)

    # Read coordinates and compute gcps
    infile = os.path.join(infile)
    fid = netCDF4.Dataset(infile, 'r')

    satellite_id, _ = fid.title.split(' ', 1)
    product_name = _product_name.format(satellite_id)

    lon = fid.variables['lon'][:]
    lat = fid.variables['lat'][:]

    # Read time
    time0 = fid.variables['time']
    dtime0 = netCDF4.num2date(time0[0], units=time0.units)
    time1 = fid.variables['sst_dtime'][0, :, :]
    bandname = 'sea_surface_temperature'
    sst = fid.variables[bandname][0, :, :]
    ql = fid.variables['quality_level'][0, :, :]
    fid.close()

    # Compute masks
    logger.info('Build masks')
    ql_mask = numpy.full(numpy.shape(ql), True, dtype='bool')
    ql_mask[numpy.where(ql > 2)] = False
    del ql

    lat_mask = numpy.full(numpy.shape(lat), False, dtype='bool')
    lat_mask[numpy.where(numpy.abs(lat) > lat_crop)] = True

    # syntool-merger will not produce a correct output if the maximal
    # longitudinal distance between two arbitrary points of the granule exceeds
    # 360 degrees (possible overlap between the beginning and the end if the
    # swath. To make sure it does not happen, input data is split into two
    # halfs and each of them is processed independently.

    mid_idx = numpy.floor(0.5 * lon.shape[0]).astype('int')

    # Part 1
    part1_slice = (slice(None, mid_idx), slice(None),)
    part1 = (time1[part1_slice], lon[part1_slice], lat[part1_slice],
             sst[part1_slice], ql_mask[part1_slice], lat_mask[part1_slice])
    i = 0
    for meta, geo, data, extra in process_part(part1, dtime0,):
        logger.info('Write geotiff')
        now = datetime.datetime.utcnow()
        meta['conversion_datetime'] = stfmt.format_time(now)
        meta['product_name'] = '{}'.format(product_name)
        meta['name'] = '{}_1_{}'.format(granule_name, i)
        meta['datagroup'] = '{}_1'.format(granule_name)
        meta['source_URI'] = infile
        meta['sensor_platform'] = satellite_id
        tifffile = stfmt.format_tifffilename(outdir, meta, create_dir=True)
        stfmt.write_geotiff(tifffile, meta, geo, data)
        i = i + 1

    # Part 2
    part2_slice = (slice(mid_idx, None), slice(None),)
    part2 = (time1[part2_slice], lon[part2_slice], lat[part2_slice],
             sst[part2_slice], ql_mask[part2_slice], lat_mask[part2_slice])
    i = 0
    for meta, geo, data, extra in process_part(part2, dtime0):
        logger.info('Write geotiff')
        now = datetime.datetime.utcnow()
        meta['conversion_datetime'] = stfmt.format_time(now)
        meta['product_name'] = '{}'.format(product_name)
        meta['name'] = '{}_2_{}'.format(granule_name, i)
        meta['datagroup'] = '{}_2'.format(granule_name)
        meta['source_URI'] = infile
        meta['sensor_platform'] = satellite_id
        tifffile = stfmt.format_tifffilename(outdir, meta, create_dir=True)
        stfmt.write_geotiff(tifffile, meta, geo, data)
        i = i + 1

    logger.info(datetime.datetime.utcnow() - t0)


if '__main__' == __name__:
    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.DEBUG)

    import sys
    infile = sys.argv[1]
    outdir = sys.argv[2]

    sentinel3_slstr_sst_eumetsat(infile, outdir)
