# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-03-01

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import pyproj
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def build_mask(f_handler):
    """"""

    bits = {'SSS_USEABLE': 0,
            'FOUR_LOOKS': 1,
            'POINTING': 2,
            'LARGE_GALAXY_CORRECTION': 3,
            'ROUGHNESS_CORRECTION': 4,
            'SST_TOO_COLD': 5,
            'LAND': 6,
            'ICE': 7,
            'HIGH_SPEED_USEABLE': 8}

    off_flags = numpy.uint16(0)
    off_flags = off_flags + numpy.uint16(1 << bits['ROUGHNESS_CORRECTION'])
    off_flags = off_flags + numpy.uint16(1 << bits['LAND'])
    off_flags = off_flags + numpy.uint16(1 << bits['ICE'])
    quality_var = f_handler.variables['quality_flag']
    quality = numpy.ma.getdata(quality_var[:])
    mask = ((quality_var._FillValue == quality) |
            (numpy.bitwise_and(quality, off_flags) > 0))
    return mask


def build_grid(f_handler):
    """"""
    lat_var = f_handler.variables['lat']
    lat = lat_var[:]
    lon_var = f_handler.variables['lon']
    lon = lon_var[:]
    sss_var = f_handler.variables['smap_sss']
    sss = sss_var[:]

    data_mask = build_mask(f_handler)

    _mask = (lon.mask | (numpy.ma.getdata(lon) == lon_var._FillValue) |
             lat.mask | (numpy.ma.getdata(lat) == lat_var._FillValue) |
             sss.mask | (numpy.ma.getdata(sss) == sss_var._FillValue) |
             data_mask)
    valid_idx = numpy.where(_mask == False)
    lons = lon[valid_idx]
    lats = lat[valid_idx]
    ssss = sss[valid_idx]

    # Not clear whether the resolution is 60km or 25km
    # Using a 1/3° resolution seems like a good compromise
    grid_shape = (540, 1080,)  # 1/3° resolution
    _grid = numpy.full(grid_shape, -1.0, dtype=numpy.float)
    grid = numpy.ma.array(_grid)
    grid.mask = True

    x_max = 180.0
    x_min = -1 * x_max
    dx = (x_max - x_min) / grid_shape[1]  # grid(lat, lon) and x <=> lon

    y_max = 90.0
    y_min = -1 * y_max
    dy = (y_max - y_min) / grid_shape[0]  # grid(lat, lon) and y <=> lat

    _xp = lons
    _yp = lats
    xp = numpy.mod(_xp - x_min, (x_max - x_min)) + x_min
    yp = numpy.mod(_yp - y_min, (y_max - y_min)) + y_min

    x_ind = numpy.floor((xp - x_min) / dx).astype('int')
    y_ind = numpy.floor((yp - y_min) / dy).astype('int')

    points = zip(x_ind, y_ind, ssss)
    for point in points:
        point_col, point_row, point_sss = point
        grid[point_row, point_col] = point_sss
        grid.mask[point_row, point_col] = False

    return grid


def write_geotiff(product_name, metadata, geolocation, band, infile, outdir):
    """"""
    now = datetime.datetime.utcnow()

    metadata['product_name'] = product_name
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'Jet Propulsion Laboratory'
    metadata['processing_center'] = ''
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'sea surface salinity'

    # Write geotiff
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)


def smap_l2_sss_jpl_v4(input_path, output_path, vmin=31.825, vmax=38.175,
                       vmin_pal=32, vmax_pal=38, write_netcdf=False):
    """"""

    f_handler = netCDF4.Dataset(input_path, 'r')

    _time_start = f_handler.time_coverage_start
    time_start = '{}000Z'.format(_time_start)
    start_dt = datetime.datetime.strptime(time_start, '%Y-%jT%H:%M:%S.%fZ')

    _time_end = f_handler.time_coverage_end
    time_end = '{}000Z'.format(_time_end)
    stop_dt = datetime.datetime.strptime(time_end, '%Y-%jT%H:%M:%S.%fZ')
    version = f_handler.product_version.lower()
    if 'v4.0' == version:
        version = 'v4'  # retro-compatibility
    grid = build_grid(f_handler)
    f_handler.close()

    if numpy.all(grid.mask):
        logger.warn('All values masked... Skipping.')
        return

    dtime, time_range = stfmt.format_time_and_range(start_dt, stop_dt,
													units='s')

    metadata = {}
    metadata['name'] = os.path.splitext(os.path.basename(input_path))[0]
    metadata['time_range'] = time_range
    metadata['datetime'] = dtime

    grid_width, grid_height = numpy.shape(grid)
    lon0 = -180.0
    dlon = 360.0 / grid_height
    lat0 = -90.0
    dlat = 180.0 / grid_width

    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['geotransform'] = [lon0, dlon, 0,
                                   lat0, 0, dlat]

	# Pack data
    band = []
    array, offset, scale = pack.ubytes_0_254(grid.data, vmin, vmax)
    if grid.mask.any():
        array[numpy.where(grid.mask)] = 255
    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin_pal,
                                         vmax_pal=vmax_pal)
    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea surface salinity',
                 'unittype': 'PSS',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    product_name = 'SMAP_L2_SSS_JPL_{}'.format(version)
    write_geotiff(product_name, metadata, geolocation, band, input_path,
                  output_path)
