# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-02-28

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def build_mask(f_handler, look):
    """"""
    iqc_bits = {'RADIO_MISSING': 0,
                'OI_ISSUE_MISSING': 1,
                'LAND_MISSING': 2,
                'ICE_MISSING': 3,
                'MLE_DIVERGENCE_MISSING': 4,
                'SUNGLINT': 5,
                'MOONGLINT': 6,
                'GALAXY': 7,
                'LAND': 8,
                'ICE': 9,
                'RESIDUAL_MLE': 10,
                'LOW_SST': 11,
                'HIGH_WIND': 12,
                'NO_IMAGER': 13,
                'RAIN': 14}

    qflags = f_handler.variables['iqc_flag'][:, :, look]
    land = f_handler.variables['gland'][:, :, look]
    ice = f_handler.variables['gice'][:,:]
    wind = f_handler.variables['winspd'][:, :]
    sss = f_handler.variables['sss_smap'][:, :, look]

    off_bits = numpy.int32(0)
    off_bits = off_bits + numpy.int32(1 << iqc_bits['SUNGLINT'])
    off_bits = off_bits + numpy.int32(1 << iqc_bits['MOONGLINT'])
    off_bits = off_bits + numpy.int32(1 << iqc_bits['GALAXY'])
    off_bits = off_bits + numpy.int32(1 << iqc_bits['RESIDUAL_MLE'])

    qflags_mask = (numpy.bitwise_and(qflags, off_bits) > 0)
    land_mask = (land > 0.01)
    ice_mask = (ice > 0.001)
    wind_mask = (wind > 20)
    sss_mask = numpy.ma.getmask(numpy.ma.masked_where((sss == -9999.0), sss))

    mask = (qflags_mask | land_mask | ice_mask | wind_mask | sss_mask)

    return mask


def write_geotiff(product_name, metadata, geolocation, band, infile, outdir):
    """"""
    now = datetime.datetime.utcnow()

    metadata['product_name'] = product_name
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'Remote Sensing Systems'
    metadata['processing_center'] = ''
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'sea surface salinity'

    # Write geotiff
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)


def extract_look(f_handler, look):
    """Look:
    0: forward
    1: backward"""
    t_units = f_handler.variables['time'].units
    t = f_handler.variables['time'][:, ::-1, look]
    lat = f_handler.variables['cellat'][:, ::-1, look]
    lon = f_handler.variables['cellon'][:, ::-1, look]
    sss = f_handler.variables['sss_smap'][:, ::-1, look]
    _data_mask = build_mask(f_handler, look)
    data_mask = _data_mask[:, ::-1]

    half = numpy.ceil(0.5 * numpy.shape(lon)[1]).astype('int')
    sides = ('left', 'right')
    for k, side in enumerate(sides):
        k0 = k * half
        k1 = (k + 1) * half
        _sss = sss[:, k0:k1]
        _lon = lon[:, k0:k1]
        _t = t[:, k0:k1]

        _mask = (numpy.ma.getmask(_sss) | data_mask[:, k0:k1])
        _mask.mask = False

        if numpy.all(_mask):
            logger.warn('All masked, skip.')
            continue
        time_idx = numpy.where(_mask == False)
        start_num = numpy.min(_t[time_idx])
        stop_num = numpy.max(_t[time_idx])
        start_dt = netCDF4.num2date(start_num, t_units)
        stop_dt = netCDF4.num2date(stop_num, t_units)

        # Look for the first column which has a non-masked value
        # Transpose so that the first index corresponds to the first valid
        # longitude (and not the first valid latitude)
        j, i = numpy.where(_mask.transpose() == False)
        first_valid_lon = _lon[i[0], j[0]]

        dlon = 0.25
        lon0_noanchor = dlon * (numpy.floor(first_valid_lon / dlon) - j[0])
        lon0 = numpy.mod(lon0_noanchor + 180.0, 360.0) - 180.0
        dlat = 0.25
        lat0 = -90.0
        geotransform = [lon0, dlon, 0, lat0, 0, dlat]

        yield (_sss, _mask, geotransform, side, start_dt, stop_dt)


def smap_l2_sss_rss_v2(input_path, output_path, vmin=31.825, vmax=38.175,
                       vmin_pal=32, vmax_pal=38, write_netcdf=False):
    """"""
    granule_filename = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_filename)

    f_handler = netCDF4.Dataset(input_path, 'r')

    metadata = {'datagroup': granule_name}

    geolocation = {'projection': stfmt.format_gdalprojection()}

    for look in (0, 1,):
        for (sss, mask, geo, side, start, stop) in extract_look(f_handler,
                                                                look):
            if numpy.all(mask):
                logger.warn('Look {} [{}]: no data.'.format(look, side))
                continue

            dtime, time_range = stfmt.format_time_and_range(start, stop,
                                                            units='s')

            metadata['datetime'] = dtime
            metadata['time_range'] = time_range
            metadata['name'] = '{}_{}_{}'.format(granule_name, look, side)
            geolocation['geotransform'] = geo

            # Pack data
            band = []
            array, offset, scale = pack.ubytes_0_254(sss.data, vmin, vmax)
            if mask.any():
                array[numpy.where(mask)] = 255
            colortable = stfmt.format_colortable('matplotlib_jet',
                                                 vmin=vmin, vmax=vmax,
                                                 vmin_pal=vmin_pal,
                                                 vmax_pal=vmax_pal)
            band.append({'array': array,
                         'scale': scale,
                         'offset': offset,
                         'description': 'sea surface salinity',
                         'unittype': 'PSS',
                         'nodatavalue': 255,
                         'parameter_range': [vmin, vmax],
                         'colortable': colortable})

            if 0 == look:
                product_name = 'SMAP_L2_SSS_RSS_v2_fwd'
            else:
                product_name = 'SMAP_L2_SSS_RSS_v2_bkd'
            write_geotiff(product_name, metadata, geolocation, band,
                          input_path, output_path)
    f_handler.close()
