# -*- coding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-02-23

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import logging
import datetime
import scipy.io
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def write_geotiff(product_name, metadata, geolocation, band, infile, outdir):
    """"""
    now = datetime.datetime.utcnow()

    metadata['product_name'] = product_name
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'Joseph Tenerelli'
    metadata['processing_center'] = ''
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'sea surface salinity'

    # Write geotiff
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)


def smap_l3_sss_jtenerelli(infile, outdir, vmin=31.825, vmax=38.175,
                           vmin_pal=32, vmax_pal=38):
    """"""
    granule_filename = os.path.basename(infile)
    granule_name, _ = os.path.splitext(granule_filename)

    # Read/Process data
    # There is an extra row and an extra colum, both filled with nan values:
    # ignore them as they overlap the first row/column and provide no
    # additional information (and they make GDAL crash...)
    logger.info('Read/Process data')
    f_handler = scipy.io.loadmat(infile)
    lat = f_handler['glat'][:-1, :-1]
    lon = f_handler['glon'][:-1, :-1]
    sss_fwd = numpy.ma.array(f_handler['sss_fore'][:-1,:-1]).transpose()
    sss_bkd = numpy.ma.array(f_handler['sss_aft'][:-1, :-1]).transpose()

    # Put longitudes in the [-180, 180] range
    lon = numpy.mod(lon + 180.0, 360.0) - 180.0

    # Center salinity data on longitude 0
    #sss_fwd = numpy.roll(sss_fwd, 720, axis=1)
    #sss_bkd = numpy.roll(sss_bkd, 720, axis=1)

    # Daily data
    _, _, _, date_str, phase, _, _ = granule_name.split('.')
    time_start = datetime.datetime.strptime(date_str, '%Y%m%d')
    time_stop = time_start + datetime.timedelta(days=1)


    # Construct metadata/geolocation/band(s)
    logger.info('Construct metadata/geolocation/band(s)')
    dtime, time_range = stfmt.format_time_and_range(time_start, time_stop,
                                                    units='h')
    lat0, dlat = lat[0, 0], lat[0, 1] - lat[0, 0]
    lon0, dlon = lon[0, 0], lon[1, 0] - lon[0, 0]

    print(numpy.shape(sss_fwd))
    print(lat0, dlat)
    print(lon0, dlon)
    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['geotransform'] = [lon0, dlon, 0,
                                   lat0, 0, dlat]

    metadata = {}
    metadata['name'] = '{}_{}'.format(granule_name, phase)
    metadata['datetime'] = dtime
    metadata['time_range'] = time_range

    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin_pal, vmax_pal=vmax_pal)

    # SMAP SSS forward look
    # -------------------------------------------------------------------------
    band = []
    data_mask = numpy.isnan(sss_fwd)
    if numpy.any(data_mask):
        sss_fwd[numpy.where(data_mask)] = 0  # avoid warning about nan values
    array, offset, scale = pack.ubytes_0_254(sss_fwd.data, vmin, vmax)
    if numpy.any(data_mask):
        array[numpy.where(data_mask)] = 255

    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea surface salinity',
                 'unittype': 'PSS',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    write_geotiff('SMAP_L3_SSS_JTEN_FWD', metadata, geolocation, band, infile,
                  outdir)

    # SMAP SSS backward look
    # -------------------------------------------------------------------------
    band = []
    data_mask = numpy.isnan(sss_bkd)
    if numpy.any(data_mask):
        sss_bkd[numpy.where(data_mask)] = 0  # avoid warning about nan values
    array, offset, scale = pack.ubytes_0_254(sss_bkd.data, vmin, vmax)
    if numpy.any(data_mask):
        array[numpy.where(data_mask)] = 255

    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea surface salinity',
                 'unittype': 'PSS',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    write_geotiff('SMAP_L3_SSS_JTEN_BKD', metadata, geolocation, band, infile,
                  outdir)
