# -*- coding: utf-8 -*-

"""
@author <sylvain.herledan@oceandatalab.com>
@date: 2018-02-23

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def write_geotiff(product_name, metadata, geolocation, band, infile, outdir):
    """"""
    now = datetime.datetime.utcnow()

    metadata['product_name'] = product_name
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'Remote Sensing Systems'
    metadata['processing_center'] = ''
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'sea surface salinity'

    # Write geotiff
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)


def smap_l3_sss_rss(infile, outdir, vmin=31.825, vmax=38.175, vmin_pal=32,
                    vmax_pal=38):
    """"""
    granule_filename = os.path.basename(infile)
    granule_name, _ = os.path.splitext(granule_filename)

    # Read/Process data
    logger.info('Read/Process data')
    f_handler = netCDF4.Dataset(infile, 'r')

    # 8 days running mean => Syntool granule coverage is the central day
    _time_start = netCDF4.num2date(f_handler.variables['time'][0],
                                   f_handler.variables['time'].units)
    time_start = _time_start - datetime.timedelta(hours=12)
    time_stop = time_start + datetime.timedelta(hours=24)
    lat = f_handler.variables['lat'][:]
    lon = f_handler.variables['lon'][:]
    smap_sss = f_handler.variables['sss_smap'][:, :]
    hycom_sss = f_handler.variables['sss_ref'][:, :]

    lon = numpy.mod(lon + 180.0, 360.0) - 180.0
    lon_ind = numpy.argsort(lon)
    lon = lon[lon_ind]
    smap_sss = smap_sss[:, lon_ind]
    hycom_sss = hycom_sss[:, lon_ind]

    version = 'v2'
    has_version = 'version' in f_handler.ncattrs()
    if has_version:
        _version, _ = f_handler.version.split(' ', 1)
        if _version.startswith('V'):
            version = _version.lower()
    f_handler.close()

    # Construct metadata/geolocation/band(s)
    logger.info('Construct metadata/geolocation/band(s)')
    dtime, time_range = stfmt.format_time_and_range(time_start, time_stop,
                                                    units='h')
    lat0, dlat = lat[0], lat[1] - lat[0]
    lon0, dlon = lon[0], lon[1] - lon[0]

    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['geotransform'] = [lon0 - 0.5 * dlon, dlon, 0,
                                   lat0 - 0.5 * dlat, 0, dlat]

    metadata = {}
    metadata['name'] = granule_name
    metadata['datetime'] = dtime
    metadata['time_range'] = time_range

    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin_pal, vmax_pal=vmax_pal)

    # SMAP SSS
    # -------------------------------------------------------------------------
    band = []
    array, offset, scale = pack.ubytes_0_254(smap_sss.data, vmin, vmax)
    if numpy.any(smap_sss.mask):
        array[smap_sss.mask] = 255

    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea surface salinity',
                 'unittype': 'PSS',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    product_name = 'SMAP_L3_SSS_RSS_{}'.format(version)
    write_geotiff(product_name, metadata, geolocation, band, infile, outdir)

    # HYCOM SSS
    # -------------------------------------------------------------------------
    band = []
    array, offset, scale = pack.ubytes_0_254(hycom_sss.data, vmin, vmax)
    if numpy.any(hycom_sss.mask):
        array[hycom_sss.mask] = 255

    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea surface salinity',
                 'unittype': 'PSS',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    product_name = 'HYCOM_L3_SSS_RSS_{}'.format(version)
    write_geotiff(product_name, metadata, geolocation, band, infile, outdir)
