# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-03-05

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import struct
import logging
import datetime
import collections
import xml.etree.ElementTree
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack
import syntool_converter

logger = logging.getLogger(__name__)

CONTROL_BITS = {'IGNORE': 0,
                'RANGE': 1,
                'SIGMA': 2,
                'CHI2': 3,
                'CHI2_P': 4,
                'CONTAMINATED': 5,
                'SUNGLINT': 6,
                'MOONGLINT': 7,
                'GAL_NOISE': 8,
                'MIXED_SCENE': 9,
                'REACH_MAXITER': 10,
                'NUM_MEAS_MIN': 11,
                'NUM_MEAS_LOW': 12,
                'MANY_OUTLIERS': 13,
                'MARQ': 14,
                'ROUGHNESS': 15,
                'FOAM': 16,
                'ECMWF': 17,
                'VALID': 18,
                'NO_SURFACE': 19,
                'RANGE_ACARD': 20,
                'SIGMA_ACARD': 21,
                'SPARE': 22,
                'USED_FARA_TEC': 23,
                'POOR_GEOPHYSICAL': 24,
                'POOR_RETRIEVAL': 25,
                'SUSPECT_RFI': 26,
                'RFI_PRONE_X': 27,
                'RFI_PRONE_Y': 28,
                'ADJUSTED_RA': 29,
                'RETRIEV_FAIL': 30,
                'SPARE2': 31}

SCIENCE_BITS = {'LAND_SEA_COAST1': 0,
                'LAND_SEA_COAST2': 1,
                'TEC_GRADIENT': 2,
                'IN_CLIM_ICE': 3,
                'ICE': 4,
                'SUSPECT_ICE': 5,
                'RAIN': 6,
                'HIGH_WIND': 7,
                'LOW_WIND': 8,
                'HIGH_SST': 9,
                'LOW_SST': 10,
                'HIGH_SSS': 11,
                'LOW_SSS': 12,
                'SEA_STATE_1': 13,
                'SEA_STATE_2': 14,
                'SEA_STATE_3': 15,
                'SEA_STATE_4': 16,
                'SEA_STATE_5': 17,
                'SEA_STATE_6': 18,
                'SST_FRONT': 19,
                'SSS_FRONT': 20,
                'ICE_ACARD': 21,
                'ECMWF_LAND': 22}


def get_version(xml_root, namespaces):
    """"""
    xmlns = '{{{}}}'.format(namespaces[None])
    source_path = './{0}Fixed_Header/{0}Source'.format(xmlns)
    version_path = '{}/{}Creator_Version'.format(source_path, xmlns)
    version_node = xml_root.find(version_path, namespaces=namespaces)
    return version_node.text


def get_temporal_coverage(xml_root, namespaces):
    """"""
    xmlns = '{{{}}}'.format(namespaces[None])
    validity_period_path = './{0}Fixed_Header/{0}Validity_Period'.format(xmlns)
    start_path = '{}/{}Validity_Start'.format(validity_period_path, xmlns)
    stop_path = '{}/{}Validity_Stop'.format(validity_period_path, xmlns)
    start_node = xml_root.find(start_path, namespaces=namespaces)
    stop_node = xml_root.find(stop_path, namespaces=namespaces)
    start_str = start_node.text[4:]
    stop_str = stop_node.text[4:]
    start_dt = datetime.datetime.strptime(start_str, '%Y-%m-%dT%H:%M:%S')
    stop_dt = datetime.datetime.strptime(stop_str, '%Y-%m-%dT%H:%M:%S')
    return (start_dt, stop_dt)


def get_spatial_coverage(xml_root, namespaces):
    """"""
    xmlns = '{{{}}}'.format(namespaces[None])
    header_path = '.'
    header_path = '{}/{}Variable_Header'.format(header_path, xmlns)
    header_path = '{}/{}Specific_Product_Header'.format(header_path, xmlns)
    header_path = '{}/{}L2_Product_Location'.format(header_path, xmlns)

    grid_south_path = '{}/{}Southernmost_Gridpoint_ID'.format(header_path,
                                                              xmlns)
    grid_south_node = xml_root.find(grid_south_path,
                                    namespaces=namespaces)

    grid_south_lat_path = '{}/{}Southernmost_Latitude'.format(header_path,
                                                              xmlns)
    grid_south_lat_node = xml_root.find(grid_south_lat_path,
                                        namespaces=namespaces)

    grid_north_path = '{}/{}Northernmost_Gridpoint_ID'.format(header_path,
                                                              xmlns)
    grid_north_node = xml_root.find(grid_north_path,
                                    namespaces=namespaces)

    grid_north_lat_path = '{}/{}Northernmost_Latitude'.format(header_path,
                                                              xmlns)
    grid_north_lat_node = xml_root.find(grid_north_lat_path,
                                        namespaces=namespaces)

    grid_west_path = '{}/{}Westernmost_Gridpoint_ID'.format(header_path,
                                                            xmlns)
    grid_west_node = xml_root.find(grid_west_path,
                                   namespaces=namespaces)

    grid_west_lon_path = '{}/{}Westernmost_Longitude'.format(header_path,
                                                             xmlns)
    grid_west_lon_node = xml_root.find(grid_west_lon_path,
                                       namespaces=namespaces)

    grid_east_path = '{}/{}Easternmost_Gridpoint_ID'.format(header_path,
                                                            xmlns)
    grid_east_node = xml_root.find(grid_east_path,
                                   namespaces=namespaces)

    grid_east_lon_path = '{}/{}Easternmost_Longitude'.format(header_path,
                                                             xmlns)
    grid_east_lon_node = xml_root.find(grid_east_lon_path,
                                       namespaces=namespaces)

    south_y = float(grid_south_node.text)
    south_lat = float(grid_south_lat_node.text)
    north_y = float(grid_north_node.text)
    north_lat = float(grid_north_lat_node.text)
    west_x = float(grid_west_node.text)
    west_lon = float(grid_west_lon_node.text)
    east_x = float(grid_east_node.text)
    east_lon = float(grid_east_lon_node.text)

    grid = {'south': south_y,
            'north': north_y,
            'west': west_x,
            'east': east_x}
    lonlat = {'south': south_lat,
              'north': north_lat,
              'west': west_lon,
              'east': east_lon}
    return (grid, lonlat)


def get_data_geometry(xml_root, namespaces, ds_name):
    """"""
    xmlns = '{{{}}}'.format(namespaces[None])
    header_path = '.'
    header_path = '{}/{}Variable_Header'.format(header_path, xmlns)
    header_path = '{}/{}Specific_Product_Header'.format(header_path, xmlns)
    header_path = '{}/{}List_of_Data_Sets'.format(header_path, xmlns)

    dset_path = '{}/{}Data_Set'.format(header_path, xmlns)
    dset_name_nodes = xml_root.findall(dset_path, namespaces=namespaces)
    for dset_node in dset_name_nodes:
        dset_name_node = dset_node.find('./{}DS_Name'.format(xmlns))
        if ds_name == dset_name_node.text:
            break
    else:
        logger.error('No dataset named {} in this file'.format(ds_name))
        raise

    num_dsr_path = './{}Num_DSR'.format(xmlns)
    dsr_size_path = './{}DSR_Size'.format(xmlns)
    ds_size_path = './{}DS_Size'.format(xmlns)
    byte_order_path = './{}Byte_Order'.format(xmlns)
    ds_offset_path = './{}DS_Offset'.format(xmlns)

    num_dsr_node = dset_node.find(num_dsr_path, namespaces=namespaces)
    num_dsr = int(num_dsr_node.text)

    dsr_size_node = dset_node.find(dsr_size_path, namespaces=namespaces)
    dsr_size = int(dsr_size_node.text)

    ds_size_node = dset_node.find(ds_size_path, namespaces=namespaces)
    ds_size = int(ds_size_node.text)

    byte_order_node = dset_node.find(byte_order_path, namespaces=namespaces)
    byte_order = 'little' if ('0123' == byte_order_node.text) else 'big'

    ds_offset_node = dset_node.find(ds_offset_path, namespaces=namespaces)
    ds_offset = int(ds_offset_node.text)

    result = {'records_count': num_dsr,
              'record_size': dsr_size,
              'dataset_size': ds_size,
              'dataset_offset': ds_offset,
              'endianess': byte_order}
    return result


def get_struct_fmt(endianess):
    """"""
    # f: float 4 bytes
    # I: unsigned int 4 bytes
    # H: unsigned short 2 bytes

    # 01 - Number of Grid_Points data set record structures [uint32]
    n_grid_points = 'I'

    record_fields = collections.OrderedDict()
    # 02 - Unique identifier of Earth fixed grid point [uint32]
    record_fields['Grid_Point_ID'] = 'I'

    # 03 - Geodetic latitude of grid point (WSG84) [float32]
    record_fields['Latitude'] = 'f'

    # 04 - Geocentric longitude of grid point [float32]
    record_fields['Longitude'] = 'f'

    # 05 - Equivalent footprint diameter (default value: 999 if no olarization
    # for this grid point) [float32]
    record_fields['Equiv_ftprt_diam'] = 'f'

    # 06 - Mean time of acquisition for all valid TB measurements of DGG point.
    # Expressed in UTC decimal days (in MJD2000 reference). Default value -999
    # if grid point not processed [float32]
    record_fields['Mean_acq_time'] = 'f'

    # 07 - Sea Surface salinity corrected for lan-sea contamination (default
    # value -999 if not processed) [float32]
    record_fields['SSS_corr'] = 'f'

    # 08 - Theoretical uncertainty computed for SSS_corr (defualt value -999 if
    # not  processed) [float32]
    record_fields['Sigma_SSS_corr'] = 'f'

    # 09 - Sea surface salinity without land-sea correction (defualt value -999
    # if not processed) [float32]
    record_fields['SSS_uncorr'] = 'f'

    # 10 - Theoretical uncertainty computed for SSS_uncorr (default value -999
    # if not processed) [float32]
    record_fields['Sigma_SSS_uncorr'] = 'f'

    # 11 - Sea surface salinity anomaly = SSS_corr - climatology (nominally
    # WOA2009 climatology, default value -999 if not processed) [float32]
    record_fields['SSS_anom'] = 'f'

    # 12 - Theoretical uncertainty computed for SSS_anom (default value -999 if
    # not processed, nominally copied from Sigma_SSS_corr) [float32]
    record_fields['Sigma_SSS_anom'] = 'f'

    # 13 - Effective_Acard retrieved with minimalist model (default value -999
    # if not processed) [float32]
    record_fields['A_card'] = 'f'

    # 14 - Theoretical uncertainty computed for Acard (default value -999 if
    # not processed) [float32]
    record_fields['Sigma_Acard'] = 'f'

    # 15 - 10 neutral wind module derived from ECMWF UN10 & VN10 (defual value
    # -999 if not processed) [float32]
    record_fields['WS'] = 'f'

    # 16 - Sea surface temperature from ECMWF (default -999 if grid point not
    # processed) [float32]
    record_fields['SST'] = 'f'

    # 17 - Brightness temperature at surface level derived with default forward
    # model and retrieved geophysical parameters, H olarization direction
    # (default value -999 if grid point not processed) [float32]
    record_fields['Tb_42.5H'] = 'f'

    # 18 - Theoretical uncertainty computed for Tb42.5H (default value -999 if
    # grid point not processed) [float32]
    record_fields['Sigma_Tb_42.5H'] = 'f'

    # 19 - Brightness temperature at surface level derived with default forward
    # model and retrieved geophysical parameters, V olarization direction
    # (default value -999 if grid point not processed) [float32]
    record_fields['Tb_42.5V'] = 'f'

    # 20 - Theoretical uncertainty computed for Tb42.5V (default value -999 if
    # grid point not processed) [float32]
    record_fields['Sigma_Tb_42.5V'] = 'f'

    # 21 - Brightness temperature at antenna level derived with default forward
    # model and retrieved geophysical parameters, X olarization direction
    # (default value -999 if grid point not processed) [float32]
    record_fields['Tb_42.5X'] = 'f'

    # 22 - Theoretical uncertainty computed for Tb42.5X (default value -999 if
    # grid point not processed) [float32]
    record_fields['Sigma_Tb_42.5X'] = 'f'

    # 23 - Brightness temperature at antenna level derived with efualt forward
    # model and retrieved geophysical parameters, Y olarization direction
    # (default value -999 if grid point not processed) [float32]
    record_fields['Tb_42.5Y'] = 'f'

    # 24 - Theoretical uncertainty computed for Tb42.5Y (default value -999 if
    # grid point not processed) [float32]
    record_fields['Sigma_42.5Y'] = 'f'

    # 25 - Control flags for SSS_corr retrieval. See table 4-20 for details.
    # Least significant bit is field #01, most significant bit is field #32
    # [uint32]
    record_fields['Control_Flags_corr'] = 'I'

    # 26 - Control flags for SSS_uncorr retrieval. See table 4-20 for details.
    # Least significant bit is field #01, most significant bit is field #32
    # [uint32]
    record_fields['Control_Flags_uncorr'] = 'I'

    # 27 - Control flags for SSS_anom retrieval. See table 4-20 for details.
    # Least significant bit is field #01, most significant bit is field #32
    # [uint32]
    record_fields['Control_Flags_anom'] = 'I'

    # 28 - Control flags for Acard retrieval. See table 4-20 for details.
    # Least significant bit is field ~01, most significant bit is field #32
    # [uint32]
    record_fields['Control_Flags_Acard'] = 'I'

    # 29 - Normalized retrieval fit quality index for SSS_corr, scaled by
    # multiplying by 100 (default value 0 if grid point not processed) [uint16]
    record_fields['Dg_chi2_corr'] = 'H'

    # 30 - Normalized retrieval fit quality index for SSS_uncorr, scaled by
    # multiplying by 100 (default value 0 if grid point not processed) [uint16]
    record_fields['Dg_chi2_uncorr'] = 'H'

    # 31 - Wind speed module retrieved with SSS_corr, scaled by multiplying by
    # 1000 (default value -999 if not processed) [uint16]
    record_fields['WS_corr'] = 'H'

    # 32 - Normalized retrieval fir quality index from olarizat model, scaled
    # by multiplying by 100 (default value 0 if grid point not processed)
    # [uint16]
    record_fields['Dg_chi2_Acard'] = 'H'

    # 33 - Normalized chi2 high value acceptability probability for SSS_corr,
    # scaled by multiplying by 1000 (default value 0 if grid point not
    # processed) [uint16]
    record_fields['Dg_chi2_P_corr'] = 'H'

    # 34 - Normalized chi2 high value acceptability probability for SSS_uncorr,
    # scaled by multiplying by 1000 (default value 0 if grid point not
    # processed) [uint16]
    record_fields['Dg_chi2_P_uncorr'] = 'H'

    # 35 - Error on wind speed module retrieved with SSS_corr, scaled by
    # multiplying by 1000 (default value -999 if not processed) [uint16]
    record_fields['Sigma_WS_corr'] = 'H'

    # 36 - Normalized chi2 high value acceptability probability with from
    # cardoids model, scaled by multiplying by 1000 (default value 0 if grid
    # point not processed) [uint16]
    record_fields['Dg_chi2_P_Acard'] = 'H'

    # 37 - Quality index for SSS_corr: lower is better (default 999 if grid
    # point not processed) [uint16]
    record_fields['Dg_quality_SSS_corr'] = 'H'

    # 38 - Quality index for SSS_uncorr: lower is better (default 999 if grid
    # point not processed) [uint16]
    record_fields['Dg_quality_SSS_uncorr'] = 'H'

    # 39 - Quality index for SSS_anom: lower is better (default 999 if grid
    # point not processed, nominally copied from Dg_quality_SSS_anom) [uint16]
    record_fields['Dg_quality_SSS_anom'] = 'H'

    # 40 - Salinity from interpolated climatology, scaled by multiplying by 100
    # [uint16]
    record_fields['SSS_climatology'] = 'H'

    # 41 - Number of iterations for the retrieval of SSS_corr (0 if not
    # processed) [ubyte]
    record_fields['Dg_num_iter_corr'] = 'B'
    # 42 - Number of iterations for the retrieval of SSS_uncorr (0 if not
    # processed) [ubyte]
    record_fields['Dg_num_iter_uncorr'] = 'B'

    # 43 - Distance to nearest coast, scaled by mutiplying by 0.05 [ubyte]
    record_fields['Coast_distance'] = 'B'

    # 44 - Number of iterations for the retrieval of Acard (0 if not processed)
    # [ubyte]
    record_fields['Dg_num_iter_Acard'] = 'B'

    # 45 - Number of measurements available in L1c product [uint16]
    record_fields['Dg_num_meas_l1c'] = 'H'

    # 46 - Number of valid measurements available for SSS retrieval [uint16]
    record_fields['Dg_num_meas_valid'] = 'H'

    # 47 - Number of valid measurements with BORDER_FOV flag raised [uint16]
    record_fields['Dg_border_fov'] = 'H'

    # 48 - Number of valid measurements with AF_FOV flag raised [uint16]
    record_fields['Dg_af_fov'] = 'H'

    # 49 - Number of measurements with UN_TAILS flag raised [uint16]
    record_fields['Dg_sun_tails'] = 'H'

    # 50 - Number of measurements with SUN_GLINT_FOV flag raised [uint16]
    record_fields['Dg_sun_glint_area'] = 'H'

    # 51 - Number of measurements with SUN_FOV flag raised [uint16]
    record_fields['Dg_sun_glint_fov'] = 'H'

    # 52 - Number of measurements with SUN_FOV flag raised [uint16]
    record_fields['Dg_sun_fov'] = 'H'

    # 53 - Number of measurements with L2 sunglint flag raised [uint16]
    record_fields['Dg_sun_glint_L2'] = 'H'

    # 54 - Number of suspected ice-contaminated measurements [uint16]
    record_fields['Dg_Suspect_ice'] = 'H'

    # 55 - Number of measurements discarded due to errors in galactic noise
    # [uint16]
    record_fields['Dg_galactic_Noise_Error'] = 'H'

    # 56 - Count measurements with specular direction toward a strong galactic
    # source [uint16]
    record_fields['Dg_sky'] = 'H'

    # 57 - Number of measurements with L2 moonglint flag raised [uint16]
    record_fields['Dg_moonglint'] = 'H'

    # 58 - Number of measurements suspected by L1 as being contaminated by RFI
    # [uint16]
    record_fields['Dg_RFI_L1'] = 'H'

    # 59 - Number of measurements suspected of being contaminated by RFI in X
    # olarization [uint16]
    record_fields['Dg_RFI_X'] = 'H'

    # 60 - Number of measurements suspected of being contaminated by RFI in Y
    # olarization [uint16]
    record_fields['Dg_RFI_Y'] = 'H'

    # 61 - Probability of grid point being contaminated by RFI, estimated from
    # AUX_DGGRFI [uint16]
    record_fields['Dg_RFI_probability'] = 'H'

    # 62 - Grid point distance from the satellite track (default value: 999 if
    # grid point not processed) [float32]
    record_fields['X_swath'] = 'f'

    # 63 - Science flag for SSS_corr retrieval. see table 4-21 for details.
    # least significant bit is field #01, most significant bit is field #32
    # [uint32]
    record_fields['Science_Flags_corr'] = 'I'

    # 64 - Science flag for SSS_uncorr retrieval. see table 4-21 for details.
    # least significant bit is field #01, most significant bit is field #32
    # [uint32]
    record_fields['Science_Flags_uncorr'] = 'I'

    # 65 - Science flag for SSS_anom retrieval. see table 4-21 for details.
    # least significant bit is field #01, most significant bit is field #32
    # [uint32]
    record_fields['Science_Flags_anom'] = 'I'

    # 66 - Science flag for Acard retrieval. see table 4-21 for details.
    # least significant bit is field #01, most significant bit is field #32
    # [uint32]
    record_fields['Science_Flags_Acard'] = 'I'

    fmt = '{}{}'.format(endianess, n_grid_points)

    record_fmt = '{}{}'.format(endianess, ''.join(record_fields.values()))

    return (fmt, record_fmt,)


def get_grid_indices(lat, lon, x_res, y_res):
    """Implement BEC method to regrid data.
    See:
    https://www.researchgate.net/publication/
    263271245_Remapping_ISEA_grid_to_lat-lon_grid_in_SMOS_context"""
    delta = 0.5 + 1.0 / (2 * numpy.cos(numpy.deg2rad(lat)))
    rel_i_min = -1 * delta + (lon + 180.0) / x_res
    rel_i_max = +1 * delta + (lon + 180.0) / x_res
    i_min = numpy.ceil(rel_i_min).astype('int')
    i_max = numpy.floor(rel_i_max).astype('int')
    rel_j = (lat + 90.0) / y_res
    j = numpy.floor(rel_j).astype('int')
    return (i_min, i_max, j)


def is_point_valid(unpacked):
    """"""
    lat_value = unpacked[1]
    sss_value = unpacked[5]
    dg_af_fov = unpacked[46]
    control_flags = unpacked[23]  # int32
    science_flags = unpacked[61]  # int32

    control_on = numpy.int32(0)
    control_on = control_on + numpy.int32(1 << CONTROL_BITS['ECMWF'])

    control_off = numpy.int32(0)
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['CHI2_P'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['SUNGLINT'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['MOONGLINT'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['REACH_MAXITER'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['NUM_MEAS_MIN'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['NUM_MEAS_LOW'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['MANY_OUTLIERS'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['MARQ'])
    control_off = control_off + numpy.int32(1 << CONTROL_BITS['SUSPECT_RFI'])

    ctrl_on_mask = (numpy.bitwise_and(control_on, control_flags) == control_on)
    ctrl_off_mask = (numpy.bitwise_and(control_off, control_flags) == 0)
    control_valid = (ctrl_on_mask & ctrl_off_mask)

    science_on = numpy.int32(0)
    science_on = science_on + numpy.int32(1 << SCIENCE_BITS['LAND_SEA_COAST1'])
    science_on = science_on + numpy.int32(1 << SCIENCE_BITS['LOW_WIND'])

    science_off = numpy.int32(0)
    science_off = science_off + numpy.int32(1 << SCIENCE_BITS['ICE'])
    science_off = science_off + numpy.int32(1 << SCIENCE_BITS['SUSPECT_ICE'])

    sci_on_mask = (numpy.bitwise_and(science_on, science_flags) == science_on)
    sci_off_mask = (numpy.bitwise_and(science_off, science_flags) == 0)
    science_valid = (sci_on_mask & sci_off_mask)

    af_fov_valid = (dg_af_fov > 130.0)

    sss_valid = False
    if sss_value != -999:
        if abs(lat_value) > 30.0:
            if (30.0 <= sss_value) and (sss_value <= 40.0):
                sss_valid = True
        else:
            if (25.0 <= sss_value) and (sss_value <= 40.0):
                sss_valid = True

    valid_point = (sss_valid & af_fov_valid & control_valid & science_valid)
    return valid_point


def parse_data_set_records(dbl_path, data_geometry, sss_grid, count_grid):
    """"""
    x_size, y_size = numpy.shape(sss_grid)
    x_res = 360.0 / x_size
    y_res = 180.0 / y_size

    endianess = '>'
    if 'little' == data_geometry['endianess']:
        endianess = '<'

    (fmt, record_fmt) = get_struct_fmt(endianess)

    header_size = struct.calcsize(fmt)
    record_size = struct.calcsize(record_fmt)

    with open(dbl_path, 'rb') as f:
        data = f.read()
    count = struct.unpack(fmt, data[:header_size])

    for k in range(0, count[0]):
        record_start = header_size + record_size * k
        record_end = record_start + record_size
        packed_record = data[record_start: record_end]
        unpacked = struct.unpack(record_fmt, packed_record)

        valid_point = is_point_valid(unpacked)
        if valid_point:
            lat_value = unpacked[1]
            lon_value = unpacked[2]
            sss_value = unpacked[5]
            (i_min, i_max, j) = get_grid_indices(lat_value, lon_value, x_res,
                                                 y_res)
            if i_min >= i_max:
                _i = numpy.mod(i_min, x_size).astype('int')
                count_grid[_i, j] += 1
                sss_grid[_i, j] += sss_value
            else:
                for i in range(i_min, i_max):  # [i_min, i_max[
                    _i = numpy.mod(i, x_size).astype('int')
                    count_grid[_i, j] += 1
                    sss_grid[_i, j] += sss_value


def parse_header(hdr_path):
    """"""
    namespaces = {None: 'http://193.146.123.163/smos/schemas',
                  'xsi': 'http://www.w3.org/2001/XMLSchema-instance'}

    xml_tree = xml.etree.ElementTree.parse(hdr_path)
    xml_root = xml_tree.getroot()

    """
    # v662
    namespaces = {None: 'http://193.146.123.163/smos/schemas',
                  'xsi': 'http://www.w3.org/2001/XMLSchema-instance'}
    # v700
    namespaces = {None: 'http://213.170.46.150/smos/schemas',
                  'xsi': 'http://www.w3.org/2001/XMLSchema-instance'}
    """
    namespaces = {None: xml_root.tag[1:xml_root.tag.find('}')],
                  'xsi': 'http://www.w3.org/2001/XMLSchema-instance'}

    # Version
    version = get_version(xml_root, namespaces)

    # Time coverage
    (start_dt, stop_dt) = get_temporal_coverage(xml_root, namespaces)

    # Space coverage
    (grid, lonlat) = get_spatial_coverage(xml_root, namespaces)

    # Read data geometry
    data_geometry = get_data_geometry(xml_root, namespaces, 'SSS_SWATH')

    return (version, start_dt, stop_dt, data_geometry)


def smos_l2_sss_dpgs(input_path, output_path, vmin=31.825, vmax=38.175,
                     vmin_pal=32, vmax_pal=38):
    """"""
    granule_dir = os.path.abspath(os.path.dirname(input_path))
    granule_filename = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_filename)

    hdr_path = os.path.join(granule_dir, '{}.HDR'.format(granule_name))
    dbl_path = os.path.join(granule_dir, '{}.DBL'.format(granule_name))

    now = datetime.datetime.utcnow()
    (version, time_start, time_stop, data_geometry) = parse_header(hdr_path)
    dtime, time_range = stfmt.format_time_and_range(time_start, time_stop,
                                                    units='s')

    metadata = {}
    metadata['product_name'] = 'SMOS_L2_SSS_DPGS_v{}'.format(version)
    metadata['name'] = granule_name
    metadata['datetime'] = dtime
    metadata['time_range'] = time_range
    metadata['source_URI'] = input_path
    metadata['source_provider'] = 'DPGS'
    metadata['processing_center'] = ''
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'sea surface salinity'

    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin,
                                         vmax=vmax,
                                         vmin_pal=vmin_pal,
                                         vmax_pal=vmax_pal)

    output_shape = (1440, 720)
    sss_grid = numpy.full(output_shape, 0.0, dtype=numpy.float)
    count_grid = numpy.full(output_shape, 0, dtype=numpy.integer)
    parse_data_set_records(dbl_path, data_geometry, sss_grid, count_grid)

    valid_indices = numpy.where(count_grid > 0)
    _sss = numpy.full(output_shape, -1000, dtype=numpy.float)
    sss = numpy.ma.array(_sss)
    sss.mask = True
    sss.mask[valid_indices] = False
    sss[valid_indices] = sss_grid[valid_indices] / count_grid[valid_indices]

    # Remove all surrounding rows and columns which contain no valid data
    # (i.e. optimize the size of the bounding box) to reduce the size of the
    # output file.
    cols_mask = numpy.logical_and.reduce(sss.mask, axis=0)
    rows_mask = numpy.logical_and.reduce(sss.mask, axis=1)
    valid_cols_ind = numpy.where(~cols_mask)
    valid_rows_ind = numpy.where(~rows_mask)

    if 0 >= len(valid_rows_ind[0]):
        logger.warn('No valid data to process')
        raise syntool_converter.EarlyExit()

    lon_slice = slice(valid_rows_ind[0][0], 1 + valid_rows_ind[0][-1])
    lat_slice = slice(valid_cols_ind[0][0], 1 + valid_cols_ind[0][-1])
    sss = sss[lon_slice, lat_slice].transpose()

    dlon = 360.0 / output_shape[0]
    lon0 = -180.0 + lon_slice.start * dlon
    dlat = 180.0 / output_shape[1]
    lat0 = -90.0 + lat_slice.start * dlat
    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['geotransform'] = [lon0, dlon, 0,
                                   lat0, 0, dlat]

    band = []
    array, offset, scale = pack.ubytes_0_254(sss.data, vmin, vmax)
    if numpy.any(sss.mask):
        array[numpy.where(sss.mask)] = 255

    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea surface salinity',
                 'unittype': 'PSS',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    # Write geotiff
    tifffile = stfmt.format_tifffilename(output_path, metadata,
                                         create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)
