# -*- coding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-02-23

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import pyproj
import netCDF4
import logging
import datetime
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt

logger = logging.getLogger(__name__)


def smos_l3_sss_cpdc(infile, outdir, vmin=31.825, vmax=38.175, vmin_pal=32,
                     vmax_pal=38):
    """"""
    granule_filename = os.path.basename(infile)
    granule_name, _ = os.path.splitext(granule_filename)

    # Read/Process data
    f_handler = netCDF4.Dataset(infile, 'r')
    lon = f_handler.variables['lon'][:]
    lat = f_handler.variables['lat'][:]
    sss = f_handler.variables['Mean_Sea_Surface_Salinity_1'][:, :]
    proj_txt = f_handler.proj4text
    f_handler.close()

    # Construct metadata/geolocation/band(s)
    now = datetime.datetime.utcnow()
    _, tail = granule_name.split('__', 1)
    start_str, stop_str, version, _ = tail.split('_', 3)
    time_start = datetime.datetime.strptime(start_str, '%Y%m%dT%H%M%S')
    time_stop = datetime.datetime.strptime(stop_str, '%Y%m%dT%H%M%S')
    is_monthly = (time_stop - time_start) > datetime.timedelta(days=15)

    if int(version) >= 317 and not is_monthly:
        # daily product with a 10-days running window
        time_start = time_start + datetime.timedelta(days=4, hours=12)
        time_stop = time_stop - datetime.timedelta(days=4, hours=12)

    dtime, time_range = stfmt.format_time_and_range(time_start, time_stop,
                                                    units='h')

    # The product was created using an EPSG:6933 grid, so the geotransform
    # should be expressed using the parameters of this grid instead of lat/lon.
    srs = stfmt.proj2srs(str(proj_txt))
    iproj = pyproj.Proj(init='epsg:4326')
    oproj = pyproj.Proj(srs.ExportToProj4())

    xx, yy = numpy.meshgrid(lon, lat)

    xp, yp = pyproj.transform(iproj, oproj, xx, yy)
    mxp = numpy.mean(xp, axis=0)
    myp = numpy.mean(yp, axis=1)
    dx = (mxp[-1] - mxp[0]) / numpy.shape(lon)[0]
    dy = (myp[-1] - myp[0]) / numpy.shape(lat)[0]
    x0 = mxp[0]
    y0 = myp[0]

    geolocation = {}
    geolocation['projection'] = srs.ExportToWkt()
    geolocation['geotransform'] = [x0, dx, 0,
                                   y0, 0, dy]

    product_name = 'SMOS_L3_SSS_CPDC_{}'.format(version)
    if is_monthly:
        product_name = '{}_monthly'.format(product_name)

    metadata = {}
    metadata['product_name'] = product_name
    metadata['name'] = granule_name
    metadata['datetime'] = dtime
    metadata['time_range'] = time_range
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'SMOS CATDS Processing Chain'
    metadata['processing_center'] = ''
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'sea surface salinity'

    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin,
                                         vmax=vmax,
                                         vmin_pal=vmin_pal,
                                         vmax_pal=vmax_pal)

    band = []
    array, offset, scale = pack.ubytes_0_254(sss.data, vmin, vmax)
    if numpy.any(sss.mask):
        array[numpy.where(sss.mask)] = 255

    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea surface salinity',
                 'unittype': 'PSS',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    # Write geotiff
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)
