# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-02-23

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import netCDF4
import logging
import datetime
import dateutil.relativedelta
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def smos_l3_sss_icdc(infile, outdir, vmin=31.825, vmax=38.175, vmin_pal=32,
                     vmax_pal=38):
    """"""
    granule_filename = os.path.basename(infile)
    granule_name, _ = os.path.splitext(granule_filename)

    # Read/Process data
    f_handler = netCDF4.Dataset(infile, 'r')
    time_units = f_handler.variables['time'].units
    t = numpy.ma.getdata(f_handler.variables['time'][:])[0]
    lon = f_handler.variables['lon'][:]
    lat = f_handler.variables['lat'][::-1]
    sss = f_handler.variables['sss'][::-1, :]
    f_handler.close()

    # Construct metadata/geolocation/band(s)
    now = datetime.datetime.utcnow()
    if 'days since 0-1-1' == time_units:
        # standard library cannot handle this... Use a workaround:
        epoch = datetime.datetime(1970, 1, 1)
        granule_delta = datetime.timedelta(days=t)
        _dt = epoch + granule_delta
        dt = _dt - dateutil.relativedelta.relativedelta(years=1970)
    else:
        dt = netCDF4.num2date(t, time_units)

    # Monthly product
    time_start = datetime.datetime(dt.year, dt.month, 1)
    time_stop = time_start + dateutil.relativedelta.relativedelta(months=1)
    time_stop = time_stop - datetime.timedelta(seconds=1)
    dtime, time_range = stfmt.format_time_and_range(time_start, time_stop,
                                                    units='h')

    lat0, dlat = lat[0], lat[1] - lat[0]
    lon0, dlon = lon[0], lon[1] - lon[0]
    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['geotransform'] = [lon0 - 0.5 * dlon, dlon, 0,
                                   lat0 - 0.5 * dlat, 0, dlat]

    metadata = {}
    metadata['product_name'] = 'SMOS_L3_SSS_ICDC'
    metadata['name'] = granule_name
    metadata['datetime'] = dtime
    metadata['time_range'] = time_range
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'ICDC'
    metadata['processing_center'] = ''
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'sea surface salinity'

    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin,
                                         vmax=vmax,
                                         vmin_pal=vmin_pal,
                                         vmax_pal=vmax_pal)

    band = []
    array, offset, scale = pack.ubytes_0_254(sss.data, vmin, vmax)
    if numpy.any(sss.mask):
        array[numpy.where(sss.mask)] = 255

    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea surface salinity',
                 'unittype': 'PSS',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    # Write geotiff
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)
