# -*- coding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-02-26

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import pyproj
import logging
import netCDF4
import datetime
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt

logger = logging.getLogger(__name__)


def smos_l3_sss_locean(infile, outdir, vmin=31.825, vmax=38.175,
                       vmin_pal=32, vmax_pal=38):
    """"""
    granule_filename = os.path.basename(infile)
    granule_name, _ = os.path.splitext(granule_filename)

    # Read/Process data
    f_handler = netCDF4.Dataset(infile, 'r')
    time_units = f_handler.variables['time'].units
    _t = f_handler.variables['time'][0]
    lat = f_handler.variables['lat'][::-1]
    lon = f_handler.variables['lon'][:]
    sss = f_handler.variables['SSS'][::-1, :]
    title = f_handler.title
    f_handler.close()

    version = '2'
    if 'SMOS SSS - LOCEAN_ACRI_v2018' == title:
        version = '3'
    elif 'SMOS SSS - LOCEAN_ACRI_v2019' == title:
        version = '4'
    elif 'SMOS SSS - LOCEAN_ACRI_v2020' == title:
        version = '5'
    elif 'SMOS SSS - LOCEAN_ACRI_v2022' == title:
        version = '7'

    # Construct metadata/geolocation/band(s)
    logger.info('Construct metadata/geolocation/band(s)')
    now = datetime.datetime.utcnow()
    _time = netCDF4.num2date(_t, time_units)
    time_start = _time - datetime.timedelta(days=2)
    time_stop = _time + datetime.timedelta(days=2)
    dtime, time_range = stfmt.format_time_and_range(time_start, time_stop,
                                                    units='h')

    # Definition for EPSG:6933 (for EASE grid)
    proj_txt = '+proj=cea +lon_0=0 +lat_ts=30 +x_0=0 +y_0=0 +ellps=WGS84'
    proj_txt += ' +units=m +no_defs'

    # The product was created using an EPSG:6933 grid, so the geotransform
    # should be expressed using the parameters of this grid instead of lat/lon.
    srs = stfmt.proj2srs(str(proj_txt))
    iproj = pyproj.Proj(init='epsg:4326')
    oproj = pyproj.Proj(srs.ExportToProj4())

    xx, yy = numpy.meshgrid(lon, lat)

    xp, yp = pyproj.transform(iproj, oproj, xx, yy)
    mxp = numpy.mean(xp, axis=0)
    myp = numpy.mean(yp, axis=1)
    dx = (mxp[-1] - mxp[0]) / numpy.shape(lon)[0]
    dy = (myp[-1] - myp[0]) / numpy.shape(lat)[0]
    x0 = mxp[0]
    y0 = myp[0]

    geolocation = {}
    geolocation['projection'] = srs.ExportToWkt()
    geolocation['geotransform'] = [x0, dx, 0,
                                   y0, 0, dy]

    metadata = {}
    metadata['product_name'] = 'SMOS_L3_SSS_LOCEAN_v{}'.format(version)
    metadata['name'] = granule_name
    metadata['datetime'] = dtime
    metadata['time_range'] = time_range
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'CEC-OS LOCEAN/IPSL/ACRI-ST'
    metadata['processing_center'] = ''
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'sea surface salinity'

    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin_pal, vmax_pal=vmax_pal)

    band = []
    array, offset, scale = pack.ubytes_0_254(sss.data, vmin, vmax)
    if numpy.any(sss.mask):
        array[numpy.where(sss.mask)] = 255

    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'sea surface salinity',
                 'unittype': 'PSS',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    # Write geotiff
    logger.info('Write geotiff')
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)
