# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-02-23

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def smos_l3u_wind_nrt(infile, outdir, vmin=0.0, vmax=50.8, vmin_pal=0.0,
                      vmax_pal=50.0, crop_lat=85.0):
    """"""
    granule_filename = os.path.basename(infile)
    granule_name, _ = os.path.splitext(granule_filename)

    # Read/Process data
    f_handler = netCDF4.Dataset(infile, 'r')
    lon = f_handler.variables['longitude'][:]
    lat = f_handler.variables['latitude'][:]
    wind = f_handler.variables['wind_speed'][0, :, :]
    meas_time = f_handler.variables['measurement_time'][0, :, :]
    time_units = f_handler.variables['measurement_time'].units
    f_handler.close()

    if numpy.all(numpy.isnan(meas_time)):
        logger.warn('No valid time in {}'.format(infile))
        return None

    # Convert longitudes from [0, 360] to [-180, 180]
    lon_above_ind = numpy.where(lon > 180.0)
    lon[lon_above_ind] = lon[lon_above_ind] - 360.0

    # Reorganize data to keep longitudes in ascending order
    lon_ind_sorted = numpy.argsort(lon)
    lon = lon[lon_ind_sorted]
    wind = wind[:, lon_ind_sorted]

    # Crop data in high latitude
    lat_ind = numpy.where(numpy.abs(lat) < crop_lat)
    lat = lat[lat_ind]
    wind = wind[lat_ind[0], :]

    # Construct metadata/geolocation/band(s)
    now = datetime.datetime.utcnow()
    time_start = netCDF4.num2date(numpy.nanmin(meas_time), time_units)
    time_stop = netCDF4.num2date(numpy.nanmax(meas_time), time_units)
    dtime, time_range = stfmt.format_time_and_range(time_start, time_stop,
                                                    units='s')

    lat0, dlat = lat[0], lat[1] - lat[0]
    lon0, dlon = lon[0], lon[1] - lon[0]
    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['geotransform'] = [lon0, dlon, 0,
                                   lat0, 0, dlat]

    metadata = {}
    metadata['product_name'] = None  # set later
    metadata['name'] = granule_name
    metadata['datetime'] = dtime
    metadata['time_range'] = time_range
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'IFREMER'
    metadata['processing_center'] = ''
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'wind speed'

    # Save granule for low speed winds |0, 50] m.s-1
    metadata['product_name'] = 'SMOS_L3U_WS_NRT'
    vmin = 0.0
    vmax = 50.8
    vmin_pal = 0.0
    vmax_pal = 50.0
    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin,
                                         vmax=vmax,
                                         vmin_pal=vmin_pal,
                                         vmax_pal=vmax_pal)

    band = []
    _data = numpy.ma.getdata(wind)
    _mask = (numpy.ma.getmask(wind) | ~numpy.isfinite(wind))
    array, offset, scale = pack.ubytes_0_254(_data, vmin, vmax)
    if numpy.any(_mask):
        array[numpy.where(_mask)] = 255

    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'retrieved wind speed',
                 'unittype': 'm.s-1',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    # Write geotiff
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)

    """
    # Save granule for higher speed winds |30, 60] m.s-1
    metadata['product_name'] = 'SMOS_L3U_HWS_NRT'
    vmin = 30.0
    vmax = 70.0
    vmin_pal = 30.0
    vmax_pal = 70.0
    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin,
                                         vmax=vmax,
                                         vmin_pal=vmin_pal,
                                         vmax_pal=vmax_pal)

    band = []
    _data = numpy.ma.getdata(wind)
    infinite_ind = numpy.where(~numpy.isfinite(_data))
    _data[infinite_ind] = 0
    low_wind_mask = _data < vmin
    _data[infinite_ind] = numpy.nan  # not required but clearer to restore

    _mask = (numpy.ma.getmask(wind) | ~numpy.isfinite(wind) | low_wind_mask)
    array, offset, scale = pack.ubytes_0_254(_data, vmin, vmax)
    if numpy.any(_mask):
        array[numpy.where(_mask)] = 255

    band.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'retrieved wind speed',
                 'unittype': 'm.s-1',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    # Write geotiff
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)
    """


if '__main__' == __name__:
    import sys

    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.DEBUG)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    smos_l3u_wind_nrt(input_path, output_path)
