#!/usr/bin/env python
# coding=utf-8

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import netCDF4
import datetime
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt


def smos_l4_oscar_current(infile, outdir,
                          vmin=0., vmax=5.08, vmin_pal=0., vmax_pal=2.):
    """
    """
    # Read/Process data
    print 'Read/Process data'
    smos = netCDF4.Dataset(infile, 'r')
    time_start_units = smos.variables['date_start'].units
    time_start = netCDF4.num2date(smos.variables['date_start'][0],
                                  time_start_units)
    time_stop_units = smos.variables['date_stop'].units
    time_stop = netCDF4.num2date(smos.variables['date_stop'][0],
                                 time_stop_units)
    if (time_stop - time_start) == datetime.timedelta(days=6):
        time_stop += datetime.timedelta(days=1)
    lat = smos.variables['latitude'][::-1]
    lon = smos.variables['longitude'][:]
    ucur = smos.variables['Zonal_component_surface_currents'][0, ::-1, :]
    vcur = smos.variables['Meridional_component_surface_currents'][0, ::-1, :]

    # Construct metadata/geolocation/band(s)
    print 'Construct metadata/geolocation/band(s)'
    dtime, time_range = stfmt.format_time_and_range(time_start, time_stop,
                                                    units='h')
    lat0, dlat = lat[0], lat[1] - lat[0]
    lon0, dlon = lon[0], lon[1] - lon[0]

    now = datetime.datetime.utcnow()
    metadata = {}
    metadata['product_name'] = 'SMOS_L4_OSCAR_current'
    metadata['name'] = os.path.splitext(os.path.basename(infile))[0]
    metadata['datetime'] = dtime
    metadata['time_range'] = time_range
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'Ifremer/CNES'
    metadata['processing_center'] = ''
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = ['current velocity', 'current direction']
    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['geotransform'] = [lon0-dlon/2., dlon, 0,
                                   lat0-dlat/2., 0, dlat]
    band = []
    data_mask = (numpy.ma.getmaskarray(ucur) | numpy.ma.getmaskarray(vcur))
    curvel = numpy.sqrt(ucur.data**2 + vcur.data**2)
    curdir = numpy.mod(numpy.arctan2(vcur.data, ucur.data)*180./numpy.pi+360.,
                       360.)
    array, offset, scale = pack.ubytes_0_254(curvel, vmin, vmax)
    if numpy.any(data_mask):
        array[numpy.where(data_mask)] = 255

    colortable = stfmt.format_colortable('matplotlib_jet', vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin_pal, vmax_pal=vmax_pal)
    band.append({'array':array, 'scale':scale, 'offset':offset,
                 'description':'current velocity', 'unittype':'m/s',
                 'nodatavalue':255, 'parameter_range':[vmin, vmax],
                 'colortable':colortable})

    array, offset, scale = pack.ubytes_0_254(curdir, 0.0, 360.0)
    if numpy.any(data_mask):
        array[numpy.where(data_mask)] = 255
    band.append({'array':array, 'scale':scale, 'offset': offset,
                 'description':'current direction', 'unittype':'deg',
                 'nodatavalue':255, 'parameter_range':[0, 360.]})

    # Write geotiff
    print 'Write geotiff'
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, band)
