# -*- encoding: utf-8 -*-

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-10-22
"""

import os
import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)

VARIABLES = {'sst': (('analysed_sst',), 271.05, 309.15, 'K',
                     'cerbere_medspiration'),
            'cool_diff': (('sea_surface_cool_skin_temperature_difference',),
                          -0.5, 0.0, 'K', 'matplotlib_Blues_r'),
            'warm_diff': (('sea_surface_warm_layer_temperature_difference',),
                          -0.04, 2.0, 'K', 'matplotlib_Reds')}


class UnsupportedVariables(ValueError):
    """Error triggered when a conversion is requested for one or several
    variables that are not supported by this reader."""

    def __init__(self, variables):
        self.variables = variables


class CannotGuessTimeRange(Exception):
    """Error triggered when the size of the time dimension is neither 1 not 24,
    i.e. no assumption can be made regarding the time range."""
    pass


def build_scalar_data(f_handler, ind_lat, ind_lon, t, vname, _min, _max, units,
                      cmap, _mask, vmin_pal=None, vmax_pal=None):
    """"""
    v_data = f_handler.variables[vname][t, ind_lat, ind_lon]
    v_data = numpy.ma.masked_invalid(v_data)
    mask = (_mask | numpy.ma.getmaskarray(v_data))
    masked_ind = numpy.where(mask)

    valid_ind = numpy.where(~numpy.ma.getmaskarray(v_data))
    logger.debug('{} min: {}'.format(vname, numpy.nanmin(v_data[valid_ind])))
    logger.debug('{} max: {}'.format(vname, numpy.nanmax(v_data[valid_ind])))

    if vmin_pal is None:
        vmin_pal = _min
    if vmax_pal is None:
        vmax_pal = _max

    colortable = stfmt.format_colortable(cmap,
                                         vmin=_min, vmax=_max,
                                         vmin_pal=vmin_pal, vmax_pal=vmax_pal)

    array, offset, scale = pack.ubytes_0_254(v_data, _min, _max)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data = []
    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': vname,
                 'unittype': units,
                 'nodatavalue': 255,
                 'parameter_range': [_min, _max],
                 'colortable': colortable})
    return data


def read_from_file(input_path, process_var):
    """"""
    granule_filename = os.path.basename(input_path)
    granule_prefix, _ = os.path.splitext(granule_filename)

    if 'all' == process_var:
        process_var = VARIABLES.keys()

    unsupported_vars = [x for x in process_var if x not in VARIABLES.keys()]
    if 0 < len(unsupported_vars):
        raise UnsupportedVariables(unsupported_vars)

    f_handler = netCDF4.Dataset(input_path, 'r')
    lon = f_handler.variables['lon'][:]
    lat = f_handler.variables['lat'][:]
    _t = f_handler.variables['time'][:]
    _t_units = f_handler.variables['time'].units

    mask_bits = {'water': 0,
                 'land': 1,
                 'optional_lake_surface': 2,
                 'sea_ice': 3,
                 'optional_river_surface': 4}

    ok_flags = numpy.int8(0)
    ok_flags = ok_flags + numpy.int8(1 << mask_bits['water'])

    nok_flags = numpy.int8(0)
    nok_flags = nok_flags + numpy.int8(1 << mask_bits['sea_ice'])

    ind_lon = numpy.where(numpy.isfinite(lon))[0]

    # Web mercator projection stops at 85.xxx in latitude, so crop to avoid
    # processing unused data
    ind_lat = numpy.where(numpy.abs(lat) < 86)[0]
    lat = lat[ind_lat]

    lon0, dlon = lon[0], numpy.mean(lon[1:] - lon[:-1])
    lat0, dlat = lat[0], numpy.mean(lat[1:] - lat[:-1])
    geo = {'projection': stfmt.format_gdalprojection(),
           'geotransform': [lon0 - 0.5 * dlon, dlon, 0,
                            lat0 - 0.5 * dlat, 0, dlat]}
    now = datetime.datetime.utcnow()

    meta = {'product_name': None,  # Set later
            'name': None,  # Set later
            'datetime': None, # Set later
            'time_range': None, # Set later
            'source_URI': input_path,
            'source_provider': 'WEkEO - Copernicus',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_software': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'parameter': None} # Set later

    dtimes = netCDF4.num2date(_t, _t_units)

    time_start = datetime.datetime.strptime(f_handler.time_coverage_start,
                                            '%Y%m%dT%H%M%SZ')
    time_stop = datetime.datetime.strptime(f_handler.time_coverage_end,
                                           '%Y%m%dT%H%M%SZ')

    for variable_name in process_var:
        fields, _min, _max, units, cmap = VARIABLES[variable_name]

        meta['product_name'] = 'OSTIA_{}'.format(variable_name)
        meta['parameter'] = variable_name

        for t, dtime in enumerate(dtimes):
            granule_name = '{}_{}_{}'.format(granule_prefix,
                                             dtime.strftime('%Y%m%d%H%M%S'),
                                             variable_name)
            meta['name'] = granule_name
            if 1 == len(dtimes):
                # Assume daily product
                _dtime, time_range = stfmt.format_time_and_range(time_start,
                                                                 time_stop,
                                                                 units='s')
                meta['datetime'] = _dtime
                meta['time_range'] = time_range
            elif 24 == len(dtimes):
                # Assume hourly product
                meta['datetime'] = stfmt.format_time(dtime)
                meta['time_range'] = ['-30m', '+30m']
            else:
                raise CannotGuessTimeRange

            tmp_mask = f_handler.variables['mask'][t, ind_lat, ind_lon]
            mask = ( (numpy.bitwise_and(tmp_mask, ok_flags) <= 0)
                   | (numpy.bitwise_and(tmp_mask, nok_flags) > 0))
            vmin_pal = None
            vmax_pal = None
            if 'sst' == variable_name:
                vmin_pal = 273
                vmax_pal = 305
            data = build_scalar_data(f_handler, ind_lat, ind_lon, t, fields[0],
                                     _min, _max, units, cmap, mask, vmin_pal,
                                     vmax_pal)

            yield (meta, geo, data, None)
    f_handler.close()


def convert_from_file(input_path, output_path, process_var='all'):
    """"""
    for (meta, geo, data, extra) in read_from_file(input_path, process_var):
        tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                               create_dir=True)
        stfmt.write_geotiff(tiff_fname, meta, geo, data)


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.DEBUG)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)
