# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import netCDF4
import datetime
import syntool_converter.utils.syntoolformat as stfmt

SUPPORTED_VERSIONS = ('v05.0','v05.1',)


class ProductVersionNotSupported(Exception):
    """"""
    def __init__(self, product_version, *args, **kwargs):
        self.product_version = product_version
        super(ProductVersionNotSupported, self).__init__(*args, **kwargs)


def remss_l4_mw(infile, outdir,
                vmin=271.05, vmax=309.15, vmin_pal=273., vmax_pal=305.):
    """
    """
    granule_fname = os.path.basename(infile)
    granule_name, _ = os.path.splitext(granule_fname)

    # Read/Process data
    f_handler = netCDF4.Dataset(infile, 'r')
    product_version = f_handler.product_version
    if product_version not in SUPPORTED_VERSIONS:
        f_handler.close()
        raise ProductVersionNotSupported(product_version)

    sst = f_handler.variables['analysed_sst'][0, ::-1, :]
    mask = f_handler.variables['mask'][0, ::-1, :]
    dtime = f_handler.variables['time'][0]
    dtime_units = f_handler.variables['time'].units
    dtime = netCDF4.num2date(dtime, dtime_units)
    dtime_str = stfmt.format_time(dtime)
    lon0 = f_handler.westernmost_longitude
    dlon = f_handler.geospatial_lon_resolution
    lat0 = f_handler.northernmost_latitude
    dlat = -1.0 * f_handler.geospatial_lat_resolution
    f_handler.close()

    now = datetime.datetime.utcnow()
    now_str = stfmt.format_time(now)
    # Construct metadata/geolocation/band(s)
    metadata = {'product_name': 'REMSS_MWOI_SST_{}'.format(product_version),
                'name': granule_name,
                'datetime': dtime_str,
                'time_range': ['-12h', '+12h'],
                'source_URI': infile,
                'source_provider': 'REMSS',
                'processing_center': '',
                'conversion_software': 'Syntool',
                'conversion_version': '0.0.0',
                'conversion_datetime': now_str,
                'parameter': 'sea surface temperature',
                'type': 'remote sensing',
                'longitude_resolution': abs(dlon),
                'latitude_resolution': abs(dlat)}

    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['geotransform'] = (lon0, dlon, 0,
                                   lat0, 0, dlat)

    mask_bits = {'water': 0,
                 'land': 1,
                 'optional_lake_surface': 2,
                 'sea_ice': 3,
                 'optional_river_surface': 4,
                 'IR_data': 5,
                 'MW_data': 6}

    # Invalidate if these flags are not on
    on_flags = numpy.byte(0)
    on_flags = on_flags + numpy.byte(1 << mask_bits['water'])
    on_flags = on_flags + numpy.byte(1 << mask_bits['MW_data'])

    data = []
    data_mask = ((sst.mask == True) |
                 (numpy.bitwise_and(mask, on_flags) != on_flags))

    offset, scale = vmin, (vmax - vmin) / 254.0
    numpy.clip(sst, vmin, vmax, out=sst)
    array = numpy.round((sst - offset) / scale).astype('uint8')
    array[numpy.where(data_mask == True)] = 255
    colortable = stfmt.format_colortable('cerbere_medspiration',
                                         vmax=vmax, vmax_pal=vmax_pal,
                                         vmin=vmin, vmin_pal=vmin_pal)
    data.append({'array': array,
                 'scale':scale,
                 'offset':offset,
                 'description':'sea surface temperature',
                 'unittype':'K',
                 'nodatavalue': 255,
                 'parameter_range': (vmin, vmax),
                 'colortable':colortable})

    # Write geotiff
    tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
    stfmt.write_geotiff(tifffile, metadata, geolocation, data)
