# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-10-22

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


class FileNotSupported(Exception):
    """Raised when the input file is or might not contain the required
    information. The "title" global attribute of CMEMS_008_046 and
    CMEMS_008_047 files should begin by "NRT" and "DT" respectively."""
    pass


def build_scalar_data(f_handler, ind_lat, ind_lon, t, vname, _min, _max, units,
                      cmap):
    """"""
    v_data = f_handler.variables[vname][t, ind_lat, ind_lon]
    v_data = numpy.ma.masked_invalid(v_data)
    mask = numpy.ma.getmaskarray(v_data)
    masked_ind = numpy.where(mask)

    # Remove mask and set dummy value to prevent issues with numpy
    # It has no side-effect as mask is saved and reapplied later
    v_data._sharedmask = False
    v_data.mask = False
    v_data[masked_ind] = 0

    valid_ind = numpy.where(~numpy.ma.getmaskarray(v_data))
    logger.debug('{} min: {}'.format(vname, numpy.nanmin(v_data[valid_ind])))
    logger.debug('{} max: {}'.format(vname, numpy.nanmax(v_data[valid_ind])))

    colortable = stfmt.format_colortable(cmap,
                                         vmin=_min, vmax=_max,
                                         vmin_pal=_min, vmax_pal=_max)

    array, offset, scale = pack.ubytes_0_254(v_data, _min, _max)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data = []
    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': vname,
                 'unit': units,
                 'nodatavalue': 255,
                 'parameter_range': [_min, _max],
                 'colortable': colortable})
    return data


def build_directional_data(f_handler, ind_lat, ind_lon, t, vnames, _min, _max,
                           units, cmap):
    """"""
    u_data = f_handler.variables[vnames[0]][t, ind_lat, ind_lon]
    u_data = numpy.ma.masked_invalid(u_data)

    v_data = f_handler.variables[vnames[1]][t, ind_lat, ind_lon]
    v_data = numpy.ma.masked_invalid(v_data)

    mask = (numpy.ma.getmaskarray(u_data) | numpy.ma.getmaskarray(v_data))
    masked_ind = numpy.where(mask)

    u_data[masked_ind] = 0
    v_data[masked_ind] = 0

    norm = numpy.hypot(u_data, v_data)
    _direction = numpy.rad2deg(numpy.arctan2(v_data, u_data))
    direction = numpy.mod(_direction, 360.0)

    valid_ind = numpy.where(~mask)
    logger.debug('{} min: {}'.format(vnames, numpy.nanmin(norm[valid_ind])))
    logger.debug('{} max: {}'.format(vnames, numpy.nanmax(norm[valid_ind])))
    colortable = stfmt.format_colortable(cmap,
                                         vmin=_min, vmax=_max,
                                         vmin_pal=_min, vmax_pal=_max)

    data = []

    array, offset, scale = pack.ubytes_0_254(norm, _min, _max)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'amplitude',
                 'unit': units,
                 'nodatavalue': 255,
                 'parameter_range': [_min, _max],
                 'colortable': colortable})

    array, offset, scale = pack.ubytes_0_254(direction, 0, 360)
    if numpy.any(masked_ind):
        array[masked_ind] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'direction',
                 'unit': 'degrees',
                 'nodatavalue': 255,
                 'parameter_range': [0, 360]})
    return data


def read_from_file(input_path, variables):
    """"""
    f_handler = netCDF4.Dataset(input_path, 'r')
    lon = f_handler.variables['longitude'][:]
    lat = f_handler.variables['latitude'][:]
    _t = f_handler.variables['time'][:]
    _t_units = f_handler.variables['time'].units

    # Make sure longitudes are expressed in [-180, 180] range
    lon[lon > 180.0] = lon[lon > 180.0] - 360.0
    ind_lon = numpy.argsort(lon)
    lon = lon[ind_lon]

    ind_lat = numpy.where(numpy.abs(lat) < 86)[0]
    lat = lat[ind_lat]

    lon0, dlon = lon[0], numpy.mean(lon[1:] - lon[:-1])
    lat0, dlat = lat[0], numpy.mean(lat[1:] - lat[:-1])
    geo = {'projection': stfmt.format_gdalprojection(),
           'geotransform': [lon0 - 0.5 * dlon, dlon, 0,
                            lat0 - 0.5 * dlat, 0, dlat]}

    now = datetime.datetime.utcnow()
    dtimes = netCDF4.num2date(_t, _t_units)

    granule_fname = os.path.basename(input_path)
    _granule_name, _ = os.path.splitext(granule_fname)

    meta = {'product_name': None,  # Set later
            'name': None,  # Set later
            'datetime': None,  # Set later
            'time_range': None,  # Set later
            'source_URI': input_path,
            'source_provider': 'CMEMS',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_version': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'parameter': None}  # Set later

    if f_handler.title.startswith('NRT'):
        product_name = 'CMEMS_008_046'
    elif f_handler.title.startswith('DT'):
        product_name = 'CMEMS_008_047'
    else:
        raise FileNotSupported()

    for variable_name in variables:
        fields, _min, _max, units, cmap = variables[variable_name]

        meta['product_name'] = '{}_{}'.format(product_name, variable_name)
        meta['parameter'] = variable_name

        for t, dtime in enumerate(dtimes):
            granule_name = '{}_{}'.format(_granule_name, variable_name,
                                          dtime.strftime('%Y%m%d%H%M%S'))
            meta['name'] = granule_name
            meta['datetime'] = stfmt.format_time(dtime)
            meta['time_range'] = ['0s', '+86399s']  # daily

            if 1 == len(fields):
                data = build_scalar_data(f_handler, ind_lat, ind_lon, t,
                                         fields[0], _min, _max, units, cmap)
            elif 2 == len(fields):
                data = build_directional_data(f_handler, ind_lat, ind_lon, t,
                                              fields, _min, _max, units, cmap)

            yield (meta, geo, data, None)
    f_handler.close()


def convert_from_file(input_path, output_path, process_var='all', vmin=None,
                      vmax=None):
    """"""
    variables = {'msla': [('sla',), -0.4, 0.4, 'm', 'matplotlib_jet'],
                 'madt': [('adt',), -1.7, 1.9, 'm', 'matplotlib_jet'],
                 'madt_current': [('ugos', 'vgos',), 0.0, 2.0, 'm.s-1',
                                  'matplotlib_jet'],
                 'msla_current': [('ugosa', 'vgosa',), 0.0, 1.0, 'm.s-1',
                                  'matplotlib_jet']}
    if 'all' != process_var:
        variables = {k: v for k, v in variables.iteritems()
                     if k in process_var}
        if vmin is not None:
            if not isinstance(vmin, list):
                vmin = [vmin]
            for k, varname in enumerate(process_var):
                variables[varname][1] = vmin[k]

        if vmax is not None:
            if not isinstance(vmax, list):
                vmax = [vmax]
            for k, varname in enumerate(process_var):
                variables[varname][2] = vmax[k]

    for (meta, geo, data, extra) in read_from_file(input_path, variables):
        tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                               create_dir=True)
        stfmt.write_geotiff(tiff_fname, meta, geo, data)


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARN)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)
