# coding=utf-8

"""
@author <sylvain.herledan@oceandatalab.com>
@date 2019-02-04

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt

numpy.seterr(all='raise')
logger = logging.getLogger(__name__)


def read_from_file(f_handler):
    """"""
    dlat = f_handler.geospatial_lat_resolution
    dlon = f_handler.geospatial_lon_resolution
    time_units = f_handler.variables['time'].units
    lon = f_handler.variables['lon'][:]
    lat = f_handler.variables['lat'][:]
    time = f_handler.variables['time'][0]
    fsle_max = f_handler.variables['fsle_max'][0, :, :]

    # Longitudes are defined in [0, 360], shift them to [-180, 180]
    lon_above180_index = numpy.where(lon > 180.0)
    lon[lon_above180_index] = lon[lon_above180_index] - 360.0
    ascending_lon_index = numpy.argsort(lon)
    lon = lon[ascending_lon_index]
    fsle_max = fsle_max[:, ascending_lon_index]

    lon0 = numpy.round(10000 * lon[0]) / 10000
    lat0 = numpy.round(10000 * lat[0]) / 10000

    dtime = netCDF4.num2date(time, time_units)
    dtime_str = stfmt.format_time(dtime)

    now = datetime.datetime.utcnow()
    meta = {'product_name': 'SSALTO_FSLE',
            'name': None,
            'datetime': dtime_str,
            'time_range': ['-43200s', '+43199s'],
            'source_URI': None,
            'source_provider': 'aviso@atimetry.fr',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_version': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'spatial_resolution': '4452',  # at equator: 111.321 km/° * 1/25°
            'parameter': 'Finite Size Lyapunov Exponents'}

    geolocation = {'projection': stfmt.format_gdalprojection(),
                   'geotransform': [lon0 - 0.5 * dlon, dlon, 0,
                                    lat0 - 0.5 * dlat, 0, dlat]}

    vmin = -1
    vmax = 0
    vmin_pal = -1
    vmax_pal = 0
    colortable = stfmt.format_colortable('matplotlib_jet_r',
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin_pal, vmax_pal=vmax_pal)

    mask = numpy.ma.getmaskarray(fsle_max)
    array, offset, scale = pack.ubytes_0_254(fsle_max, vmin, vmax)
    if numpy.any(mask):
        array[numpy.where(mask)] = 255

    data = []
    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'FSLE',
                 'name': 'fsle_max',
                 'unittype': 'day-1',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    yield (meta, geolocation, data, None)


def convert_from_file(input_path, output_path):
    """"""
    granule_filename = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_filename)
    f_handler = netCDF4.Dataset(input_path, 'r')
    for (meta, geolocation, data, extra) in read_from_file(f_handler):
        meta['name'] = granule_name
        meta['source_URI'] = input_path
        tifffile = stfmt.format_tifffilename(output_path, meta,
                                             create_dir=True)
        stfmt.write_geotiff(tifffile, meta, geolocation, data)
    f_handler.close()


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARNING)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)
