# coding=utf-8

"""
@author <sylvain.herledan@oceandatalab.com>
@date 2019-02-04

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# Note: this reader has been written for experimental products available at:
# ftp-access.aviso.altimetry.fr:/duacs-experimental/dt-phy-grids/altimetry_drifters/
# ftp-access.aviso.altimetry.fr:/duacs-experimental/dt-phy-grids/altimetry_sst/
# ftp-access.aviso.altimetry.fr:/duacs-experimental/dt-phy-grids/dynamic_interpolation/

import os
import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt

logger = logging.getLogger(__name__)


def _process_directional(u, v, vmin, vmax, var_name):
    """"""
    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin, vmax_pal=vmax)

    data = []
    mask = (numpy.ma.getmaskarray(u) | numpy.ma.getmaskarray(v))
    valid_ind = numpy.where(~mask)
    module = u.copy()
    module[valid_ind] = numpy.sqrt(u[valid_ind] ** 2 + v[valid_ind] ** 2)
    direction = numpy.ma.array(numpy.ndarray(shape=numpy.shape(u),
                                             dtype=numpy.float32))
    direction[valid_ind] = numpy.arctan2(v[valid_ind], u[valid_ind])
    direction[valid_ind] = numpy.mod(numpy.rad2deg(direction[valid_ind]),
                                     360.0)

    logger.debug('Max {}: {}'.format(var_name, numpy.max(module)))
    module.mask = mask
    array, offset, scale = pack.ubytes_0_254(module, vmin, vmax)
    if numpy.any(mask):
        array[numpy.where(mask)] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': '',
                 'name': var_name,
                 'unittype': 'm.s-1',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    vmin = 0.0
    vmax = 360.0
    direction.mask = mask
    array, offset, scale = pack.ubytes_0_254(direction, vmin, vmax)
    if numpy.any(mask):
        array[numpy.where(mask)] = 255
    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': '',
                 'name': '{}_direction'.format(var_name),
                 'unittype': 'degree',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax]})

    extra = {'variable': var_name}

    return data, extra


def _process_scalar(d, vmin, vmax, var_name):
    """"""
    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin, vmax_pal=vmax)

    data = []
    mask = numpy.ma.getmaskarray(d)

    logger.debug('Min {}: {}'.format(var_name, numpy.min(d)))
    logger.debug('Max {}: {}'.format(var_name, numpy.max(d)))
    array, offset, scale = pack.ubytes_0_254(d, vmin, vmax)
    if numpy.any(mask):
        array[numpy.where(mask)] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': '',
                 'name': 'adt',
                 'unittype': 'm',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    extra = {'variable': var_name}

    return data, extra


def read_from_file(f_handler):
    """"""
    dlat = f_handler.geospatial_lat_resolution
    dlon = f_handler.geospatial_lon_resolution
    time_units = f_handler.variables['time'].units
    lon = f_handler.variables['longitude'][:]
    lat = f_handler.variables['latitude'][:]
    time = f_handler.variables['time'][0]
    title = f_handler.title

    # Longitudes are defined in [0, 360], shift them to [-180, 180]
    lon_above180_index = numpy.where(lon > 180.0)
    lon[lon_above180_index] = lon[lon_above180_index] - 360.0
    ascending_lon_index = numpy.argsort(lon)
    lon = lon[ascending_lon_index]

    has_gosa = ('ugosa' in f_handler.variables and
                'vgosa' in f_handler.variables)
    has_gos = ('ugos' in f_handler.variables and
               'vgos' in f_handler.variables)
    has_sla = 'sla' in f_handler.variables
    has_adt = 'adt' in f_handler.variables
    xtot = 'eastward_eulerian_current_velocity'
    ytot = 'northward_eulerian_current_velocity'
    has_total = (xtot in f_handler.variables and ytot in f_handler.variables)

    if has_gosa:
        x_geostr_anomaly = f_handler.variables['ugosa'][0, :, :]
        y_geostr_anomaly = f_handler.variables['vgosa'][0, :, :]
        x_geostr_anomaly = x_geostr_anomaly[:, ascending_lon_index]
        y_geostr_anomaly = y_geostr_anomaly[:, ascending_lon_index]
        if 'drifters Gulf of Mexico' in title:
            x_geostr_anomaly = x_geostr_anomaly / 100.
            y_geostr_anomaly = y_geostr_anomaly / 100.
    if has_gos:
        x_geostr_velocity = f_handler.variables['ugos'][0, :, :]
        y_geostr_velocity = f_handler.variables['vgos'][0, :, :]
        x_geostr_velocity = x_geostr_velocity[:, ascending_lon_index]
        y_geostr_velocity = y_geostr_velocity[:, ascending_lon_index]
    if has_sla:
        sla = f_handler.variables['sla'][0, :, :]
        sla = sla[:, ascending_lon_index]
        # 2br : correct for wrong unit in combined drifter product
        if 'drifters Gulf of Mexico' in title:
            sla = sla / 100.
    if has_adt:
        adt = f_handler.variables['adt'][0, :, :]
        adt = adt[:, ascending_lon_index]

    if has_total:
        x_eulerian_velocity = f_handler.variables[xtot][0, :, :]
        y_eulerian_velocity = f_handler.variables[ytot][0, :, :]
        x_eulerian_velocity = x_eulerian_velocity[:, ascending_lon_index]
        y_eulerian_velocity = y_eulerian_velocity[:, ascending_lon_index]

    lon0 = numpy.round(10000 * lon[0]) / 10000
    lat0 = numpy.round(10000 * lat[0]) / 10000

    dtime = netCDF4.num2date(time, time_units)
    dtime_str = stfmt.format_time(dtime)

    now = datetime.datetime.utcnow()
    meta = {'product_name': None,
            'name': None,
            'datetime': dtime_str,
            'time_range': ['+0s', '+86399s'],
            'source_URI': None,
            'source_provider': 'aviso@atimetry.fr',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_version': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'spatial_resolution': '0',
            'parameter': None}

    geolocation = {'projection': stfmt.format_gdalprojection(),
                   'geotransform': [lon0 - 0.5 * dlon, dlon, 0,
                                    lat0 - 0.5 * dlat, 0, dlat]}

    if has_adt:
        # ADT
        # ---------------------------------------------------------------------
        vmin = -1.7
        vmax = 1.9
        var_name = 'adt'

        data, extra = _process_scalar(adt, vmin, vmax, var_name)

        yield (meta, geolocation, data, extra)

    if has_sla:
        # SLA
        # ---------------------------------------------------------------------
        vmin = -0.4
        vmax = 0.4
        var_name = 'sla'

        data, extra = _process_scalar(sla, vmin, vmax, var_name)

        yield (meta, geolocation, data, extra)

    if has_gosa:
        # Geostrophic anomaly
        # ---------------------------------------------------------------------
        vmin = 0.0
        vmax = 1.0
        var_name = 'geostrophic_anomaly'
#
        data, extra = _process_directional(x_geostr_anomaly, y_geostr_anomaly,
                                           vmin, vmax, var_name)
#
        yield (meta, geolocation, data, extra)

    if has_gos:
        # Geostrophic velocity
        # ---------------------------------------------------------------------
        vmin = 0.0
        vmax = 2.0
        var_name = 'geos_velocity'

        data, extra = _process_directional(x_geostr_velocity,
                                           y_geostr_velocity,
                                           vmin, vmax, var_name)

        yield (meta, geolocation, data, extra)

    if has_total:
        # Total velocity
        # ---------------------------------------------------------------------
        vmin = 0.0
        vmax = 2.0
        var_name = 'eul_velocity'

        data, extra = _process_directional(x_eulerian_velocity,
                                           y_eulerian_velocity,
                                           vmin, vmax, var_name)

        yield (meta, geolocation, data, extra)


def convert_from_file(input_path, output_path, product_type):
    """"""
    granule_filename = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_filename)
    f_handler = netCDF4.Dataset(input_path, 'r')
    for (meta, geolocation, data, extra) in read_from_file(f_handler):
        meta['product_name'] = '{}_{}'.format(product_type, extra['variable'])
        meta['name'] = granule_name
        meta['source_URI'] = input_path
        meta['parameter'] = extra['variable']
        tifffile = stfmt.format_tifffilename(output_path, meta,
                                             create_dir=True)
        stfmt.write_geotiff(tifffile, meta, geolocation, data)
    f_handler.close()


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARNING)

    numpy.seterr(all='raise')

    product_type = sys.argv[1]
    input_path = sys.argv[2]
    output_path = sys.argv[3]

    try:
        convert_from_file(input_path, output_path, product_type)
    except:
        logger.error('Failed to process {}'.format(input_path),
                     exc_info=sys.exc_info())
        sys.exit(1)
