# coding=utf-8

"""
@author <sylvain.herledan@oceandatalab.com>
@date 2019-02-04

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import netCDF4
import logging
import datetime
from scipy import interpolate
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt

numpy.seterr(all='raise')
logger = logging.getLogger(__name__)


class UnknownParameter(Exception):
    """"""
    def __init__(self, param, *args, **kwargs):
        """"""
        self.parameter = param


def read_from_file(f_handler, m_handler):
    """"""
    lon = f_handler.variables['NbLongitudes'][:]
    lat = f_handler.variables['NbLatitudes'][:]
    values = f_handler.variables['Grid_0001'][:, :]

    # Longitudes are defined in [0, 360], shift them to [-180, 180]
    lon_above180_index = numpy.where(lon > 180.0)
    lon[lon_above180_index] = lon[lon_above180_index] - 360.0
    ascending_lon_index = numpy.argsort(lon)
    lon = lon[ascending_lon_index]
    values = values[ascending_lon_index, :].transpose()  # GDAL expects lat/lon
    print(values)
    if m_handler is not None:
        mask, mlon, mlat = read_mask(m_handler)
        dlon = numpy.mean(mlon[1:] - mlon[:-1])
        dlat = numpy.mean(mlat[1:] - mlat[:-1])
        lon0 = numpy.round(10000 * mlon[0]) / 10000 - 0.5 * dlon
        lat0 = numpy.round(10000 * mlat[0]) / 10000 - 0.5 * dlat
    else:
        dlon = numpy.mean(lon[1:] - lon[:-1])
        dlat = numpy.mean(lat[1:] - lat[:-1])
        lon0 = numpy.round(10000 * lon[0]) / 10000
        lat0 = numpy.round(10000 * lat[0]) / 10000

    original_name = f_handler.OriginalName
    _, _, vname, date_str, _ = original_name.split('_', 4)
    vname = str(vname)
    print(vname, date_str)

    if vname not in ('mswh', 'mwind',):
        raise UnknownParameter(vname)

    dtime = datetime.datetime.strptime(date_str, '%Y%m%d')
    dtime_str = stfmt.format_time(dtime)

    now = datetime.datetime.utcnow()
    meta = {'product_name': 'SSALTO_NRT_{}'.format(vname),
            'name': None,
            'datetime': dtime_str,
            'time_range': ['0s', '+86399s'],  # daily
            'source_URI': None,
            'source_provider': 'aviso@atimetry.fr',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_version': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'spatial_resolution': '111321',  # at equator: 111.321 km/°
            'parameter': vname}

    geolocation = {'projection': stfmt.format_gdalprojection(),
                   'geotransform': [lon0, dlon, 0,
                                    lat0, 0, dlat]}

    vmin = None
    vmax = None
    units = None
    colortable = None
    if 'mwind' == vname:
        vmin = 0.0
        vmax = 25.4
        units = 'm.s-1'
        # Same settings as ECMWF product processed automatically in OVL
        colortable = stfmt.format_colortable('noaa_wind',
                                             vmin=vmin, vmax=vmax,
                                             vmin_pal=0.0, vmax_pal=50 * 0.514)
    elif 'mswh' == vname:
        vmin = 0.0
        vmax = 8.0
        units = 'm'
        # Same settings as WW3 product processed automatically in OVL
        colortable = stfmt.format_colortable('matplotlib_jet',
                                             vmin=vmin, vmax=vmax,
                                             vmin_pal=vmin, vmax_pal=vmax)


    values = numpy.ma.masked_invalid(values)
    mask_lr = numpy.ma.getmaskarray(values)
    if m_handler is not None:
        _val = interpolate.RectBivariateSpline(lat, lon, values, kx=1, ky=1, s=0)
        mlon, mlat = numpy.meshgrid(mlon, mlat)
        values = _val.ev(mlat, mlon)
        _teval = interpolate.RectBivariateSpline(lat, lon, mask_lr, kx=1, ky=1, s=0)
        mask_lr2 = _teval.ev(mlat, mlon)
    else:
        mask_lr2 = mask_lr
        mask = mask_lr
    array, offset, scale = pack.ubytes_0_254(values, vmin, vmax)
    if numpy.any(mask):
        array[numpy.where(mask)] = 255
    if numpy.any(mask_lr2):
        array[numpy.where(mask_lr2)] = 255
    data = []
    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': vname,
                 'name': vname,
                 'unittype': units,
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    yield (meta, geolocation, data, None)


def read_mask(m_handler):
    mlon = m_handler.variables['longitude'][:] #-0.25
    mlat = m_handler.variables['latitude'][:] #-0.25
    values = m_handler.variables['sla'][0, :, :]
    values = numpy.ma.masked_invalid(values)
    mask = numpy.ma.getmaskarray(values)
    # Longitudes are defined in [0, 360], shift them to [-180, 180]
    lon_above180_index = numpy.where(mlon > 180.0)
    mlon[lon_above180_index] = mlon[lon_above180_index] - 360.0
    ascending_lon_index = numpy.argsort(mlon)
    mlon = mlon[ascending_lon_index]
    mask = mask[:, ascending_lon_index]  # GDAL expects lat/lon
    return mask, mlon, mlat


def convert_from_file(input_path, output_path, landmaskpath=None):
    """"""
    granule_filename = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_filename)
    f_handler = netCDF4.Dataset(input_path, 'r')
    if landmaskpath is not None:
        m_handler = netCDF4.Dataset(landmaskpath, 'r')
    else:
        m_handler = None
    for (meta, geolocation, data, extra) in read_from_file(f_handler, m_handler):
        meta['name'] = granule_name
        meta['source_URI'] = input_path
        tifffile = stfmt.format_tifffilename(output_path, meta,
                                             create_dir=True)
        stfmt.write_geotiff(tifffile, meta, geolocation, data)
    f_handler.close()
    if m_handler is not None:
        m_handler.close()


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARNING)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)
