# -*- encoding: utf-8 -*-
# vim: ts=4:sts=4:sw=4:et

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-03-15

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy
import pyproj
import logging
import datetime

logger = logging.getLogger(__name__)
in_proj = pyproj.Proj(init='epsg:4326')
out_proj = pyproj.Proj(init='epsg:3857')


def _bbox_shape(xidl, western_area, eastern_area):
    """"""
    tlbr = [-20037508.34, -20037508.34, 20037508.34, 20037508.34]
    shape_wkt = None
    bbox_wkt = None
    w_bbox = None
    e_bbox = None

    if xidl:
        w_min_lon = tlbr[1]
        w_max_lon = numpy.max([p[0] for p in western_area])
        w_min_lat = numpy.min([p[1] for p in western_area])
        w_max_lat = numpy.max([p[1] for p in western_area])
        w_bbox = 'POLYGON(({} {},{} {},{} {},{} {},{} {}))'.format(
                                                    w_min_lon, w_max_lat,
                                                    w_max_lon, w_max_lat,
                                                    w_max_lon, w_min_lat,
                                                    w_min_lon, w_min_lat,
                                                    w_min_lon, w_max_lat)

        e_min_lon = numpy.min([p[0] for p in eastern_area])
        e_max_lon = tlbr[3]
        e_min_lat = numpy.min([p[1] for p in eastern_area])
        e_max_lat = numpy.max([p[1] for p in eastern_area])
        e_bbox = 'POLYGON(({} {},{} {},{} {},{} {},{} {}))'.format(
                                                    e_min_lon, e_max_lat,
                                                    e_max_lon, e_max_lat,
                                                    e_max_lon, e_min_lat,
                                                    e_min_lon, e_min_lat,
                                                    e_min_lon, e_max_lat)
        shape_wkt = 'GEOMETRYCOLLECTION ({},{})'.format(w_bbox, e_bbox)
        min_lon = w_min_lon
        max_lon = e_max_lon
        min_lat = numpy.min([w_min_lat, e_min_lat])
        max_lat = numpy.max([w_max_lat, w_max_lat])
        bbox_wkt = 'POLYGON(({} {},{} {},{} {},{} {},{} {}))'.format(
                                                    min_lon, max_lat,
                                                    max_lon, max_lat,
                                                    max_lon, min_lat,
                                                    min_lon, min_lat,
                                                    min_lon, max_lat)
    else:
        western_area.extend(eastern_area)
        min_lon = numpy.min([p[0] for p in western_area])
        max_lon = numpy.max([p[0] for p in western_area])
        min_lat = numpy.min([p[1] for p in western_area])
        max_lat = numpy.max([p[1] for p in western_area])
        shape_wkt = 'POLYGON(({} {},{} {},{} {},{} {},{} {}))'.format(
                                                    min_lon, max_lat,
                                                    max_lon, max_lat,
                                                    max_lon, min_lat,
                                                    min_lon, min_lat,
                                                    min_lon, max_lat)
        bbox_wkt = shape_wkt

    return shape_wkt, bbox_wkt, w_bbox, e_bbox


def handle_dateline(p1, p2):
    """ """
    tlbr = [-20037508.34, -20037508.34, 20037508.34, 20037508.34]
    margin = 1
    extent = tlbr[3] - tlbr[1]
    # If one point is in the upper quarter of the map
    # and the other is in the lower one, consider that the dateline is
    # crossed
    x1, y1 = p1
    x2, y2 = p2
    if abs(x1 - x2) > 0.5 * extent:
        if 0 < x1:
            p = (y2 - y1) / (x2 - (x1 - extent))
            yi = y1 + (tlbr[3] - x1) * p
            right_limit = (tlbr[3] - margin, yi)
            left_limit = (tlbr[1] + margin, yi)
            return (p1, right_limit, left_limit, p2)
        else:
            p = (y2 - y1) / (x2 - (x1 + extent))
            yi = y1 + (tlbr[1] - x1) * p
            left_limit = (tlbr[1] + margin, yi)
            right_limit = (tlbr[3] - 1, yi)
            return (p1, left_limit, right_limit, p2)
    return (p1, p2)


def get_continuous_slices(points):
    """"""
    points_count = len(points)
    slices = []
    start = None
    for i in range(0, points_count):
        if None in (points[i][0], points[i][1]):
            if start is not None:
                _slice = slice(start, i)
                slices.append(_slice)
            start = None
        elif start is None:
            start = i
    if start is not None:
        _slice = slice(start, None)
        slices.append(_slice)
    return slices


def create_geojson_feature(times, points, fields):
    """"""
    geometry = {'coordinates': points, 'type': 'LineString'}
    feature = {'geometry': geometry, 'properties': fields, 'type': 'Feature'}
    return feature


def create_features(times, points, obs, obs_id, points_start, points_count):
    """"""
    feature_points = points[points_start:(1 + points_start + points_count)]
    lon, lat = zip(*feature_points)
    x, y = pyproj.transform(in_proj, out_proj, lon, lat)
    bbox = (numpy.min(x), numpy.max(x), numpy.min(y), numpy.max(y))

    result = []
    idl_crossed = False
    for i in range(0, points_count - 1):
        props = {'fields': {}}
        for field_name in obs:
            if field_name in ('lon', 'lat', 'date'):
                props[field_name] = obs[field_name][obs_id]
            else:
                props['fields'][field_name] = obs[field_name][obs_id]

        origin = (x[i], y[i])
        target = (x[i + 1], y[i + 1])
        _points = handle_dateline(origin, target)

        if 2 == len(_points):
            # No IDL issue
            _feat = create_geojson_feature(times[obs_id], _points, props)
            result.append(_feat)
        elif 4 == len(_points):
            # IDL has been crossed and feature must be split to avoid issues
            _feat = create_geojson_feature(times[obs_id], _points[0:2], props)
            result.append(_feat)
            _feat = create_geojson_feature(times[obs_id], _points[2:4], props)
            result.append(_feat)
            idl_crossed = True
    return result, bbox, idl_crossed


def update_bbox(ref, added):
    """"""
    x_min = min(ref[0], added[0])
    x_max = max(ref[1], added[1])
    y_min = min(ref[2], added[2])
    y_max = max(ref[3], added[3])
    return (x_min, x_max, y_min, y_max)


def process_slice(times, locations, obs):
    """"""
    # Observations are considered valid between halfway from the previous
    # location and halfway to the next location.
    # => add middle points between original locations
    geod = pyproj.Geod(ellps='WGS84')
    count = len(locations)
    points = []
    for i in range(0, count - 1):
        cur_point = locations[i]
        next_point = locations[i + 1]
        points.append(cur_point)
        mid_point = geod.npts(cur_point[0], cur_point[1], next_point[0],
                              next_point[1], npts=1, radians=False)[0]
        points.append(mid_point)
    points.append(locations[-1])

    lons = [p[0] for p in points]
    lats = [p[1] for p in points]
    proj_x, proj_y = pyproj.transform(in_proj, out_proj, lons, lats)
    proj_points = zip(proj_x, proj_y)
    east = [p for p in proj_points if 0 <= p[0]]
    west = [p for p in proj_points if 0 > p[0]]

    # First and last observations have only two points while the others have
    # three
    features, bbox, idl = create_features(times, points, obs, 0, 0, 2)
    points_start = 1  # second point in 0-indexed array
    for i in range(1, count - 1):
        _features, _bbox, _idl = create_features(times, points, obs, i,
                                                 points_start, 3)
        features.extend(_features)
        idl = (idl or _idl)
        bbox = update_bbox(bbox, _bbox)
        points_start = points_start + 2
    _features, _bbox, _idl = create_features(times, points, obs, count - 1,
                                             points_start, 2)
    features.extend(_features)
    idl = (idl or _idl)
    bbox = update_bbox(bbox, _bbox)

    return (features, bbox, east, west, idl)


def process_data(times, locations, obs):
    """"""
    time_fmt = '%Y-%m-%d %H:%M:%S'
    bbox = (float('inf'), float('-inf'), float('inf'), float('-inf'))
    eastern_area = []
    western_area = []
    idl = False
    features = []
    slices = get_continuous_slices(locations)
    for current_slice in slices:
        slice_times = times[current_slice]
        slice_dt = [datetime.datetime.utcfromtimestamp(t) for t in slice_times]
        if 2 > len(slice_times):
            logger.warn('Skip slice with less than 2 values')
            continue
        slice_locations = locations[current_slice]
        slice_obs = {}
        for field_name in obs:
            slice_obs[field_name] = obs[field_name][current_slice]
        lon, lat = zip(*slice_locations)
        slice_obs['lon'] = lon
        slice_obs['lat'] = lat
        slice_obs['date'] = [dt.strftime(time_fmt) for dt in slice_dt]

        _features, _bbox, _east, _west, _idl = process_slice(slice_times,
                                                             slice_locations,
                                                             slice_obs)
        features.extend(_features)
        bbox = update_bbox(bbox, _bbox)
        eastern_area.extend(_east)
        western_area.extend(_west)
        idl = (idl or _idl)

    start_dt = datetime.datetime.utcfromtimestamp(times[0])
    stop_dt = datetime.datetime.utcfromtimestamp(times[-1])
    start_str = start_dt.strftime(time_fmt)
    stop_str = stop_dt.strftime(time_fmt)
    geojson_ = {'type': 'FeatureCollection',
                'features': features,
                'bbox': [bbox[0], bbox[2], bbox[1], bbox[3]],
                'crs': {'type': 'name',
                        'properties': {'name': 'urb:ogc:def:crs:EPSG:3857'}},
                'properties': {'timeRange': [start_str, stop_str]}}
    return (geojson_, bbox, eastern_area, western_area, idl)
