#!/usr/bin/env python
# coding=utf-8

"""
Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import logging
import netCDF4
import httplib
import datetime
from syntool_converter.utils.http import download_file

logger = logging.getLogger(__name__)


# OC product (http://oceancolor.gsfc.nasa.gov/cms/)
class VIIRSL2File(object):
    """ """
    def __init__(self, viirsl2fname):
        """ """
        self._dset = netCDF4.Dataset(viirsl2fname, 'r')

    def close(self):
        """ """
        self._dset.close()

    def read_chlora(self):
        """ """
        return self._dset.groups['geophysical_data'].variables['chlor_a'][:]

    def read_lon(self):
        """ """
        return self._dset.groups['navigation_data'].variables['longitude'][:]

    def read_lat(self):
        """ """
        return self._dset.groups['navigation_data'].variables['latitude'][:]

    def read_attributes(self):
        """ """
        attrs = {}
        attrs['start_time'] = datetime.datetime.strptime(self._dset.time_coverage_start,
                                                         '%Y-%m-%dT%H:%M:%S.%fZ')
        attrs['stop_time'] = datetime.datetime.strptime(self._dset.time_coverage_end,
                                                        '%Y-%m-%dT%H:%M:%S.%fZ')
        attrs['pass'] = self._dset.startDirection
        attrs['platform'] = self._dset.platform
        return attrs


def get_data(platform, dtime, param, conn=None, download_dir=None):
    """"""
    if param not in ('OC', 'SST', 'SST3', 'IOP', 'PICT'):
        raise ParameterNotSupported(param)

    file_name = 'V{}.L2_{}_{}.nc'.format(dtime.strftime('%Y%j%H%M%S'),
                                         platform.upper(), param)
    file_path = '/cgi/getfile/{}'.format(file_name)
    file_url = 'https://oceandata.sci.gsfc.nasa.gov{}'.format(file_path)

    if download_dir is None:
        return file_path, file_url, conn

    if conn is None:
        conn = httplib.HTTPSConnection('oceandata.sci.gsfc.nasa.gov',
                                       timeout=180)

    file_remote_path = file_path
    file_path = os.path.join(download_dir, file_name)
    if not os.path.exists(file_path):
        logger.info('Downloading {}'.format(file_url))
        raw_data = download_file(conn, file_remote_path, tries=10)
        with open(file_path, 'wb') as f:
            f.write(raw_data)

    return file_path, file_url, conn
