#!/usr/bin/env python
# coding=utf-8

"""
Copyright (C) 2014-2020 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import re
import sys
import math
import pyproj
import logging
import datetime
import netCDF4
import numpy
import numpy.ma as ma
from scipy.signal import find_peaks
import syntool_converter
import syntool_converter.utils.pack as pack
import syntool_converter.utils.viirs as viirs
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.resample as resample
from scipy.ndimage.morphology import binary_opening

logger = logging.getLogger(__name__)

## For old VIIRS L2 OC (ie. V*.L2_NPP_OC.nc)
# swath size : ~576.00km x ~3104.04km
# 100km ngcps=(7, 32)
## For new VIIRS L2 OC (ie. V*.L2_SNPP_OC.nc)
# swath size : ~2436.00km x ~3104.04km
# 100km ngcps=(26, 32)
# viirs_chlora('V2015062114540', '/mnt/data/syntool_inputs/synergy/',
#              download_dir='/mnt/data/viirs')

def flag_data(quality, mask):
    bits = {'Substitute_Cal': 0, # 1
            'Out_of_Range': 1, # 2
	    'Saturation': 2, # 4
	    'Temp_not_Nominal': 3, # 16
	    'Bowtie_Deleted': 8, # 256
	    'Missing_EV': 9, # 512
	    'Cal_Fail': 10, # 1024
	    'Dead_Detector': 11 #2 048
	       }
    list_flag = ('Substitute_Cal', 'Out_of_Range',  'Saturation',
		'Temp_not_Nominal', 'Bowtie_Deleted','Missing_EV',
		'Cal_Fail', 'Dead_Detector')
    off_invalid = numpy.uint32(0)
    for iflag in list_flag:
	off_invalid = off_invalid + numpy.uint32(1 << bits[iflag])
    contrast_mask = (mask | numpy.bitwise_and(quality, off_invalid)>0)
    if contrast_mask.all():
        logger.info('No data')
        raise syntool_converter.EarlyExit()
    return contrast_mask


def convert_from_file(infile, outdir, geoloc_path=None,
                 vmin=None, vmax=None,
                 ngcps=(26, 32), open_iterations=1,
                 nprocs=1, write_netcdf=False, mission='SNPP'):
    """
    """

    infileid = None
    if geoloc_path is None:
        logger.error('Please Provide Geolocation file')
        raise Exception('')
    # Read data
    logger.info('Read data')
    handler = netCDF4.Dataset(infile, 'r')
    data = handler['observation_data']['I02'][:]
    quality_flag = handler['observation_data']['I02_quality_flags'][:]
    fmt = "%Y-%m-%dT%H:%M:%S.%fZ"
    start_time = datetime.datetime.strptime(handler.time_coverage_start, fmt)
    stop_time = datetime.datetime.strptime(handler.time_coverage_end, fmt)
    handler.close()

    # Read coordinates
    handler = netCDF4.Dataset(geoloc_path, 'r')
    lon = handler['geolocation_data']['longitude'][:]
    lat = handler['geolocation_data']['latitude'][:]
    handler.close()

    # Mask  data
    mask = ma.getmaskarray(data)
    contrast_mask = flag_data(quality_flag, mask)
    data.data[contrast_mask] = numpy.nan

    # GCPs for resampling and geotiff georeference
    scansize = 32
    resol = 375
    dtime0 = datetime.datetime.utcnow()
    gcps = resample.get_gcps_from_bowtie(lon, lat, scansize, ngcps=ngcps)
    dtime = datetime.datetime.utcnow() - dtime0
    logger.info('Get GCPs from bowtie swath : {}'.format(dtime))

    gcplon, gcplat, gcpnpixel, gcpnline = gcps
    rspysize = lon.shape[0]
    geod = pyproj.Geod(ellps='WGS84')
    mid = abs(gcpnline[:, 0] - 0.5).argmin()
    xdists = geod.inv(gcplon[mid, :-1], gcplat[mid, :-1],
                      gcplon[mid, 1:], gcplat[mid, 1:])[2]
    xdist = numpy.sum(xdists) / abs(gcpnpixel[mid, -1] - gcpnpixel[mid, 0])
    rspxsize = numpy.round(xdist / resol).astype('int') + 1
    gcpline = gcpnline * rspysize
    gcppixel = gcpnpixel * rspxsize

    # Resample with LinearNDInterpolator in output space
    dtime0 = datetime.datetime.utcnow()
    pix, lin = resample.get_points_from_gcps(gcplon, gcplat, gcppixel,
                                             gcpline, rspxsize, rspysize,
                                             1, lon, lat, nprocs=nprocs) - 0.5
    dtime = datetime.datetime.utcnow() - dtime0
    logger.info('Get input coordinates in new grid : {}'.format(dtime))
    dtime0 = datetime.datetime.utcnow()
    rspdata = resample.resample_bowtie_linear(pix, lin, data.data, scansize,
                                                rspxsize, rspysize, show=False)
    dtime = datetime.datetime.utcnow() - dtime0
    logger.info('Interpolate in new grid : {}'.format(dtime))

    rspdata = rspdata[::-1, ::-1]
    rspmask = ma.getmaskarray(rspdata)
    gcppixel = rspxsize - gcppixel
    gcpline = rspysize - gcpline
    if vmin is None:
        vmin = numpy.percentile(rspdata[~rspmask], 0.1)
    if vmax is None:
        count, division = numpy.histogram(rspdata[~rspmask], bins=100)
        peaks, props = find_peaks(-count, width=0.1)
        vmax = division[peaks[0]]
    print(vmin, vmax)

    # Construct metadata/geolocation/band(s)
    logger.info('Construct metadata/geolocation/band(s)')
    now = datetime.datetime.utcnow()
    granule_name, _ = os.path.splitext(os.path.basename(infile))
    metadata = {}
    (dtime, time_range) = stfmt.format_time_and_range(start_time,
                                                      stop_time,
                                                      units='ms')
    metadata['product_name'] = 'Brightness_contrast_VIIRS_{}'.format(mission)
    metadata['name'] = granule_name
    metadata['datetime'] = dtime
    metadata['time_range'] = time_range
    metadata['source_URI'] = infile
    metadata['source_provider'] = 'NOAA'
    metadata['processing_center'] = 'OceanDataLab'
    metadata['conversion_software'] = 'Syntool'
    metadata['conversion_version'] = '0.0.0'
    metadata['conversion_datetime'] = stfmt.format_time(now)
    metadata['parameter'] = 'brightness_contrast'
    metadata['type'] = 'remote sensing'
    metadata['sensor_type'] = 'radiometer'
    metadata['sensor_name'] = 'VIIRS'
    metadata['sensor_platform'] = mission
    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    gcpheight = numpy.zeros(gcppixel.shape)
    geolocation['gcps'] = stfmt.format_gdalgcps(gcplon, gcplat, gcpheight,
                                                gcppixel, gcpline)
    band = []
    array, offset, scale = pack.ubytes_0_254(rspdata.data, vmin, vmax)
    if ma.any(rspmask):
        array[numpy.where(rspmask == True)] = 255

    band.append({'array':array, 'scale':scale, 'offset':offset,
                 'description':'brightness contrast',
                 'unittype':'',
                 'nodatavalue':255, 'parameter_range':[vmin, vmax],
                 })

    if write_netcdf is False:
        # Write geotiff
        logger.info('Write geotiff')
        tifffile = stfmt.format_tifffilename(outdir, metadata, create_dir=True)
        stfmt.write_geotiff(tifffile, metadata, geolocation, band)
        # Write projected png/kml
    elif write_netcdf is True:
        logger.info('Write netcdf')
        ncfile = stfmt.format_ncfilename(outdir, metadata, create_dir=True)
        band[0]['name'] = 'i02'
        band[0]['long_name'] = 'Brightness contrast'
        band[0]['standard_name'] = 'radiance'
        band[0]['unittype'] = ''
        metadata['spatial_resolution'] = 375.
        stfmt.write_netcdf(ncfile, metadata, geolocation, band, 'swath',
                           ngcps=gcplon.shape)

