# -*- encoding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2018-10-22

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import numpy
import netCDF4
import logging
import datetime
import syntool_converter.utils.syntoolformat as stfmt
import syntool_converter.utils.pack as pack

logger = logging.getLogger(__name__)


def read_from_file(input_path):
    """"""
    f_handler = netCDF4.Dataset(input_path, 'r')
    lon = f_handler.variables['longitude'][:]
    lat = f_handler.variables['latitude'][:]
    _t = f_handler.variables['time'][:]
    _t_units = f_handler.variables['time'].units

    lon0, dlon = lon[0], numpy.mean(lon[1:] - lon[:-1])
    lat0, dlat = lat[0], numpy.mean(lat[1:] - lat[:-1])
    geo = {'projection': stfmt.format_gdalprojection(),
           'geotransform': [lon0, dlon, 0, lat0, 0, dlat]}

    now = datetime.datetime.utcnow()

    t_fmt = '%Y%m%dT%H%M%S'
    dtimes = netCDF4.num2date(_t, _t_units)
    granule_name = '{}_ascat_mean_wind'.format(dtimes[0].strftime(t_fmt))

    meta = {'product_name': 'ASCAT_MEAN_WIND',
            'name': granule_name,
            'datetime': stfmt.format_time(dtimes[0]),
            'time_range': ['-12h', '+12h'],
            'source_URI': input_path,
            'source_provider': 'IFREMER',
            'processing_center': '',
            'conversion_software': 'Syntool',
            'conversion_software': '0.0.0',
            'conversion_datetime': stfmt.format_time(now),
            'parameter': 'wind'}


    data = []

    v_data = f_handler.variables['wind_speed'][0, 0, :, :]
    vmin = 0.0
    vmax = 25.4
    colortable = stfmt.format_colortable('matplotlib_jet',
                                         vmin=vmin, vmax=vmax,
                                         vmin_pal=vmin, vmax_pal=vmax)
    array, offset, scale = pack.ubytes_0_254(v_data, vmin, vmax)
    if numpy.any(v_data.mask):
        array[numpy.where(v_data.mask)] = 255
    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'wind speed',
                 'unit': 'm.s-1',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    u_data = f_handler.variables['eastward_wind'][0, 0, :, :]
    v_data = f_handler.variables['northward_wind'][0, 0, :, :]
    n_mask = (u_data.mask | v_data.mask)
    n_data = numpy.rad2deg(numpy.arctan2(v_data, u_data))
    vmin = 0.0
    vmax = 360.0
    array, offset, scale = pack.ubytes_0_254(n_data, vmin, vmax)
    if numpy.any(n_mask):
        array[numpy.where(n_mask)] = 255
    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': 'wind direction',
                 'unit': 'degrees',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax]})

    yield (meta, geo, data, None)
    f_handler.close()


def convert_from_file(input_path, output_path):
    """"""
    for (meta, geo, data, extra) in read_from_file(input_path):
        tiff_fname = stfmt.format_tifffilename(output_path, meta,
                                               create_dir=True)
        stfmt.write_geotiff(tiff_fname, meta, geo, data)


if '__main__' == __name__:
    import sys

    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARN)

    input_path = sys.argv[1]
    output_path = sys.argv[2]

    convert_from_file(input_path, output_path)
