# -*- encoding: utf-8 -*-

"""
@author: <lucile.gaultier@oceandatalab.com>
@author: <sylvain.herledan@oceandatalab.com>
@date: 2019-05-14

Copyright (C) 2014-2021 OceanDataLab

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import os
import numpy
import logging
import netCDF4
import datetime
import syntool_converter
import syntool_converter.utils.pack as pack
import syntool_converter.utils.syntoolformat as stfmt

logger = logging.getLogger(__name__)

GDAL_MAX_GCPS = 8000


def build_gcps(slice_lat, slice_lon):
    """"""
    dgcp = 64.
    min_abs_lat = numpy.min(numpy.abs(slice_lat))
    max_abs_lat = numpy.max(numpy.abs(slice_lat))

    # Increase GCPs density on high latitudes
    if numpy.all(70 < min_abs_lat):
        dgcp = 32.
    if 87.0 < max_abs_lat:
        dgcp = 16.
    if 88.0 < max_abs_lat:
        dgcp = 8.
    if 89.0 < max_abs_lat:
        dgcp = 4.

    slice_shape = numpy.array(slice_lon.shape)
    ngcps = (numpy.ceil(slice_shape / dgcp) + 1.).astype('int32')

    # Adapt GCPs density so the GCPs counter doesn't exceed what GDAL is
    # able to handle (~8000)
    while GDAL_MAX_GCPS < ngcps[0] * ngcps[1]:
        dgcp = dgcp * 2.0
        ngcps = (numpy.ceil(slice_shape / dgcp) + 1.).astype('int32')

    pix = numpy.linspace(0, slice_lon.shape[1] - 1, num=ngcps[1]).round()
    lin = numpy.linspace(0, slice_lon.shape[0] - 1, num=ngcps[0]).round()
    pix2d, lin2d = numpy.meshgrid(pix.astype('int32'),
                                  lin.astype('int32'))
    _gcp_lon = slice_lon[lin2d, pix2d]
    gcp_lat = slice_lat[lin2d, pix2d]
    gcp_pix = pix2d + 0.5
    gcp_lin = lin2d + 0.5
    gcp_hei = numpy.zeros(ngcps)

    # Recreate the GCP longitudes grid using the following assumption:
    # "Longitudinal distance between two adjacent GCPs should never exceed 180
    # degrees"
    #
    # Maybe there is a cleaner way to achieve longitudinal continuity even
    # across the IDL but at least this method is simple and produces the
    # expected result.
    gcp_lon = numpy.full(numpy.shape(_gcp_lon), _gcp_lon[0, 0],
                         dtype=_gcp_lon.dtype)

    _dlon0_dy = _gcp_lon[1:, 0] - _gcp_lon[:-1, 0]
    dlon0_dy = numpy.mod(_dlon0_dy + 180.0, 360.0) - 180.0
    for y in range(dlon0_dy.shape[0]):
        gcp_lon[y + 1, 0] = gcp_lon[y, 0] + dlon0_dy[y]

    _dlon_dx = _gcp_lon[:, 1:] - _gcp_lon[:, :-1]
    dlon_dx = numpy.mod(_dlon_dx + 180.0, 360.0) - 180.0
    for x in range(dlon_dx.shape[1]):
        gcp_lon[:, x + 1] = gcp_lon[:, x] + dlon_dx[:, x]

    if 120.0 <= (numpy.max(gcp_lon) - numpy.min(gcp_lon)):
        raise LongitudinalExtentTooWide()

    geolocation = {}
    geolocation['projection'] = stfmt.format_gdalprojection()
    geolocation['gcps'] = stfmt.format_gdalgcps(gcp_lon, gcp_lat, gcp_hei,
                                                gcp_pix, gcp_lin)

    return geolocation


def sar_l2c_owi(infile, outdir):
    """"""
    basename = os.path.basename(infile)
    granule_name, _ = os.path.splitext(basename)

    infile = os.path.join(infile)
    fid = netCDF4.Dataset(infile, 'r')

    # Read coordinates
    lon = fid.variables['owiLon'][:]
    lat = fid.variables['owiLat'][:]

    # Read time
    start_time = datetime.datetime.strptime(fid.firstMeasurementTime,
                                            '%Y-%m-%dT%H:%M:%SZ')
    end_time = datetime.datetime.strptime(fid.lastMeasurementTime,
                                          '%Y-%m-%dT%H:%M:%SZ')

    # Read variables
    norm = numpy.ma.array(fid.groups['owiInversionTables_UV'].variables['owiWindSpeed_Tab_dualpol_2steps'][:])
    #direction = numpy.ma.array(fid.groups['owiInversionTables_UV'].variables['owiWindDirection_Tab_dualpol_2steps'][:])
    direction = numpy.ma.array(fid.groups['owiPreProcessing'].variables['wind_streaks_orientation'][:])
    fid.close()

    # Direction is provided with meteo convention
    direction = 270 - direction
    direction = numpy.mod(direction, 360.0)

    geo = build_gcps(lat, lon)

    data = []
    mask = (numpy.ma.getmaskarray(norm) | numpy.ma.getmaskarray(direction))
    colortable = stfmt.format_colortable('faozi',
                                         vmin=0.0, vmax=69.964,
                                         vmin_pal=0.0, vmax_pal=69.964)


    vmin = 0.0
    vmax = 69.964
    norm.mask = mask
    array, offset, scale = pack.ubytes_0_254(norm, vmin, vmax)
    if numpy.any(mask):
        array[numpy.where(mask)] = 255

    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': '',
                 'name': 'wind_speed',
                 'unittype': 'm.s-1',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax],
                 'colortable': colortable})

    vmin = 0.0
    vmax = 360.0
    direction.mask = mask
    array, offset, scale = pack.ubytes_0_254(direction, vmin, vmax)
    if numpy.any(mask):
        array[numpy.where(mask)] = 255
    data.append({'array': array,
                 'scale': scale,
                 'offset': offset,
                 'description': '',
                 'name': 'wind_direction',
                 'unittype': 'degree',
                 'nodatavalue': 255,
                 'parameter_range': [vmin, vmax]})

    now = datetime.datetime.utcnow()
    (dtime, time_range) = stfmt.format_time_and_range(start_time, end_time,
                                                      units='s')

    platform = 'Sentinel-1'
    if granule_name.startswith('rs2'):
        platform = 'RADARSAT-2'
    meta = {}
    meta['product_name'] = 'SAR_L2C_OWI'
    meta['name'] = granule_name
    meta['datetime'] = dtime
    meta['time_range'] = time_range
    meta['source_URI'] = infile
    meta['source_provider'] = 'IFREMER'
    meta['processing_center'] = 'OceanDataLab'
    meta['conversion_software'] = 'Syntool'
    meta['conversion_version'] = '0.0.0'
    meta['conversion_datetime'] = None  # set later
    meta['parameter'] = ['wind_speed', 'wind_direction']
    meta['type'] = 'remote sensing'
    meta['sensor_platform'] = platform
    meta['conversion_datetime'] = stfmt.format_time(now)

    tifffile = stfmt.format_tifffilename(outdir, meta, create_dir=True)
    stfmt.write_geotiff(tifffile, meta, geo, data)



if '__main__' == __name__:
    # Setup logging
    main_logger = logging.getLogger()
    main_logger.handlers = []
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.DEBUG)

    import sys
    infile = sys.argv[1]
    outdir = sys.argv[2]

    sar_l2_owi(infile, outdir)
